// Angle.cs
// Static class of angle routines for my SOFA DLL.
// Copyright 2018 by Paul S. Hirose.

// revision history
#if false
2007-10-08 Rewrote to call my .NET Framework SofaJplNet DLL instead of Win32 sofajpl.dll.
2007-11-25 Added self-test method, which ran OK.
2008-01-30 Added an attribute to ensure the assembly is CLS compliant.
2008-02-10 Moved the preceding to its own file.
2008-06-13 Added horizontal parallax and semidiameter methods.
2008-06-19 DmsToRad() and HmsToRad() added and test OK.
2008-06-26 Era00() and EquaOrig06a() added. Test OK.
2008-06-28 GAST methods added. Test OK.
2008-07-02 Apparent semidiameter method added.
2008-07-03 Change semidiameter computation to use arc sine, not arc tangent, to match the
Almanac algorithm (even though the Explanatory Supp. says to use arc tan). The arc sin way
makes more sense to me: it gives the angle between the center and the limb.
Also changed arguments to the horizontal parallax, semidiameter, and apparent SD methods
so they're more convenient to call. E.g., Earth radius is an assumed value.
2012-03-31 Added DmsToDegrees(). Self-test OK.
2012-10-13 Added some shorter name aliases to the angle and time conversion methods. Still
not done.
2012-10-14 Eliminated the AU in favor of the km. The former varies a tiny amount between
the various JPL ephemerides, while the latter is constant. Removed GastMica().
2012-10-15 Eliminated the special semidiameter method for the Moon. Redundant. Self
test revision mostly done. Resume at GAST94.
2012-10-16 Self test brought up to date and runs OK.

2014-10-07  Began modifications to support my new SOFA DLL. The only changes are from upper to lower
case to match the SOFA spellings.
2014-10-09  Minor change to DmsToDegrees() comment. Same for related methods.
All Earth rotation angle and sidereal time methods disabled.
Self test needs to be checked yet.
2014-10-17  Self test revised and passes.
2015-04-14  Added HmsToDays(). Test OK.
2015-05-11 Earth rotation angle and sidereal time methods enabled. Not sure why I disabled them
on 2014-10-09. Calls to the SOFA routines are not as convenient.

2018-06-04 Added Gast06b().
#endif

using System;


namespace HirosePS.SofaJpl
{
    /// <summary>
    /// Miscellaneous functions and constants applicable to angles.
    /// </summary>
    /// <remarks>Unit conversions (degrees, hours radians). Pi/2, pi, 2pi, complements and
    /// supplements. Sexagesimal to decimal. Normalization to the intervals [0, 2pi) and [-pi, +pi).
    /// Semidiameter and horizontal parallax. Sidereal time and Earth rotation angle.</remarks>
    public static class Angle
    {
        /// <summary>
        /// The constant pi divided by 2.
        /// </summary>
        /// <value>pi divided by 2</value>
        public const double HalfPi = Math.PI / 2.0;

        /// <summary>
        /// The constant pi.
        /// </summary>
        /// <value>pi</value>
        public const double Pi = Math.PI;

        /// <summary>
        /// The constant 2 times pi.
        /// </summary>
        /// <value>2 times pi</value>
        public const double TwoPi = Math.PI * 2.0;


        /// <summary>
        /// Convert sexagesimal angle to decimal.
        /// </summary>
        /// <remarks>If one or more parameters is negative, all are assumed to be negative, and the
        /// result is negative. The parameters are not restricted to their conventional ranges,
        /// e.g., minutes may exceed 60.
        /// </remarks>
        /// <param name="deg">degrees</param>
        /// <param name="min">minutes</param>
        /// <param name="sec">seconds</param>
        /// <returns>decimal degrees</returns>
        public static double DmsToDegrees(double deg, double min, double sec)
        {
            double d = Math.Abs(deg) + (Math.Abs(min) + Math.Abs(sec) / 60.0) / 60.0;
            return (deg < 0.0 || min < 0.0 || sec < 0.0) ? -d : d;
        }


        /// <summary>
        /// Convert sexagesimal angle to decimal.
        /// </summary>
        /// <remarks>If one or more parameters is negative, all are assumed to be negative, and the
        /// result is negative. The parameters are not restricted to their conventional ranges,
        /// e.g., minutes may exceed 60.
        /// </remarks>
        /// <param name="deg">degrees</param>
        /// <param name="min">minutes</param>
        /// <param name="sec">seconds</param>
        /// <returns>decimal degrees</returns>
        public static double DmsToDeg(double deg, double min, double sec)
        {
            return DmsToDegrees(deg, min, sec);
        }


        /// <summary>
        /// Convert sexagesimal time to decimal.
        /// </summary>
        /// <remarks>If one or more parameters is negative, all are assumed to be negative, and the
        /// result is negative. The parameters are not restricted to their conventional ranges,
        /// e.g., minutes may exceed 60.
        /// </remarks>
        /// <param name="hour">hours</param>
        /// <param name="min">minutes</param>
        /// <param name="sec">seconds</param>
        /// <returns>decimal hours</returns>
        public static double HmsToHours(double hour, double min, double sec)
        {
            return DmsToDegrees(hour, min, sec);
        }


        /// <summary>
        /// Convert sexagesimal time to decimal days.
        /// </summary>
        /// <param name="hour">hours</param>
        /// <param name="min">minutes</param>
        /// <param name="sec">seconds</param>
        /// <returns>decimal days</returns>
        /// <remarks>If one or more parameters is negative, all are assumed to be negative, and the
        /// result is negative. The parameters are not restricted to their conventional ranges,
        /// e.g., minutes may exceed 60.
        /// </remarks>
        public static double HmsToDays(double hour, double min, double sec) {
            return DmsToDegrees(hour, min, sec) / 24.0;
        }


        /// <summary>
        /// Convert degrees to radians.
        /// </summary>
        /// <param name="degrees">degrees</param>
        /// <returns>radians</returns>
        public static double DegreesToRadians(double degrees)
        {
            return degrees / 180.0 * Pi;
        }

        /// <summary>
        /// Convert degrees to radians.
        /// </summary>
        /// <param name="degrees">degrees</param>
        /// <returns>radians</returns>
        public static double DegToRad(double degrees)
        {
            return DegreesToRadians(degrees);
        }

        /// <summary>
        /// Convert hours to radians. (12 hours = pi radians)
        /// </summary>
        /// <param name="hours">hours</param>
        /// <returns>radians</returns>
        public static double HoursToRadians(double hours)
        {
            return hours / 12.0 * Pi;
        }

#if false
        /// <summary>
        /// Convert hours to radians. (12 hours = pi radians)
        /// </summary>
        /// <param name="hours">hours</param>
        /// <returns>radians</returns>
        public static double HourToRad(double hours)
        {
            return HoursToRadians(hours);
        }
#endif

        /// <summary>
        /// Convert gons to radians. (200 gon = pi radians)
        /// </summary>
        /// <param name="gons">gons</param>
        /// <returns>radians</returns>
        public static double GonsToRadians(double gons)
        {
            return gons / 200.0 * Math.PI;
        }
        
        /// <summary>
        /// Convert degrees, minutes, seconds to radians.
        /// </summary>
        /// <remarks>To indicate a negative angle, make at least one parameter negative.
        /// </remarks>
        /// <param name="deg">degrees</param>
        /// <param name="min">minutes</param>
        /// <param name="sec">seconds</param>
        /// <returns>radians</returns>
        public static double DmsToRad(double deg, double min, double sec)
        {
            return DegreesToRadians(DmsToDegrees(deg, min, sec));
        }

        /// <summary>
        /// Convert hours, minutes, seconds to radians. (12 hours = pi radians)
        /// </summary>
        /// <remarks>To indicate a negative time, make at least one parameter negative.
        /// </remarks>
        /// <param name="hour">hours</param>
        /// <param name="min">minutes</param>
        /// <param name="sec">seconds</param>
        /// <returns>radians</returns>
        public static double HmsToRad(double hour, double min, double sec)
        {
            return HoursToRadians(HmsToHours(hour, min, sec));
        }

        /// <summary>
        /// Convert radians to degrees.
        /// </summary>
        /// <param name="radians">radians</param>
        /// <returns>degrees</returns>
        public static double RadiansToDegrees(double radians)
        {
            return radians * 180.0 / Math.PI;
        }

        /// <summary>
        /// Convert radians to degrees.
        /// </summary>
        /// <param name="radians">radians</param>
        /// <returns>degrees</returns>
        public static double RadToDeg(double radians)
        {
            return RadiansToDegrees(radians);
        }

        /// <summary>
        /// Convert radians to hours. (12 hours = pi radians)
        /// </summary>
        /// <param name="radians">radians</param>
        /// <returns>hours</returns>
        public static double RadiansToHours(double radians)
        {
            return radians / Pi * 12.0;
        }

        /// <summary>
        /// Convert radians to gons. (200 gon = pi radians)
        /// </summary>
        /// <param name="radians">radians</param>
        /// <returns>gons</returns>
        public static double RadiansToGons(double radians)
        {
            return radians / Math.PI * 200.0;
        }

        /// <summary>
        /// Complement: pi/2 radians minus <paramref name="angle"/>.
        /// </summary>
        /// <remarks>The result is not necessarily in the interval [0, 2pi).</remarks>
        /// <param name="angle">radians</param>
        /// <returns>complement of <paramref name="angle"/> (radians)</returns>
        public static double Complement(double angle)
        {
            return HalfPi - angle;
        }

        /// <summary>
        /// Supplement: pi radians minus <paramref name="angle"/>.
        /// </summary>
        /// <remarks>The result is not necessarily in the interval [0, 2pi).</remarks>
        /// <param name="angle">radians</param>
        /// <returns>supplement of <paramref name="angle"/> (radians)</returns>
        public static double Supplement(double angle)
        {
            return Math.PI - angle;
        }

        /// <summary>
        /// Explement: 2pi radians minus <paramref name="angle"/>.
        /// </summary>
        /// <remarks>The result is not necessarily in the interval [0, 2pi).</remarks>
        /// <param name="angle">radians</param>
        /// <returns>explement of <paramref name="angle"/> (radians)</returns>
        public static double Explement(double angle)
        {
            return TwoPi - angle;
        }

        /// <summary>
        /// Normalize to plus: put an angle in the interval [0, 2pi) radians.
        /// </summary>
        /// <param name="angle">radians</param>
        /// <returns><paramref name="angle"/> normalized to [0, 2pi) radians</returns>
        public static double NormPlus(double angle)
        {
            return Sofa.Anp(angle);
        }

        /// <summary>
        /// Normalize to plus or minus: put an angle in the interval [-pi, +pi) radians.
        /// </summary>
        /// <param name="angle">radians</param>
        /// <returns><paramref name="angle"/> normalized to [-pi, +pi) radians</returns>
        public static double NormPlusMinus(double angle)
        {
            return Sofa.Anpm(angle);
        }


        /// <summary>
        /// Horizontal parallax.
        /// </summary>
        /// <remarks>Both parameters must use the same unit of length.</remarks>
        /// <param name="distance">distance to body</param>
        /// <param name="earthRadius">Earth radius</param>
        /// <returns>horizontal parallax (radians)</returns>
        public static double HorizontalParallax(double distance, double earthRadius)
        {
            return Math.Asin(earthRadius / distance);
        }


        /// <summary>
        /// Semidiameter of a body.
        /// </summary>
        /// <remarks>The formula is an arc sine. Both parameters must use the same unit
        /// of length. For example, if distance is given in Earth radii, so must the
        /// body radius.</remarks>
        /// <param name="distance">distance to the body</param>
        /// <param name="bodyRadius">body's radius</param>
        /// <returns>semidiameter (radians)</returns>
        public static double Semidiameter(double distance, double bodyRadius)
        {
            return Math.Asin(bodyRadius / distance);
        }


        // With appropriate argument units, Semidiameter() can do anything this can do.
#if false
        /// <summary>
        /// Apparent semidiameter of the Moon, given its distance.
        /// </summary>
        /// <remarks>Formula from the eclipse explanation in the Almanac (k = .2725076).
        /// </remarks>
        /// <param name="distance">distance to the Moon (Earth radii)</param>
        /// <returns>apparent Moon semidiameter (radians)</returns>
        public static double MoonApparentSemidiameter(double distance)
        {
            return Math.Asin(.2725076 / distance);
        }
#endif


        // Earth rotation angle and sidereal time methods.

        
        /// <summary>
        /// Earth rotation angle (IAU 2000 model).
        /// </summary>
        /// <param name="ut1">UT1</param>
        /// <returns>earth rotation angle (radians)</returns>
        public static double Era00(JulianDate ut1)
        {
            return Sofa.Era00(ut1.DatePart, ut1.TimePart);
        }


        /// <summary>
        /// Equation of the origins (earth rotation angle minus GAST), IAU 2006 precession
        /// and IAU 2000A nutation.
        /// </summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>equation of the origins (radians)</returns>
        public static double EquaOrig06a(JulianDate tt)
        {
            return Sofa.Eo06a(JulianDate.J2000Base.ToDouble(), tt.ToJ2000());
        }


#if false
        /// <summary>
        /// Greenwich apparent sidereal time, compatible with the USNO MICA 2.0 program.
        /// </summary>
        /// <remarks>This is compatible with MICA 2.0 at the present epoch, to the limits
        /// of MICA's resolution. The simplifying assumption that TT = UT1 introduces only
        /// about 2 microseconds of error.</remarks>
        /// <param name="ut1">UT1</param>
        /// <returns>Greenwich apparent sidereal time (radians)</returns>
        public static double GastMica(JulianDate ut1)
        {
            // mean obliquity
            double epsa = Sofa.Obl80(JulianDate.J2000Base.ToDouble(), ut1.ToJ2000());

            // nutation in longitude and obliquity
            double dpsi, deps;
            Sofa.Nut80(JulianDate.J2000Base.ToDouble(), ut1.ToJ2000(), out dpsi, out deps);

            return Sofa.Gmst82(ut1.DatePart, ut1.TimePart) + dpsi * Math.Cos(epsa + deps);
        }
#endif


        /// <summary>
        /// Greenwich apparent sidereal time, consistent with the IAU 1982/94 resolutions.
        /// </summary>
        /// <remarks>
        /// <para>Accuracy has been compromised for the sake of convenience in that UT is
        /// used instead of TDB (or TT) to compute the equation of the equinoxes.</para>
        /// <para>This GAST must be used only in conjunction with contemporaneous
        /// IAU standards such as 1976 precession, 1980 obliquity and 1982
        /// nutation. It is not compatible with the IAU 2000 resolutions.</para>
        /// </remarks>
        /// <param name="ut1">UT1</param>
        /// <returns>Greenwich apparent sidereal time (radians)</returns>
        public static double Gast94(JulianDate ut1)
        {
            return Sofa.Gst94(ut1.DatePart, ut1.TimePart);
        }


        /// <summary>
        /// Greenwich apparent sidereal time, consistent with the IAU 2000 resolutions
        /// (IAU 2000A nutation model).
        /// </summary>
        /// <remarks>For more speed and less accuracy, a routine with the truncated IAU
        /// 2000B nutation model is also provided.</remarks>
        /// <param name="ut1">UT1</param>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>Greenwich apparent sidereal time (radians)</returns>
        public static double Gast00a(JulianDate ut1, JulianDate tt)
        {
            return Sofa.Gst00a(ut1.DatePart, ut1.TimePart, tt.DatePart, tt.TimePart);
        }


        /// <summary>
        /// Greenwich apparent sidereal time, consistent with the IAU 2000 resolutions
        /// (IAU 2000B nutation model).
        /// </summary>
        /// <remarks>
        /// <para>Accuracy has been compromised for the sake of speed and convenience in
        /// two respects: 1) UT is used instead of TDB (or TT) to compute the precession
        /// component of GMST and the equation of the equinoxes. This results in errors of
        /// order 0.1 mas at present. 2) The IAU 2000B abridged nutation model (McCarthy
        /// and Luzum, 2001) is used, introducing errors of up to 1 mas.</para>
        /// <para>For more accuracy and less speed, a routine with the full IAU 2000A
        /// nutation model is also provided.</para>
        /// </remarks>
        /// <param name="ut1">UT1</param>
        /// <returns>Greenwich apparent sidereal time (radians)</returns>
        public static double Gast00b(JulianDate ut1)
        {
            return Sofa.Gst00b(ut1.DatePart, ut1.TimePart);
        }


        /// <summary>
        /// Greenwich apparent sidereal time, consistent with the IAU 2000 and 2006
        /// resolutions.
        /// </summary>
        /// <remarks>This GAST is compatible with the IAU 2000/2006 resolutions and must
        /// be used only in conjunction with IAU 2006 precession and IAU 2000A nutation.
        /// </remarks>
        /// <param name="ut1">UT1</param>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>Greenwich apparent sidereal time (radians)</returns>
        public static double Gast06a(JulianDate ut1, JulianDate tt)
        {
            return Sofa.Gst06a(ut1.DatePart, ut1.TimePart, tt.DatePart, tt.TimePart);
        }


        /// <summary>
        /// Greenwich apparent sidereal time (IAU 2006 precession and 2000B nutation).
        /// </summary>
        /// <param name="ut1">UT1</param>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>Greenwich apparent sidereal time (radians)</returns>
        public static double Gast06b(JulianDate ut1, JulianDate tt) {
            // The SOFA routines are most accurate if the components of UT1 and TT are in
            // the J2000 format.
            double j2000Base = JulianDate.J2000Base.ToDouble();       // JD at J2000.0
            double ut1_J2000 = ut1.ToJ2000();       // days after J2000.0
            double tt_J2000 = tt.ToJ2000();         // days after J2000.0

            // Fukushima-Williams angles for frame bias and precession.
            double gamb, phib, psib, epsa;
            Sofa.Pfw06(j2000Base, tt_J2000, out gamb, out phib, out psib, out epsa);

            // IAU 2000B nutation components. Per the "SOFA Tools for Earth Attitude" document,
            // the small adjustments that are formally required with IAU 2000A nutation may be
            // neglected with the 2000B model.
            double dpsi, deps;
            Sofa.Nut00b(j2000Base, tt_J2000, out dpsi, out deps);

            // Equinox based nutation x precession x bias matrix.
            double[,] rnpb = new double[3, 3];
            Sofa.Fw2m(gamb, phib, psib + dpsi, epsa + deps, rnpb);

            return Sofa.Gst06(ut1.DatePart, ut1.TimePart, tt.DatePart, tt.TimePart, rnpb);
        }


        #region self-test code
#if DEBUG

        /// <summary>
        /// Self-test method.
        /// </summary>
        public static void Test()
        {
            // multiples of pi
            equalsApprox(HalfPi, 3.1415926535897932384626433832795 / 2.0, 1e-12, "HalfPi");
            equalsApprox(Pi, 3.1415926535897932384626433832795, 1e-12, "Pi");
            equalsApprox(TwoPi, 3.1415926535897932384626433832795 * 2.0, 1e-12, "TwoPi");

            // DmsToDegrees()
            equalsApprox(DmsToDegrees(0, 0, 0), 0.0, 0.0,
                "Angle.DmsToDegrees(0,0,0)");
            equalsApprox(DmsToDegrees(110.0, 0, 0), 110.0, 0.0,
                "Angle.DmsToDegrees(110,0,0)");
            equalsApprox(DmsToDegrees(0.0, 15.0, 0.0), .25, 0.0,
                "Angle.DmsToDegrees(0,15,0)");
            equalsApprox(DmsToDegrees(0, 0, 1.0), 1.0 / 3600.0, 1e-9,
                "Angle.DmsToDegrees(0,0,1)");
            equalsApprox(DmsToDegrees(351.0, 4.0, 52.0), 351.081111111, 2e-9,
                "Angle.DmsToDegrees(351.0,4.0,52.0)");
            equalsApprox(DmsToDegrees(-351.0, 4.0, 52.0), -351.081111111, 2e-9,
                "Angle.DmsToDegrees(-351.0,4.0,52.0)");
            equalsApprox(DmsToDegrees(-351.0, -4.0, -52.0), -351.081111111, 2e-9,
                "Angle.DmsToDegrees(-351.0,-4.0,-52.0)");

            // DmsToDeg() is simply a shell for the above, so one test is sufficient.
            // Likewise HmsToHours().
            equalsApprox(DmsToDeg(-351.0, -4.0, -52.0), -351.081111111, 2e-9,
                "Angle.DmsToDeg(-351.0,-4.0,-52.0)");
            equalsApprox(HmsToHours(-351.0, -4.0, -52.0), -351.081111111, 2e-9,
                "Angle.HmsToHours(-351.0,-4.0,-52.0)");

            // Unit conversions
            assertTrue(Math.Abs(Angle.DegreesToRadians(-431.273) - (-7.52713)) < .00001);
            assertTrue(Math.Abs(Angle.DegToRad(-431.273) - (-7.52713)) < .00001);
            assertTrue(Math.Abs(Angle.RadiansToDegrees(-7.52713) - (-431.273)) < .001);
            assertTrue(Math.Abs(Angle.RadToDeg(-7.52713) - (-431.273)) < .001);
            assertTrue(Math.Abs(Angle.GonsToRadians(-479.192) - (-7.52713)) < .00001);
            assertTrue(Math.Abs(Angle.RadiansToGons(-7.52713) - (-479.192)) < .001);
            assertTrue(Math.Abs(Angle.RadiansToHours(-7.52713) - (-28.7515)) < .0001);
            assertTrue(Math.Abs(Angle.HoursToRadians(-28.7515) - (-7.52713)) < .00001);
#if false   // method deleted
            assertTrue(Math.Abs(Angle.HourToRad(-28.7515) - (-7.52713)) < .00001);
#endif

            // DMS to radians
            assertTrue(Math.Abs(DmsToRad(114.0, 37.0, 40.0) - 2.000637) < 5e-6);
            assertTrue(Math.Abs(DmsToRad(-114, 37, 40.0) + 2.000637) < 5e-6);
            assertTrue(Math.Abs(DmsToRad(-114, -37, -40) + 2.000637) < 5e-6);
            assertTrue(Math.Abs(DmsToRad(0.0, 0, -40.0) + .000194) < 5e-6);

            // HMS to radians
            assertTrue(Math.Abs(HmsToRad(15.0, 43.0, 44.5) - 4.117853) < 7e-6);
            assertTrue(Math.Abs(HmsToRad(-15.0, 43.0, 44.5) + 4.117853) < 7e-6);
            assertTrue(Math.Abs(HmsToRad(15, -43, 44.5) + 4.117853) < 7e-6);
            assertTrue(Math.Abs(HmsToRad(0.0, -43, 44.5) + .190859) < 7e-6);

            // HMS to days. The method's functionality comes from DmsToDegrees(), so a check that
            // degrees are converted to days is sufficient.
            equalsApprox(HmsToDays(24, 0, 0), 1.0, 1e-9, "HmsToDays()");

            // complement, supplement, explement
            assertTrue(Math.Abs(Angle.Complement(5.950) - (-4.379)) < .001);
            assertTrue(Math.Abs(Angle.Supplement(-5.958) - 9.100) < .001);
            assertTrue(Math.Abs(Angle.Explement(-2.984) - 9.267) < .001);

            // normalize
            assertTrue(Math.Abs(Angle.NormPlus(33.917) - 2.501) < .001);
            assertTrue(Math.Abs(Angle.NormPlus(-33.917) - 3.782) < .001);
            assertTrue(Math.Abs(Angle.NormPlusMinus(33.917) - 2.501) < .001);
            assertTrue(Math.Abs(Angle.NormPlusMinus(-33.917) - (-2.501)) < .001);

            // horizontal parallax, semidiameter

            // values from Almanac 2012 Feb 15
            const double moonRadius = .272399;     // wrt Earth radius (p. D3)
            const double moonDistance = 58.251;
            double moonHP = DmsToRad(0, 59, 1.12);
            double moonSD = DmsToRad(0, 16, 4.55);

            // Test the HP and SD methods.
            equalsApprox(moonHP, HorizontalParallax(58.251, 1.0), DmsToRad(0, 0, .1),
                "HorizontalParallax()");
            equalsApprox(moonSD, Semidiameter(moonDistance, moonRadius),
                DmsToRad(0, 0, .1), "Semidiameter()");


            // Rotation angle and sidereal time methods.

            // earth rotation angle
            assertTrue(Math.Abs(Era00(new JulianDate(2006, 1, 0, 0.0)) - DmsToRad(
                99, 26, 39.6076)) < DmsToRad(0, 0, .0001));

            // equation of the origins
            assertTrue(Math.Abs(Angle.EquaOrig06a(new JulianDate(2006, 1, 0, 0.0)) -
                DmsToRad(0, -4, 34.6091)) < DmsToRad(0, 0, .001));

            // GAST 1994 (SOFA test)
            equalsApprox(Gast94(new JulianDate(2400000.5, 53736.0)), 1.754166136020645203,
                1e-12, "Gast94()");

            // GAST (MICA)
#if false
            assertTrue(Math.Abs(NormPlusMinus(GastMica(ut1) - micaGast))
                < HmsToRad(0, 0, .0001));
#endif

            // GAST 2000A (SOFA test)
            equalsApprox(Gast00a(new JulianDate(2400000.5, 53736.0), new JulianDate(
                2400000.5, 53736.0)), 1.754166138018281369, 1e-12, "Gast00a()");

            // GAST 2000B (SOFA test)
            equalsApprox(Gast00b(new JulianDate(2400000.5, 53736.0)),
                1.754166136510680589, 1e-12, "Gast00b()");

            // GAST 2006A (SOFA test)
            equalsApprox(Gast06a(new JulianDate(2400000.5, 53736.0), new JulianDate(
                2400000.5, 53736.0)), 1.754166137675019159, 1e-12, "Gast06a()");

            // GAST 2006B. Should be within a couple mas of 2000A.
            equalsApprox(Gast06b(new JulianDate(2400000.5, 53736.0), new JulianDate(
                2400000.5, 53736.0)), 1.754166137675019159, SofaJpl.Angle.DmsToRad(0, 0, .002),
                "Gast06b()");
        }


        /// <summary>
        /// Throw exception if argument is false.
        /// </summary>
        /// <param name="assertion">true = OK</param>
        private static void assertTrue(bool assertion)
        {
            if (!assertion)
                throw new Exception("Assertion was false.");
        }


        /// <summary>
        /// Throw exception if argument is true.
        /// </summary>
        /// <param name="assertion">false = OK</param>
        private static void assertFalse(bool assertion)
        {
            if (assertion)
                throw new Exception("Assertion was true.");
        }


        private static void equalsApprox(double a, double b, double tolerance,
            string message)
        {
            if (Math.Abs(a - b) > tolerance)
                throw new Exception(message);
        }

#endif  // end self test code
        #endregion
    }
}
