﻿// Atmosphere.cs
// Troposphere and astronomical refraction class.
// Copyright 2018 by Paul S. Hirose

// revision history
#if false

2015-07-05 Created. (Actually, took the code from my Lunar3 program.)
2015-07-15 Rewrote so the internal pressure unit is pascals throughout program. The user still sees
millibars. Self test OK.
2015-08-30 MillibarsPerInchHg property returns float now, not double.
2015-09-05 Refract() and Unrefract() have Spherical overloads.

2018-04-18 Added a property to get the molecular mass.
#endif


using System;



namespace HirosePS.SofaJpl {
/// <summary>
/// The atmospheric conditions at the observer.
/// </summary>
/// <remarks>
/// <para>The main purpose is to apply refraction, which as accomplished with the Cassini formula,
/// or Astronomical Almanac formula at low altitude. To prevent discontinuity where the models
/// change, a weighted average is taken in the transition zone (altitude 10 - 15 degrees).</para>
/// <para>The class also contains functions to compute pressure and density in a standard atmosphere
/// gradient layer or isothermal layer.</para>
/// </remarks>
public class Atmosphere {
    // The transition zone where the low altitude refraction model gradually takes over (degrees).
    const float _upperTransitionDeg = 15;
    const float _lowerTransitionDeg = 10;

    const float _co2Default = 400;          // default CO2 level, ppm
    const float _lambdaDefault = .55f;      // default light wavelength (microns)

    // Max number of iterations in the refraction solution before the method throws
    // an exception.
    const int _maxIterations = 10;

    // US Standard Atmosphere 1976 constants
    const float _g0 = 9.80665f;         // accel of gravity
    const float _p0 = 101325.0f;        // sea level pressure, pascals
    const float _t0 = 288.15f;          // sea level temperature, K
    const float _rStar = 8314.32f;      // universal gas constant
    const float _a0 = -.0065f;          // lapse rate, Kelvins per meter
    const float _mDry = 28.9644f;       // molecular mass of dry air
    const float _r0 = _rStar / _mDry;           // gas constant for dry air
    const float _rho0 = _p0 / (_t0 * _r0);      // sea level density
    const float EARTH_RADIUS = 6356766.0f;      // effective radius of Earth (meters)

    const float _kMinusC = 273.15f;      // offset between Kelvin and Celsius scales
    const float _mbPerInch = 1013.25f / (760.0f / 25.4f);   // millibars per inch of mercury
    const float _pascalsPerMillibar = 100.0f;

    const float _mWater = 18.016f;      // molecular mass, water

    // The transition zone where the low altitude refraction model gradually takes over (radians).
    static float _upperTransition;
    static float _lowerTransition;
    static float _transitionWidth;     // difference between upper and lower

    // conditions at station where altimeter setting was observed
    float _stationHeight;       // meters (geometric)
    float _stationTemperature;  // Celsius
    float _altimeterSetting;    // pascals
    float _stationPressure;     // pascals
    float _humidity;            // percent relative or dew point (Celsius)
    bool _humidityIsRelative;   // true if humidity is relative
    float _pv;                  // water vapor partial pressure (pascals)
    float _stationDensity;      // kg per cubic meter

    // observer data (not necessarily identical to station data)
    float _observerHeight;          // geometric height (meters)
    float _observerTemperature;     // Celsius
    float _observerPressure;        // pascals
    float _observerDensity;         // kilograms per cubic meter
    float _m;           // Molecular mass of the air, including water vapor.
    double _Robs;       // distance from Earth center to observer (meters)
    double _h;          // height of equivalent homogenous atmosphere (meters)
    double _n;          // index of refraction


    static Atmosphere() {
        _upperTransition = (float)Angle.DegreesToRadians(_upperTransitionDeg);
        _lowerTransition = (float)Angle.DegreesToRadians(_lowerTransitionDeg);
        _transitionWidth = _upperTransition - _lowerTransition;
    }


    /// <summary>
    /// Create atmosphere conditions at observer. Valid only within the troposphere.
    /// </summary>
    /// <param name="observerHeight">observer geometric height above sea level (meters)</param>
    /// <param name="observerTemperature">temperature at observer (Celsius)</param>
    /// <param name="altimeterSetting">altimeter setting (millibars)</param>
    /// <param name="humidity">relative humidity (percent) or dew point (Celsius)</param>
    /// <param name="humidityIsRelative">true if <paramref name="humidity"/> is
    /// relative humidity</param>
    /// <param name="stationHeight">height of station where altimeter setting was measured
    /// (geometric height above sea level, meters)</param>
    /// <param name="stationTemperature">temperature at station where altimeter setting was measured
    /// (Celsius)</param>
    public Atmosphere(float observerHeight, float observerTemperature, float altimeterSetting,
        float humidity, bool humidityIsRelative, float stationHeight, float stationTemperature) {

        _observerHeight      = observerHeight;
        _observerTemperature = observerTemperature;
        _altimeterSetting    = altimeterSetting * _pascalsPerMillibar;
        _humidity            = humidity;
        _humidityIsRelative  = humidityIsRelative;
        _stationHeight       = stationHeight;
        _stationTemperature  = stationTemperature;

        // geopotential heights of observer and station
        float hObs = (float)GeometricToGeopotential(_observerHeight);
        float hSta = (float)GeometricToGeopotential(_stationHeight);

        // pressure at the weather station (pascals)
        _stationPressure = (float)Pressure(hSta, _altimeterSetting, _t0, _a0, _mDry);

        // water vapor partial pressure at the weather station (pascals)
        if (humidityIsRelative) {
            if (_humidity < 0.0f || _humidity > 100.0f)
                throw new ArgumentException("Relative humidity must be 0 to 100.");
            _pv = .01f * _humidity * (float)SaturationVaporPressure(_stationTemperature);
        } else {
            if (_humidity > _stationTemperature)
                throw new ArgumentException("Dew point must not exceed temperature.");
            _pv = (float)SaturationVaporPressure(_humidity);    // sat. pressure at dew point
        }

        // Water vapor partial pressure (pascals) equivalent to its mole fraction, at the station.
        // This is slightly different from _pv, due to failure to behave as a perfect gas.
        float pvEquiv = _pv * EnhancementFactor(_stationPressure, _stationTemperature);

        // dry air partial pressure at station (pascals)
        float pd = _stationPressure - pvEquiv;

        // Molecular mass of the air, including water vapor. This is assumed constant between the
        // weather station and observer.
        _m = (_mWater * pvEquiv + _mDry * pd) / _stationPressure;

        // gas constant of the air, including water vapor
        float rAir = _rStar / _m;

        // temperature rate between station and observer (Kelvins per geopotential meter)
        float a = (_observerTemperature - _stationTemperature) / (hObs - hSta);

        // air density at station, including water vapor
        _stationDensity = _stationPressure / (rAir * (_stationTemperature + _kMinusC));

        // conditions at the observer
        if (_observerHeight != _stationHeight) {
            _observerPressure = (float)Pressure(hObs - hSta, _stationPressure,
                _stationTemperature + _kMinusC, a, _m);
            _observerDensity = _observerPressure / (rAir * (_observerTemperature + _kMinusC));
        } else {
            _observerPressure = _stationPressure;
            _observerDensity  = _stationDensity;
        }

        // distance from Earth center to observer
        _Robs = _observerHeight + EARTH_RADIUS;

        // Compute equivalent height of homogenous atmosphere.
        _h = _observerPressure / (_g0 * _observerDensity);

        // Index of refraction for the assumed wavelength of light. Assume water vapor mole
        // fraction is the same at the station and observer.
        _n = IndexOfRefraction(_observerTemperature, _observerPressure, pvEquiv / _stationPressure,
            _co2Default, _lambdaDefault);
    }


    /// <summary>
    /// Create atmosphere conditions at observer. Valid only within the troposphere.
    /// </summary>
    /// <param name="observerHeight">observer geometric height above sea level (meters)</param>
    /// <param name="observerTemperature">temperature at observer (Celsius)</param>
    /// <param name="altimeterSetting">altimeter setting at observer (millibars)</param>
    /// <param name="humidity">relative humidity (percent) or dew point (Celsius)</param>
    /// <param name="humidityIsRelative">true if <paramref name="humidity"/> is
    /// relative humidity</param>
    public Atmosphere(float observerHeight, float observerTemperature, float altimeterSetting,
        float humidity, bool humidityIsRelative) :
    this(observerHeight, observerTemperature, altimeterSetting, humidity, humidityIsRelative,
        observerHeight, observerTemperature) { }


    /// <summary>
    /// Create atmosphere conditions at observer. Valid only within the troposphere.
    /// </summary>
    /// <param name="observerHeight">observer geometric height above sea level (meters)</param>
    /// <param name="observerTemperature">temperature at observer (Celsius)</param>
    /// <remarks>Standard altimeter setting and zero humidity are assumed.</remarks>
    public Atmosphere(float observerHeight, float observerTemperature) :
        this(observerHeight, observerTemperature, _p0 / _pascalsPerMillibar, 0, true) { }


    /// <summary>
    /// Create atmosphere conditions at observer. Valid only within the troposphere.
    /// </summary>
    /// <param name="observerHeight">observer geometric height above sea level (meters)</param>
    /// <remarks>1976 U.S. standard atmosphere is assumed.</remarks>
    public Atmosphere(float observerHeight) :
        this(observerHeight, _a0 * (float)GeometricToGeopotential(observerHeight) + _t0 - _kMinusC) { }



    /// <summary>
    /// Conversion factor: millibars per inch of mercury.
    /// </summary>
    public static float MillibarsPerInchHg { get { return _mbPerInch; } }

    /// <summary>
    /// Geometric height (meters) of the station where the altimeter setting was measured.
    /// </summary>
    public float StationHeight { get { return _stationHeight; } }

    /// <summary>
    /// Geometric height (meters) of the observer.
    /// </summary>
    public float ObserverHeight { get { return _observerHeight; } }

    /// <summary>
    /// Pressure (millibars) at the station where the altimeter setting was measured.
    /// </summary>
    public float StationPressure { get { return _stationPressure / _pascalsPerMillibar; } }

    /// <summary>
    /// Pressure (millibars) at the observer.
    /// </summary>
    public float ObserverPressure { get { return _observerPressure / _pascalsPerMillibar; } }

    /// <summary>
    /// Temperature (Celsius) at the station where the altimeter setting was measured.
    /// </summary>
    public float StationTemperature { get { return _stationTemperature; } }

    /// <summary>
    /// Temperature (Celsius) at the observer.
    /// </summary>
    public float ObserverTemperature { get { return _observerTemperature; } }

    /// <summary>
    /// Density (kg/m^3) at the station where the altimeter setting was measured.
    /// </summary>
    public float StationDensity { get { return _stationDensity; } }

    /// <summary>
    /// Density (kg/m^3) at the observer.
    /// </summary>
    public float ObserverDensity { get { return _observerDensity; } }

    /// <summary>
    /// Rho: actual air density divided by standard sea level density, at the station where the
    /// altimeter setting was measured.
    /// </summary>
    public float StationRho { get { return _stationDensity / _rho0; } }

    /// <summary>
    /// Rho: actual air density divided by standard sea level density, at the observer.
    /// </summary>
    public float ObserverRho { get { return _observerDensity / _rho0; } }

    /// <summary>
    /// Get altimeter setting (millibars).
    /// </summary>
    public float AltimeterSetting { get { return _altimeterSetting / _pascalsPerMillibar; } }

    /// <summary>
    /// Get relative humidity or dew point.
    /// </summary>
    /// <remarks>The HumidityIsRelative property indicates whether the returned value is
    /// relative humidity or dew point.</remarks>
    public float Humidity { get { return _humidity; } }

    /// <summary>
    /// Get a flag which indicates whether the Humidity property returns relative
    /// humidity or dew point.
    /// </summary>
    public bool HumidityIsRelative { get { return _humidityIsRelative; } }

    /// <summary>
    /// Get index of refraction of the air.
    /// </summary>
    public double IndexOfRefraction() { return _n; }

    /// <summary>
    /// Get water vapor partial pressure (millibars).
    /// </summary>
    /// <remarks>Does not include the enhancement factor correction. I.e., the water vapor mole
    /// fraction is slightly greater than implied by its partial pressure due to failure to behave
    /// as a perfect gas.</remarks>
    public double WaterVaporPartialPressure { get { return _pv / _pascalsPerMillibar; } }


    /// <summary>
    /// Get the wavelength of light (nanometers) used for index of refraction computation.
    /// </summary>
    public float Wavelength {
        // convert from microns to nm
        get { return 1000f * (float)_lambdaDefault; }
    }


    /// <summary>
    /// Get molecular mass of the air, including water vapor.
    /// </summary>
    public float MolecularMass { get { return _m; } }



    /// <summary>
    /// Saturation vapor pressure of water at a given temperature.
    /// </summary>
    /// <param name="celsius">temperature (Celsius)</param>
    /// <returns>saturation pressure (pascals)</returns>
    /// <remarks>IAPWS formulas, Peter H. Huang, via Stone and Zimmerman, NIST Web site.
    /// </remarks>
    public static double SaturationVaporPressure(double celsius) {
        // Equations recommended by the International Association for the Properties of
        // Water and Steam (IAPWS) described by Huang, via Stone and Zimmerman on the
        // NIST Web page, where they say it's "generally accepted as the best formula
        // available."
        const double k1 = 1.16705214528E+03;
        const double k2 = -7.24213167032E+05;
        const double k3 = -1.70738469401E+01;
        const double k4 = 1.20208247025E+04;
        const double k5 = -3.23255503223E+06;
        const double k6 = 1.49151086135E+01;
        const double k7 = -4.82326573616E+03;
        const double k8 = 4.05113405421E+05;
        const double k9 = -2.38555575678E-01;
        const double k10 = 6.50175348448E+02;

        const double a1 = -13.928169;
        const double a2 = 34.7078238;

        double t = celsius + 273.15;

        if (celsius >= 0.0) {
            double omega = t + k9 / (t - k10);
            double omegaSquared = omega * omega;
            double a = omegaSquared + k1 * omega + k2;
            double b = k3 * omegaSquared + k4 * omega + k5;
            double c = k6 * omegaSquared + k7 * omega + k8;
            double x = -b + Math.Sqrt(b * b - 4.0 * a * c);
            double twoCOverX = 2 * c / x;
            return 1e6 * twoCOverX * twoCOverX * twoCOverX * twoCOverX;
        }
        else {
            double theta = t / 273.16;
            double y = a1 * (1.0 - Math.Pow(theta, -1.5))
                + a2 * (1.0 - Math.Pow(theta, -1.25));
            return 611.657 * Math.Exp(y);
        }
    }

    /// <summary>
    /// Get the enhancement factor, by which the water vapor mole fraction does not
    /// exactly equal its partial pressure fraction.
    /// </summary>
    /// <param name="pascals">air pressure (pascals)</param>
    /// <param name="celsius">air temperature (Celsius)</param>
    /// <returns>water vapor mole fraction divided by its partial pressure fraction
    /// </returns>
    /// <remarks>This value differs slightly from unity (about 1.004 at room temperature)
    /// due to imperfect behavior of the gases.</remarks>
    public static float EnhancementFactor(float pascals, float celsius) {
        // From Jack A. Stone and Jay H. Zimmerman, at the NIST Web site, op cit.
        const float alpha = 1.00062f;
        const float beta = 3.14e-8f;
        const float gamma = 5.60e-7f;

        return alpha + beta * pascals + gamma * celsius * celsius;
    }

    /// <summary>
    /// Index of refraction, given pressure, temperature, water vapor mole fraction,
    /// carbon dioxide concentration, and wavelength.
    /// </summary>
    /// <param name="t">air temperature (Celsius)</param>
    /// <param name="p">air pressure (pascals)</param>
    /// <param name="xv">water vapor mole fraction</param>
    /// <param name="co2ppm">carbon dioxide concentration (ppm)</param>
    /// <param name="lambda">wavelength of light (microns)</param>
    /// <returns>index of refraction</returns>
    /// <remarks>Implements the Ciddor model. See Stone and Zimmerman at the NIST Web page.
    /// </remarks>
    public static double IndexOfRefraction(double t, double p,
        double xv, double co2ppm, double lambda) {

        const double w0 = 295.235;
        const double w1 = 2.6422;
        const double w2 = -0.03238;
        const double w3 = 0.004028;
        const double k0 = 238.0185;
        const double k1 = 5792105.0;
        const double k2 = 57.362;
        const double k3 = 167917.0;
        const double a0 = 1.58123e-6;
        const double a1 = -2.9331e-8;
        const double a2 = 1.1043e-10;
        const double b0 = 5.707e-6;
        const double b1 = -2.051e-8;
        const double c0 = 1.9898e-4;
        const double c1 = -2.376e-6;
        const double d = 1.83e-11;
        const double e = -0.765e-8;
        const double pR1 = 101325.0;
        const double TR1 = 288.15;
        const double Za = 0.9995922115;
        const double Rhovs = 0.00985938;
        const double R = 8.314472;
        const double Mv = 0.018015;

        double S = 1.0 / (lambda * lambda);
        double ras = 1e-8 * ((k1 / (k0 - S)) + (k3 / (k2 - S)));
        double rvs = 1.022e-8 * (w0 + (w1 + (w2 + w3 * S) * S) * S);
        double Ma = 0.0289635 + 1.2011e-8 * (co2ppm - 400.0);
        double raxs = ras * (1.0 + 5.34e-7 * (co2ppm - 450));

        double T = t + 273.15;
        double xv2 = xv * xv;
        double pOverT = p / T;
        double Zm = 1.0 - pOverT * (a0 + (a1 + a2 * t) * t + (b0 + b1 * t) * xv
            + (c0 + c1 * t) * xv2) + pOverT * pOverT * (d + e * xv2);
        double Rhoaxs = pR1 * Ma / (Za * R * TR1);
        double Rhov = xv * p * Mv / (Zm * R * T);
        double Rhoa = (1.0 - xv) * p * Ma / (Zm * R * T);

        return 1.0 + (Rhoa / Rhoaxs) * raxs + (Rhov / Rhovs) * rvs;
    }

    /// <summary>
    /// Calculate refraction with Cassini's formula.
    /// </summary>
    /// <param name="refAltitude">refracted altitude (radians)</param>
    /// <returns>deflection of the ray due to refraction (radians)</returns>
    /// <remarks>as described by Andrew T. Young at his Web site</remarks>
    public double CassiniRefraction(double refAltitude) {
        // Cassini's model, as described by Andrew T. Young at his Web site.
        double sinZ1 = _Robs / (_Robs + _h)
            * Math.Sin(SofaJpl.Angle.Complement(refAltitude));
        return Math.Asin(_n * sinZ1) - Math.Asin(sinZ1);
    }

    /// <summary>
    /// Calculate refraction with the Bennett formula.
    /// </summary>
    /// <param name="refAlt">refracted altitude (radians)</param>
    /// <returns>deflection of the ray due to refraction (radians)</returns>
    /// <remarks>Formula via Meeus 1991. Includes the additional correction.
    /// </remarks>
    public double BennettRefraction(double refAlt) {
        double altDeg = Angle.RadiansToDegrees(refAlt);
        double deg = altDeg + 7.31 / (altDeg + 4.4);
        double refMin = .28 * this.ObserverPressure / (this.ObserverTemperature + _kMinusC)
            / Math.Tan(Angle.DegreesToRadians(deg));
        refMin += -.06 * Math.Sin(Angle.DegreesToRadians(14.7 * refMin + 13));
        return Angle.DmsToRad(0, refMin, 0);
    }


    /// <summary>
    /// Calculate refraction with a second degree low altitude formula, applicable for
    /// altitudes less than 15 degrees.
    /// </summary>
    /// <param name="refAlt">refracted altitude (radians)</param>
    /// <returns>refraction (radians)</returns>
    /// <remarks>Seidelmann, Explanatory Supplement, 1992, equation 3.283-2. To my eye, output
    /// looks reasonable down to refracted altitude -2 degrees. Below that I wouldn't trust it.
    /// </remarks>
    public double LowAltRefraction(double refAlt) {
        double a = Angle.RadiansToDegrees(refAlt);
        return Angle.DegreesToRadians(
            this.ObserverPressure / (_kMinusC + this.ObserverTemperature)
            * (.1594 + (.0196 + .00002 * a) * a) / (1.0 + (.505 + .0845 * a) * a));
    }

    /// <summary>
    /// Get a combined result of the Cassini and low altitude refraction models.
    /// </summary>
    /// <param name="refAlt">refracted altitude (radians)</param>
    /// <returns>refraction (radians)</returns>
    /// <remarks>Fades out the Cassini and fades in the low altitude model over a
    /// transition zone of a few degrees. Not reliable below -2 degrees refracted altitude.
    /// </remarks>
    public double CompositeRefraction(double refAlt) {
        if (refAlt > _upperTransition)
            return CassiniRefraction(refAlt);
        if (refAlt < _lowerTransition)
            return LowAltRefraction(refAlt);

        // interpolating factor
        double p = (refAlt - _lowerTransition) / _transitionWidth;

        // Return a weighted mean of both models.
        double low = LowAltRefraction(refAlt);
        double cassini = CassiniRefraction(refAlt);
        return (1.0 - p) * low + p * cassini;
    }


    /// <summary>
    /// Apply refraction of this Atmosphere to the unrefracted altitude.
    /// </summary>
    /// <param name="unrefAltitude">unrefracted altitude (radians)</param>
    /// <param name="accuracy">desired accuracy (radians)</param>
    /// <returns>refracted altitude (radians)</returns>
    /// <remarks>Solution is iterative, so it is necessary to specify the desired
    /// accuracy. Not reliable at altitude below -2 degrees.</remarks>
    public double Refract(double unrefAltitude, double accuracy) {
        double oldRefAlt, newRefAlt;
        int i = 0;      // iteration counter

        // Begin by assuming the refracted altitude is the same as unrefracted.
        newRefAlt = unrefAltitude;

        // Then improve that approximation.
        do {
            if (++i > _maxIterations)
                throw new Exception(
                    "Refraction solution needed too many iterations to converge.");
            oldRefAlt = newRefAlt;
            newRefAlt = unrefAltitude + CompositeRefraction(oldRefAlt);
        } while (Math.Abs(oldRefAlt - newRefAlt) > accuracy);

        return newRefAlt;
    }


    /// <summary>
    /// Apply refraction of this Atmosphere to the phi (altitude) angle of a Spherical object.
    /// The original object does not change.
    /// </summary>
    /// <param name="sph">object to receive refraction</param>
    /// <param name="accuracy">desired accuracy (radians)</param>
    /// <returns>new Spherical object, with refraction</returns>
    /// <remarks>Solution is iterative, so it is necessary to specify the desired
    /// accuracy. Not reliable at altitude below -2 degrees.</remarks>
    public Spherical Refract(Spherical sph, double accuracy) {
        return new Spherical(sph.Theta, this.Refract(sph.Alt, accuracy), sph.Radius);
    }


    /// <summary>
    /// Apply refraction of this Atmosphere to compute unrefracted altitude from refracted altitude.
    /// </summary>
    /// <param name="refractedAltitude">refracted altitude (radians)</param>
    /// <returns>unrefracted altitude (radians)</returns>
    /// <remarks>Not reliable at altitude below -2 degrees.</remarks>
    public double Unrefract(double refractedAltitude) {
        return refractedAltitude - CompositeRefraction(refractedAltitude);
    }


    /// <summary>
    /// Apply this Atmosphere to remove refraction from the phi (altitude) angle of a Spherical
    /// object. The original object does not change.
    /// </summary>
    /// <param name="sph">refracted object</param>
    /// <returns>new Spherical unrefracted object</returns>
    /// <remarks>Not reliable at altitude below -2 degrees.</remarks>
    public Spherical Unrefract(Spherical sph) {
        return new Spherical(sph.Theta, this.Unrefract(sph.Alt), sph.Radius);
    }


    // no longer required
#if false
    /// <summary>
    /// Compute unrefracted altitude from refracted altitude.
    /// </summary>
    /// <param name="refractedAltitude">refracted altitude (radians)</param>
    /// <param name="accuracy">(ignored)</param>
    /// <returns>unrefracted altitude (radians)</returns>
    /// <remarks>The purpose of this method is to hide the corresponding method in the
    /// base class.</remarks>
    public double Unrefract(double refractedAltitude, double accuracy) {
        return Unrefract(refractedAltitude);
    }
#endif


    /// <summary>
    /// Return refracted semidiameter of a body at a given position angle.
    /// </summary>
    /// <param name="unrefSD">unrefracted semidiameter (radians)</param>
    /// <param name="unrefEl">unrefracted elevation of body center (radians)</param>
    /// <param name="positionAngle">position angle on limb (0 is toward zenith) (radians)
    /// </param>
    /// <param name="accuracy">desired accuracy (radians)</param>
    /// <returns>refracted position angle (radians)</returns>
    public double RefractedSemidiameter(double unrefSD, double unrefEl,
        double positionAngle, double accuracy) {

        // vectors to body center and the desired point on limb, centered at nadir
        SofaJpl.Vector centerVec = new HirosePS.SofaJpl.Vector(0, 0, -1);
        SofaJpl.Vector limbVec = new HirosePS.SofaJpl.Vector(
            new SofaJpl.Spherical(positionAngle, unrefSD - SofaJpl.Angle.HalfPi));

        // Rotate both vectors to their relative positions in the horizontal system.
        SofaJpl.RMatrix rmat = new HirosePS.SofaJpl.RMatrix()
            .YRotate(SofaJpl.Angle.HalfPi + unrefEl);
        SofaJpl.Spherical centerSph = new HirosePS.SofaJpl.Spherical(rmat * centerVec);
        SofaJpl.Spherical limbSph = new HirosePS.SofaJpl.Spherical(rmat * limbVec);

        // Apply refraction.
        centerSph = new HirosePS.SofaJpl.Spherical(centerSph.Theta,
            this.Refract(centerSph.Phi, accuracy));
        limbSph = new HirosePS.SofaJpl.Spherical(limbSph.Theta,
            this.Refract(limbSph.Phi, accuracy));

        // Return separation angle between refracted spherical coordinates.
        return centerSph.SeparationAngle(limbSph);
    }



    // static methods



    /// <summary>
    /// Compute atmosphere pressure at a given height above the base of a gradient layer or
    /// isothermal layer.
    /// </summary>
    /// <param name="h">geopotential height above base of layer (meters)</param>
    /// <param name="p_base">pressure at base of layer (any unit)</param>
    /// <param name="t_base">temperature at base of layer (Kelvins)</param>
    /// <param name="a">temperature rate within layer (K/m, negative if decrease with height)
    /// </param>
    /// <param name="mm">molecular mass of the air</param>
    /// <returns>pressure (same units as <paramref name="p_base"/>)</returns>
    public static double Pressure(double h, double p_base, double t_base, double a, double mm) {
        double R = _rStar / mm;     // gas constant of the atmosphere
        if (a == 0.0)	/* isothermal layer */
            return p_base * Math.Exp(-(_g0 / (R * t_base)) * h);
        else			/* gradient layer */
            return p_base * Math.Pow(1.0 + (h * a) / t_base, -_g0 / (a * R));
    }


    /// <summary>
    /// Compute atmosphere density at a given height above the base of a gradient layer or
    /// isothermal layer.
    /// </summary>
    /// <param name="h">geopotential height above base of layer (meters)</param>
    /// <param name="rho_base">density at base of layer (any units)</param>
    /// <param name="t_base">temperature at base of layer (Kelvins)</param>
    /// <param name="a">temperature rate within layer (K/m, negative if decrease with height)
    /// </param>
    /// <param name="mm">molecular mass of the air</param>
    /// <returns>density (same units as <paramref name="rho_base"/>)</returns>
    public static double Density(double h, double rho_base, double t_base, double a, double mm) {
        double R = _rStar / mm;     // gas constant of the atmosphere
        if (a == 0.0)		/* isothermal layer */
            return rho_base * Math.Exp(-_g0 / (R * t_base) * h);
        else				/* gradient layer */
            return rho_base * Math.Pow(1.0 + (h * a) / t_base, -(_g0 / (a * R) + 1.0));
    }


    /// <summary>
    /// Compute geopotential height above the base of a gradient layer or isothermal layer.
    /// </summary>
    /// <param name="p">pressure (Pascals)</param>
    /// <param name="p_base">pressure at base of layer (Pascals)</param>
    /// <param name="t_base">temperature at base of layer (Kelvins)</param>
    /// <param name="a">temperature rate within layer (K/m, negative if decrease with height)
    /// </param>
    /// <param name="mm">molecular mass of the air</param>
    /// <returns>geopotential height above base of layer (meters)</returns>
    /// <remarks>The two pressure parameters may actually use any unit, but it must be the same for
    /// both.</remarks>
    public static double GeopotientialHeight(double p, double p_base, double t_base, double a,
        double mm) {

        double R = _rStar / mm;     // gas constant of the atmosphere
        if (a == 0.0)	/* isothermal layer */
            // return p_base * Math.Exp(-(_g0 / (R * t_base)) * h);
            return -R * t_base / _g0 * Math.Log(p / p_base);
        else			/* gradient layer */
            // return p_base * Math.Pow(1.0 + (h * a) / t_base, -_g0 / (a * R));
            return (Math.Pow(p / p_base, -a * R / _g0) - 1.0) * t_base / a;
    }


    /// <summary>
    /// convert geometric height to geopotential height
    /// </summary>
    /// <param name="z">geometric height (m)</param>
    /// <returns>geopotential height (m)</returns>
    public static double GeometricToGeopotential(double z) {
        return EARTH_RADIUS * z / (EARTH_RADIUS + z);
    }


    /// <summary>
    /// Convert from geopotential height to geometric height.
    /// </summary>
    /// <param name="h">geopotential height (m)</param>
    /// <returns>geometric height (m)</returns>
    public static double GeopotentialToGeometric(double h) {
        return EARTH_RADIUS * h / (EARTH_RADIUS - h);
    }



    // test code
#if DEBUG


    /// <summary>
    /// Compare the Cassini and low altitude refraction algorithms.
    /// </summary>
    /// <param name="refAlt">refracted altitude (radians)</param>
    /// <returns>Cassini minus low altitude formula (radians)</returns>
    public float CassiniMinusLow(float refAlt) {

        return (float)(CassiniRefraction(refAlt) - LowAltRefraction(refAlt));
    }


    /// <summary>
    /// Self test the class. Runs silently if OK.
    /// </summary>
    public static void Test() {
        Atmosphere atm;
        Spherical sph1, sph2;
        double accuracy, tolerance, altitude, refraction, correction;
        float p, t;
        const float kAt0C = 273.15f;

        // test properties with standard atmosphere, observer 1000 meters altitude

        atm = new Atmosphere((float)GeopotentialToGeometric(1000), 15 - 6.5f, 1013.25f, 0, true, 0, 15);

        equalsApprox(atm.StationPressure, 1013.25f, .01f);
        equalsApprox(atm.ObserverPressure, 8.98746e+2f, .001);

        equalsApprox(atm.StationTemperature, 15f, 0);
        equalsApprox(atm.ObserverTemperature, 15 - 6.5f, 0);

        equalsApprox(atm.StationDensity, 1.22500, 1e-5);
        equalsApprox(atm.ObserverDensity, 1.11164, 1e-4);

        equalsApprox(atm.StationRho, 1.0, 1e-5);
        equalsApprox(atm.ObserverRho, 9.07463e-1, 1e-4);

        equalsApprox(atm.Humidity, 0, 0);
        if (atm.HumidityIsRelative != true)
            throw new Exception("HumidityIsRelative failed.");

        equalsApprox(atm.WaterVaporPartialPressure, 0, 1);
        equalsApprox(atm.Wavelength, 550, 1);
        equalsApprox(atm.AltimeterSetting, 1013.25, .01);
        equalsApprox(atm.IndexOfRefraction(), 1.000252125, 1e-6);
        equalsApprox(atm.MolecularMass, _mDry, 0f);


        // Same test, with constructor that takes height, temperature at observer, altimeter
        // setting at observer, and humidity.

        atm = new Atmosphere((float)GeopotentialToGeometric(1000), 15 - 6.5f, 1013.25f, 0, true);

        equalsApprox(atm.StationPressure, 8.98746e+2f, .01f);
        equalsApprox(atm.ObserverPressure, 8.98746e+2f, .001);

        equalsApprox(atm.StationTemperature, 15 - 6.5f, 0);
        equalsApprox(atm.ObserverTemperature, 15 - 6.5f, 0);

        equalsApprox(atm.StationDensity, 1.11164, 1e-5);
        equalsApprox(atm.ObserverDensity, 1.11164, 1e-4);

        equalsApprox(atm.StationRho, 9.07463e-1, 1e-5);
        equalsApprox(atm.ObserverRho, 9.07463e-1, 1e-4);

        equalsApprox(atm.Humidity, 0, 0);
        if (atm.HumidityIsRelative != true)
            throw new Exception("HumidityIsRelative failed.");

        equalsApprox(atm.WaterVaporPartialPressure, 0, 1);
        equalsApprox(atm.Wavelength, 550, 1);
        equalsApprox(atm.AltimeterSetting, 1013.25, .01);
        equalsApprox(atm.IndexOfRefraction(), 1.000252125, 1e-6);


        // Same test, with constructor that takes height and temperature at observer.

        atm = new Atmosphere((float)GeopotentialToGeometric(1000), 15 - 6.5f);

        equalsApprox(atm.StationPressure, 8.98746e+2f, .01f);
        equalsApprox(atm.ObserverPressure, 8.98746e+2f, .001);

        equalsApprox(atm.StationTemperature, 15 - 6.5f, 0);
        equalsApprox(atm.ObserverTemperature, 15 - 6.5f, 0);

        equalsApprox(atm.StationDensity, 1.11164, 1e-5);
        equalsApprox(atm.ObserverDensity, 1.11164, 1e-4);

        equalsApprox(atm.StationRho, 9.07463e-1, 1e-5);
        equalsApprox(atm.ObserverRho, 9.07463e-1, 1e-4);

        equalsApprox(atm.Humidity, 0, 0);
        if (atm.HumidityIsRelative != true)
            throw new Exception("HumidityIsRelative failed.");

        equalsApprox(atm.WaterVaporPartialPressure, 0, 1);
        equalsApprox(atm.Wavelength, 550, 1);
        equalsApprox(atm.AltimeterSetting, 1013.25, .01);
        equalsApprox(atm.IndexOfRefraction(), 1.000252125, 1e-6);


        // Same test, with constructor that takes height only.

        atm = new Atmosphere((float)GeopotentialToGeometric(1000));

        equalsApprox(atm.StationPressure, 8.98746e+2f, .01f);
        equalsApprox(atm.ObserverPressure, 8.98746e+2f, .001);

        equalsApprox(atm.StationTemperature, 15 - 6.5f, 0);
        equalsApprox(atm.ObserverTemperature, 15 - 6.5f, 0);

        equalsApprox(atm.StationDensity, 1.11164, 1e-5);
        equalsApprox(atm.ObserverDensity, 1.11164, 1e-4);

        equalsApprox(atm.StationRho, 9.07463e-1, 1e-5);
        equalsApprox(atm.ObserverRho, 9.07463e-1, 1e-4);

        equalsApprox(atm.Humidity, 0, 0);
        if (atm.HumidityIsRelative != true)
            throw new Exception("HumidityIsRelative failed.");

        equalsApprox(atm.WaterVaporPartialPressure, 0, 1);
        equalsApprox(atm.Wavelength, 550, 1);
        equalsApprox(atm.AltimeterSetting, 1013.25, .01);
        equalsApprox(atm.IndexOfRefraction(), 1.000252125, 1e-6);


        // pressure, temperature, altitude methods, in the troposphere

        equalsApprox(Pressure(2000, 101325, _t0, -.0065, _mDry), 7.94952e+4, 1.0);
        equalsApprox(Density(2000, _rho0, _t0, -.0065, _mDry), 1.00649, 2e-5);

        equalsApprox(Pressure(-500, 101325, _t0, -.0065, _mDry), 1.07478e+5, 1.0);
        equalsApprox(Density(-500, _rho0, _t0, -.0065, _mDry), 1.28490, 2e-5);

        equalsApprox(GeopotientialHeight(7.94952e+4, 101325, _t0, -.0065, _mDry), 2000, 1);
        equalsApprox(GeopotientialHeight(1.07478e+5, 101325, _t0, _a0, _mDry), -500, 1);

        // pressure, temperature, altitude methods, in the stratosphere

        equalsApprox(Pressure(5000.0, 2.26321e+4, -56.5 + _kMinusC, 0, _mDry), 1.02875e+4, 1);
        equalsApprox(GeopotientialHeight(1.02875e+4, 2.26321e+4, -56.5 + _kMinusC, 0, _mDry),
            5000, 1);
        equalsApprox(Density(5000, 3.63918e-1, -56.5 + _kMinusC, 0, _mDry), 1.65420e-1, 1e-5);


        // Refraction tests.
        atm = new Atmosphere(0.0f, 10.0f, 1010f, 0.0f, true);
        tolerance = HirosePS.SofaJpl.Angle.DmsToRad(0, .1, 0);

        // Test Bennett refraction at sea level, 10 C, 1010 mb, against Almanac table.
        altitude = Angle.DmsToRad(10.0, 8.0, 0);
        refraction = atm.BennettRefraction(altitude);
        equalsApprox(refraction, HirosePS.SofaJpl.Angle.DmsToRad(0, 5.25, 0), tolerance);
        altitude = Angle.DmsToRad(42, 44, 0);
        equalsApprox(atm.BennettRefraction(altitude),
            Angle.DmsToRad(0, 1.05, 0), tolerance);

        // Compare Cassini and Bennett.
#if false
        // This should fail.
        altitude = Angle.DegreesToRadians(10);
        equalsApprox(atm.cassiniRefraction(altitude),
            atm.bennettRefraction(0), tolerance);
#endif
        // This should pass.
        for (int i = 10; i <= 90; ++i) {
            altitude = Angle.DegreesToRadians(i);
            equalsApprox(atm.CassiniRefraction(altitude),
                atm.BennettRefraction(altitude), tolerance);
        }

        // Compare low altitude method and Bennett in standard sea level conditions and
        // at low density too.
        atm = new Atmosphere(0f, 10f, 1010f, 50f, true);
        for (int i = 0; i <= 15; i += 5) {
            altitude = Angle.DegreesToRadians(i);
            refraction = atm.BennettRefraction(altitude);
            equalsApprox(atm.LowAltRefraction(altitude), refraction, .05 * refraction);
        }
        atm = new Atmosphere(2000f, 30f, 1000f, 50f, true);
        for (int i = 0; i <= 15; i += 5) {
            altitude = Angle.DegreesToRadians(i);
            refraction = atm.BennettRefraction(altitude);
            equalsApprox(atm.LowAltRefraction(altitude), refraction, .05 * refraction);
        }

        // Compare Refract() and Unrefract() to Nautical Almanac tables in standard
        // conditions.
        atm = new Atmosphere(0, 10, 1010, 50, true);
        accuracy = SofaJpl.Angle.DmsToRad(0, .02, 0);
        tolerance = accuracy * 2;

        roundTripTest(atm, SofaJpl.Angle.DmsToRad(90, 0, 0), 0.0, accuracy, tolerance);
        roundTripTest(atm, SofaJpl.Angle.DmsToRad(42, 44, 0),
            Angle.DmsToRad(0, 1.05, 0), accuracy, tolerance);
        roundTripTest(atm, SofaJpl.Angle.DmsToRad(25, 14, 0),
            Angle.DmsToRad(0, 2.05, 0), accuracy, tolerance);
        roundTripTest(atm, SofaJpl.Angle.DmsToRad(17, 28, 0),
            Angle.DmsToRad(0, 3.05, 0), accuracy, tolerance);
        roundTripTest(atm, SofaJpl.Angle.DmsToRad(13, 13, 0),
            Angle.DmsToRad(0, 4.05, 0), accuracy, tolerance);
        roundTripTest(atm, SofaJpl.Angle.DmsToRad(10, 33, 0),
            Angle.DmsToRad(0, 5.05, 0), accuracy, tolerance);

        // Error tolerance at low altitude is 2% of the correct refraction.
        refraction = Angle.DmsToRad(0, 6.6, 0);
        roundTripTest(atm, SofaJpl.Angle.DmsToRad(8, 0, 0), refraction, accuracy,
            .02 * refraction);
        refraction = Angle.DmsToRad(0, 8.5, 0);
        roundTripTest(atm, SofaJpl.Angle.DmsToRad(6, 0, 0), refraction, accuracy,
            .02 * refraction);
        refraction = Angle.DmsToRad(0, 11.8, 0);
        roundTripTest(atm, SofaJpl.Angle.DmsToRad(4, 0, 0), refraction, accuracy,
            .02 * refraction);

        // Compare Refract() and Unrefract() to Nautical Almanac table in greater than
        // standard air density. Instead of the "additional corrections" Almanac table,
        // I compute the correction factor for nonstandard conditions here.
        t = -7.5f;
        p = 1050;
        correction = (p / 1010) * ((10 + kAt0C) / (t + kAt0C));
        atm = new Atmosphere(0, t, p, 50, true);
        accuracy = SofaJpl.Angle.DmsToRad(0, .02, 0);
        tolerance = SofaJpl.Angle.DmsToRad(0, .1, 0);

        roundTripTest(atm, SofaJpl.Angle.DmsToRad(90, 0, 0), correction * 0.0,
            accuracy, tolerance);
        roundTripTest(atm, SofaJpl.Angle.DmsToRad(42, 44, 0),
            correction * Angle.DmsToRad(0, 1.05, 0), accuracy, tolerance);
        roundTripTest(atm, SofaJpl.Angle.DmsToRad(25, 14, 0),
            correction * Angle.DmsToRad(0, 2.05, 0), accuracy, tolerance);
        roundTripTest(atm, SofaJpl.Angle.DmsToRad(17, 28, 0),
            correction * Angle.DmsToRad(0, 3.05, 0), accuracy, tolerance);
        roundTripTest(atm, SofaJpl.Angle.DmsToRad(13, 13, 0),
            correction * Angle.DmsToRad(0, 4.05, 0), accuracy, tolerance);
        roundTripTest(atm, SofaJpl.Angle.DmsToRad(10, 33, 0),
            correction * Angle.DmsToRad(0, 5.05, 0), accuracy, tolerance);

        // Error tolerance at low altitude is 2% of the correct refraction.
        refraction = correction * Angle.DmsToRad(0, 6.6, 0);
        roundTripTest(atm, SofaJpl.Angle.DmsToRad(8, 0, 0), refraction, accuracy,
            .02 * refraction);
        refraction = correction * Angle.DmsToRad(0, 8.5, 0);
        roundTripTest(atm, SofaJpl.Angle.DmsToRad(6, 0, 0), refraction, accuracy,
            .02 * refraction);
        refraction = correction * Angle.DmsToRad(0, 11.8, 0);
        roundTripTest(atm, SofaJpl.Angle.DmsToRad(4, 0, 0), refraction, accuracy,
            .02 * refraction);

        // Test the method that returns the weighted mean of Cassini and low altitude
        // refraction. Sea level standard conditions.
        atm = new Atmosphere(0, 15, 1013.25f, 0, true);
        altitude = Angle.DegreesToRadians(15.01);
        equalsApprox(atm.CompositeRefraction(altitude), atm.CassiniRefraction(altitude),
            0.0);
        altitude = Angle.DegreesToRadians(9.99);
        equalsApprox(atm.CompositeRefraction(altitude), atm.LowAltRefraction(altitude),
            0.0);
        altitude = Angle.DegreesToRadians(13);
        equalsApprox(atm.CompositeRefraction(altitude),
            .4 * atm.LowAltRefraction(altitude) +
            .6 * atm.CassiniRefraction(altitude), 5e-7);

        // Test Refract() and Unrefract().
        altitude = Angle.DegreesToRadians(10);   // refracted
        equalsApprox(atm.Refract(atm.Unrefract(altitude), 5e-7), altitude, 5e-7);
        refraction = altitude - atm.Unrefract(altitude);
        equalsApprox(refraction, atm.CompositeRefraction(altitude), 5e-7);

        // Refract() and Unrefract() round trip, both ways.
        atm = new Atmosphere(
            500, 35, (float)(29.5 * SofaJpl.Atmosphere.MillibarsPerInchHg), 50, true);
        altitude = Angle.DegreesToRadians(16);
        accuracy = Angle.DegreesToRadians(.00005);
        tolerance = accuracy * 2;
        equalsApprox(atm.Refract(atm.Unrefract(altitude), accuracy), altitude, tolerance);
        equalsApprox(atm.Unrefract(atm.Refract(altitude, accuracy)), altitude, tolerance);

        // Refract() and Unrefract() overloads that take a Spherical parameter should exactly
        // match the overloads that take just an angle.
        altitude = Angle.DegreesToRadians(10);      // assume this is unrefracted
        sph1 = new Spherical(0, altitude);
        sph2 = atm.Refract(sph1, accuracy);
        equalsApprox(sph2.Alt, atm.Refract(sph1.Alt, accuracy), 0);
        sph1 = atm.Unrefract(sph2);
        equalsApprox(sph1.Alt, atm.Unrefract(sph2.Alt), 0);

        // Water vapor saturation pressure. Correct values from Stone & Zimmerman on
        // the NIST Web page op cit.
        equalsApprox(SaturationVaporPressure(-10), 260, 2.6);
        equalsApprox(SaturationVaporPressure(20), 2339, 23);
        equalsApprox(SaturationVaporPressure(40), 7384, 74);

        // index of refraction method vs. table by Stone & Zimmerman at NIST Web page.
        p = 101325.0f;          // air pressure (palcals)
        t = 20;                 // air temperature (Celsius)
        double lambda = .633;   // wavelength (microns)
        double xv = 0;          // water vapor mole fraction
        double co2 = 450;       // carbon dioxide concentration (ppm)

        equalsApprox(IndexOfRefraction(t, p, xv, co2, lambda), 1.000271800, 1e-9);
        p = 60000;
        equalsApprox(IndexOfRefraction(t, p, xv, co2, lambda), 1.000160924, 1e-9);
        p = 120000;
        equalsApprox(IndexOfRefraction(t, p, xv, co2, lambda), 1.000321916, 1e-9);
        t = 50;
        p = 100000;
        equalsApprox(IndexOfRefraction(t, p, xv, co2, lambda), 1.000243285, 1e-9);
        t = 5; p = 100000;
        equalsApprox(IndexOfRefraction(t, p, xv, co2, lambda), 1.000282756, 1e-9);
        t = -40; p = 100000;
        equalsApprox(IndexOfRefraction(t, p, xv, co2, lambda), 1.000337580, 1e-9);

        t = 50; p = 120000;
        xv = EnhancementFactor(p, t) * SaturationVaporPressure(t) / p;
        equalsApprox(IndexOfRefraction(t, p, xv, co2, lambda), 1.000287924, 1e-9);
        t = 40; p = 120000;
        xv = .75 * EnhancementFactor(p, t) * SaturationVaporPressure(t) / p;
        equalsApprox(IndexOfRefraction(t, p, xv, co2, lambda), 1.000299418, 1e-9);
        t = 20; p = 100000;
        xv = EnhancementFactor(p, t) * SaturationVaporPressure(t) / p;
        equalsApprox(IndexOfRefraction(t, p, xv, co2, lambda), 1.000267394, 1e-9);

        t = 40; p = 110000; lambda = 1.7;
        xv = EnhancementFactor(p, t) * SaturationVaporPressure(t) / p;
        equalsApprox(IndexOfRefraction(t, p, xv, co2, lambda), 1.000270247, 1e-9);
        t = 20; p = 101325; xv = 0;
        equalsApprox(IndexOfRefraction(t, p, xv, co2, lambda), 1.000268479, 1e-9);

        t = 40; p = 110000; lambda = .3;
        xv = EnhancementFactor(p, t) * SaturationVaporPressure(t) / p;
        equalsApprox(IndexOfRefraction(t, p, xv, co2, lambda), 1.000289000, 1e-9);
        t = 20; p = 101325; xv = 0;
        equalsApprox(IndexOfRefraction(t, p, xv, co2, lambda), 1.000286581, 1e-9);
        t = -40; p = 120000;
        equalsApprox(IndexOfRefraction(t, p, xv, co2, lambda), 1.000427233, 1e-9);

        // Verify RefractedSemidiameter().

        p = 1013.25f;
        t = 15;
        tolerance = SofaJpl.Angle.DmsToRad(0, 0, .1);
        accuracy = tolerance / 2.0;
        atm = new Atmosphere(0, (float)t, (float)p, 50, true);
        double unrefEl = SofaJpl.Angle.DmsToRad(20, 0, 0);
        double unrefSD = SofaJpl.Angle.DmsToRad(0, 16, 0);
        // vertical semidiameter at lower limb
        double lowerRefSD = atm.Refract(unrefEl, accuracy)
            - atm.Refract(unrefEl - unrefSD, accuracy);
        double refSD = atm.RefractedSemidiameter(unrefSD, unrefEl, Math.PI, accuracy);
        equalsApprox(refSD, lowerRefSD, tolerance);
        // Chauvenet says horizontal contraction is a quarter second.
        refSD = atm.RefractedSemidiameter(unrefSD, unrefEl, SofaJpl.Angle.HalfPi,
            accuracy);
        equalsApprox(refSD, unrefSD - SofaJpl.Angle.DmsToRad(0, 0, .25), tolerance);
        // Per Chauvenet approximate formula, at 45 deg position angle the contraction
        // is half the vertical contraction. Per my independent computation, his
        // approximation is accurate to .12 arcsec in these conditions.
        refSD = atm.RefractedSemidiameter(unrefSD, unrefEl, Math.PI * .75, accuracy);
        equalsApprox(refSD, unrefSD - (unrefSD - lowerRefSD) * .5, 1.5 * tolerance);
    }


    /// <summary>
    /// Verify Refract() and Unrefract() are inverses of each other, and refraction
    /// is the expected value.
    /// </summary>
    /// <param name="atm">atmosphere</param>
    /// <param name="refractedAltitude">refracted altitude (radians)</param>
    /// <param name="correctRefraction">expected refraction (radians)</param>
    /// <param name="accuracy">desired accuracy (radians)</param>
    /// <param name="tolerance">max error allowed (radians)</param>
    private static void roundTripTest(Atmosphere atm, double refractedAltitude,
        double correctRefraction, double accuracy, double tolerance) {

        double refraction = refractedAltitude
            - atm.Unrefract(refractedAltitude);
        equalsApprox(refraction, correctRefraction, tolerance);
        equalsApprox(atm.Refract(refractedAltitude - refraction, accuracy),
            refractedAltitude, 2.0 * accuracy);
    }


    private static void equalsApprox(double a, double b, double tolerance) {
        if (Math.Abs(a - b) > tolerance)
            throw new Exception(string.Format(
                "a = {0:e2}, b = {1:e2}, tolerance = {2:e2}, error = {3:e2}",
                a, b, tolerance, a - b));
    }

#endif
}       // end class Atmosphere
}       // end namespace HirosePS.SofaJpl
