﻿// Body.cs
// Base class for all celestial bodies in SofaJpl.
// Copyright 2018 by Paul S. Hirose

// revision history
#if false

2015-06-26 Created.
2015-09-05  Changed TopocentricGeometric() to return a PVVector so it's consistent with the other
methods that return geometric place.

2018-04-18 Removed "protected" from Radius set property so user can control radius of the body.
#endif


using System;


namespace HirosePS.SofaJpl {
/// <summary>
/// Base class of all celestial bodies.
/// </summary>
/// <remarks>An instance of the class can provide the geometric ITRS
/// position and velocity of the body with respect to the barycenter, heliocenter, geocenter, or
/// topocenter at a specified epoch. For the geocenter and topocenter, astrometric and apparent
/// place are available too.</remarks>
public abstract class Body {

    // Calculated physical constants.
    private static double _c;           // speed of light, km/day
    private static double _muOverC2;    // mu over c squared, km

    // properties
    private Ephemeris _ephemeris;
    private string _name;
    private bool _isSun, _isPlanet, _isEarth;
    private double _radius;     // km


    static Body() {
        // Physical constants, 2016 Almanac, Section K.
        const double secPerDay = 86400.0;
        const double cKmPerSec = 299792.458;   // speed of light, km/s
        // constant of gravitation, km3kg-1day-2
        const double G = 6.67428e-20 * secPerDay * secPerDay;
        const double massSun = 1.9884e30;               // mass of Sun, kg

        // speed of light, km/day
        _c = cKmPerSec * secPerDay;

        // mu over c squared, km
        _muOverC2 = (G * massSun) / (_c * _c);

        // Check for blunders in the calculated constants and properties.
        if (Math.Abs(cLight - 2.59021e10) > 1e5)
            throw new Exception("Internal error: cLight property is wrong.");
        if (Math.Abs(muOverC2 - 1.4766) > .001)
            throw new Exception("Internal error: muOverC2 property is wrong.");
    }


    /// <summary>
    /// Get or set the name of the body.
    /// </summary>
    public string Name {
        get { return _name; }
        set { _name = value; }
    }

    /// <summary>
    /// Get the name of the body
    /// </summary>
    /// <returns>body name</returns>
    public override string ToString() { return _name; }

    /// <summary>
    /// Get the solar system ephemeris.
    /// </summary>
    public Ephemeris Ephemeris {
        get { return _ephemeris; }
        protected set { _ephemeris = value; }
    }

    /// <summary>
    /// True if the body is the Sun.
    /// </summary>
    public bool IsSun {
        get { return _isSun; }
        protected set { _isSun = value; }
    }

    /// <summary>
    /// True if the body is a solar system body, including the solar system barycenter and Earth
    /// Moon barycenter.
    /// </summary>
    public bool IsSolarSystemBody {
        get { return _isPlanet; }
        protected set { _isPlanet = value; }
    }

    /// <summary>
    /// True if the body is Earth.
    /// </summary>
    public bool IsEarth {
        get { return _isEarth; }
        protected set { _isEarth = value; }
    }

    /// <summary>
    /// Get the radius of the body (km).
    /// </summary>
    public double Radius {
        get { return _radius; }
        set { _radius = value; }
    }

    /// <summary>
    /// Speed of light, km / day.
    /// </summary>
    protected static double cLight { get { return _c; } }

    /// <summary>
    /// Mu over c squared, km.
    /// </summary>
    protected static double muOverC2 { get { return _muOverC2; } }



    /// <summary>
    /// Barycentric geometric position and velocity.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>barycentric position (km) and velocity (km/day), in the same
    /// coordinate reference system as the ephemeris or star catalog</returns>
    public abstract PVVector Barycentric(JulianDate tdb);


    /// <summary>
    /// Heliocentric geometric position and velocity.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>heliocentric geometric position (km) and velocity (km/day), in the
    /// same coordinate reference system as the ephemeris or star catalog</returns>
    public abstract PVVector Heliocentric(JulianDate tdb);


    /// <summary>
    /// Geocentric geometric position and velocity.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>geocentric geometric position (km) and velocity (km/day), in the same coordinate
    /// reference system as the ephemeris or star catalog</returns>
    public abstract PVVector GeocentricGeometric(JulianDate tdb);


    /// <summary>
    /// Geocentric astrometric position.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>geocentric astrometric position (km) in the same coordinate reference
    /// system as the ephemeris or star catalog</returns>
    public abstract Vector GeocentricAstrometric(JulianDate tdb);


    /// <summary>
    /// Geocentric apparent position.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="deflect">true = include light deflection due to gravity of
    /// the Sun.</param>
    /// <returns>geocentric apparent position (km), in the same coordinate reference
    /// system as the ephemeris or star catalog</returns>
    public abstract Vector GeocentricApparent(JulianDate tdb, bool deflect);


    /// <summary>
    /// Geocentric apparent position, including light deflection due to gravity of
    /// the Sun.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>geocentric apparent position (km), in the same coordinate reference
    /// system as the ephemeris or star catalog</returns>
    public abstract Vector GeocentricApparent(JulianDate tdb);


    // Replaced be a method that reterns a PVVector.
#if false
    /// <summary>
    /// Topocentric geometric position.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observerP">observer position (km) with respect to the geocenter,
    /// in the same coordinate reference system as the ephemeris or star catalog
    /// </param>
    /// <returns>topocentric geometric position (km) in the same coordinate reference
    /// system as the ephemeris or star catalog</returns>
    public abstract Vector TopocentricGeometric(JulianDate tdb, Vector observerP);
#else
    /// <summary>
    /// Topocentric geometric position and velocity.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observerPV">observer position (km) and velocity (km/day) with respect to the
    /// geocenter, in the same coordinate reference system as the ephemeris or star catalog</param>
    /// <returns>topocentric geometric position (km) and velocity (km/day) in the same coordinate
    /// reference system as the ephemeris or star catalog</returns>
    public abstract PVVector TopocentricGeometric(JulianDate tdb, PVVector observerPV);
#endif


    /// <summary>
    /// Topocentric astrometric position.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observerP">observer position (km) with respect to the geocenter,
    /// in the same coordinate reference system as the ephemeris or star catalog
    /// </param>
    /// <returns>topocentric astrometric position (km) in the same coordinate
    /// reference system as the ephemeris or star catalog</returns>
    public abstract Vector TopocentricAstrometric(JulianDate tdb, Vector observerP);


    /// <summary>
    /// Topocentric apparent position.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observer">observer's position (km) and velocity (km/day) with
    /// respect to the geocenter, in the same coordinate reference system as the
    /// ephemeris or star catalog</param>
    /// <param name="deflect">true = include light deflection due to solar gravity
    /// </param>
    /// <returns>topocentric apparent position (km) in the same coordinate reference
    /// system as the ephemeris or star catalog</returns>
    public abstract Vector TopocentricApparent(JulianDate tdb, PVVector observer,
        bool deflect);


    /// <summary>
    /// Topocentric apparent position, including light deflection due to
    /// solar gravity.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observer">observer's position (km) and velocity (km/day) with
    /// respect to the geocenter, in the same coordinate reference system as the
    /// ephemeris or star catalog</param>
    /// <returns>topocentric apparent position (km) in the same coordinate reference
    /// system as the ephemeris or star catalog</returns>
    public abstract Vector TopocentricApparent(JulianDate tdb, PVVector observer);



    
    // begin self-test code
#if DEBUG


    /// <summary>
    /// Throw exception if a boolean expression is false.
    /// </summary>
    /// <param name="assertion">boolean expression</param>
    /// <param name="message">message to display if assertion is false.</param>
    public static void assert(bool assertion, string message) {
        if (!assertion)
            throw new Exception(String.Concat("assertion failed: ", message));
    }


    /// <summary>
    /// Throw exception if a boolean expression is false.
    /// </summary>
    /// <param name="assertion">boolean expression</param>
    public static void assert(bool assertion) {
        assert(assertion, "reason unknown");
    }


    /// <summary>
    /// Throw exception if the separation angle between two vectors exceeds a given angle.
    /// </summary>
    /// <param name="v1">first vector</param>
    /// <param name="v2">second vector</param>
    /// <param name="tolerance">max allowable separation angle (radians)</param>
    /// <param name="message">message to display with exception</param>
    public static void testSeparationAngle(Vector v1, Vector v2, double tolerance,
        string message) {
        double sep = v1.SeparationAngle(v2);
        if (sep > tolerance)
            throw new Exception(string.Concat(string.Format(
                "separation angle = {0:e1} arcsec: ", Angle.RadToDeg(sep) * 3600.0,
                message)));
    }


    /// <summary>
    /// Throw exception if the separation angle between a vector and a Spherical instance exceeds a
    /// given angle.
    /// </summary>
    /// <param name="v1">vector</param>
    /// <param name="s1">spherical coordinates</param>
    /// <param name="tolerance">max allowable separation angle (radians)</param>
    /// <param name="message">message to display with exception</param>
    public static void testSeparationAngle(Vector v1, Spherical s1, double tolerance,
        string message) {
        testSeparationAngle(v1, new Vector(s1), tolerance, message);
    }


    /// <summary>
    /// Throw exception if two doubles fail to agree within a given tolerance.
    /// </summary>
    /// <param name="a">first value</param>
    /// <param name="b">second value</param>
    /// <param name="tolerance">max allowable difference</param>
    /// <param name="message">message to display with exception</param>
    public static void equalsApprox(double a, double b, double tolerance,
        string message) {
        if (Math.Abs(a - b) > tolerance)
            throw new Exception(string.Concat("doubles out of tolerance: ", message));
    }


    /// <summary>
    /// Throw exception if the root mean square difference between two vectors exceeds a tolerance.
    /// </summary>
    /// <param name="v1">first vector</param>
    /// <param name="v2">second vector</param>
    /// <param name="rmsTolerance">max RMS difference</param>
    /// <param name="message">message to display with exception</param>
    public static void equalsApprox(Vector v1, Vector v2, double rmsTolerance,
        string message) {
        double deltaX = v1.X - v2.X;
        double deltaY = v1.Y - v2.Y;
        double deltaZ = v1.Z - v2.Z;
        if (Math.Sqrt((deltaX * deltaX + deltaY * deltaY + deltaZ * deltaZ) / 3.0)
            > rmsTolerance)
            throw new Exception(message);
    }

#endif    // end self-test code

}       // end class Body
}       // end namespace HirosePS.SofaJpl
