// Duration.cs
// Class to represent an amount of time.
// Copyright 2018 by Paul S. Hirose.

// revision history
#if false
2007-11-17 Changed several properties (ToDouble et al) to methods.
2007-11-24 Self-test method completed and ran OK.
2007-11-28 Self-test method checks that CompareTo() throws exception for wrong type.
2007-12-05 CompareTo() and Equals() now null safe. Test() needs to be rewritten to
check that.
2007-12-06 All relational methods and operators test OK.
2007-12-20 Added a few simple methods: ToDays(), ToHours(), etc.
2007-12-29 Added a couple static constants for the TT-TAI and TAI-GPS offsets.
2008-01-14 Added Abs().
2009-01-06 Added FromHours(), FromMinutes(), FromSeconds(), FromRadians().
2010-01-31 Added a documentation comment to make clear the ToRadians() return value is
modulo 2pi, so is always positive.
2012-05-05 Added constructors taking days hours minutes seconds, and hours minutes sec.

2014-10-07  Now a support class to my IAU SOFA DLL. Appears to require no changes.
2014-10-13  Improved XML comments. Added several constants. Improved constructor that takes days,
hours, min, sec so only one param need be negative.
2014-10-18  All modifications complete. Self test OK.
2014-11-08  Added a static version of Abs().
#endif


using System;


namespace HirosePS.SofaJpl
{
    /// <summary>
    /// A Duration object represents an amount of time.
    /// </summary>
    /// <remarks>
    /// <para>The internal representation is a pair of double precision variables: one for whole
    /// days and one for the fraction of a day. Resolution is better than one nanosecond in a span
    /// of trillions of years. In practice a Duration object is almost always added to or subtracted
    /// from a JulianDate, so the useful range is "only" a few million years.</para>
    /// <para>The class also contains the fixed offsets between TT, GPS time, and TAI</para>
    /// <para>Instances of this class are immutable.</para>
    /// </remarks>
    public class Duration : IComparable<Duration>, IComparable, IEquatable<Duration>
    {
        // internally, the duration is represented as integral and fractional parts, both
        // in days
        private double integer, fraction;



        // constants
        private const double _hoursPerDay = 24.0;
        private const double _minutesPerDay = _hoursPerDay * 60.0;
        private const double _secondsPerDay = _minutesPerDay * 60.0;
        private const double _daysPerJulianCentury = 36525.0;
        private const double _ttMinusTai = 32.184;
        private const double _taiMinusGps = 19.0;

        /// <summary>
        /// Time scale difference: Terrestrial Time minus International Atomic Time.
        /// </summary>
        public static readonly Duration TTMinusTai = new Duration(_ttMinusTai /
            _secondsPerDay);

        /// <summary>
        /// Time scale difference: Terrestrial Time minus GPS Time.
        /// </summary>
        public static readonly Duration TTMinusGps = new Duration((_ttMinusTai +
            _taiMinusGps) / _secondsPerDay);



        // constructors


        /// <summary>
        /// Construct an instance of duration zero.
        /// </summary>
        public Duration() {
            this.integer = 0.0;
            this.fraction = 0.0;
        }


        /// <summary>
        /// Copy an instance.
        /// </summary>
        /// <param name="diff">The instance to copy</param>
        public Duration(Duration diff)
        {
            this.integer = diff.integer;
            this.fraction = diff.fraction;
        }


        /// <summary>
        /// Construct from two double precision values (days) which sum to the desired duration.
        /// </summary>
        /// <param name="integerDays">duration (days) part 1
        /// </param>
        /// <param name="fractionalDays">duration (days) part 2
        /// </param>
        /// <remarks>It is not necessary that the parameters be integer days and fractional days,
        /// but that format gives maximum resolution. One or both may be negative.</remarks>
        public Duration(double integerDays, double fractionalDays)
        {
            integer = integerDays;
            fraction = fractionalDays;
            this.normalize();
        }


        /// <summary>
        /// Construct from one double precision value (days).
        /// </summary>
        /// <param name="days">duration in days</param>
        public Duration(double days)
            : this(days, 0.0)
        {}


        /// <summary>
        /// Construct from a .NET Framework TimeSpan.
        /// </summary>
        /// <param name="span">a .NET Framework TimeSpan</param>
        public Duration(TimeSpan span)
        {
            this.integer = span.Days;
            this.fraction = (double)(span.Ticks - TimeSpan.TicksPerDay * this.integer)
                / (double)TimeSpan.TicksPerDay;
            this.normalize();
        }


        /// <summary>
        /// Construct from days, hours, minutes, seconds.
        /// </summary>
        /// <param name="days">days</param>
        /// <param name="hours">hours</param>
        /// <param name="minutes">minutes</param>
        /// <param name="seconds">seconds</param>
        /// <remarks>If any parameter is negative, it's the same as all negative parameters, and
        /// the result is negative.</remarks>
        public Duration(int days, int hours, int minutes, double seconds)
        {
            this.integer = days;
            this.fraction = Angle.HmsToHours(hours, minutes, seconds) / _hoursPerDay;
            // If the components have different signs, make both negative.
            if (this.integer < 0.0 != this.fraction < 0.0)
                if (this.integer < 0.0)
                    this.fraction = -this.fraction;
                else
                    this.integer = -this.integer;
            this.normalize();
        }


        /// <summary>
        /// Construct from hours, minutes, seconds.
        /// </summary>
        /// <param name="hours">hours</param>
        /// <param name="minutes">minutes</param>
        /// <param name="seconds">seconds</param>
        /// <remarks>If any parameter is negative, it's the same as all negative parameters, and
        /// the result is negative.</remarks>
        public Duration(int hours, int minutes, double seconds)
            : this(0, hours, minutes, seconds)
        { }




        // properties


        /// <summary>
        /// The integer (whole days) part.
        /// </summary>
        public double IntegerPart { get { return this.integer; } }


        /// <summary>
        /// The fractional part (always positive even if the instance is negative).
        /// </summary>
        public double FractionPart { get { return this.fraction; } }




        // methods


        /// <summary>
        /// Express the value of this instance in double precision floating point.
        /// </summary>
        /// <remarks>The resolution is about 2.0e-16 times the value.</remarks>
        /// <returns>days</returns>
        public double ToDouble()
        {
            return this.integer + this.fraction;
        }


        /// <summary>
        /// Convert value to days.
        /// </summary>
        /// <remarks>The resolution is about 2.0e-16 times the value.</remarks>
        /// <returns>days</returns>
        public double ToDays()
        {
            return ToDouble();
        }


        /// <summary>
        /// Convert value to hours.
        /// </summary>
        /// <returns>hours</returns>
        public double ToHours()
        {
            return ToDays() * _hoursPerDay;
        }


        /// <summary>
        /// Convert value to minutes.
        /// </summary>
        /// <returns>minutes</returns>
        public double ToMinutes()
        {
            return ToDays() * _minutesPerDay;
        }


        /// <summary>
        /// Convert value to seconds.
        /// </summary>
        /// <returns>seconds</returns>
        public double ToSeconds()
        {
            return ToDays() * _secondsPerDay;
        }


        /// <summary>
        /// Convert the fractional days component to radians (1 day = 2pi radians).
        /// </summary>
        /// <remarks>The returned value is modulo 2pi, so is always positive, even when
        /// the instance is negative.</remarks>
        /// <returns>radians in interval [0, 2pi)</returns>
        public double ToRadians()
        {
            return fraction * Angle.TwoPi;
        }

        
        /// <summary>
        /// Convert to Julian centuries of 36,525 days.
        /// </summary>
        /// <returns>Julian centuries</returns>
        public double ToJulianCenturies()
        {
            return this.ToDays() / _daysPerJulianCentury;
        }


        /// <summary>
        /// String representation of the integer and fractional parts in f1 and f6 formats
        /// respectively.
        /// </summary>
        /// <returns>String</returns>
        public override string ToString()
        {
            return string.Format("{0:f1} + {1:f6}", IntegerPart, FractionPart);
        }


        /// <summary>
        /// Convert to a .NET Framework TimeSpan object.
        /// </summary>
        /// <remarks>This is a lossy conversion: a .NET Framework TimeSpan has
        /// less resolution.
        /// </remarks>
        /// <returns>new .NET Framework TimeSpan</returns>
        public TimeSpan ToTimeSpan()
        {
            return new TimeSpan((long)(this.integer * TimeSpan.TicksPerDay
                + this.fraction * TimeSpan.TicksPerDay));
        }


        /// <summary>
        /// Convert hours to a Duration.
        /// </summary>
        /// <param name="hours">hours (negative is legal)</param>
        /// <returns>new Duration</returns>
        public static Duration FromHours(double hours)
        {
            return new Duration(hours / _hoursPerDay);
        }


        /// <summary>
        /// Convert minutes of time to a Duration.
        /// </summary>
        /// <param name="minutes">minutes of time (negative is legal)</param>
        /// <returns>new Duration</returns>
        public static Duration FromMinutes(double minutes)
        {
            return new Duration(minutes / _minutesPerDay);
        }


        /// <summary>
        /// Convert seconds of time to a Duration.
        /// </summary>
        /// <param name="seconds">seconds of time (negative is legal)</param>
        /// <returns>new Duration</returns>
        public static Duration FromSeconds(double seconds)
        {
            return new Duration(seconds / _secondsPerDay);
        }


        /// <summary>
        /// Convert radians to a Duration (1 day = 2pi radians).
        /// </summary>
        /// <param name="radians">radians (negative is legal)</param>
        /// <returns>new Duration</returns>
        public static Duration FromRadians(double radians)
        {
            return new Duration(radians / Angle.TwoPi);
        }


        /// <summary>
        /// Absolute value.
        /// </summary>
        /// <param name="d">a Duration instance</param>
        /// <returns>new Duration instance with absolute value</returns>
        public static Duration Abs(Duration d) {
            // The fractional part is not tested because it's never negative.
            return (d.integer < 0.0) ? d.Negate() : new Duration(d);
        }

        /// <summary>
        /// Absolute value.
        /// </summary>
        /// <returns>absolute value</returns>
        public Duration Abs()
        {
            return Abs(this);
        }

        /// <summary>
        /// Construct the negative of an instance.
        /// </summary>
        /// <returns>negative</returns>
        public Duration Negate()
        {
            return new Duration(-this.integer, -this.fraction);
        }

        /// <summary>
        /// Construct the negative of an instance.
        /// </summary>
        /// <param name="dur">Duration</param>
        /// <returns>negative</returns>
        public static Duration operator -(Duration dur)
        {
            return dur.Negate();
        }

        /// <summary>
        /// Add two Durations.
        /// </summary>
        /// <param name="rhs">Duration to add</param>
        /// <returns>sum</returns>
        public Duration Add(Duration rhs)
        {
            return new Duration(this.integer + rhs.integer, this.fraction + rhs.fraction);
        }
        
        /// <summary>
        /// Add two Durations.
        /// </summary>
        /// <param name="lhs">Duration</param>
        /// <param name="rhs">Duration</param>
        /// <returns>sum</returns>
        public static Duration operator+(Duration lhs, Duration rhs)
        {
            return lhs.Add(rhs);
        }
        
        /// <summary>
        /// Subtract one Duration from another. Result can be negative.
        /// </summary>
        /// <param name="rhs">Duration to subtract</param>
        /// <returns>difference</returns>
        public Duration Subtract(Duration rhs)
        {
            return new Duration(this.integer - rhs.integer, this.fraction - rhs.fraction);
        }

        /// <summary>
        /// Subtract one Duration from another. Result can be negative.
        /// </summary>
        /// <param name="lhs">Duration to subtract from</param>
        /// <param name="rhs">Duration to subtract</param>
        /// <returns>difference</returns>
        public static Duration operator -(Duration lhs, Duration rhs)
        {
            return lhs.Subtract(rhs);
        }

        /// <summary>
        /// Multiply Duration by a scalar.
        /// </summary>
        /// <param name="rhs">scaler</param>
        /// <returns>product</returns>
        public Duration Multiply(double rhs)
        {
            return new Duration(this.integer * rhs, this.fraction * rhs);
        }

        /// <summary>
        /// Multiply Duration by a scalar.
        /// </summary>
        /// <param name="lhs">Duration</param>
        /// <param name="rhs">scalar</param>
        /// <returns>product</returns>
        public static Duration operator *(Duration lhs, double rhs)
        {
            return lhs.Multiply(rhs);
        }

        /// <summary>
        /// Multiply Duration by a scalar.
        /// </summary>
        /// <param name="lhs">scalar</param>
        /// <param name="rhs">Duration</param>
        /// <returns>product</returns>
        public static Duration operator *(double lhs, Duration rhs)
        {
            return rhs.Multiply(lhs);
        }

        /// <summary>
        /// Divide Duration by a scalar.
        /// </summary>
        /// <param name="rhs">scalar</param>
        /// <returns>quotient</returns>
        public Duration Divide(double rhs) {
            return new Duration(this.integer / rhs, this.fraction / rhs);
        }

        /// <summary>
        /// Divide Duration by a scalar.
        /// </summary>
        /// <param name="lhs">Duration</param>
        /// <param name="rhs">scalar</param>
        /// <returns>quotient</returns>
        public static Duration operator /(Duration lhs, double rhs) {
            return new Duration(lhs.Divide(rhs));
        }

        /// <summary>
        /// Divide one Duration by another.
        /// </summary>
        /// <param name="rhs">denominator</param>
        /// <returns>quotient</returns>
        public double Divide(Duration rhs)
        {
            return this.ToDouble() / rhs.ToDouble();
        }

        /// <summary>
        /// Divide one Duration by another.
        /// </summary>
        /// <param name="lhs">numerator</param>
        /// <param name="rhs">denominator</param>
        /// <returns>quotient</returns>
        public static double operator /(Duration lhs, Duration rhs) {
            return lhs.Divide(rhs);
        }

        /// <summary>
        /// Compare value: this instance vs. another.
        /// </summary>
        /// <param name="rhs">the other instance</param>
        /// <returns>less than zero, zero, greater than zero, if this instance is less than,
        /// equal to, greater than rhs, respectively
        /// </returns>
        public int CompareTo(Duration rhs)
        {
            if ((object)rhs == null)
                return 1;
            int i = this.integer.CompareTo(rhs.integer);
            return (i != 0) ? i : this.fraction.CompareTo(rhs.fraction);
        }

        /// <summary>
        /// Test for value equality.
        /// </summary>
        /// <param name="rhs">another instance of this class</param>
        /// <returns>true if values are equal</returns>
        public bool Equals(Duration rhs)
        {
            return this.CompareTo(rhs) == 0;
        }

        /// <summary>
        /// Test for value equality.
        /// </summary>
        /// <param name="obj">object to test against</param>
        /// <returns>true if values are equal</returns>
        public override bool Equals(object obj)
        {
            if (obj == null || obj.GetType() != this.GetType())
                return false;
            return this.CompareTo((Duration)obj) == 0;
        }

        
        #region IComparable Members

        /// <summary>
        /// Value compare: this instance vs. an object.
        /// </summary>
        /// <param name="obj">Duration (or derived from)</param>
        /// <returns>less than zero, zero, greater than zero, if this instance is less than,
        /// equal to, greater than obj, respectively</returns>
        public int CompareTo(object obj)
        {
            if (obj == null)
                return 1;
            if (this.GetType() != obj.GetType())
                throw new ArgumentException("CompareTo() expected Duration type.");
            return this.CompareTo((Duration)obj);
        }

        #endregion


        /// <summary>
        /// Compare two instances for value equality.
        /// </summary>
        /// <param name="lhs">left hand side of ==</param>
        /// <param name="rhs">right hand side of ==</param>
        /// <returns>true if equal</returns>
        public static bool operator ==(Duration lhs, Duration rhs)
        {
            if ((object)lhs == null) {
                if ((object)rhs == null)
                    return true;
                return false;
            }
            return lhs.CompareTo(rhs) == 0;
        }


        /// <summary>
        /// Test two instances for value inequality.
        /// </summary>
        /// <param name="lhs">left hand side</param>
        /// <param name="rhs">right hand side</param>
        /// <returns>true if values are unequal</returns>
        public static bool operator !=(Duration lhs, Duration rhs)
        {
            return !(lhs == rhs);
        }


        /// <summary>
        /// Less than operator.
        /// </summary>
        /// <param name="lhs">left hand side</param>
        /// <param name="rhs">right hand side</param>
        /// <returns>true if <paramref name="lhs"/> is less than <paramref name="rhs"/></returns>
        public static bool operator <(Duration lhs, Duration rhs)
        {
            return lhs.CompareTo(rhs) < 0;
        }


        /// <summary>
        /// Less than or equal operator
        /// </summary>
        /// <param name="lhs">left hand side</param>
        /// <param name="rhs">right hand side</param>
        /// <returns>true if <paramref name="lhs"/> &lt;= <paramref name="rhs"/></returns>
        public static bool operator <=(Duration lhs, Duration rhs)
        {
            return lhs.CompareTo(rhs) <= 0;
        }


        /// <summary>
        /// Greater than or equal operator.
        /// </summary>
        /// <param name="lhs">left hand side</param>
        /// <param name="rhs">right hand side</param>
        /// <returns>true if <paramref name="lhs"/> &gt;= <paramref name="rhs"/></returns>
        public static bool operator >=(Duration lhs, Duration rhs)
        {
            return lhs.CompareTo(rhs) >= 0;
        }


        /// <summary>
        /// Greater than operator.
        /// </summary>
        /// <param name="lhs">left hand side</param>
        /// <param name="rhs">right hand side</param>
        /// <returns>true if <paramref name="lhs"/> &gt; <paramref name="rhs"/></returns>
        public static bool operator >(Duration lhs, Duration rhs)
        {
            return lhs.CompareTo(rhs) > 0;
        }


        /// <summary>
        /// Value-based hash of this instance.
        /// </summary>
        /// <returns>hash code</returns>
        public override int GetHashCode()
        {
            return integer.GetHashCode() ^ fraction.GetHashCode();
        }

#if DEBUG
        /// <summary>
        /// Normalize the representation to put the integer and fraction parts in their
        /// proper ranges without changing their sum.
        /// </summary>
#endif
        private void normalize()
        {
            double part1 = this.integer;
            double part2 = this.fraction;
            double part1_floor = Math.Floor(part1);
            double part2_floor = Math.Floor(part2);

            // remove the fraction part from part1
            double fraction1 = part1 - part1_floor;	    // always positive
            part1 = part1_floor;

            // remove the integer part from part2
            part2 -= part2_floor;				// always positive

            // put the integer and fraction parts in their correct places
            this.integer = part1 + part2_floor;
            this.fraction = part2 + fraction1;
            if (this.fraction >= 1.0)
            {
                this.fraction -= 1.0;
                this.integer += 1.0;
            }
        }


        #region self-test code
#if DEBUG

        /// <summary>
        /// Self-test method.
        /// </summary>
        public static void Test()
        {
            Duration dur1, dur2;
            const double oneUsec = 1e-6 / _secondsPerDay;
            const double oneNsec = 1e-9 / _secondsPerDay;
            const double oneGyr = 1e9 * _daysPerJulianCentury;

            // constants
            equalsApprox(_hoursPerDay, 24.0, 0.0);
            equalsApprox(_minutesPerDay, 1440.0, 0.0);
            equalsApprox(_secondsPerDay, 86400.0, 0.0);
            equalsApprox(_daysPerJulianCentury, 36525.0, 0.0);
            equalsApprox(_ttMinusTai, 32.184, 1e-9);
            equalsApprox(_taiMinusGps, 19.0, 1e-9);

            // time scale offsets
            TTMinusTai.EqualApprox(0.0, _ttMinusTai / _secondsPerDay, oneNsec);
            TTMinusGps.EqualApprox(0.0, (_taiMinusGps + _ttMinusTai) / _secondsPerDay, oneNsec);


            // constructor with no parameters
            dur1 = new Duration();
            dur1.EqualApprox(0.0, 0.0, 0.0);

            // The 2-value constructor gets most attention since other constructors call it.

            // near 0
            dur1 = new Duration(0.0, oneUsec);
            dur1.EqualApprox(0.0, oneUsec, oneNsec);
            dur1 = new Duration(0.0, -oneUsec);
            dur1.EqualApprox(-1.0, 1.0 - oneUsec, oneNsec);

            // extreme positive
            dur1 = new Duration(oneGyr, oneUsec);
            dur1.EqualApprox(oneGyr, oneUsec, oneNsec);
            dur1 = new Duration(oneGyr, -oneUsec);
            dur1.EqualApprox(oneGyr - 1.0, 1.0 - oneUsec, oneNsec);

            // extreme negative
            dur1 = new Duration(-oneGyr, oneUsec);
            dur1.EqualApprox(-oneGyr, oneUsec, oneNsec);
            dur1 = new Duration(-oneGyr, -oneUsec);
            dur1.EqualApprox(-oneGyr - 1.0, 1.0 - oneUsec, oneNsec);

            // test normalization
            dur1 = new Duration(9.604297, 4.856360);
            dur1.EqualApprox(14.0, .460657, oneNsec);
            dur1 = new Duration(-9.604297, 4.856360);
            dur1.EqualApprox(-5.0, .252063, oneNsec);
            dur1 = new Duration(9.604297, -4.856360);
            dur1.EqualApprox(4.0, .747937, oneNsec);

            // copy constructor
            dur2 = new Duration(dur1);
            dur2.EqualApprox(4.0, .747937, oneNsec);

            // single-value constructor
            dur1 = new Duration(14.460657);
            dur1.EqualApprox(14.0, .460657, oneNsec);

            // construct from a TimeSpan
            TimeSpan ts1 = new TimeSpan(10, 56, 9);
            dur1 = new Duration(ts1);
            dur1.EqualApprox(0.0, .455660, .000001);

            // construct from days, hours, minutes, seconds
            dur1 = new Duration(3, 23, 59, 60.0 - 1e-6);
            dur1.EqualApprox(3, 1.0 - oneUsec, oneNsec);
            dur1 = new Duration(-3, 23, 59, 60.0 - 1e-6);
            dur1.EqualApprox(-4.0, oneUsec, oneNsec);
            dur1 = new Duration(3, 0, 0, -1e-6);
            dur1.EqualApprox(-4.0, 1.0 - oneUsec, oneNsec);

            // Construct from hours, minutes, seconds.
            dur1 = new Duration(23, 59, 60.0 - 1e-6);
            dur1.EqualApprox(0.0, 1.0 - oneUsec, oneNsec);

            // properties
            dur1 = new Duration(9.604297, -4.856360);
            dur1.EqualApprox(dur1.IntegerPart, dur1.FractionPart, oneNsec);

            // ToDouble()
            dur1 = new Duration(9.604297, 4.856360);
            assertTrue(Math.Abs(dur1.ToDouble() - 14.460657) < 1e-12);

            // ToDays()
            assertTrue(Math.Abs(dur1.ToDays() - 14.460657) < 1e-12);

            // ToHours()
            assertTrue(Math.Abs(dur1.ToHours() - 14.460657 * _hoursPerDay) < 1e-10);

            // ToMinutes()
            assertTrue(Math.Abs(dur1.ToMinutes() - 14.460657 * _minutesPerDay) < 1e-8);

            // ToSeconds()
            assertTrue(Math.Abs(dur1.ToSeconds() - 14.460657 * _secondsPerDay) < 1e-6);

            // ToRadians()
            assertTrue(Math.Abs(dur1.ToRadians() - .460657 * 2.0 * Math.PI) < 1e-12);

            // ToJulianCenturies()
            assertTrue(Math.Abs(dur1.ToJulianCenturies() - 14.460657 / _daysPerJulianCentury)
                < 1e-17);

            // ToString()
            assertTrue(dur1.ToString().Equals("14.0 + 0.460657"));

            // Convert to a .NET Framework TimeSpan object.
            ts1 = dur1.ToTimeSpan();
            assertTrue(Math.Abs(ts1.TotalDays - 14.460657) < 1e-12);

            // FromHours()
            dur1 = FromHours(24.0);
            assertTrue(Math.Abs(dur1.ToDays() - 1.0) < 1e-13);

            // FromMinutes()
            dur1 = FromMinutes(_minutesPerDay);
            assertTrue(Math.Abs(dur1.ToDays() - 1.0) < 1e-13);

            // FromSeconds()
            dur1 = FromSeconds(_secondsPerDay);
            assertTrue(Math.Abs(dur1.ToDays() - 1.0) < 1e-13);

            // FromRadians()
            dur1 = FromRadians(Angle.TwoPi);
            assertTrue(Math.Abs(dur1.ToDays() - 1.0) < 1e-13);


            // Abs() static & instance methods. First, absolute value of a positive...
            dur1 = new Duration(1.5);
            assertTrue(Duration.Abs(dur1).IntegerPart == 1.0 &&
                Math.Abs(Duration.Abs(dur1).FractionPart - .5) < 1e-13);
            Duration.Abs(dur1).EqualApprox(1.0, .5, 1e-13);
            assertTrue(dur1.Abs().IntegerPart == 1.0 && Math.Abs(dur1.Abs().FractionPart - .5)
                < 1e-13);
            dur1.Abs().EqualApprox(1.0, .5, 1e-13);

            // ... then a negative.
            dur1 = new Duration(-1.5);
            assertTrue(Duration.Abs(dur1).IntegerPart == 1.0 &&
                Math.Abs(Duration.Abs(dur1).FractionPart - .5) < 1e-13);
            Duration.Abs(dur1).EqualApprox(1.0, .5, 1e-13);
            assertTrue(dur1.Abs().IntegerPart == 1.0 && Math.Abs(dur1.Abs().FractionPart - .5)
                < 1e-13);
            dur1.Abs().EqualApprox(1.0, .5, 1e-13);

            // Negate() method, - unary operator
            dur1 = new Duration(9.604297, 4.856360);
            dur2 = dur1.Negate();
            dur2.EqualApprox(-15.0, .539343, 1e-13);
            dur2 = -dur1;
            dur2.EqualApprox(-15.0, .539343, 1e-13);

            // Add() method and + operator
            dur1 = new Duration(2.0, .379133);
            dur2 = new Duration(4.0, .589546);
            Duration dur3 = dur2.Add(dur1);
            dur3.EqualApprox(6.0, .968679, 1e-13);
            dur3 = dur1 + dur2;
            dur3.EqualApprox(6.0, .968679, 1e-13);

            // Subtract() method and - operator
            dur3 = dur1.Subtract(dur2);
            dur3.EqualApprox(-3.0, .789587, 1e-13);
            dur3 = dur1 - dur2;
            dur3.EqualApprox(-3.0, .789587, 1e-13);

            // Multiply
            dur3 = dur1.Multiply(.945253);
            dur3.EqualApprox(2.0, .248883, .000001);
            dur3 = dur1 * .945253;
            dur3.EqualApprox(2.0, .248883, .000001);
            dur3 = .945253 * dur1;
            dur3.EqualApprox(2.0, .248883, .000001);

            // Divide by another Duration, and by scalar.
            assertTrue(Math.Abs(dur1.Divide(dur2) - .518381) < .000001);
            dur3 = dur1.Divide(.518381);
            dur3.EqualApprox(4.0, .589546, .000001);
            assertTrue(Math.Abs(dur1 / dur2 - .518381) < .000001);
            dur3 = dur1 / .518381;
            dur3.EqualApprox(4.0, .589546, .000001);

            // CompareTo(Duration)
            assertTrue(dur1.CompareTo(dur2) < 0);
            assertTrue(dur2.CompareTo(dur1) > 0);
            Duration durNull = null;
            assertTrue(dur2.CompareTo(durNull) > 0);
            Duration dur1a = new Duration(2.0, .379133);  // same as dur1
            assertTrue(dur1.CompareTo(dur1a) == 0);

            // Equals(Duration)
            assertTrue(dur1.Equals(dur1a));
            assertFalse(dur1.Equals(dur2));
            assertFalse(dur1.Equals(durNull));

            // Equals(Object)
            object obj1 = dur1;
            object obj2 = dur2;
            assertTrue(dur1.Equals(obj1));
            assertFalse(dur1.Equals(obj2));
            assertFalse(dur1.Equals(durNull));  // null object
            assertFalse(dur1.Equals(ts1));      // type mismatch

            // CompareTo(object)
            assertTrue(dur1.CompareTo(obj2) < 0);
            assertTrue(dur2.CompareTo(obj1) > 0);
            assertTrue(dur1.CompareTo(obj1) == 0);
            object objNull = durNull;
            assertTrue(dur1.CompareTo(objNull) > 0);
            try {
                Console.WriteLine("Duration.CompareTo(object) returned {0}", dur1.CompareTo(ts1));
                Console.WriteLine("That was an error: failed to detect wrong type.");
            } catch {
                ;
            }

            // relational operators

            assertTrue(dur1 == dur1a);
            assertFalse(dur1 == dur2);
            Duration durNull2 = null;
            assertTrue(durNull == durNull2);
            assertFalse(durNull == dur1);

            assertTrue(dur1 != dur2);
            assertFalse(dur1 != dur1a);
            assertFalse(durNull != durNull2);
            assertTrue(durNull != dur1);

            assertTrue(dur1 < dur2);
            assertFalse(dur1 < dur1a);
            assertFalse(dur2 < dur1);

            assertTrue(dur1 <= dur2);
            assertTrue(dur1 <= dur1a);
            assertFalse(dur2 <= dur1);

            assertFalse(dur1 >= dur2);
            assertTrue(dur1 >= dur1a);
            assertTrue(dur2 >= dur1);

            assertFalse(dur1 > dur2);
            assertFalse(dur1 > dur1a);
            assertTrue(dur2 > dur1);


            // GetHashCode
            assertFalse(dur1.GetHashCode() == dur2.GetHashCode());
        }


        /// <summary>
        /// Throw exception if double is not correct within specified tolerance.
        /// </summary>
        /// <param name="a">computed value</param>
        /// <param name="b">correct value</param>
        /// <param name="tolerance">max allowable difference (absolute value)</param>
        private static void equalsApprox(double a, double b, double tolerance) {
            if (Math.Abs(a - b) > tolerance)
                throw new Exception("Double out of tolerance.");
        }

        /// <summary>
        /// Throw exception if this instance isn't correct within a certain tolerance.
        /// </summary>
        /// <param name="days">expected whole days</param>
        /// <param name="frac">expected fractional days</param>
        /// <param name="tolerance">tolerance on fractional days</param>
        public void EqualApprox(double days, double frac, double tolerance)
        {
            double error = Math.Abs((this.integer - days) + (this.fraction - frac));
            if (error > tolerance)
                throw new Exception("Duration out of tolerance.");
        }


        /// <summary>
        /// Throw exception if argument is false.
        /// </summary>
        /// <param name="assertion">true = OK</param>
        private static void assertTrue(bool assertion)
        {
            if (!assertion)
                throw new Exception("Assertion was false.");
        }


        /// <summary>
        /// Throw exception if argument is true.
        /// </summary>
        /// <param name="assertion">false = OK</param>
        private static void assertFalse(bool assertion)
        {
            if (assertion)
                throw new Exception("Assertion was true.");
        }

#endif      // end class self-test code
        #endregion
    }
}
