﻿// Ephemeris.cs
// Base class for solar system ephemeris.
// Copyright 2018 by Paul S. Hirose

// revision history
#if false
2015-05-10 Created.
#endif



namespace HirosePS.SofaJpl {
/// <summary>
/// Solar system ephemeris base class.
/// </summary>
public abstract class Ephemeris {

    // This can be changed via a property.
    private double _kmPerAu = 1.49597870700e8;      // from 2015 Almanac p. K6


    /// <summary>
    /// Get or change the default kilometers per au.
    /// </summary>
    public double KmPerAu {
        get { return _kmPerAu; }
        set { _kmPerAu = value; }
    }


    /// <summary>
    /// Solar system barycentric geometric position and velocity of the designated body.
    /// </summary>
    /// <param name="tdb">epoch (TDB)</param>
    /// <param name="targetBody">target body designator</param>
    /// <returns>position (km) and velocity (km/day)</returns>
    public abstract PVVector Barycentric(JulianDate tdb, int targetBody);


    /// <summary>
    /// Heliocentric geometric position and velocity of the designated body.
    /// </summary>
    /// <param name="tdb">epoch (TDB)</param>
    /// <param name="targetBody">target body designator</param>
    /// <returns>position (km) and velocity (km/day)</returns>
    public abstract PVVector Heliocentric(JulianDate tdb, int targetBody);


    /// <summary>
    /// Geocentric geometric position and velocity of the designated body.
    /// </summary>
    /// <param name="tdb">epoch (TDB)</param>
    /// <param name="targetBody">target body designator</param>
    /// <returns>position (km) and velocity (km/day)</returns>
    public abstract PVVector Geocentric(JulianDate tdb, int targetBody);


    /// <summary>
    /// Earth barycentric position and velocity.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>Earth barycentric position (km) and velocity (km/day), in the same
    /// coordinate reference system as the ephemeris.</returns>
    public abstract PVVector EarthBaryPV(JulianDate tdb);


    /// <summary>
    /// Sun barycentric position and velocity.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>Sun barycentric position (km), and velocity (km/day), in the same coordinate
    /// reference system as the ephemeris.</returns>
    public abstract PVVector SunBaryPV(JulianDate tdb);

    
    /// <summary>
    /// Get the name of the ephemeris.
    /// </summary>
    /// <returns>ephemeris name</returns>
    public abstract override string ToString();


}   // end class Ephemeris
}   // end namespace HirosePS.SofaJpl
