﻿// HipparcosCatalog.cs
// A class where Hipparcos star catalog data can be looked up by the star's designation.
// Copyright 2018 by Paul S. Hirose

// revision history
#if false

2015-07-19 Created
2015-07-21 Self test OK.
2015-08-21 Added XML comments to star catalog entry class to stop compiler warnings.
2015-09-03 Added a constructor with parameters to specify file names and initial dictionary
capacities. Renamed classes to HipparcosCatalog and HipparcosCatalogEntry.
2015-09-05 Added GetStar().

2018-04-18 Increased default size of star name dictionary. Added code in self test to show number of
entries in both dictionaries.
#endif


using System;
using System.Collections.Generic;


namespace HirosePS.SofaJpl {

/// <summary>
/// A subset of the Hipparcos star catalog (complete to mag. 3), searchable by star designation.
/// </summary>
/// <remarks>The catalog data are from the re-reduction (van Leeuwen, 2007). The star designation
/// (name) file and data file are XML and can be modified by the user.</remarks>
public class HipparcosCatalog {

    // Default initial sizes of the dictionary that converts a name to the corresponding Hipparcos
    // catalog number, and the dictionary that converts a Hipparcos catalog number to the
    // corresponding catalog data. The default files use half this capacity.
    const int _nameDictionarySize = 568;
    const int _dataDictionarySize = 330;

    // files of data for the dictionaries
    const string _nameFile = "star_names.vot";
    const string _dataFile = "star_cat.vot";

    // the dictionaries
    Dictionary<string, int> _nameDictionary;
    Dictionary<int, HipparcosCatalogEntry> _dataDictionary;



    /// <summary>
    /// Construct from named XML files, with specified initial sizes for the name and data
    /// dictionaries.
    /// </summary>
    /// <param name="nameFile">name of the file that cross references star name to Hipparcos
    /// catalog number</param>
    /// <param name="dataFile">name of the file that cross references Hipparcos catalog number to
    /// catalog data</param>
    /// <param name="nameDictionarySize">initial maximum number of star names</param>
    /// <param name="dataDictionarySize">initial maximum number of stars</param>
    /// <remarks>
    /// <para>The star name and star data dictionaries automatically resize when their capacity is
    /// exceeded. However, this requires copying the data into a new larger dictionary
    /// and discarding the old dictionary. That can happen more than once, since the system
    /// doesn't know appropriate final sizes of the dictionaries. By specifying large enough initial
    /// capacities you can avoid the expense of resizing.</para>
    /// <para>Normally the name dictionary has more entries than the data dictionary since there can
    /// be multiple names for the same star.</para>
    /// </remarks>
    public HipparcosCatalog(string nameFile, string dataFile, int nameDictionarySize,
        int dataDictionarySize) {

        _nameDictionary = new Dictionary<string, int>(nameDictionarySize,
            StringComparer.OrdinalIgnoreCase);
        readStarNames(nameFile, _nameDictionary);

        _dataDictionary = new Dictionary<int, HipparcosCatalogEntry>(dataDictionarySize);
        readStarCat(dataFile, _dataDictionary);
    }


    /// <summary>
    /// Construct from the default XML files. Use default initial sizes for the name and data
    /// dictionaries.
    /// </summary>
    public HipparcosCatalog()
        : this(_nameFile, _dataFile, _nameDictionarySize, _dataDictionarySize)
    {
#if false
        _nameDictionary = new Dictionary<string, int>(_nameDictionarySize,
            StringComparer.OrdinalIgnoreCase);
        readStarNames(_nameFile, _nameDictionary);

        _dataDictionary = new Dictionary<int, StarCatalogEntry>(_dataDictionarySize);
        readStarCat(_dataFile, _dataDictionary);
#endif
    }


    /// <summary>
    /// Get the Dictionary that converts a star designation to Hipparcos catalog number.
    /// </summary>
    /// <remarks>The returned value can be used to add more (name, number) pairs to the
    /// dictionary.</remarks>
    public Dictionary<string, int> NameDictionary { get { return _nameDictionary; } }


    /// <summary>
    /// Get the Dictionary that converts a Hipparcos catalog number to the corresponding catalog
    /// entry.
    /// </summary>
    /// <remarks>The returned value can be used to add more stars.</remarks>
    public Dictionary<int, HipparcosCatalogEntry> DataDictionary { get { return _dataDictionary; } }


    /// <summary>
    /// Construct a HipparcosStar object, given a star name and an Ephemeris object.
    /// </summary>
    /// <param name="name">star designation</param>
    /// <param name="eph">solar system ephemeris</param>
    /// <returns>HipparcosStar object</returns>
    /// <remarks>The star designation must be a complete and exact match (except not case sensitive)
    /// to an entry in the star name dictionary.</remarks>
    public HipparcosStar GetStar(string name, Ephemeris eph) {
        HipparcosCatalogEntry entry = GetCatalogEntry(name);
        return new HipparcosStar(entry.RightAscension, entry.Declination, entry.Parallax,
            entry.RightAscensionPM, entry.DeclinationPM, eph, name, entry.CatalogNumber);
    }


    /// <summary>
    /// Get the star catalog entry, given the star designation (name).
    /// </summary>
    /// <param name="starDesignation">name of star (not case sensitive)</param>
    /// <returns>the matching Hipparcos catalog entry</returns>
    /// <exception cref="KeyNotFoundException"> if star designation not found</exception>
    /// <remarks>The star designation must exactly match an entry in the name catalog, except that
    /// it's not case sensitive.</remarks>
    public HipparcosCatalogEntry GetCatalogEntry(string starDesignation) {
        return _dataDictionary[_nameDictionary[starDesignation]];
    }


    /// <summary>
    /// Get the default capacity of the star name dictionary.
    /// </summary>
    public static int DefaultNameDictionaryCapacity { get { return _nameDictionarySize; } }


    /// <summary>
    /// Get the default capacity of the star data dictionary.
    /// </summary>
    public static int DefaultDataDictionaryCapacity { get { return _dataDictionarySize; } }



    // Go to the first row of the table data in a VOTable XML file.
    private void goToVotableRow1(System.Xml.XmlReader reader) {
        // Find <VOTABLE>.
        while (reader.Read() && !reader.IsStartElement("VOTABLE"))
            ;
        if (!reader.IsStartElement("VOTABLE"))
            throw new Exception("Failed to find <VOTABLE>.");

        // Find the first row of table data.
        if (!reader.ReadToDescendant("RESOURCE"))
            throw new Exception("Failed to find <RESOURCE>.");
        if (!reader.ReadToDescendant("TABLE"))
            throw new Exception("Failed to find <TABLE>.");
        if (!reader.ReadToDescendant("DATA"))
            throw new Exception("Failed to find <DATA>.");
        if (!reader.ReadToDescendant("TABLEDATA"))
            throw new Exception("Failed to find <TABLEDATA>.");
        if (!reader.ReadToDescendant("TR"))
            throw new Exception("Failed to find first row of table.");
    }


#if DEBUG
    /// <summary>
    /// Load the star name dictionary with the name (the key) and the Hipparcos catalog number
    /// of each star in the specified XML file.
    /// </summary>
    /// <param name="filename">name of the VOTable XML file with the input data</param>
    /// <param name="nameDictionary">the dictionary that receives the data</param>
#endif
    private void readStarNames(string filename, Dictionary<string, int> nameDictionary) {
        // Create the XML reader.
        System.Xml.XmlReader reader = System.Xml.XmlReader.Create(filename);

        // Go to first <TR> of table.
        goToVotableRow1(reader);

        // Process the table rows.
        string designation, hipStr;
        int hipInt;
        float mag;
        do {
            // Extract the designation and Hipparcos number. Magnitude isn't used in this
            // application.
            if (!reader.ReadToDescendant("TD"))
                throw new Exception("Failed to find designation <TD>.");
            designation = reader.ReadElementContentAsString().ToUpper();
            if (!reader.IsStartElement("TD"))
                throw new Exception("Can't find Hipparcos number <TD>.");
            hipStr = reader.ReadElementContentAsString();
            if (!reader.IsStartElement("TD"))
                throw new Exception("Can't find magnitude.");
            mag = reader.ReadElementContentAsFloat();

            // If the item has a Hipparcos catalog number, add it to the dictionary.
            if (hipStr != string.Empty) {
                hipInt = int.Parse(hipStr.Substring(4));    // ignore "HIP "

                // Add designation and Hipparcos number to table
                nameDictionary.Add(designation, hipInt);
            }

            // Go past the other <TD>s in this <TR>.
            while (reader.ReadToNextSibling("TD"))
                ;
            // Go to next <TR>.
            reader.ReadToNextSibling("TR");
        } while (reader.IsStartElement("TR"));
    }


#if DEBUG
    /// <summary>
    /// Load the star data dictionary with the Hipparcos catalog number (the key) and the data
    /// for each star in the specified XML file.
    /// </summary>
    /// <param name="filename">name of VOTable XML file</param>
    /// <param name="dataDictionary">the dictionary</param>
#endif
    private void readStarCat(string filename, Dictionary<int, HipparcosCatalogEntry> dataDictionary) {
        // Create the XML reader.
        System.Xml.XmlReader reader = System.Xml.XmlReader.Create(filename);

        // Go to first <TR> of table.
        goToVotableRow1(reader);

        // Process the rows.
        double ra, dec;
        float mag, pmra, pmdec, para;
        int hip;
        byte solutionType, components;
        // Process the table rows.
        do {
            if (!reader.ReadToDescendant("TD"))
                throw new Exception("Failed to find first <TD> in row.");
            hip = reader.ReadElementContentAsInt();
            solutionType = (byte)reader.ReadElementContentAsInt();
            components = (byte)reader.ReadElementContentAsInt();
            ra = reader.ReadElementContentAsDouble();
            dec = reader.ReadElementContentAsDouble();
            para = reader.ReadElementContentAsFloat();
            pmra = reader.ReadElementContentAsFloat();
            pmdec = reader.ReadElementContentAsFloat();
            mag = reader.ReadElementContentAsFloat();

            // Add the star to the catalog.
            dataDictionary.Add(hip, new HipparcosCatalogEntry(ra, dec, pmra, pmdec, para, mag,
                hip, solutionType, components));

            // Go past any other <TD>s in this <TR>.
            if (reader.IsStartElement())
                while (reader.ReadToNextSibling("TD"))
                    ;
            // Go to next <TR>.
            reader.ReadToNextSibling("TR");
        } while (reader.IsStartElement("TR"));
    }




    // self test code
#if DEBUG

    /// <summary>
    /// Run the class self test.
    /// </summary>
    public static void Test() {

        HipparcosCatalogEntry polaris;
        Ephemeris eph = new SofaEphemeris();
        Star star1, star2;
        JulianDate tdb;
        PVVector pv1, pv2;
        int i;

        HipparcosCatalog cat = new HipparcosCatalog();

        // Verify correct initial dictionary sizes.
        i = cat.NameDictionary.Count;
        if (i > _nameDictionarySize)
            throw new Exception("Default name dictionary size too small.");
        Console.WriteLine("{0:d} entries in name dictionary", i);
        i = cat.DataDictionary.Count;
        if (i > _dataDictionarySize)
            throw new Exception("Default data dictionary size too small.");
        Console.WriteLine("{0:d} entries in data dictionary", i);

        polaris = cat.GetCatalogEntry("name polaris");
        if (polaris.RightAscension != 037.94614300)
            throw new Exception("Polaris RA wrong.");

        // Verify correct action if designation not found.
        try {
            polaris = cat.GetCatalogEntry("polaris");
            throw new Exception("Failed to throw exception for designation not found.");
        } catch (KeyNotFoundException) {
            ;
        }

        // Add an alternate designation for Polaris, verify it works.
        Dictionary<string, int> nameDictionary = cat.NameDictionary;
        nameDictionary["Pole Star"] = polaris.CatalogNumber;
        polaris = cat.GetCatalogEntry("pole star");
        if (polaris.RightAscension != 037.94614300)
            throw new Exception("New designation didn't work.");

        // Test GetStar().
        tdb = new JulianDate(2015, 9, 5, 22, 25, 0);
        star1 = cat.GetStar("name Polaris", eph);
        star2 = new Star(JulianDate.FromJulianEpoch(1991.25), polaris.RightAscension / 15,
            polaris.Declination, polaris.Parallax, polaris.RightAscensionPM, polaris.DeclinationPM,
            0, eph, "Polaris");
        pv1 = star1.Barycentric(tdb);
        pv2 = star2.Barycentric(tdb);
        pv1.EqualsApprox(pv2, pv2.Position.Modulus() * 1e-6, pv2.Velocity.Modulus() * 1e-6);
    }

#endif      // end self test code




    /// <summary>
    /// Helper class to represent a star catalog entry.
    /// </summary>
    public class HipparcosCatalogEntry : IComparable {
        double _rightAscension;     // degrees
        double _declination;        // degrees
        float _rightAscensionPM;    // great circle mas/year
        float _declinationPM;       // mas/year
        float _parallax;            // mas
        float _magnitude;
        int _hipNumber;             // Hipparcos catalog number
        byte _solutionType;         // type of solution
        byte _components;           // number of components


        /// <summary>
        /// Construct a catalog entry with data from the Hipparcos second reduction.
        /// </summary>
        /// <param name="rightAscension">right ascension (degrees)</param>
        /// <param name="declination">declination (degrees)</param>
        /// <param name="rightAscensionPM">proper motion in right ascension
        /// (great circle mas/year)</param>
        /// <param name="declinationPM">proper motion in declination (mas/year)</param>
        /// <param name="parallax">parallax (mas)</param>
        /// <param name="magnitude">visual magnitude</param>
        /// <param name="hipNumber">Hipparcos catalog number</param>
        /// <param name="solutionType">solution type</param>
        /// <param name="components">number of components</param>
        public HipparcosCatalogEntry(double rightAscension, double declination,
            float rightAscensionPM, float declinationPM, float parallax,
            float magnitude, int hipNumber, byte solutionType, byte components) {
            _rightAscension = rightAscension;
            _rightAscensionPM = rightAscensionPM;
            _declination = declination;
            _declinationPM = declinationPM;
            _parallax = parallax;
            _magnitude = magnitude;
            _hipNumber = hipNumber;
            _solutionType = solutionType;
            _components = components;
        }

        /// <summary>
        /// Construct a phony instance as a search aid.
        /// </summary>
        /// <param name="hipNumber">Hipparcos catalog number</param>
        public HipparcosCatalogEntry(int hipNumber) {
            _hipNumber = hipNumber;
        }


        /// <summary>
        /// Right ascension (degrees).
        /// </summary>
        public double RightAscension { get { return _rightAscension; } }

        /// <summary>
        /// Declination (degrees).
        /// </summary>
        public double Declination { get { return _declination; } }

        /// <summary>
        /// Proper motion in right ascension (great circle mas/year).
        /// </summary>
        public double RightAscensionPM { get { return _rightAscensionPM; } }

        /// <summary>
        /// Proper motion in declination (mas/year).
        /// </summary>
        public double DeclinationPM { get { return _declinationPM; } }

        /// <summary>
        /// Trigonometric parallax (mas).
        /// </summary>
        public double Parallax { get { return _parallax; } }

        /// <summary>
        /// Visual magnitude.
        /// </summary>
        public double Magnitude { get { return _magnitude; } }

        /// <summary>
        /// Hipparcos catalog number.
        /// </summary>
        public int CatalogNumber { get { return _hipNumber; } }

        /// <summary>
        /// Bit flags for peculiarities.
        /// </summary>
        /// <remarks>0 = single star, 1 = double star, 2 = variable in system with
        /// amplitude &gt; 0.2 mag, 4 = astrometry refers to photocenter, 
        /// 8 = measurements concern the secondary (fainter) in the double system</remarks>
        public int Peculiarities { get { return _solutionType / 10; } }

        /// <summary>
        /// The type of solution adopted.
        /// </summary>
        /// <remarks>1 = stochastic, 3 = VIM, 5 = 5-parameter, 7 = 7-parameter,
        /// 9 = 9-parameter</remarks>
        public int SolutionType { get { return _solutionType - Peculiarities * 10; } }

        /// <summary>
        /// Number of components in this catalog entry.
        /// </summary>
        public int Components { get { return _components; } }


        /// <summary>
        /// Test star catalog entries for equal Hipparcos numbers.
        /// </summary>
        /// <param name="obj">StarCatalogEntry</param>
        /// <returns>true if Hipparcos numbers are equal</returns>
        public override bool Equals(object obj) {
            if (obj is HipparcosCatalogEntry)
                return this._hipNumber.Equals(((HipparcosCatalogEntry)obj)._hipNumber);
            return false;
        }


        /// <summary>
        /// Get hash code of this entry.
        /// </summary>
        /// <returns>hash of Hipparcos catalog number</returns>
        public override int GetHashCode() {
            return this._hipNumber.GetHashCode();
        }


        #region IComparable Members

        /// <summary>
        /// Compare star catalog entries
        /// </summary>
        /// <param name="obj">StarCatalogEntry</param>
        /// <returns>comparison of the HIP numbers</returns>
        public int CompareTo(object obj) {
            if (obj == null)
                return 1;
            if (obj is HipparcosCatalogEntry)
                return this._hipNumber.CompareTo(((HipparcosCatalogEntry)obj)._hipNumber);
            else
                throw new ArgumentException("Object is not a StarCatalogEntry.");
        }

        #endregion

    }   // end nested class starCatalogEntry
}       // end class StarCatalog
}       // end namespace HirosePS.SofaJpl
