﻿// HipparcosStar.cs
// Copyright 2018 by Paul S. Hirose

// log of changes
#if false

2008-06-02 Began work.
2008-06-09 Finished; self-test OK.
2008-06-11 Self-test moved to base class.
2008-11-20 Corrected improper base class initialization. No difference in net result,
but style is better.
2008-11-23 Tiny correction to kmPerAu to bring it in line with the Hipparcos catalog.
Too small to make any practical difference, though.
2012-03-10 Had to rewrite constructors to account for the changes to the Star
constructor.
2012-03-11 Edited comments.
2012-03-14 Minor improvements to XML comments.
2012-03-21 Ditto.
2012-03-26 Self-test code move back to the derived classes for easier access to the
intermediate values for test purposes. Test is OK.
2012-04-30 kmPerAU now comes from a property in the Body class. Self-test OK.
2012-05-05 Removed IBody interface as redundant.
2012-12-23 Star name added to constructor parameter list. Self test OK.
2013-03-19 Identified the stars in the radial velocity table.

2015-09-04 Transplanted to SofaJpl 2.0. The constructor with a radial velocity parameter has been
deleted. I see no reason to not rely on automatic table lookup.

2018-04-20 Numerous fixes to self test. I don't know how it passed before, but OK now.
#endif

using System;
using System.Collections.Generic;
using System.Text;


namespace HirosePS.SofaJpl {
    /// <summary>
    /// Derived from Star, this class is identical except that the constructors take
    /// Hipparcos catalog units (R.A. in degrees, etc.) for the convenience of the
    /// programmer.
    /// </summary>
    public class HipparcosStar : Star {
        private const double degreesPerHour = 15.0;

        // the stars in the Hipparcos catalog that have significant radial velocity,
        // sorted by catalog number, with velocity in km/s
        private static SortedList<int, double> radVelList =
            new SortedList<int, double>(25);

        // base epoch of the Hipparcos catalog
        private static JulianDate _baseEpoch =
            new JulianDate(JulianDate.FromJulianEpoch(1991.25));


        static HipparcosStar() {
            // initialize radial velocity table from Table 1.2.3 of Hipparcos catalog

            radVelList.Add(439, 22.9);      // HD 225213, mag 8.6, par 230.4
            radVelList.Add(3829, -38.0);    // VAN MAANEN STAR, mag 12.4, par 235 mas
            radVelList.Add(5336, -98.1);    // mu. Cas, mag 5.1, par 132 mas
            radVelList.Add(15510, 86.7);    // e Eri, mag 4.3, par 165
            radVelList.Add(19849, -42.7);   // omi02 Eri A, mag 4.4, par 201
            radVelList.Add(24186, 245.5);   // Kapteyn's star, mag 8.9, par 256
            radVelList.Add(26857, 105.6);   // V1352 Ori, mag 11.5, par 172
            radVelList.Add(54035, -84.3);   // 95735, mag 7.5, par 393
            radVelList.Add(54211, 68.8);    // LHS 38, mag 8.8, par 206
            radVelList.Add(57939, -99.1);   // LHS 44, mag 6.4, par 110
            radVelList.Add(70890, -16.0);   // PROXIMA CENTAURI, mag 11.0, par 772

            radVelList.Add(71681, -18.1);   // alf02 Cen, mag 1.3, par 797
            radVelList.Add(71683, -26.2);   // alf Cen A, mag 0.0, par 755
            radVelList.Add(74234, 308.0);   // LDS 52, mag 9.4, par 35
            radVelList.Add(74235, 294.3);   // LHS 53, mag 9.1, par 35
            radVelList.Add(86990, -115.0);  // LHS 454, mag 10.8, par 171
            radVelList.Add(87937, -111.0);  // Barnard's star, mag 9.5, par 548
            radVelList.Add(99461, -129.8);  // LHS 486, mag 5.3, par 166
            radVelList.Add(104214, -64.8);  // 61 Cyg A, mag 5.2, par 287
            radVelList.Add(104217, -64.3);  // 61 Cyg B, mag 6.0, par 286
            radVelList.Add(108870, -40.4);  // eps Ind A, mag 4.7, par 276
        }


        // This constructor has a radial velocity parameter, but I've decided to rely on
        // automatic lookup in the table.
#if false
        /// <summary>
        /// Construct a star using the units of the Hipparcos catalog.
        /// </summary>
        /// <remarks><paramref name="sun"/> and <paramref name="earth"/> are the
        /// basis for computing gravitational light deflection and retardation by the Sun,
        /// and annual parallax and aberration due to Earth's motion.</remarks>
        /// <param name="ra">right ascension (degrees)</param>
        /// <param name="dec">declination (degrees)</param>
        /// <param name="parallax">parallax in milli arc seconds (mas)</param>
        /// <param name="raPM">proper motion in right ascension (great circle mas / Julian
        /// year)</param>
        /// <param name="decPM">proper motion in declination (mas/Jy)</param>
        /// <param name="radVel">radial velocity (km/s receding)</param>
        /// <param name="sun">supplies the Sun's geometric position (AU) and velocity
        /// (AU/day) in the BCRS; null is OK if neither heliocentric origin nor apparent
        /// place will be requested</param>
        /// <param name="earth">supplies Earth's geometric position (AU) and velocity
        /// (AU/day) in the BCRS; null is OK if neither geocentric nor topcentric origin
        /// will be requested</param>
        /// <param name="name">star name</param>
        public HipparcosStar(double ra, double dec, double parallax, double raPM,
            double decPM, double radVel, SolarSystemBody sun, SolarSystemBody earth,
            string name)
            : base(_baseEpoch, ra / degreesPerHour, dec, parallax, raPM, decPM, radVel,
            sun, earth, name) { }
#endif


        /// <summary>
        /// Construct a Star from the native units of the Hipparcos catalog, e.g., RA in degrees.
        /// Radial velocity is supplied automatically.
        /// </summary>
        /// <param name="ra">right ascension (degrees)</param>
        /// <param name="dec">declination (degrees)</param>
        /// <param name="parallax">parallax in milli arc seconds (mas)</param>
        /// <param name="raPM">proper motion in right ascension (great circle mas / Julian
        /// year)</param>
        /// <param name="decPM">proper motion in declination (mas/Jy)</param>
        /// <param name="eph">solar system ephemeris for parallax, aberration, and light
        /// deflection</param>
        /// <param name="name">star name</param>
        /// <param name="catalogNumber">Hipparcos catalog number</param>
        /// <remarks>Hipparcos catalog number is required to look up radial velocity in a table of
        /// the 21 stars with significant velocity (Hipparcos catalog table 1.2.3). If the star
        /// is not in the table, its velocity is set to 0.
        /// </remarks>
        public HipparcosStar(double ra, double dec, double parallax, double raPM,
            double decPM, Ephemeris eph, string name, int catalogNumber)

            : base(_baseEpoch, ra / degreesPerHour, dec, parallax, raPM, decPM,
            lookUpRadialVel(catalogNumber), eph, name) { }



        private static double lookUpRadialVel(int catalogNumber) {
            double rv;
            radVelList.TryGetValue(catalogNumber, out rv);
            return rv;
        }


        // self-test code
#if DEBUG
        /// <summary>
        /// Run the class self test.
        /// </summary>
        new public static void Test() {
            // seconds per day
            const double secPerDay = 86400.0;

            // JPL ephemeris
            string ephName = "binp2000_2099.406";

            // open the ephemeris
            Ephemeris eph = new JplEphemeris(ephName);

            // km per AU
            double kmPerAU = eph.KmPerAu;

            // base epoch of Hipparcos catalog
            JulianDate tdb1 = JulianDate.FromJulianEpoch(1991.25);

            // create fake star with a catalog number matching an entry in the radial
            // velocity table, verify the velocity lookup is OK
            HipparcosStar star1 = new HipparcosStar(293.08577692, 69.66540172, 173.41,
                598.43, -1738.81, eph, "fake", 24186);
            assert(Math.Abs(star1.RadVel / secPerDay - 245.5) < .1);

            // Test space motion value with a fake star moving 10 au / year in each proper
            // motion coordinate.
            star1 = new HipparcosStar(293.08577692, 69.66540172, 1.0,
                10.0, 10.0, eph, "fake", 0);
            assert(Math.Abs(star1.Barycentric(tdb1).Velocity.Modulus() / kmPerAU * 365.25
                - Math.Sqrt(10 * 10 + 10 * 10)) < .01);

            // verify zero parallax puts the star at the correct distance
            star1 = new HipparcosStar(293.08577692, 69.66540172, 0.0,
                0.0, 0.0, eph, "fake", 0);
            double starDistance = star1.Barycentric(tdb1).Position.Modulus();
            double correctDistance = kmPerAU / star1.MinimumParallax;
            assert(Math.Abs(starDistance - correctDistance) / correctDistance < .0001);
            assert(star1.MinimumParallax != 0.0);

            // create the real star sigma Dra (Hipparcos new reduction)
            star1 = new HipparcosStar(293.08577729, 69.66540164, 173.77,
                598.07, -1738.40, eph, "sig Dra", 96100);

            // this star's velocity should be zero
            assert(star1.RadVel == 0.0);

            // assign the star to a Body object to test the virtual methods
            Body body1 = star1;


            // barycentric position (mean equator at mid-year), from 2012 Almanac bright star table
            tdb1 = new JulianDate(2456110.5, .125);
            double ra1, dec1;
            ra1 = Angle.HoursToRadians(19.0 + (32.0 + 20.0 / 60.0) / 60.0);
            dec1 = Angle.DegreesToRadians((69.0 + (40.0 + 57.0 / 60.0) / 60.0));
            Vector vec1 = new Vector(new Spherical(ra1, dec1, 1.0));
            Vector vec2 = RMatrix.Precess06(tdb1) * body1.Barycentric(tdb1).Position;
            double sepAngle = Angle.RadiansToDegrees(vec1.SeparationAngle(vec2)) * 3600.0;
            assert(sepAngle < 1.0);

            // heliocentric position should be nearly identical to barycentric
            vec1 = RMatrix.Precess06(tdb1) * body1.Heliocentric(tdb1).Position;
            sepAngle = Angle.RadiansToDegrees(vec1.SeparationAngle(vec2)) * 3600.0;
            assert(sepAngle < .1);

            // MICA no longer gives geocentric geometric position, so use Tinyac
            tdb1 = new JulianDate(2016, 7, 1, 0);
            // geocentric geometric position (ICRS)
            ra1 = Angle.DegreesToRadians(293.09789);
            dec1 = Angle.DegreesToRadians(69.65325);
            vec1 = new Vector(new Spherical(ra1, dec1));
            vec2 = body1.GeocentricGeometric(tdb1).Position;
            sepAngle = Angle.RadiansToDegrees(vec1.SeparationAngle(vec2)) * 3600.0;
            assert(sepAngle < .05);

            // Verify the previous geocentric geometric place is practically equal to astrometric
            vec1 = body1.GeocentricAstrometric(tdb1);
            sepAngle = Angle.RadiansToDegrees(vec1.SeparationAngle(vec2)) * 3600.0;
            assert(sepAngle < .001);

            // Compare with Mica 2.2.2.
            tdb1 = new JulianDate(2009, 7, 1, 0);
            ra1 = Angle.HmsToRad(19, 32, 22.690);
            dec1 = Angle.DmsToRad(69, 39, 23.93);
            vec1 = new Vector(new Spherical(ra1, dec1));
            vec2 = body1.GeocentricAstrometric(tdb1);
            sepAngle = Angle.RadiansToDegrees(vec1.SeparationAngle(vec2)) * 3600.0;

            // rotation matrix, CRS to true equator & equinox
            RMatrix crsToTrue = RMatrix.GcrsToTrue06b(tdb1);

            // geocentric apparent position (true RA/dec of date) compared to MICA 2.2.2
            ra1 = Angle.HmsToRad(19, 32, 23.756);
            dec1 = Angle.DmsToRad(69, 40, 32.58);
            vec1 = new Vector(new Spherical(ra1, dec1));
            vec2 = crsToTrue * body1.GeocentricApparent(tdb1);
            sepAngle = Angle.RadiansToDegrees(vec1.SeparationAngle(vec2)) * 3600.0;
            assert(sepAngle < .05);

            // topocenter (Griffith Obs.) position and velocity in the GCRS, assuming zero delta T
            // and zero polar motion
            Topocenter gp1 = new Topocenter(
                Angle.DegreesToRadians(34.0 + (7.0 + 06.0 / 60.0) / 60.0),
                Angle.DegreesToRadians(-(118.0 + (17.0 + 54.0 / 60.0) / 60.0)), 357);
            double[,] rotArr = new double[3, 3];
            Sofa.C2t06a(tdb1.DatePart, tdb1.TimePart, tdb1.DatePart, tdb1.TimePart,
                0.0, 0.0, rotArr);
            PVVector observerPV = gp1.ToPVVector(new RMatrix(rotArr).Transpose());

            // topocentric geometric -- to the accuracy of the Hipparcos catalog, this is
            // identical to geocentric geometric position already checked, with allowance for
            // topocenter velocity
            PVVector pv1 = body1.GeocentricGeometric(tdb1) - observerPV;
            PVVector pv2 = body1.TopocentricGeometric(tdb1, observerPV);

            pv1.EqualsApprox(pv2, 1e-6 * pv1.Position.Modulus(), 1e-6 * pv1.Velocity.Modulus());

            // topocentric astrometric is identical to topocentric geometric position
            pv1 = body1.TopocentricGeometric(tdb1, observerPV);
            vec2 = body1.TopocentricAstrometric(tdb1, observerPV.Position);
            sepAngle = Angle.RadiansToDegrees(pv1.Position.SeparationAngle(vec2)) * 3600.0;
            assert(sepAngle < .05);

            //  topocentric apparent (equator of date), per MICA 2.2.2
            body1 = new HipparcosStar(296.06896508, 37.35426967, 11.70,
                72.58, 35.14, eph, "15 Cyg", 97118);
            ra1 = Angle.HoursToRadians(19.7442484);
            dec1 = Angle.DegreesToRadians(37.376587);
            vec1 = new Vector(new Spherical(ra1, dec1));
            vec2 = crsToTrue * body1.TopocentricApparent(tdb1, observerPV);
            sepAngle = Angle.RadiansToDegrees(vec1.SeparationAngle(vec2)) * 3600.0;
            assert(sepAngle < .05);
        }
#endif  // end self test code
    }
}
