﻿// JplBody.cs
// A body in a JPL planetary ephemeris. Part of SofaJpl.
// Copyright 2018 by Paul S. Hirose

// revision history
#if false

2015-06-28  Created
2015-06-29  Finished, and self test passes. However, the tests (which were written for DE406)
won't all pass with DE422. Specifically, the astrometric place of Pluto is different enough to
fail the tests. You must use DE406
2015-08-30  Modified self test because Topocenter height unit changed to meters.
#endif


using System;


namespace HirosePS.SofaJpl {

/// <summary>
/// A body in a JPL planetary ephemeris.
/// </summary>
public class JplBody : SolarSystemBody {

    // enumeration that identifies body
    private JplEphemeris.Body _body;


    /// <summary>
    /// Create a JPL ephemeris body.
    /// </summary>
    /// <param name="body">identity of body</param>
    /// <param name="ephemeris">the JPL ephemeris</param>
    /// <param name="accuracy">light time correction accuracy (radians)</param>
    /// <remarks>The "body" can also be nutation, libration, etc. if present in the ephemeris.
    /// </remarks>
    public JplBody(JplEphemeris.Body body, JplEphemeris ephemeris, double accuracy) {
        // The Body enumeration includes nutation et al. Such "bodies" aren't allowed here.
        if (body > JplEphemeris.Body.Sun)
            throw new ArgumentOutOfRangeException("body", "must be an actual body");
        _body = body;
        Name = body.ToString();
        Ephemeris = ephemeris;

        IsSun = body == JplEphemeris.Body.Sun;
        IsEarth = body == JplEphemeris.Body.Earth;
        IsSolarSystemBody = true;
        Radius = SolarSystemBody.BodyRadius(Name);

        Accuracy = accuracy;
    }



    /// <summary>
    /// Barycentric geometric position and velocity.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>barycentric ICRS position (km) and velocity (km/day)</returns>
    public override PVVector Barycentric(JulianDate tdb) {
        return Ephemeris.Barycentric(tdb, (int)_body);
    }

    /// <summary>
    /// Heliocentric geometric position and velocity.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>heliocentric geometric ICRS position (km) and velocity (km/day)</returns>
    public override PVVector Heliocentric(JulianDate tdb) {
        return Ephemeris.Heliocentric(tdb, (int)_body);
    }

    /// <summary>
    /// Geocentric geometric position and velocity.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>geocentric geometric ICRS position (km) and velocity (km/day)</returns>
    public override PVVector GeocentricGeometric(JulianDate tdb) {
        return ((JplEphemeris)Ephemeris).Read(tdb, _body, JplEphemeris.Body.Earth);
    }


#if DEBUG        // begin self-test code

    /// <summary>
    /// Run the class self test.
    /// </summary>
    public static void Test() {
        Vector vec1, vec2;
        PVVector pv1;
        JulianDate tdb1, tdb2, tt1, ut1;

        tdb1 = new JulianDate(2008, 5, 30, 15.0 / 24.0);
        ut1 = tdb1 - new Duration((65.184 + 0.42781) / 86400.0);

        // desired accuracy
        double accuracy = Angle.DegreesToRadians(.1 / 3600.0);

        // The ephemeris. DE422 is sufficiently different with respect to Pluto that some tests
        // would fail. So it's good old DE406 for the tests.
        string ephName = "binp2000_2099.406";

        // open the ephemeris
        JplEphemeris jplEph = new JplEphemeris(ephName);

        // construct the test bodies
        JplBody sun = new JplBody(JplEphemeris.Body.Sun, jplEph, accuracy);
        JplBody earth = new JplBody(JplEphemeris.Body.Earth, jplEph, accuracy);
        JplBody moon = new JplBody(JplEphemeris.Body.Moon, jplEph, accuracy);
        JplBody pluto = new JplBody(JplEphemeris.Body.Pluto, jplEph, accuracy);

        // Test Name property.
        if (moon.Name != "Moon")
            throw new Exception("JPL body has wrong name.");

        // earthBaryPV()
        pv1 = jplEph.EarthBaryPV(tdb1);
        pv1.EqualsApprox(earth.Barycentric(tdb1), 0, 0, "EarthBaryPV");

        // SunBaryPV()
        pv1 = jplEph.SunBaryPV(tdb1);
        pv1.EqualsApprox(sun.Barycentric(tdb1), 0, 0, "SunBaryPV()");

        // Compare Barycentric() with a call to Read().
        pv1 = jplEph.Read(tdb1, JplEphemeris.Body.Moon, JplEphemeris.Body.SolarSysBary);
        moon.Barycentric(tdb1).EqualsApprox(pv1,
            accuracy * pv1.Position.Modulus(), accuracy * pv1.Velocity.Modulus());

        // Compare Heliocentric() with a direct call to Read().
        pv1 = jplEph.Read(tdb1, JplEphemeris.Body.Moon, JplEphemeris.Body.Sun);
        moon.Heliocentric(tdb1).EqualsApprox(pv1,
            accuracy * pv1.Position.Modulus(), accuracy * pv1.Velocity.Modulus());

        // geocentric geometric vs. MICA
        vec1 = new Vector(362500.877, 54784.593, 50823.740);
        vec2 = moon.GeocentricGeometric(tdb1).Position;
        vec2.EqualsApprox(vec1, accuracy * vec1.Modulus(), "geocentric geometric");

        // Geocentric astrometric (compared to Almanac). The distance in the Almanac
        // is the geometric distance of date.
        tdb2 = new JulianDate(2012, 1, -4, 0.0);
        vec1 = new Vector(new Spherical(Angle.HmsToRad(18, 29, 31.302),
            Angle.DmsToRad(-19, 20, 3.80), 33.131728 * 1.495978707e8));
        vec2 = pluto.GeocentricAstrometric(tdb2).Unit();
        vec2 *= pluto.GeocentricGeometric(tdb2).Position.Modulus();
        vec2.EqualsApprox(vec1, Angle.DmsToRad(0, 0, .02) * vec1.Modulus(),
            "geocentric astrometric");

        // rotation matrix, GCRS to true equator/equinox of date, IAU 2006/00B
        RMatrix gcrsToTrue = RMatrix.GcrsToTrue06b(tdb1);

        // geocentric apparent, true equatorial (compare to HORIZONS)
        vec1 = new Vector(new Spherical(Angle.DegreesToRadians(8.70473),
            Angle.DegreesToRadians(7.94005), .002473917462109 * jplEph.KmPerAu));
        vec2 = gcrsToTrue * moon.GeocentricApparent(tdb1);
        vec2.EqualsApprox(vec1, accuracy * vec1.Modulus(), "geocentric apparent");

        // topocenter position and velocity in the GCRS
        Topocenter gp1 = new Topocenter(
            Angle.DegreesToRadians(34.0), Angle.DegreesToRadians(-118.0), 400.0);
        RMatrix itrsToGcrs = RMatrix.GcrsToItrs06b(tdb1, ut1, 0, 0).Transpose();
        PVVector observer = gp1.ToPVVector(itrsToGcrs);

        // topocentric geometric - compare Pluto geometric and astrometric -
        // should be close - there's a decent check of topocentric astrometric
        // place later.
        vec1 = pluto.TopocentricAstrometric(tdb1, observer.Position);
        vec2 = pluto.TopocentricGeometric(tdb1, observer).Position;
        vec2.EqualsApprox(vec1, Angle.DmsToRad(0, 0, 5) * vec1.Modulus(),
            "Pluto topocentric geometric");

        // topocentric astrometric vs. JPL HORIZONS
        vec1 = new Vector(new Spherical(Angle.DegreesToRadians(270.24699),
            Angle.DegreesToRadians(-17.01810), 30.5246742778393 * jplEph.KmPerAu));
        pluto.TopocentricAstrometric(tdb1, observer.Position).EqualsApprox(
            vec1, accuracy * vec1.Modulus(), "topocentric astrometric");

        // topocentric apparent vs. JPL HORIZONS
        vec1 = new Vector(new Spherical(Angle.DegreesToRadians(8.90048),
            Angle.DegreesToRadians(7.49968), .002436651279780 * jplEph.KmPerAu));
        vec2 = gcrsToTrue * moon.TopocentricApparent(tdb1, observer);
        vec2.EqualsApprox(vec1, accuracy * vec1.Modulus(), "topocentric apparent");


        // 2016 Astronomical Almanac planetary reduction example p. B68. The ephemeris is DE430,
        // but DE422 is close enough.
        jplEph = new JplEphemeris("binp2000_2099.422");
        tt1 = new JulianDate(2016, 4, 17, 12, 0, 0) + Duration.FromSeconds(68);
        tdb1 = tt1 + Duration.FromSeconds(18.54e-9);
        Body venus = new JplBody(JplEphemeris.Body.Venus, jplEph, 1e-9);
        testSeparationAngle(venus.GeocentricApparent(tdb1),
            new Vector(.968541778, .237268519, .075036485), 2e-9,
            "testJplEphBody(): Almanac apparent place");
    }
#endif  // end self-test code
}       // end class JplBody
}       // end namespace HirosePS.SofaJpl
