﻿// JplEphemeris.cs
// Implements JPL ephemeris.
// Copyright 2018 by Paul S. Hirose

// revision history
#if false
2015-05-19 Created.
2015-06-19  TESTPO is good on DE406. One bug was my original Read() (the equivalent to JPL DPLEPH).
The rewrite not only works right, but also is cleaner code.

Another point of confusion is a comment in the Fortran code for INTERP:

T(1) IS DP FRACTIONAL TIME IN INTERVAL COVERED BY
COEFFICIENTS AT WHICH INTERPOLATION IS WANTED
(0 .LE. T(1) .LE. 1).  T(2) IS DP LENGTH OF WHOLE
INTERVAL IN INPUT TIME UNITS.

I understood the "interval" in the T(1) and T(2) comments to be the same thing. But actually the
former is what I call the subrecord duration (8 days for the Moon in DE406). The latter is what I
call the record duration (64 days in DE406).

2015-06-23  The Cut and Splice methods are done, as are the override methods, and the entire self
test is OK with DE422. I think the class is complete.

2015-06-26  Not quite complete. Added ToString(). Test OK.

2015-08-21  Added XML comments to each enumeration member to stop compiler warnings.

2017-08-14  An exception occurred in interpolate() if the first call after opening an ephemeris
had an interpolating factor of zero. I added a test for Chebychev polynomial count.

2018-02-21  In the processing of the header, added code to avert an exception if the last line of
constants is padded with dummy values. Test good.
#endif


using System;
using System.Collections.Generic;
using System.IO;


namespace HirosePS.SofaJpl {

    /// <summary>
    /// A class to get position and velocity of solar system bodies from a specified JPL ephemeris.
    /// </summary>
public class JplEphemeris : Ephemeris {

    // Filename of the ephemeris for ToString().
    private string _name;

    // Number of coefficients per record. This does not include the two Julian dates that
    // begin every record in the ASCII ephemeris.
    private int _coeffPerRecord;

    // First and last Julian date of the ephemeris. These are .5 less than the actual dates.
    private int _firstJd, _lastJd;

    // Number of days per record of coefficients.
    private int _daysPerRecord;

    // Constants from the ephemeris ASCII header.
    private Dictionary<string, double> _constants;

    // Constants needed by the class.
    private double _earthMass;      // Earth mass divided by Moon mass.

    // Format of the Chebychev polynomial coefficients in each record. This is the Group 1050 table
    // in the header, plus a fourth row which is the number of coordinates each "body" needs.
    // For example, a planet has 3 coordinates (x, y, z) but nutation has two (delta psi, delta
    // epsilon).
    private int[,] _chebyFormat;

    // Chebychev polynomials for position and rate.
    private double[] _chebyPositionPoly, _chebyVelocityPoly;

    // Number of Chebychev polynomials that have been computed.
    private int _chebyPositionPolyCount, _chebyVelocityPolyCount;

    // The full set of Chebychev coefficients from the binary file.
    private double[] _chebyCoeff;

    // Positions and velocities of the ephemeris bodies. Each column corresponds to one of the
    // ChebychevBody enumeration. The six rows are xyz and xyz rates.
    private double[,] _state;


    // Two enumerations of the solar system bodies are necessary because what's convenient for the
    // user isn't always what's in the ephemeris. For instance, the ephemeris has the Earth-Moon
    // barycenter but not the Earth itself.


    /// <summary>
    /// An enumeration of the celestial "bodies". (Some, such as nutation, are not actual bodies.)
    /// </summary>
    public enum Body {
        /// <summary>
        /// Mercury
        /// </summary>
        Mercury,
        /// <summary>
        /// Venus
        /// </summary>
        Venus,
        /// <summary>
        /// Earth
        /// </summary>
        Earth,
        /// <summary>
        /// Mars
        /// </summary>
        Mars,
        /// <summary>
        /// Jupiter
        /// </summary>
        Jupiter,
        /// <summary>
        /// Saturn
        /// </summary>
        Saturn,
        /// <summary>
        /// Uranus
        /// </summary>
        Uranus,
        /// <summary>
        /// Neptune
        /// </summary>
        Neptune,
        /// <summary>
        /// Pluto
        /// </summary>
        Pluto,
        /// <summary>
        /// Moon
        /// </summary>
        Moon,
        /// <summary>
        /// Sun
        /// </summary>
        Sun,
        /// <summary>
        /// solar system barycenter
        /// </summary>
        SolarSysBary,
        /// <summary>
        /// Earth-Moon barycenter
        /// </summary>
        EarthMoonBary,
        /// <summary>
        /// nutation angles: d(psi) and d(epsilon) (radians)
        /// </summary>
        Nutation,
        /// <summary>
        /// lunar libration angles: phi, theta, psi (radians)
        /// </summary>
        Libration,
        /// <summary>
        /// Lunar mantle angular velocity: omega_x,omega_y,omega_z
        /// </summary>
        LunarMantleAngularVelocity,
        /// <summary>
        /// TT-TDB
        /// </summary>
        TTMinusTdb
    };


    /// <summary>
    /// An enumeration of the bodies that have Chebychev coefficients in the ephemeris.
    /// </summary>
    /// <remarks>This enumeration indexes into the columns of the Group 1050 table in the ASCII
    /// header.
    /// </remarks>
    public enum ChebychevBody {
        /// <summary>
        /// Mercury
        /// </summary>
        Mercury,
        /// <summary>
        /// Venus
        /// </summary>
        Venus,
        /// <summary>
        /// Earth-Moon barycenter
        /// </summary>
        EarthMoonBary,
        /// <summary>
        /// Mars
        /// </summary>
        Mars,
        /// <summary>
        /// Jupiter
        /// </summary>
        Jupiter,
        /// <summary>
        /// Saturn
        /// </summary>
        Saturn,
        /// <summary>
        /// Uranus
        /// </summary>
        Uranus,
        /// <summary>
        /// Neptune
        /// </summary>
        Neptune,
        /// <summary>
        /// Pluto
        /// </summary>
        Pluto,
        /// <summary>
        /// Moon
        /// </summary>
        Moon,
        /// <summary>
        /// Sun
        /// </summary>
        Sun,
        /// <summary>
        /// nutation angles: d(psi) and d(epsilon) (radians)
        /// </summary>
        Nutation,
        /// <summary>
        /// lunar libration angles: phi, theta, psi (radians)
        /// </summary>
        Libration,
        /// <summary>
        /// Lunar mantle angular velocity: omega_x,omega_y,omega_z
        /// </summary>
        LunarMantleAngularVelocity,
        /// <summary>
        /// TT-TDB
        /// </summary>
        TTMinusTdb
    };



    /// <summary>
    /// Construct an instance, given the filename of a binary ephemeris.
    /// </summary>
    /// <param name="filename">binary ephemeris filename</param>
    /// <remarks>The binary file format is unique to SofaJpl and not compatible with other software.
    /// </remarks>
    public JplEphemeris(string filename) {
        using (FileStream fs = System.IO.File.OpenRead(filename))
        using (BinaryReader reader = new BinaryReader(fs)) {
            int numCoeff;       // number of Cheby coefficients in file

            // Create binary serializer.
            System.Runtime.Serialization.IFormatter formatter =
                new System.Runtime.Serialization.Formatters.Binary.BinaryFormatter();

            // Read the data.
            _coeffPerRecord = reader.ReadInt32();     // number of coefficients per record
            _firstJd = reader.ReadInt32();
            _lastJd = reader.ReadInt32();
            _daysPerRecord = reader.ReadInt32();
            _constants = (Dictionary<string, double>)formatter.Deserialize(fs);
            _chebyFormat = (int[,])formatter.Deserialize(fs);

            numCoeff = (int)(fs.Length - fs.Position) / sizeof(double);
            _chebyCoeff = new double[numCoeff];
            for (int i = 0; i < numCoeff; ++i)
                _chebyCoeff[i] = reader.ReadDouble();
        }

        // The columns of _state correspond to the columns of _chebyFormat.
        _state = new double[6, _chebyFormat.GetUpperBound(1) + 1];

        // Get max number of Cheby coefficients per coordinate. That sets the length of the
        // polynomial arrays.
        int max = 0;
        for (int i = _chebyFormat.GetUpperBound(1); i >= 0; --i)
            if (_chebyFormat[1, i] > max)
                max = _chebyFormat[1, i];
        _chebyPositionPoly = new double[max];
        _chebyVelocityPoly = new double[max];

        // Initialize the invariant Chebychev polynomials.
        _chebyPositionPoly[0] = 1.0;
        _chebyVelocityPoly[1] = 1.0;
        _chebyPositionPolyCount = 1;
        _chebyVelocityPolyCount = 2;

        // Some important constants.
        _earthMass = GetConstant("EMRAT");
        KmPerAu = _constants["AU"];

        // Remember ephemeris filename.
        _name = filename;
    }



    // Properties


    /// <summary>
    /// Get the first epoch of this ephemeris.
    /// </summary>
    public JulianDate FirstEpoch { get { return new JulianDate(_firstJd, .5); } }

    /// <summary>
    /// Get the last epoch of this ephemeris.
    /// </summary>
    public JulianDate LastEpoch { get { return new JulianDate(_lastJd, .5); } }

    /// <summary>
    /// Get a Dictionary collection which contains the header constants.
    /// </summary>
    public Dictionary<string, double> Constants { get { return _constants; } }

    /// <summary>
    /// Get the amount of time covered by each ephemeris data record.
    /// </summary>
    public Duration RecordDuration { get { return new Duration(_daysPerRecord); } }

    /// <summary>
    /// Get the number of coefficients per record.
    /// </summary>
    /// <remarks>This is two less than what the ASCII header says
    /// because I don't include the Julian dates of the record's time span.</remarks>
    public int CoefficientsPerRecord { get { return _coeffPerRecord; } }

    /// <summary>
    /// Return the array of Chebychev polynomial coefficients in this instance.
    /// </summary>
    /// <remarks>Does not include the two Julian dates that begin each record of the ASCII
    /// ephemerides.
    /// </remarks>
    public double[] Coefficients { get { return _chebyCoeff; } }

    /// <summary>
    /// Get the [4, n] integer array that indicates the format of the Chebychev coefficients of
    /// the n bodies in each record.
    /// </summary>
    public int[,] RecordFormat { get { return _chebyFormat; } }

    /// <summary>
    /// Get the state array. Its columns correspond to the ChebychevBody enumeration. The 6 rows
    /// are xyz position and xyz velocity.
    /// </summary>
    public double[,] State { get { return _state; } }



    // Instance methods


    /// <summary>
    /// Get barycentric geometric ICRS position and velocity of a specified body.
    /// </summary>
    /// <param name="tdb">epoch (TDB)</param>
    /// <param name="targetBody">from the Body enumeration</param>
    /// <returns>ICRS position and velocity (km and km/day)</returns>
    public override PVVector Barycentric(JulianDate tdb, int targetBody) {
        return Read(tdb, (Body)targetBody, Body.SolarSysBary);
    }

    /// <summary>
    /// Get heliocentric geometric ICRS position and velocity of a specified body.
    /// </summary>
    /// <param name="tdb">epoch (TDB)</param>
    /// <param name="targetBody">from the Body enumeration</param>
    /// <returns>ICRS position and velocity (km and km/day)</returns>
    public override PVVector Heliocentric(JulianDate tdb, int targetBody) {
        return Read(tdb, (Body)targetBody, Body.Sun);
    }

    /// <summary>
    /// Get geocentric geometric ICRS position and velocity of a specified body.
    /// </summary>
    /// <param name="tdb">epoch (TDB)</param>
    /// <param name="targetBody">from the Body enumeration</param>
    /// <returns>ICRS position and velocity (km and km/day)</returns>
    public override PVVector Geocentric(JulianDate tdb, int targetBody) {
        return Read(tdb, (Body)targetBody, Body.Earth);
    }

    /// <summary>
    /// Get barycentric geometric ICRS position and velocity of Earth.
    /// </summary>
    /// <param name="tdb">epoch (TDB)</param>
    /// <returns>ICRS position and velocity (km and km/day)</returns>
    public override PVVector EarthBaryPV(JulianDate tdb) {
        return Barycentric(tdb, (int)JplEphemeris.Body.Earth);
    }

    /// <summary>
    /// Get barycentric geometric ICRS position and velocity of the Sun.
    /// </summary>
    /// <param name="tdb">epoch (TDB)</param>
    /// <returns>ICRS position and velocity (km and km/day)</returns>
    public override PVVector SunBaryPV(JulianDate tdb) {
        return Barycentric(tdb, (int)JplEphemeris.Body.Sun);
    }

    
    /// <summary>
    /// Get filename of ephemeris.
    /// </summary>
    /// <returns>ephemeris filename</returns>
    public override string ToString() {
        return _name;
    }


    /// <summary>
    /// Get the position and velocity of a body from the ephemeris. This overload is for those
    /// "bodies" (such as nutation or libration) that are not points in the solar system.
    /// </summary>
    /// <param name="tdb">epoch (TDB)</param>
    /// <param name="target">target "body"</param>
    /// <returns>"position" and "velocity"</returns>
    /// <exception cref="ArgumentOutOfRangeException"> if epoch is out of range or body is not
    /// legal</exception>
    /// <remarks>Angles are radians and radians/day. Unused components in the result are zero.
    /// </remarks>
    public PVVector Read(JulianDate tdb, Body target) {
        // Commands for GetState(). Length equals number of bodies that have Cheby coefficients.
        int[] stateList = new int[_chebyFormat.GetUpperBound(1) + 1];
        PVVector result;

        switch (target) {
        case Body.Nutation:
            result = getState(tdb, ChebychevBody.Nutation, stateList);
            break;
        case Body.Libration:
            result = getState(tdb, ChebychevBody.Libration, stateList);
            break;
        case Body.LunarMantleAngularVelocity:
            result = getState(tdb, ChebychevBody.LunarMantleAngularVelocity, stateList);
            break;
        case Body.TTMinusTdb:
            result = getState(tdb, ChebychevBody.TTMinusTdb, stateList);
            break;
        default:
            throw new ArgumentOutOfRangeException("target",
                string.Format("{0} not allowed in this overload.", target));
        }
        return result;
    }


    /// <summary>
    /// Get position and velocity of a target body with respect to an origin body. Only for bodies
    /// that are solar system points. Not for "bodies" such as nutation or libration, which have
    /// their own overload.
    /// </summary>
    /// <param name="tdb">epoch (TDB)</param>
    /// <param name="target">target body</param>
    /// <param name="origin">origin body</param>
    /// <returns>position and velocity</returns>
    /// <exception cref="ArgumentOutOfRangeException"> if epoch is out of range or body is not
    /// legal.</exception>
    /// <remarks>Output units are km and km/day.</remarks>
    public PVVector Read(JulianDate tdb, Body target, Body origin) {

        // Commands for GetState(). Length equals number of bodies that have Cheby coefficients.
        int[] stateList = new int[_chebyFormat.GetUpperBound(1) + 1];
        PVVector targetPV, originPV;    // target & origin, barycentric position & velocity

        if (target >= Body.Nutation)
            throw new ArgumentOutOfRangeException("target", "Must be a solar system point.");
        if (origin >= Body.Nutation)
            throw new ArgumentOutOfRangeException("origin", "Must be a solar system point.");
        if (target == origin)
            return new PVVector();      // zero position and velocity

        // The geocentric Moon or selenocentric Earth require only one body from the ephemeris,
        // since it contains geocentric Moon coordinates.
        if (origin == Body.Earth && target == Body.Moon)
            return getState(tdb, ChebychevBody.Moon, stateList);
        if (origin == Body.Moon && target == Body.Earth)
            return getState(tdb, ChebychevBody.Moon, stateList).Negate();

        // Do the cases where one body is Moon or Earth. Since neither barycentric Moon nor
        // barycentric Earth is in the ephemeris, we must get them indirectly.
        // First, if either body is the target, make it the origin to simplify program logic.
        // The "negate" flag restores the correct sense before exit.
        bool negate = false;
        if (target == Body.Earth) {
            target = origin;
            origin = Body.Earth;
            negate = true;
        }
        if (target == Body.Moon) {
            target = origin;
            origin = Body.Moon;
            negate = true;
        }
        // If either if() above was true (they can't both be true), enter the if() below.
        if (origin == Body.Earth || origin == Body.Moon) {
            // Select required bodies and call the ephemeris. Ignore the SS bary or EM bary if they
            // are targets, since neither casts properly from Body to ChebychevBody, and anyway it
            // would be redundant. 
            if (target != Body.SolarSysBary && target != Body.EarthMoonBary)
                stateList[(int)target] = 2;         // 2 = get position & velocity
            stateList[(int)ChebychevBody.Moon] = 2;
            stateList[(int)ChebychevBody.EarthMoonBary] = 2;
            // Call the ephemeris, compute the origin and target vectors, return difference.
            GetState(tdb, stateList);
            PVVector embPV = StateToPVVector(ChebychevBody.EarthMoonBary);      // barycentric
            PVVector moonPV = StateToPVVector(ChebychevBody.Moon);      // geocentric
            originPV = embPV - moonPV / (1.0 + _earthMass);     // barycentric Earth PV
            if (origin == Body.Moon)
                originPV += moonPV;                             // barycentric Moon PV
            // Get vector from SS bary to target.
            if (target == Body.SolarSysBary)
                targetPV = new PVVector();        // null vector
            else if (target == Body.EarthMoonBary)
                targetPV = embPV;
            else
                targetPV = StateToPVVector((ChebychevBody)target);
            // Return vector from origin to target, with reversed sense if the bodies were swapped
            // earlier.
            if (negate)
                return originPV - targetPV;
            return targetPV - originPV;
        }

        // If execution reaches this point, the Earth and Moon are not involved. However, one body
        // may be the Earth-Moon barycenter. If so, change it to Earth. With that trick, all bodies
        // cast properly from Body to ChebychevBody.
        if (target == Body.EarthMoonBary)
            target = Body.Earth;
        if (origin == Body.EarthMoonBary)
            origin = Body.Earth;

        // If the solar system barycenter is the origin, we need only one body from the ephemeris.
        if (origin == Body.SolarSysBary)
                return getState(tdb, (ChebychevBody)target, stateList);
        // Less common but equally simple is the reverse case: SS barycenter is the target.
        if (target == Body.SolarSysBary)
                return getState(tdb, (ChebychevBody)origin, stateList).Negate();

        // The remaining cases require two bodies from the ephemeris.
        stateList[(int)target] = 2;     // 2 = get position and velocity
        stateList[(int)origin] = 2;
        GetState(tdb, stateList);
        targetPV = StateToPVVector((ChebychevBody)target);
        originPV = StateToPVVector((ChebychevBody)origin);
        return targetPV - originPV;
    }


    /// <summary>
    /// Copy position and velocity of a body in the state array to a PVVector.
    /// </summary>
    /// <param name="chebychevBody">the body</param>
    /// <returns>position and velocity</returns>
    /// <remarks>
    /// <para>The method is smart enough to know that TTMinusTdb, for instance, only requires
    /// two numbers to be copied and the rest set to zero.</para>
    /// <para>The state array provides the functionality of the original JPL code. However, it's
    /// usually simpler to call the Read() method.</para>
    /// </remarks>
    public PVVector StateToPVVector(ChebychevBody chebychevBody) {
        int column = (int)chebychevBody;
        PVVector result;

        switch (_chebyFormat[3, (int) chebychevBody]) {
        case 1:     // single component, e.g., TT-TDB
            result = new PVVector(
                new Vector(_state[0, column], 0, 0),
                new Vector(_state[3, column], 0, 0));
            break;
        case 2:     // two components, e.g., nutation
            result = new PVVector(
                new Vector(_state[0, column], _state[1, column], 0),
                new Vector(_state[3, column], _state[4, column], 0));
            break;
        default:    // three components, e.g., a planet
            result = new PVVector(
            new Vector(_state[0, column], _state[1, column], _state[2, column]),
            new Vector(_state[3, column], _state[4, column], _state[5, column]));
            break;
        }
        return result;
    }


    // "Lite" version of GetState(). Does only one body and returns result.
    private PVVector getState(JulianDate tdb, ChebychevBody body, int[] stateList) {
        stateList[(int)body] = 2;
        GetState(tdb, stateList);
        return StateToPVVector(body);
    }


    /// <summary>
    /// Put barycentric positions and velocities of the selected bodies in the state array.
    /// </summary>
    /// <param name="tdb">epoch (TDB)</param>
    /// <param name="stateList">Elements correspond to the bodies in the ChebychevBody enumeration.
    /// The value in each element indicates what to compute for that body: 0 = nothing, 1 =
    /// position, 2 = position and velocity.</param>
    /// <returns>Array whose columns correspond to the ChebychevBody enumeration and rows are
    /// xyz position and velocity in that order.</returns>
    /// <exception cref="Exception"> if stateList[] length is not equal to the number of bodies in
    /// the ephemeris (nutation etc. counts as a body).</exception>
    /// <exception cref="ArgumentOutOfRangeException"> if epoch is outside the ephemeris</exception>
    /// <remarks>GetState() is equivalent to STATE in the original JPL code. Generally, it's easier
    /// to call Read().</remarks>
    public double[,] GetState(JulianDate tdb, int[] stateList) {
        int maxIndex = _chebyFormat.GetUpperBound(1);
        if (maxIndex != stateList.Length - 1)
            throw new Exception("Number of columns in _chebyFormat[] not equal to length of stateList[]");
        for (int column = maxIndex; column >= 0; --column) {
            if (stateList[column] > 0) {
                interpolate(tdb, column, stateList);
            }
        }
        return _state;
    }


    // Interpolate position for the body designated by variable "column" (which corresponds to
    // the ChebychevBody enumeration). If stateList[column] is greater than 1, velocity is also
    // interpolated. Result is in the _state[] array, which is unchanged except for the column
    // of the requested body. If velocity is not requested, the last three rows of the
    // specified column in _state[] ) are zero.
    private void interpolate(JulianDate tdb, int column, int[] stateList) {
        double twoT, interpolatingFactor;
        int index;      // index into Chebychev coeff. array

        // number of Cheby polynomials required by this body
        int requiredPolys = _chebyFormat[1, column];

        // Get the interpolating factor and the index to the first Chebychev coefficient applicable
        // to this time and body.
        index = GetIndex(tdb, (ChebychevBody)column, out interpolatingFactor);
        twoT = 2.0 * interpolatingFactor;

        // Number of components (or "coordinates") for this body. E.g., Mercury has x, y, and z,
        // but nutation has only delta psi and delta epsilon.
        int numCoords = _chebyFormat[3, column];

        if (_chebyPositionPoly[1] != interpolatingFactor || _chebyPositionPolyCount < 2) {
            // Interpolating factor has changed since last call, or no polynomials have been
            // computed yet (except for the one initialized by the constructor). In either case,
            // make the required number of new position polynomials from scratch.
            _chebyPositionPoly[1] = interpolatingFactor;
            // We now have the first two Chebychev position polynomials. (The 0th is always 1, so
            // the constructor initialized it.) Recursively compute the rest.
            for (int i = 2; i < requiredPolys; ++i)
                _chebyPositionPoly[i] = twoT * _chebyPositionPoly[i - 1] -
                    _chebyPositionPoly[i - 2];
            _chebyPositionPolyCount = requiredPolys;
        } else {
            // Interpolating factor is the same as last time. However, the last body may not have
            // needed as many position polynomials as this body. Ensure there are enough.
            for (int i = _chebyPositionPolyCount; i < requiredPolys; ++i)
                _chebyPositionPoly[i] = twoT * _chebyPositionPoly[i - 1] -
                    _chebyPositionPoly[i - 2];
            if (requiredPolys > _chebyPositionPolyCount)
                _chebyPositionPolyCount = requiredPolys;
        }

        // There are enough polynomials. Interpolate for position.
        for (int row = 0; row < numCoords; ++row) {
            _state[row, column] = 0.0;      // sum will accumulate here
            // Begin with the highest degree polynomial and work downward.
            for (int n = requiredPolys - 1; n >= 0; --n)
                _state[row, column] += _chebyCoeff[index + row * requiredPolys + n] *
                    _chebyPositionPoly[n];
        }
        // There aren't necessarily three "position" components for each "body". For instance,
        // there are only two for nutation. Fill the unused positions with zeros.
        for (int row = numCoords; row < 3; ++row)
            _state[row, column] = 0.0;


        // Now check if velocity was requested.
        if (stateList[column] > 1) {
            // As with position, ensure velocity polynomials are up to date and go to a high
            // enough degree.
            if (_chebyVelocityPoly[2] != 2.0 * twoT || _chebyVelocityPolyCount < 3) {
                // Do the 2nd rate polynomial. (0th isn't used, 1st is constant and initialized in
                // the constructor.)
                _chebyVelocityPoly[2] = 2.0 * twoT;
                // Recursively compute the rest.
                for (int i = 3; i < requiredPolys; ++i)
                    _chebyVelocityPoly[i] = twoT * _chebyVelocityPoly[i - 1] +
                    2.0 * _chebyPositionPoly[i - 1] - _chebyVelocityPoly[i - 2];
                _chebyVelocityPolyCount = requiredPolys;
            }
            else {
                for (int i = _chebyVelocityPolyCount; i < requiredPolys; ++i)
                    _chebyVelocityPoly[i] = twoT * _chebyVelocityPoly[i - 1] +
                    2.0 * _chebyPositionPoly[i - 1] - _chebyVelocityPoly[i - 2];
                _chebyVelocityPolyCount = requiredPolys;
                if (requiredPolys > _chebyVelocityPolyCount)
                    _chebyVelocityPolyCount = requiredPolys;
            }

            // There are enough polynomials. Interpolate for velocity.
            int subrecords = _chebyFormat[2, column];
            double velocityFactor = (2.0 * subrecords) / _daysPerRecord;
            for (int row = 3; row < numCoords + 3; ++row) {
                _state[row, column] = 0.0;      // sum will accumulate here
                // Begin with the highest degree polynomial and work downward.
                for (int n = requiredPolys - 1; n > 0; --n)
                    _state[row, column] += _chebyCoeff[index + (row - 3) * requiredPolys + n] *
                        _chebyVelocityPoly[n];
                _state[row, column] *= velocityFactor;
            }
            // Zero the unused velocity components.
            for (int row = numCoords + 3; row < 6; ++row)
                _state[row, column] = 0.0;
        } else {
            // Velocity not requested; zero all velocity components.
            for (int row = 3; row < 6; ++row)
                _state[row, column] = 0.0;
        }    // end velocity code
    }   // end interpolate()


    /// <summary>
    /// Get the value of the named constant from the ephemeris.
    /// </summary>
    /// <param name="name">name of constant (case sensitive)</param>
    /// <returns>value of constant</returns>
    /// <exception cref="KeyNotFoundException"> if name is not found</exception>
    public double GetConstant(string name) {
        return _constants[name];
    }


    /// <summary>
    /// Get the index to the first applicable element in the array of Chebychev coefficients,
    /// and the Chebychev interpolating factor, given a body and time.
    /// </summary>
    /// <param name="tdb">epoch (TDB)</param>
    /// <param name="body">the body</param>
    /// <param name="chebychevTime">Chebychev interpolating factor in the interval [-1, +1]</param>
    /// <returns>index into the Chebychev coefficient array</returns>
    /// <exception cref="ArgumentOutOfRangeException"> if time is outside the file.</exception>
    public int GetIndex(JulianDate tdb, ChebychevBody body, out double chebychevTime) {
        int i, days, daysPerSubrecord, record, subRecord, coeffPerSubrecord, remainder;
        JulianDate beginSubrecord;

        double doubleTdb = tdb.ToDouble() - .5;
        if (doubleTdb < _firstJd || doubleTdb > _lastJd)
            throw new ArgumentOutOfRangeException("tdb", "is outside the span of the file.");
        // int epoch = (int)Math.Floor(tdb.ToDouble() - .5);
        int epoch = (int)Math.Floor(doubleTdb);

        i = (int)body;
        if (i > _chebyFormat.GetUpperBound(1) || _chebyFormat[1, i] == 0)
            throw new ArgumentOutOfRangeException("body", "is not in the ephemeris.");

        daysPerSubrecord = _daysPerRecord / _chebyFormat[2, i];
        coeffPerSubrecord = _chebyFormat[1, i] * _chebyFormat[3, i];
        days = epoch - _firstJd;            // whole days since file initial epoch
        record = Math.DivRem(days, _daysPerRecord, out remainder);
        subRecord = remainder / daysPerSubrecord;

        // JD when the applicable subrecord becomes effective. Remember, _firstJd is an integer and
        // .5 less than the true epoch, so re-introduce the missing half day.
        beginSubrecord = new JulianDate(_firstJd + record * _daysPerRecord +
            subRecord * daysPerSubrecord, .5);

        // Chebychev interpolating factor in the interval [-1, +1], where -1 is when the start
        // of the subrecord's effectivty and +1 is the end.
        chebychevTime = (tdb - beginSubrecord).ToDays() / daysPerSubrecord * 2.0 - 1.0;

        // I subtract 3 because 1) the ASCII header Group 1050 table assumes the first numbers in
        // each record are the begin and end JDs of that record (but I omit them), and 2) the
        // table assumes the Fortran convention where the first index of an array is 1, not 0.
        return record * _coeffPerRecord + _chebyFormat[0, i] - 3 + subRecord * coeffPerSubrecord;
    }


    /// <summary>
    /// Verify accuracy of the ephemeris routines with the tests in the JPL testpo file.
    /// </summary>
    /// <param name="testpo">filename of applicable JPL "testpo" file</param>
    /// <exception cref="Exception"> if any error exceeds the tolerance.</exception>
    /// <remarks>Error of every 100th test is displayed in a console window. Values are au,
    /// au/day, or radians, depending on the quantity tested.</remarks>
    public void TestEph(string testpo) {

        // Display one result per 100.
        const int sampleRate = 100;

        // max permissible error
        const double tolerance = 1e-13;

        // Open the file of test cases.
        using (StreamReader sr = new StreamReader(testpo)) {
            double[,] pvArray;
            JulianDate librationEpoch = new JulianDate(2440400.5, 0);
            JulianDate jd1;
            const double daysPerJulianYear = 365.25;
            double computedValue, correctValue, error;
            int target, origin, component, row, column;
            string s;
            jplFile file = new jplFile(sr);

            // Move to first data line.
            while (file.GetNextToken() != "EOT")
                file.GetNextLine();

            // Process the test file one line at a time.

            int lineCount = 0;      // triggers status report every 100 lines

            while (!file.EOF()) {
                // Ignore the DE number and calendar date.
                file.GetNextToken(); file.GetNextToken();
                s = file.GetNextToken();
                jd1 = new JulianDate(double.Parse(s), 0);
                if (jd1 < FirstEpoch || jd1 >= LastEpoch) {
                    file.GetNextLine();     // discard this line
                    continue;
                }
                target = file.GetNextInt() - 1;
                origin = file.GetNextInt() - 1;
                component = file.GetNextInt();
                correctValue = file.GetNextDouble();  // file values are au and au/day, or radians

                // Get position and velocity from the applicable ephemeris method. For solar system
                // points, convert from km and km/day to au and au/day.
                if ((Body)target < Body.Nutation)
                    pvArray = (Read(jd1, (Body)target, (Body)origin) / KmPerAu).ToArray();
                else
                    pvArray = Read(jd1, (Body)target).ToArray();

                // Select the component to test in the PV array; compute its error.
                if (target != (int)Body.Nutation) {
                    if (component < 4) {
                        row = 0;    // position
                        column = component - 1;
                    } else {
                        row = 1;    // velocity
                        column = component - 4;
                    }
                } else {    // nutation has its own format
                    if (component < 3) {
                        row = 0;    // angle
                        column = component - 1;
                    } else {
                        row = 1;    // angle rate
                        column = component - 3;
                    }
                }
                computedValue = pvArray[row, column];
                error = Math.Abs(computedValue - correctValue);

                // Because rapid error accumulation in libration psi is normal, apply a scale factor
                // as a function of time. (100 radians per year)
                if (target == (int)Body.Libration || component == 3) {
                    double daysSinceLibrationEpoch = (jd1 - librationEpoch).ToDays();
                    error /= (1.0 + 100.0 * Math.Abs(daysSinceLibrationEpoch / daysPerJulianYear));
                }

                // Report result.
                if (error > tolerance)
                    throw new Exception(string.Format(
                        "TESTPO failed at {0:f1}: target {1}, origin {2}, component {3}, " +
                        "error {4:e2}",
                        jd1.DatePart, (Body)target, (Body)origin, component, error));
                else
                    if (lineCount == 0) {
                        Console.WriteLine("{0:f1}: target {1}, origin {2}, component {3}, " +
                            "error {4:e2}",
                            jd1.DatePart, (Body)target, (Body)origin, component, error);
                        lineCount = sampleRate - 1;
                    }
                    else
                        --lineCount;
            }       // end TESTPO processing loop
        }       // end using statement for TESTPO stream reader
    }       // end TestEph() method



    // static methods


    /// <summary>
    /// Create a JPL binary ephemeris file from ASCII header and ephemeris files.
    /// </summary>
    /// <param name="header">ASCII header filename</param>
    /// <param name="asciiEphemeris">ASCII ephemeris filename</param>
    /// <param name="binaryEpemeris">binary ephemeris filename</param>
    /// <exception cref="Exception"> if any error occurs.</exception>
    /// <remarks>The binary file format is unique to SofaJpl and not compatible with other software.
    /// </remarks>
    public static void AsciiToBinary(string header, string asciiEphemeris,
        string binaryEpemeris) {

        Dictionary<string, double> constants;
        string lines;
        int[,] group1050;
        int numConstants;       // number of constants
        int coeffPerRecord;     // number of coefficients per record
        int numBodies;          // number of "bodies" (columns in Group 1050)
        int daysPerRecord;      // number of days per coefficient record
        long jdPosition;        // file position where first & last JD are written
        int headerFirstJd, headerLastJd;     // ephemeris time span (from header, minus .5)
        jplFile file1, file2;     // ASCII file parsers
        FileStream fs;
        System.Runtime.Serialization.IFormatter formatter;

        // Initialize header file parser.
        lines = System.IO.File.ReadAllText(header);
        file1 = new jplFile(new StringReader(lines));
        file2 = new jplFile(new StringReader(lines));

        // Get number of Chebychev coefficients per record. I don't count the two Julian dates that
        // begin each record.
        if (file1.TokenCount < 4 || !file1.Token(2).Equals("NCOEFF=",
            StringComparison.InvariantCultureIgnoreCase))
            throw new Exception("First line of header doesn't contain NCOEFF value.");
        coeffPerRecord = int.Parse(file1.Token(3)) - 2;

        // Find Group 1030 (ephemeris time span and days per record).
        file1.FindGroup("1030");
        headerFirstJd = (int)Math.Floor(file1.GetNextDouble() - .5);
        headerLastJd  = (int)Math.Floor(file1.GetNextDouble() - .5);
        daysPerRecord = (int)file1.GetNextDouble();

        // Move file1 to GROUP 1040 (names of the constants), get number of constant names.
        file1.FindGroup("1040");
        numConstants = int.Parse(file1.GetNextToken());

        // Move file2 to Group 1041 (values of the constants), verify number of values.
        file2.FindGroup("1041");
        if (numConstants != int.Parse(file2.GetNextToken()))
            throw new Exception("Item counts for Groups 1040 and 1041 are different.");

        // Now run file1 and file2 in parallel to get (name, value) pairs from the header and put
        // them in a Dictionary collection.
        string str1;
        constants = new Dictionary<string, double>(numConstants);
        for (int i = 0; i < numConstants; ++i) {
            str1 = file1.GetNextToken();
            constants.Add(str1, file2.GetNextDouble());
        }

        // If the number of constants is not a multiple of 3, the last data line is padded with
        // dummy values. Move past them. Harmless if there's no padding.
        file2.GetNextLine();

        // Verify the number of names and number of values equal the expected count.
        if (
            !file1.GetNextToken().Equals("GROUP", StringComparison.InvariantCultureIgnoreCase) ||
            !file1.GetNextToken().Equals("1041",  StringComparison.InvariantCultureIgnoreCase) ||
            !file2.GetNextToken().Equals("GROUP", StringComparison.InvariantCultureIgnoreCase) ||
            !file2.GetNextToken().Equals("1050",  StringComparison.InvariantCultureIgnoreCase))
            throw new Exception(string.Format("Expected {0:d} items in Group 1040 and 1041.",
                numConstants));

        // At this point the file2 token array contains the Group 1050 title. Force it to the next
        // non-blank line.
        file2.GetNextLine();

        // Parse the Group 1050 table.
        numBodies = file2.TokenCount;
        group1050 = new int[4, numBodies];
        for (int row = 0; row < 3; ++row)
            for (int column = 0; column < numBodies; ++column)
                group1050[row, column] = int.Parse(file2.GetNextToken());

        // I add a fourth row for the number of "coordinates" applicable to each "body". For
        // example, Mercury has 3 coordinates, but nutation has 2, and TT-TDB has only 1.
        for (int column = 0; column < numBodies; ++column) {
            switch (column) {
            case (int)ChebychevBody.TTMinusTdb:
                group1050[3, (int)ChebychevBody.TTMinusTdb] = 1;
                break;
            case (int)ChebychevBody.Nutation:
                group1050[3, (int)ChebychevBody.Nutation] = 2;
                break;
            default:
                group1050[3, column] = 3;
                break;
            }
        }

        // Create binary serializer.
        formatter = new System.Runtime.Serialization.Formatters.Binary.BinaryFormatter();

        // Open the ASCII ephemeris, create the output file and a binary writer.
        using (StreamReader sr = new StreamReader(asciiEphemeris))
        using (fs = new FileStream(binaryEpemeris, FileMode.Create))
        using (BinaryWriter bw = new BinaryWriter(fs)) {
            // first and last JD of ephemeris and record
            int ephFirstJd, recordFirstJd, recordLastJd;
            int oldRecordNumber, newRecordNumber;
            bool firstRecord;

            // Write number of coefficients per record.
            bw.Write(coeffPerRecord);

            // First and last epoch of ephemeris, and days per record. Initially, the first two are
            // dummy values. Later they're changed to the time span of the binary file.
            jdPosition = fs.Position;
            bw.Write(0);      // first epoch
            bw.Write(0);      // last epoch
            bw.Write(daysPerRecord);

            // Write the constants and the table that contains the format of the ephemeris data.
            formatter.Serialize(fs, constants);
            formatter.Serialize(fs, group1050);

            // Preceding data come from the header. Now prepare to read the ephemeris itself.

            // Initialize the parser to the start of the ASCII ephemeris.
            file1 = new jplFile(sr);

            // This do-loop runs once per data record until end of file.
            firstRecord = true;
            oldRecordNumber = 0;
            recordLastJd = ephFirstJd = 0;      // prevents compiler warning
            do {
                newRecordNumber = int.Parse(file1.GetNextToken());
                if (newRecordNumber != oldRecordNumber + 1)
                    throw new Exception("record " + newRecordNumber + " is not in sequence.");
                oldRecordNumber = newRecordNumber;
                file1.GetNextLine();      // discard rest of line; coeff. per record known already

                // Verify contiguous JDs in adjacent records, and correct time span.
                if (firstRecord) {       // only for first record in file
                    firstRecord = false;
                    recordFirstJd = (int)Math.Floor(file1.GetNextDouble() - .5);
                    recordLastJd  = (int)Math.Floor(file1.GetNextDouble() - .5);
                    ephFirstJd = recordFirstJd;
                } else {
                    recordFirstJd = (int)Math.Floor(file1.GetNextDouble() - .5);
                    if (recordFirstJd != recordLastJd)
                        throw new Exception("First JD in record " + newRecordNumber +
                            " doesn't match last JD in previous record.");
                    recordLastJd = (int)Math.Floor(file1.GetNextDouble() - .5);
                }
                if (recordLastJd - recordFirstJd != daysPerRecord)
                    throw new Exception(string.Format(
                        "record {0:d} does not span {1:d} days.", newRecordNumber, daysPerRecord));

                // This for-loop runs once per Chebychev coefficient in the record.
                for (int i = coeffPerRecord; i > 0; --i)
                    bw.Write(file1.GetNextDouble());
                // Bypass any padding zeros in the last line of the record.
                file1.GetNextLine();
            } while (!file1.EOF());    // end do-loop which runs once per record.

            // Verify the time span of the binary file is within the span of the ephemeris
            // (specified in the header).
            if (ephFirstJd < headerFirstJd || recordLastJd > headerLastJd)
                throw new Exception("Time span of the ASCII ephemeris not consistent with header");

            // Go back and update the first and last JDs to the actual time span of this file.
            fs.Position = jdPosition;
            bw.Write(ephFirstJd);
            bw.Write(recordLastJd);
        }       // end all using statements; files are closed
    }


    /// <summary>
    /// Make a new ephemeris which covers a shorter time span than the original.
    /// </summary>
    /// <param name="inputEph">original binary ephemeris</param>
    /// <param name="begin">first epoch of the output ephemeris</param>
    /// <param name="end">last epoch of the output ephemeris</param>
    /// <param name="outputEph">name of output ephemeris</param>
    /// <remarks>The output covers at least the requested time, and usually a little since
    /// a whole number of records must be copied. If a file with the output name already exists, it
    /// will be deleted.</remarks>
    public static void Cut(string inputEph, JulianDate begin, JulianDate end, string outputEph) {

        if ((begin - end) >= new Duration())
            throw new Exception("End epoch must be after begin epoch.");

        // Open the files. Create the binary reader & writer.
        using (FileStream inputFs = System.IO.File.OpenRead(inputEph))
        using (BinaryReader reader = new BinaryReader(inputFs))
        using (FileStream outputFs = new FileStream(outputEph, FileMode.Create))
        using (BinaryWriter writer = new BinaryWriter(outputFs)) {

            // Create binary serializer.
            System.Runtime.Serialization.IFormatter formatter =
                new System.Runtime.Serialization.Formatters.Binary.BinaryFormatter();

            // Read header data.
            int coeffPerRecord = reader.ReadInt32();
            int firstJd = reader.ReadInt32();
            int lastJd = reader.ReadInt32();
            int daysPerRecord = reader.ReadInt32();
            Dictionary<string, double> constants =
                (Dictionary<string, double>)formatter.Deserialize(inputFs);
            int[,] chebyFormat = (int[,])formatter.Deserialize(inputFs);

            // Verify the requested begin and end times are OK. Remember, the first and last
            // JD from the header are integers and .5 day less than the true values. I convert the
            // parameters to variables which follow the same convention.
            double beginJD = begin.ToDouble() - .5;
            double endJD = end.ToDouble() - .5;
            if (beginJD < firstJd)
                throw new ArgumentOutOfRangeException("begin", "is outside the span of the file.");
            if (endJD > lastJd)
                throw new ArgumentOutOfRangeException("end", "is outside the span of the file.");

            // Perpare to copy the applicable Chebychev coefficients for the requested time span.
            // First, calculate whole days since input file first epoch.
            int beginDays = (int)Math.Floor(beginJD) - firstJd;
            int endDays = (int)Math.Floor(endJD) - firstJd;

            // Calculate the numbers of the records that contain the begin and end epochs.
            int beginRecord = beginDays / daysPerRecord;
            int endRecord   = endDays   / daysPerRecord;

            // Write the header data to output file, with the first and last JD are corrected to
            // the new time span of the ephemeris.
            writer.Write(coeffPerRecord);
            writer.Write(firstJd + beginRecord * daysPerRecord);
            writer.Write(firstJd + (endRecord + 1) * daysPerRecord);
            writer.Write(daysPerRecord);
            formatter.Serialize(outputFs, constants);
            formatter.Serialize(outputFs, chebyFormat);

            // Input file is at the first (0th) record of Cheby coefficients. Go to first record to
            // copy, then copy all applicable coefficients. This works even if one record is
            // sufficient for the requested time span.
            inputFs.Seek(beginRecord * coeffPerRecord * sizeof(double), SeekOrigin.Current);
            for (int i = endRecord - beginRecord; i >= 0; --i)
                for (int j = coeffPerRecord; j > 0; --j)
                    writer.Write(reader.ReadDouble());
        }
    }


    /// <summary>
    /// Combine two JPL binary ephemerides (whose time spans must abut or overlap) into one.
    /// </summary>
    /// <param name="inputEph1">filename of first binary input ephemeris</param>
    /// <param name="inputEph2">filename of second binary input ephemeris</param>
    /// <param name="outputEph">filename of output binary ephemeris</param>
    /// <remarks>The order of the input files doesn't matter. They must have the same DE number.
    /// There must be no time gap between them, but overlap is OK. (This includes the case where one
    /// file is entirely within the time span of the other.) The output file has the combined time
    /// span of the inputs.</remarks>
    /// <exception cref="Exception"> if an error occurs.</exception>
    public static void Splice(string inputEph1, string inputEph2, string outputEph) {
        using (FileStream inputFs1 = System.IO.File.OpenRead(inputEph1))
        using (FileStream inputFs2 = System.IO.File.OpenRead(inputEph2))
        using (BinaryReader reader1 = new BinaryReader(inputFs1))
        using (BinaryReader reader2 = new BinaryReader(inputFs2)) {

            // Create binary serializer.
            System.Runtime.Serialization.IFormatter formatter =
                new System.Runtime.Serialization.Formatters.Binary.BinaryFormatter();

            // Read header data from eph 1.
            int coeffPerRecord1 = reader1.ReadInt32();
            int firstJd1 = reader1.ReadInt32();
            int lastJd1 = reader1.ReadInt32();
            int daysPerRecord1 = reader1.ReadInt32();
            Dictionary<string, double> constants1 =
                (Dictionary<string, double>)formatter.Deserialize(inputFs1);
            int[,] chebyFormat1 = (int[,])formatter.Deserialize(inputFs1);

            // Read header data from eph 2.
            int coeffPerRecord2 = reader2.ReadInt32();
            int firstJd2 = reader2.ReadInt32();
            int lastJd2 = reader2.ReadInt32();
            int daysPerRecord2 = reader2.ReadInt32();
            Dictionary<string, double> constants2 =
                (Dictionary<string, double>)formatter.Deserialize(inputFs2);
            int[,] chebyFormat2 = (int[,])formatter.Deserialize(inputFs2);

            // Verify same DE number and no time gap.
            if (constants1["DENUM"] != constants2["DENUM"])
                throw new Exception("Files have different DE numbers.");
            if ((firstJd1 <= firstJd2) && (lastJd1 < firstJd2) ||
                (firstJd2 <= firstJd1) && (lastJd2 < firstJd1)) {
                    throw new Exception("Time gap between input files.");
            }

            // Open output file.
            using (FileStream outputFs = new FileStream(outputEph, FileMode.Create))
            using (BinaryWriter writer = new BinaryWriter(outputFs)) {
                int numCoeff, index;
                if (firstJd1 <= firstJd2) {     // file #1 is earlier than #2

                    // Write header data. Take care with the last JD, since it's possible file #2
                    // is entirely within the time span of #1. I don't consider that an error.
                    writer.Write(coeffPerRecord1);
                    writer.Write(firstJd1);
                    writer.Write((lastJd2 > lastJd1) ? lastJd2 : lastJd1);
                    writer.Write(daysPerRecord1);
                    formatter.Serialize(outputFs, constants1);
                    formatter.Serialize(outputFs, chebyFormat1);

                    // Write all coefficients of file #1.
                    numCoeff = (int)(inputFs1.Length - inputFs1.Position) / sizeof(double);
                    for (int i = numCoeff; i > 0; --i)
                        writer.Write(reader1.ReadDouble());

                    // File #2 is positioned at its first coefficient. However, we don't necessarily
                    // begin reading there since the files can overlap in time.
                    if (lastJd2 <= lastJd1)
                        return;
                    index = (lastJd1 - firstJd2) / daysPerRecord2 * coeffPerRecord2 *
                        sizeof(double);
                    inputFs2.Seek(index, SeekOrigin.Current);

                    // Write coefficients from file #2.
                    numCoeff = (int)(inputFs2.Length - inputFs2.Position) / sizeof(double);
                    for (int i = numCoeff; i > 0; --i)
                        writer.Write(reader2.ReadDouble());
                }
                else {    // file #2 is earlier than #1
                    writer.Write(coeffPerRecord2);
                    writer.Write(firstJd2);
                    writer.Write((lastJd1 > lastJd2) ? lastJd1 : lastJd2);
                    writer.Write(daysPerRecord2);
                    formatter.Serialize(outputFs, constants2);
                    formatter.Serialize(outputFs, chebyFormat2);

                    // Write all coefficients of file #2.
                    numCoeff = (int)(inputFs2.Length - inputFs2.Position) / sizeof(double);
                    for (int i = numCoeff; i > 0; --i)
                        writer.Write(reader2.ReadDouble());

                    // Position file #1 correctly.
                    if (lastJd1 <= lastJd2)
                        return;
                    index = (lastJd2 - firstJd1) / daysPerRecord1 * coeffPerRecord1 *
                        sizeof(double);
                    inputFs1.Seek(index, SeekOrigin.Current);

                    // Write coefficients from file #1.
                    numCoeff = (int)(inputFs1.Length - inputFs1.Position) / sizeof(double);
                    for (int i = numCoeff; i > 0; --i)
                        writer.Write(reader1.ReadDouble());
                }    // end the file #2 earlier than #1 branch
            }
        }
    }



    // self-test code
#if DEBUG

    /// <summary>
    /// Run the class self test.
    /// </summary>
    /// <param name="asciiToBinary">if true, build the binary ephemerides required by the tests
    /// </param>
    public static void Test(bool asciiToBinary) {
        JplEphemeris jplEph;
        int[,] chebyFormat;
        Dictionary<string, double> constants;
        JulianDate jd1;
        Duration zeroDays = new Duration();
        double[] coefficients;
        PVVector pv1, pv2;
        string jplDir = @"C:\Users\Stan\Documents\astro\jpl\";


        // Build 2000-2099 DE422 binary. Run TestEph() on it. This will test nutation and libration
        // angles, which are not in DE406.
        if(asciiToBinary)
            AsciiToBinary(jplDir + "header.422", jplDir + "ascp2000.422", "binp2000_2099.422");
        jplEph = new JplEphemeris("binp2000_2099.422");
        jplEph.TestEph(jplDir + "testpo.422");

        // Investigate the exception that is thrown at 2017-01-01 0 h with DE 422.
        jplEph = new JplEphemeris("binp2000_2099.422");
        jd1 = new JulianDate(2017, 1, 1, 0, 0, 0);
        jplEph.Read(jd1, Body.Moon, Body.SolarSysBary);

        // Ensure I have fixed the over-zealous error checking that throws an exception when the
        // last data line in the header is padded with zeroes to make the number of constants a
        // multiple of 3.
        if (asciiToBinary)
            AsciiToBinary(jplDir + "header572.430", jplDir + "ascp1950.430", "binp1950_2049.430");
       

        // Some self tests of other classes rely on a DE406 ephemeris, so build one and use it for
        // the remaining checks.

        // Build DE406 ephemerides for the 20th and 21st centuries.
        if (asciiToBinary) {
            AsciiToBinary(jplDir + "header.406", jplDir + "ascp1900.406", "binp1900_1999.406");
            AsciiToBinary(jplDir + "header.406", jplDir + "ascp2000.406", "binp2000_2099.406");
        }

        // Cut them to 1960-1999 and 2000-2070 ephemerides, then splice them.
        if (asciiToBinary) {
            Cut("binp1900_1999.406", new JulianDate(1960, 1, 1, 0), new JulianDate(2000, 1, 1, 0),
                "binp1960_1999.406");
            Cut("binp2000_2099.406", new JulianDate(2000, 1, 1, 0), new JulianDate(2070, 1, 1, 0),
                "binp2000_2070.406");
            Splice("binp1960_1999.406", "binp2000_2070.406", "binp1960_2070.406");
        }

        // Open the spliced ephemeris. Test ToString().
        jplEph = new JplEphemeris("binp1960_2070.406");
        if (jplEph.ToString() != "binp1960_2070.406")
            throw new Exception("ToString() failed.");

        // properties

        if (jplEph.FirstEpoch != new JulianDate(2.436880500000000e+06, 0))
            throw new Exception("FirstEpoch property failed.");

        if (jplEph.LastEpoch != new JulianDate(2.477136500000000e+06, 0))
            throw new Exception("LastEpoch property failed.");

        constants = jplEph.Constants;
        if (constants["EMRAT"] != 0.813005600000000044e+02)
            throw new Exception("Constants dictionary failed.");

        if (jplEph.KmPerAu != 0.149597870691000015e+09)
            throw new Exception("KmPerAu property failed.");

        if (jplEph.RecordDuration != new Duration(64))
            throw new Exception("RecordDuration property failed.");

        if (jplEph.CoefficientsPerRecord != 726)
            throw new Exception("CoefficientsPerRecord property failed.");

        // To test the Coefficients property verify the last Chebychev coefficient in the ASCII
        // ephemeris is present.
        coefficients = jplEph.Coefficients;
        if (coefficients[(jplEph._lastJd - jplEph._firstJd) / jplEph._daysPerRecord *
            jplEph._coeffPerRecord - 1] != -2.305359337157142e-03)
            throw new Exception("Last Cheby coefficient was wrong.");

        chebyFormat = jplEph.RecordFormat;
        if (chebyFormat[3, (int)ChebychevBody.Moon] != 3)
            throw new Exception("RecordFormat property failed.");

        if (jplEph.State != jplEph._state)
            throw new Exception("State property failed.");


        // JPL TESTPO
        jplEph.TestEph(jplDir + "testpo.406");


        // The override methods.
        jd1 = new JulianDate(1989, 11, 29, 1, 0, 0);
        jplEph.Barycentric(jd1, (int)Body.Mercury).EqualsApprox(jplEph.Read(jd1, Body.Mercury,
            Body.SolarSysBary), 0, 0, "Barycentric() failed.");
        jplEph.Heliocentric(jd1, (int)Body.Mercury).EqualsApprox(jplEph.Read(jd1, Body.Mercury,
            Body.Sun), 0, 0, "Heliocentric() failed.");
        jplEph.Geocentric(jd1, (int)Body.Sun).EqualsApprox(jplEph.Read(jd1, Body.Sun,
            Body.Earth), 0, 0, "Geocentric() failed.");
        pv1 = jplEph.EarthBaryPV(jd1);
        pv2 = jplEph.Barycentric(jd1, (int)Body.Earth);
        pv1.EqualsApprox(pv2, 0, 0, "EarthBaryPV() failed.");
        pv1 = jplEph.SunBaryPV(jd1);
        pv2 = jplEph.Barycentric(jd1, (int)Body.Sun);
        pv1.EqualsApprox(pv2, 0, 0, "SunBaryPV() failed.");

    }       // end Test() method


    // Alternate method to get index to first applicable coefficient and the interpolating factor.
    private static int getIndex(JplEphemeris eph, JulianDate tdb, ChebychevBody body,
        out double interpFactor) {

        JulianDate recFirstEpoch, subrecFirstEpoch;
        Duration ephTime, recTime, subrecTime, subrecDuration;
        int column, index, rec, subrec, subrecLength, offsetInRec, coeffPerCoord, subrecPerRec,
            coordPerSubrec;
        int[,] chebyFormat = eph.RecordFormat;

        column = (int)body;
        offsetInRec = chebyFormat[0, column] - 3;
        coeffPerCoord = chebyFormat[1, column];
        subrecPerRec = chebyFormat[2, column];
        coordPerSubrec = chebyFormat[3, column];

        ephTime = tdb - eph.FirstEpoch;         // time since eph first epoch
        rec = (int)Math.Floor(ephTime / eph.RecordDuration);     // number of applicable record
        recFirstEpoch = eph.FirstEpoch + rec * eph.RecordDuration;
        recTime = tdb - recFirstEpoch;      // time since first epoch of record
        subrecDuration = eph.RecordDuration / subrecPerRec;
        subrec = (int)Math.Floor(recTime / subrecDuration);
        subrecFirstEpoch = recFirstEpoch + subrec * subrecDuration;
        subrecTime = tdb - subrecFirstEpoch;
        subrecLength = coeffPerCoord * coordPerSubrec;

        index = rec * eph.CoefficientsPerRecord + offsetInRec + subrec * subrecLength;
        interpFactor = subrecTime / subrecDuration * 2.0 - 1.0;

        return index;
    }

#endif    // end self-test code



    // Helper class to extract lines and space delimited tokens from ASCII ephemeris files.
    private class jplFile {
        // The ASCII file.
        private TextReader _tr;
        // Holds one line separated into space delimited tokens, one token per array element.
        private string[] _tokens;
        // Index of next token to fetch.
        private int _tokenIndex;

        /// <summary>
        /// Initialize to first non-blank line of file.
        /// </summary>
        /// <param name="tr">a TextReader instance</param>
        public jplFile(TextReader tr) {
            string s;
            while ((s = tr.ReadLine()) != null)
                if (!string.IsNullOrWhiteSpace(s))
                    break;
            if (s == null)
                throw new Exception("File is empty.");
            this._tr = tr;
            _tokens = s.Split((char[])null, StringSplitOptions.RemoveEmptyEntries);
            _tokenIndex = 0;
        }

        
        /// <summary>
        /// Discard the tokens from the most recent line, go to the next non-blank line and get a
        /// new set of space delimited tokens.
        /// </summary>
        /// <returns>number of tokens in the line, or 0 if end of file</returns>
        public int GetNextLine() {
            string s;
            while ((s = _tr.ReadLine()) != null)
                if (!string.IsNullOrWhiteSpace(s))
                    break;
            if (s == null) {
                _tokens = new string[] {};
                return 0;
            }
            _tokens = s.Split((char[])null, StringSplitOptions.RemoveEmptyEntries);
            _tokenIndex = 0;
            return _tokens.Length;
        }


        /// <summary>
        /// Find the specified group in a JPL header.
        /// </summary>
        /// <param name="groupNumber">group number</param>
        /// <exception cref="Exception"> if requested group is not found</exception>
        public void FindGroup(string groupNumber) {
            int i;
            while ((i = GetNextLine()) != 0) {
                if (_tokens.Length < 2)
                    continue;
                if (_tokens[0].Equals("GROUP", StringComparison.InvariantCultureIgnoreCase) &&
                    _tokens[1].Equals(groupNumber, StringComparison.InvariantCultureIgnoreCase))
                    break;
            }
            if (i == 0)
                throw new Exception("Group " + groupNumber + " not found.");
            GetNextLine();      // find next non-blank line
        }


        /// <summary>
        /// Get the token at the token pointer. If no more tokens in this line, get the first token
        /// in the next non-blank line.
        /// </summary>
        /// <returns>next token</returns>
        /// <exception cref="Exception"> if attempt to read past last token in file</exception>
        public string GetNextToken() {
            if (_tokenIndex >= _tokens.Length)
                if (GetNextLine() == 0)
                    throw new Exception("No more tokens in file.");
            return _tokens[_tokenIndex++];
        }


        /// <summary>
        /// Convert the next token to an int.
        /// </summary>
        /// <returns>token converted to int</returns>
        public int GetNextInt() {
            return int.Parse(GetNextToken());
        }


        /// <summary>
        /// Convert the next token to a double, including replacement of Fortran 'D' with 'E'.
        /// </summary>
        /// <returns>token converted to double</returns>
        /// <remarks>On exit the token pointer has been advanced one token, and the line
        /// pointer is on the line after the line that contained the tokens.</remarks>
        /// <exception cref="Exception"> if attempt to read past last token in file</exception>
        public double GetNextDouble() {
            string s = GetNextToken();
            s = s.Replace('D', 'E');
            s = s.Replace('d', 'e');
            return double.Parse(s);
        }


        /// <summary>
        /// Get number of elements in the most recently tokenized line. This includes tokens
        /// that have been fetched already.
        /// </summary>
        public int TokenCount { get { return _tokens.Length; } }


        /// <summary>
        /// Return the specified element of the tokens array.
        /// </summary>
        /// <param name="index">0 = first token</param>
        /// <returns>the specified token</returns>
        /// <remarks>This method has no side effects.</remarks>
        /// <exception cref="ArgumentOutOfRangeException"> if index is out of bounds</exception>
        public string Token(int index) {
            if (index >= _tokens.Length)
                throw new ArgumentOutOfRangeException("index", "out of bounds");
            return _tokens[index];
        }


        /// <summary>
        /// Detect that the line pointer and token pointer are both beyond the last element of
        /// their arrays.
        /// </summary>
        /// <returns>true if no more data</returns>
        public bool EOF() {
            return _tokens.Length == 0;
        }

    }   // end class jplFile
}       // end class JplEphemeris
}       // end namespace SofaJpl
