// JulianDate.cs
// Julian date class.
// Copyright 2018 by Paul S. Hirose

// revision history
#if false
// 2007-10-27 Finished removing unsafe code.
// 2007-11-11 Improved ToDateTime() accuracy a tiny bit by adding .5 tick.
// 2007-11-15 Added a class self-test method.
// 2007-11-28 Fixed the way the self-test verifies exceptions are properly thrown.
// 2007-12-03 In accordance with Microsoft's recommendations, changed this to a class.
// (because the object is not smaller than 16 bytes)
// 2007-12-04 CompareTo(JulianDate) and related methods have been made null safe.
// 2008-01-14 Added an MJD method and the MJD base epoch.
// 2008-01-18 Added constructors taking sexagesimal time.
// 2008-02-10 Added J2000 base epoch.
// 2008-02-11 Added ToDouble().
// 2008-02-21 Added ToJ2000().
// 2008-04-29 Added FromBesselianEpoch(), FromJulianEpoch(), FromMjd(). Fixed mistake in
// TdbMinusTT() (distance was in meters, not km).
// 2008-05-13 Removed calls to C++ calendar and Julian date routines. That's now handled in-house.
// 2008-05-15 Retooled the above change a little bit.
// 2008-06-13 Added property to get J2000 base epoch as a double.
// 2008-06-18 Minor improvement to TdbMinusTT() XML comment.
// 2008-10-29 New tests in extreme past; all OK.
// 2008-12-09 Added constructor that takes a TimeFields.

2014-10-07 Began adapting this former SofaJpl class to my new SOFA DLL. Required changing upper
case to lower to match spelling in the SOFA DLL. Omitted ISO 8601 support.
2014-10-19 Commented out the TAI and GPS time conversions. These can be handled by the
Duration class. Much more thorough checks of day number method.
2014-10-21 Improvements to XML comments and self test. Added ToTimeFields(). Still need to
implement my idea to make the methods virtual.
2014-10-22 All methods and properties are virtual now. Good self test.
2014-11-05 Made DateToDayNumber() public and added public IsLeapYear(). These changes support
revisions to TimeFields class.
2015-06-26 Removed "virtual" keyword.
2015-07-27 Added methods to convert TT to TDB, TCG, TCB, and inverse.

2017-10-26 Added AtMidnight(), AfterMidnight(), and a ToTimeFields() 1-parameter overload.
2018-02-20 AfterMidnight() is a peculiar name. Change it to TimeOfDay().
2018-02-20 Constructors throw exception if fractional day or time of day is negative.
2018-04-20 Renamed AtMidnight() to AtZeroHours() since I thought the former might be ambiguous.
2018-08-06 The offsets (Gregorian and Julian) between my day number system and the Julian date are
computed at run time by the static constructor. Eliminates a couple magic numbers from the code.
#endif


using System;


namespace HirosePS.SofaJpl
{
    /// <summary>
    /// A Julian date.
    /// </summary>
    /// <remarks>
    /// <para>Instances of this class are immutable.</para>
    /// <para>Any time in the years -5,880,000 to 5,870,000 (approx.), Gregorian or Julian
    /// calendar, is legal and will be represented to nanosecond accuracy.</para>
    /// <para>The internal representation is normalized so one part is an integral value plus
    /// .5, and the other part is in the interval [0, 1). I.e., the parts correspond to
    /// date and time of day. Both parts are type Double.</para>
    /// </remarks>
    public class JulianDate : IComparable<JulianDate>, IEquatable<JulianDate>, IComparable
    {
        // The internal representation of a Julian date.

        // part1 is at midnight (ends in .5) and 0 <= part2 < 1
        private double part1;
        private double part2;


        private const double _secPerDay = 86400.0;


        // static constants

        /// <summary>
        /// Julian date 2400000.5, when the Modified Julian date equals zero.
        /// </summary>
        public static readonly JulianDate MjdBase;

        /// <summary>
        /// The J2000 base epoch: 2000 Jan 1 12h UT.
        /// </summary>
        public static readonly JulianDate J2000Base;


        // For conversion between calendar dates and Julian dates I use an integer day number which
        // is zero on 1 BC December 31 of the Julian or Gregoriann calendar, as applicable.
        // The variables below are the quantities which, when added to the day number, plus the
        // fraction of a day past UT midnight, yield the JD. Initialized by the static constructor.
        private static readonly double _gregorianOffset, _julianOffset;

        
        // constructors


        static JulianDate() {
            // Julian date at 2012 Jan 14 0h UT (Gregorian) and Jan 1 (Julian)
            double jd = 2455927.0 + 14.0 - .5;

            // The offsets between JD and my day number system, in both calendars.
            _gregorianOffset = jd - DateToDayNumber(2012, 1, 14, false);
            _julianOffset = jd - DateToDayNumber(2012, 1, 1, true);

            MjdBase = new JulianDate(2400000.5, 0.0);
            J2000Base = new JulianDate(2000, 1, 1, .5);
        }


        /// <summary>
        /// Make a copy of this JulianDate instance.
        /// </summary>
        /// <param name="jd">instance to copy</param>
        public JulianDate(JulianDate jd)
        {
            this.part1 = jd.part1;
            this.part2 = jd.part2;
        }


        /// <summary>
        /// Construct from a .NET Framework System.DateTime instance.
        /// </summary>
        /// <param name="dt">.NET System.DateTime</param>
        /// <remarks>The result is Greenwich time, even if the DateTime object was constructed with
        /// local time.</remarks>
        public JulianDate(DateTime dt)
        {
            // Obtain fractional days via a roundabout method in order to fully exploit the time
            // resolution of the DateTime instance.
            DateTime t = dt.ToUniversalTime();
            DateTime midnight = t.Date;         // the DateTime at 0 hours of date
            TimeSpan timeOfDay = t - midnight;
            double fracDays = (double)timeOfDay.Ticks / TimeSpan.TicksPerDay;

            calendarToJd(t.Year, t.Month, t.Day, fracDays, false);
        }


        /// <summary>
        /// Construct from a Gregorian or Julian calendar date and fractional days.
        /// </summary>
        /// <param name="year">Year: 0 = 1 BC, -1 = 2 BC, etc.</param>
        /// <param name="month">Month [1 - 12]</param>
        /// <param name="day">Day of month (any positive or negative integer)</param>
        /// <param name="fracDay">Fractional day (must be non-negative; &gt;=0 and &lt;1 for max
        /// resolution)</param>
        /// <param name="julianCalendar">true if Julian calendar date</param>
        /// <exception cref="ArgumentOutOfRangeException"> if the month is outside the interval
        /// [1, 12] or <paramref name="fracDay"/> is negative</exception>
        /// <remarks>
        /// <para>The algorithm itself has unlimited time span. In practice, the limits
        /// in the future and past are where a Julian date overflows an integer.</para>
        /// <para>Day of month is not restricted to the normal range. E.g., January 180 = 180th day
        /// of this year, and January -1 = December 30 in the previous year. However,
        /// <paramref name="fracDay"/> must be non-negative.</para>
        /// </remarks>
        public JulianDate(
            int year,
            int month,
            int day,
            double fracDay,
            bool julianCalendar)
        {
            calendarToJd(year, month, day, fracDay, julianCalendar);
        }


        /// <summary>
        /// Construct from a Gregorian calendar date and fractional days.
        /// </summary>
        /// <param name="year">Year: 0 = 1 BC, -1 = 2 BC, etc.</param>
        /// <param name="month">Month [1 - 12]</param>
        /// <param name="day">Day of month (any positive or negative integer)</param>
        /// <param name="fracDay">Fractional day (must be non-negative; &gt;=0 and &lt;1 for max
        /// resolution)</param>
        /// <exception cref="ArgumentOutOfRangeException"> if the month is outside the interval
        /// [1, 12] or <paramref name="fracDay"/> is negative</exception>
        /// <remarks>
        /// <para>The algorithm itself has unlimited time span. In practice, the limits
        /// in the future and past are where a Julian date overflows an integer.</para>
        /// <para>Day of month is not restricted to the normal range. E.g., January 180 = 180th day
        /// of this year, and January -1 = December 30 in the previous year. However,
        /// <paramref name="fracDay"/> must be non-negative.</para>
        /// </remarks>
        public JulianDate(int year, int month, int day, double fracDay)
            : this(year, month, day, fracDay, false) { }


        /// <summary>
        /// Construct from a Gregorian or Julian calendar date and hours, minutes, seconds.
        /// </summary>
        /// <param name="year">Year: 0 = 1 BC, -1 = 2 BC, etc.</param>
        /// <param name="month">Month [1 - 12]</param>
        /// <param name="day">Day of month (any positive or negative integer)</param>
        /// <param name="hour">hours</param>
        /// <param name="minute">minutes</param>
        /// <param name="second">seconds</param>
        /// <param name="julianCalendar">true if Julian calendar date</param>
        /// <exception cref="ArgumentOutOfRangeException"> if the month is outside the interval
        /// [1, 12], or hour, minute, or second is negative</exception>
        /// <remarks>
        /// <para>The algorithm itself has unlimited time span. In practice, the limits
        /// in the future and past are where a Julian date overflows an integer.</para>
        /// <para>Day of month is not restricted to the normal range. E.g., January 180 = 180th day
        /// of this year, and January -1 = December 30 in the previous year. However, hour, minute,
        /// and second must be non-negative.</para>
        /// </remarks>
        public JulianDate(int year, int month, int day, int hour, int minute, double second,
            bool julianCalendar)
            : this(year, month, day, Angle.HmsToHours(hour, minute, second) / 24.0, julianCalendar)
        { }


        /// <summary>
        /// Construct from a Gregorian calendar date and hours, minutes, seconds.
        /// </summary>
        /// <param name="year">Year: 0 = 1 BC, -1 = 2 BC, etc.</param>
        /// <param name="month">Month [1 - 12]</param>
        /// <param name="day">Day of month (any positive or negative integer)</param>
        /// <param name="hour">hours</param>
        /// <param name="minute">minutes</param>
        /// <param name="second">seconds</param>
        /// <exception cref="ArgumentOutOfRangeException"> if the month is outside the interval
        /// [1, 12], or hour, minute, or second is negative</exception>
        /// <remarks>
        /// <para>The algorithm itself has unlimited time span. In practice, the limits
        /// in the future and past are where a Julian date overflows an integer.</para>
        /// <para>Day of month is not restricted to the normal range. E.g., January 180 = 180th day
        /// of this year, and January -1 = December 30 in the previous year. However, hour, minute,
        /// and second must be non-negative.</para>
        /// </remarks>
        public JulianDate(int year, int month, int day, int hour, int minute, double second)
            : this(year, month, day, hour, minute, second, false)
        { }


#if false
        /// <summary>
        /// Initialize from a ParsedIso8601 object, which has date and time extracted from a
        /// String in ISO 8601 format. That object also controls whether the Gregorian or
        /// Julian calendar is used.
        /// </summary>
        /// <param name="parsed">ParsedIso8601</param>
        public JulianDate(ParsedIso8601 parsed)
            : this(parsed.Year, parsed.Month, parsed.Day,
            (parsed.Hour + (parsed.Minute + parsed.Second / 60.0) / 60.0 ) / 24.0,
            parsed.Julian)
        { }
#endif


        /// <summary>
        /// Construct from two double precision floating point values whose sum is the Julian date.
        /// </summary>
        /// <param name="jd1">Julian date part 1</param>
        /// <param name="jd2">Julian date part 2</param>
        /// <remarks>Maximum resolution is achieved if <paramref name="jd1"/> is at midnight
        /// (ends in .5) and <paramref name="jd2"/> is the time of day (between 0 and 1).</remarks>
        public JulianDate(double jd1, double jd2)
        {
            this.part1 = jd1;
            this.part2 = jd2;
            this.Normalize();
        }


        /// <summary>
        /// Construct from a TimeFields instance.
        /// </summary>
        /// <param name="tf">TimeFields instance</param>
        public JulianDate(TimeFields tf)
            : this(tf.Year, tf.Month, tf.Day, tf.Hms.ToDouble() / 24.0, tf.Julian)
        { }




        // properties


        /// <summary>
        /// Get the date portion of this instance.
        /// </summary>
        /// <value>the Julian date of this instance at 0 hours (always an integral value plus
        /// exactly .5).</value>
        public double DatePart { get { return part1; } }

        /// <summary>
        /// Get the fractional day portion of this instance.
        /// </summary>
        /// <value>the fractional day part of this instance, in the interval [0, 1).</value>
        public double TimePart { get { return part2; } }
        



        // methods


        /// <summary>
        /// Construct a String with the value of both components of this instance.
        /// </summary>
        /// <returns>String with the Julian date components</returns>
        public override string ToString()
        {
            return String.Format("{0:f1} + {1:f6}", this.part1, this.part2);
        }


        /// <summary>
        /// Returns the Julian date as a Double.
        /// </summary>
        /// <returns>Julian date</returns>
        public double ToDouble()
        {
            return this.DatePart + this.TimePart;
        }


        /// <summary>
        /// Return a new instance with the same date but time of day = 0 h.
        /// </summary>
        /// <returns>new JulianDate at 0 h</returns>
        public JulianDate AtZeroHours()
        {
            return new JulianDate(this.part1, 0.0);
        }


        /// <summary>
        /// Return the time of day in a Duration instance.
        /// </summary>
        /// <returns>a Duration in the interval [0, 1) days</returns>
        public Duration TimeOfDay() {
            return new Duration(this.TimePart);
        }


        /// <summary>
        /// Construct an instance of a .NET System.DateTime object, Kind = Utc.
        /// </summary>
        /// <remarks>The JulianDate class can represent dates BC, but a .NET System.DateTime cannot.
        /// </remarks>
        /// <returns>A new .NET Framework DateTime instance.</returns>
        public DateTime ToDateTime()
        {
            // Convert to Gregorian calendar date. Ignore the time of day.
            TimeFields date = new TimeFields(new JulianDate(this.DatePart, 0.0), 1, false);

            // Construct a DateTime object equal to 0 hours on that date.
            DateTime dt = new DateTime(date.Year, date.Month, date.Day, 0, 0, 0,
                DateTimeKind.Utc);

            // Construct a TimeSpan equal to the time of day.
            TimeSpan ts = new TimeSpan((long)(this.TimePart * TimeSpan.TicksPerDay + .5));

            // Return the DateTime plus the TimeSpan. This roundabout algorithm is necessary
            // because it's not possible to directly construct a DateTime to its full 100
            // nanosecond resolution. (The best you can do is 1 millisecond.)
            return dt + ts;
        }


        /// <summary>
        /// Convert UT1 to TT.
        /// </summary>
        /// <param name="deltaT">delta T (TT-UT1)</param>
        /// <returns>new instance in TT time scale</returns>
        public JulianDate Ut1ToTT(Duration deltaT)
        {
            return this + deltaT;
        }


        /// <summary>
        /// Convert TT to UT1.
        /// </summary>
        /// <param name="deltaT">delta T (TT-UT1)</param>
        /// <returns>new instance in UT1 time scale</returns>
        public JulianDate TTToUt1(Duration deltaT)
        {
            return this - deltaT;
        }


        /// <summary>
        /// Convert TT (Terrestrial Time) to TDB (Barycentric Dynamical Time).
        /// </summary>
        /// <returns>a new instance in TDB</returns>
        /// <remarks>The diurnal component of TDB-TT is ignored.</remarks>
        public JulianDate TTToTdb() {
            double tdbMinusTT = Sofa.Dtdb(JulianDate.J2000Base.ToDouble(), this.ToJ2000(),
                0, 0, 0, 0);
            return new JulianDate(this.DatePart, this.TimePart + tdbMinusTT / _secPerDay);
        }


        /// <summary>
        /// Convert TT (Terrestrial Time) to TCG (Geocentric Coordinate Time).
        /// </summary>
        /// <returns>a new instance in TCG</returns>
        public JulianDate TTToTcg() {
            double tcgDate, tcgTime;
            Sofa.Tttcg(this.DatePart, this.TimePart, out tcgDate, out tcgTime);
            return new JulianDate(tcgDate, tcgTime);
        }


        /// <summary>
        /// Convert TT (Terrestrial Time) to TCB (Barycentric Coordinate Time).
        /// </summary>
        /// <returns>a new instance in TCB</returns>
        /// <remarks>The diurnal component of TCB-TT is ignored.</remarks>
        public JulianDate TTToTcb() {
            double tcbDate, tcbTime, tdbMinusTT;
            tdbMinusTT = Sofa.Dtdb(JulianDate.J2000Base.ToDouble(), this.ToJ2000(), 0, 0, 0, 0);
            Sofa.Tdbtcb(this.DatePart, this.TimePart + tdbMinusTT / _secPerDay,
                out tcbDate, out tcbTime);
            return new JulianDate(tcbDate, tcbTime);
        }


        /// <summary>
        /// Convert TDB (Barycentric Dynamical Time) to TT (Terrestrial Time).
        /// </summary>
        /// <returns>a new instance in TT</returns>
        /// <remarks>The diurnal component of TDB-TT is ignored.</remarks>
        public JulianDate TdbToTT() {
            double tdbMinusTT = Sofa.Dtdb(JulianDate.J2000Base.ToDouble(), this.ToJ2000(),
                0, 0, 0, 0);
            return new JulianDate(this.DatePart, this.TimePart - tdbMinusTT / _secPerDay);
        }


        /// <summary>
        /// Convert TCG (Geocentric Coordinate Time) to TT (Terrestrial Time).
        /// </summary>
        /// <returns>a new instance in TT</returns>
        public JulianDate TcgToTT() {
            double ttDate, ttTime;
            Sofa.Tcgtt(this.DatePart, this.TimePart, out ttDate, out ttTime);
            return new JulianDate(ttDate, ttTime);
        }


        /// <summary>
        /// Convert TCB (Barycentric Coordinate Time) to TT (Terrestrial Time).
        /// </summary>
        /// <returns>a new instance in TT</returns>
        /// <remarks>The diurnal component of TCB-TT is ignored.</remarks>
        public JulianDate TcbToTT() {
            double tdbDate, tdbTime, tdbMinusTT;
            Sofa.Tcbtdb(this.DatePart, this.TimePart, out tdbDate, out tdbTime);
            JulianDate tdb = new JulianDate(tdbDate, tdbTime);
            tdbMinusTT = Sofa.Dtdb(JulianDate.J2000Base.ToDouble(), tdb.ToJ2000(), 0, 0, 0, 0);
            return new JulianDate(tdbDate, tdbTime - tdbMinusTT / _secPerDay);
        }


        /// <summary>
        /// Besselian Epoch. The result is in the same time scale as this instance.
        /// </summary>
        /// <returns>Besselian Epoch</returns>
        public double ToBesselianEpoch()
        {
            double t1 = this.part1;
            double t2 = this.part2;
            normalizeToJ2000(ref t1, ref t2);  // normalize to J2000 format
            return Sofa.Epb(t1, t2);
        }


        /// <summary>
        /// Convert Besselian epoch to a JulianDate instance.
        /// </summary>
        /// <param name="bEpoch">Besselian epoch</param>
        /// <returns>new JulianDate instance</returns>
        public static JulianDate FromBesselianEpoch(double bEpoch)
        {
            double jd1, jd2;
            Sofa.Epb2jd(bEpoch, out jd1, out jd2);
            return new JulianDate(jd1, jd2);
        }


        /// <summary>
        /// Julian Epoch. The result is in the same time scale as this instance.
        /// </summary>
        /// <returns>Julian Epoch (e.g. 1996.8)</returns>
        public double ToJulianEpoch()
        {
            double t1 = this.part1;
            double t2 = this.part2;
            normalizeToJ2000(ref t1, ref t2);  // normalize to J2000 format
            return Sofa.Epj(t1, t2);
        }


        /// <summary>
        /// Convert Julian epoch to a JulianDate instance.
        /// </summary>
        /// <param name="jEpoch">Julian epoch</param>
        /// <returns>new JulianDate instance</returns>
        public static JulianDate FromJulianEpoch(double jEpoch)
        {
            double jd1, jd2;
            Sofa.Epj2jd(jEpoch, out jd1, out jd2);
            return new JulianDate(jd1, jd2);
        }


        /// <summary>
        /// Convert this instance to the modified Julian date.
        /// </summary>
        /// <returns>Julian date minus 2400000.5</returns>
        public double ToMjd()
        {
            return new Duration(this - MjdBase).ToDouble();
        }


        /// <summary>
        /// Convert a modified Julian date to a JulianDate instance.
        /// </summary>
        /// <param name="mjd">modified Julian date</param>
        /// <returns>new JulianDate</returns>
        public static JulianDate FromMjd(double mjd)
        {
            return new JulianDate(2400000.5, mjd);
        }


        /// <summary>
        /// Number of days after the origin of the J2000 epoch.
        /// </summary>
        /// <returns>Julian date minus (2000 Jan 1.5)</returns>
        public double ToJ2000()
        {
            return new Duration(this - J2000Base).ToDays();
        }


        /// <summary>
        /// Convert to a TimeFields instance in the Julian or Gregorian calendar.
        /// </summary>
        /// <param name="resolution">desired roundoff resolution (1/hours), e.g., 3600.0 for
        /// 1-second resolution</param>
        /// <param name="julianCalendar">true if Julian calendar date is desired</param>
        /// <returns>TimeFields corresponding to this JulianDate</returns>
        public TimeFields ToTimeFields(double resolution, bool julianCalendar) {
            return new TimeFields(this, resolution, julianCalendar);
        }


        /// <summary>
        /// Convert to a TimeFields instance in the Gregorian calendar.
        /// </summary>
        /// <param name="resolution">desired roundoff resolution (1/hours), e.g., 3600.0 for
        /// 1-second resolution</param>
        /// <returns>TimeFields (Gregorian calendar) corresponding to this JulianDate</returns>
        public TimeFields ToTimeFields(double resolution) {
            return this.ToTimeFields(resolution, false);
        }



#if false
        /// <summary>
        /// Barycentric Dynamical Time minus Terrestrial Time. Formally, the instance should be
        /// TDB, but there is no practical effect if it's TT. The topocentric
        /// terms (which contribute 2 microseconds max) are omitted.
        /// </summary>
        /// <returns>TDB - TT</returns>
        public Duration TdbMinusTT()
        {
            double date1 = this.part1;
            double date2 = this.part2;
            normalizeToJ2000(ref date1, ref date2);    // normalize to J2000 format
            return new Duration(Sofa.Dtdb(date1, date2,
                0.0, 0.0, 0.0, 0.0) / 86400.0);
        }
#endif


#if false
        /// <summary>
        /// Barycentric Dynamical Time minus Terrestrial Time. Formally, the instance should be
        /// TDB, but there is no practical effect if it's TT. The topocentric
        /// terms (which contribute 2 microseconds max) are included.
        /// </summary>
        /// <returns>TDB - TT</returns>
        public Duration TdbMinusTT(JulianDate ut1, Topocenter pos)
        {
            double date1 = this.part1;
            double date2 = this.part2;
            normalizeToJ2000(ref date1, ref date2);    // normalize to J2000 format
            double lon = Math.Atan2(pos.Y, pos.X);      // east longitude
            // distance from Earth spin axis, and north from equator, km
            double u = Math.Sqrt(pos.X * pos.X + pos.Y * pos.Y) / 1000.0;
            double v = pos.Z / 1000.0;
            return new Duration(Sofa.DTDB(date1, date2, ut1.TimePart, lon, u, v) / 86400.0);
        }
#endif


        /// <summary>
        /// Compare by value: this JulianDate vs. another. Result is less than 0, 0, or greater
        /// than 0, if this instance is less than, equal to, or greater than the other,
        /// respectively.
        /// </summary>
        /// <param name="rhs">JulianDate</param>
        /// <returns>result of value comparison</returns>
        public int CompareTo(JulianDate rhs)
        {
            if ((object)rhs == null)
                return 1;
            int i = this.part1.CompareTo(rhs.part1);
            if (i != 0)
                return i;
            return this.part2.CompareTo(rhs.part2);
        }


        /// <summary>
        /// Compare by value: this JulianDate vs. an object. Result is less than 0, 0, or greater
        /// than 0, if this instance is less than, equal to, or greater than the object,
        /// respectively.
        /// </summary>
        /// <param name="obj">the object to compare to</param>
        /// <returns>result of value comparison</returns>
        /// <exception cref="ArgumentException">if <paramref name="obj"/> is wrong type.</exception>
        public int CompareTo(object obj)
        {
            if (obj == null)
                return 1;
            if (this.GetType() != obj.GetType())
                throw new ArgumentException("JulianDate CompareTo(object): type mismatch.");
            return this.CompareTo((JulianDate)obj);
        }
        

        /// <summary>
        /// Test for value equality: this JulianDate instance vs. another.
        /// </summary>
        /// <param name="rhs">JulianDate</param>
        /// <returns>true if values are equal</returns>
        public bool Equals(JulianDate rhs)
        {
            return this.CompareTo(rhs) == 0;
        }


        /// <summary>
        /// Test for value equality: this instance vs. an object.
        /// </summary>
        /// <param name="obj">JulianDate (or derived from)</param>
        /// <returns>Boolean</returns>
        public override bool Equals(object obj)
        {
            if (obj == null || this.GetType() != obj.GetType())
                return false;
            return this.CompareTo((JulianDate)obj) == 0;
        }


        /// <summary>
        /// Hash this instance into an integer.
        /// </summary>
        /// <returns>Hash code based on the value of this instance.</returns>
        public override int GetHashCode()
        {
            return this.part1.GetHashCode() ^ this.part2.GetHashCode();
        }


        /// <summary>
        /// Add a negative or positive Duration to this JulianDate instance.
        /// </summary>
        /// <param name="rhs">Duration to add to this instance</param>
        /// <returns>new JulianDate: this JulianDate + <paramref name="rhs"/></returns>
        public JulianDate Add(Duration rhs)
        {
            return new JulianDate(this.part1 + rhs.IntegerPart, this.part2 + rhs.FractionPart);
        }
        

        /// <summary>
        /// Operator to add a JulianDate and a Duration.
        /// </summary>
        /// <param name="lhs">JulianDate</param>
        /// <param name="rhs">Duration</param>
        /// <returns>new JulianDate: <paramref name="lhs"/> + <paramref name="rhs"/></returns>
        public static JulianDate operator +(JulianDate lhs, Duration rhs)
        {
            return lhs.Add(rhs);
        }
        

        /// <summary>
        /// Operator to add a Duration and a JulianDate.
        /// </summary>
        /// <param name="lhs">Duration</param>
        /// <param name="rhs">JulianDate</param>
        /// <returns>new JulianDate: lhs + rhs</returns>
        public static JulianDate operator +(Duration lhs, JulianDate rhs)
        {
            return rhs.Add(lhs);
        }


        /// <summary>
        /// Subtract a Duration from this JulianDate instance.
        /// </summary>
        /// <param name="rhs">Duration to subtract</param>
        /// <returns>new JulianDate: this JulianDate minus <paramref name="rhs"/></returns>
        public JulianDate Subtract(Duration rhs)
        {
            return new JulianDate(this.part1 - rhs.IntegerPart, this.part2 - rhs.FractionPart);
        }
        

        /// <summary>
        /// Operator to subtract a Duration from a JulianDate.
        /// </summary>
        /// <param name="lhs">JulianDate</param>
        /// <param name="rhs">Duration to subtract</param>
        /// <returns>new JulianDate: <paramref name="lhs"/> minus <paramref name="rhs"/></returns>
        public static JulianDate operator -(JulianDate lhs, Duration rhs)
        {
            return lhs.Subtract(rhs);
        }
        

        /// <summary>
        /// The Duration (time difference) between this JulianDate instance and another.
        /// </summary>
        /// <param name="rhs">the other JulianDate</param>
        /// <returns>this instance minus <paramref name="rhs"/> (positive if this instance is after
        /// rhs)</returns>
        public Duration Subtract(JulianDate rhs)
        {
            return new Duration(this.part1 - rhs.part1, this.part2 - rhs.part2);
        }


        /// <summary>
        /// Operator to subtract one JulianDate from another.
        /// </summary>
        /// <param name="lhs">JulianDate</param>
        /// <param name="rhs">JulianDate</param>
        /// <returns><paramref name="lhs"/> - <paramref name="rhs"/> (positive if lhs
        /// is after rhs)</returns>
        public static Duration operator -(JulianDate lhs, JulianDate rhs)
        {
            return lhs.Subtract(rhs);
        }


        /// <summary>
        /// Operator "less than" to compare two instances by value.
        /// </summary>
        /// <param name="lhs">left JulianDate</param>
        /// <param name="rhs">right JulianDate</param>
        /// <returns>true if <paramref name="lhs"/> &lt; <paramref name="rhs"/></returns>
        public static bool operator <(JulianDate lhs, JulianDate rhs)
        {
            return lhs.CompareTo(rhs) < 0;
        }


        /// <summary>
        /// Operator "less than or equal to" to compare two instances by value.
        /// </summary>
        /// <param name="lhs">left JulianDate</param>
        /// <param name="rhs">right JulianDate</param>
        /// <returns>true if <paramref name="lhs"/> &lt;= <paramref name="rhs"/></returns>
        public static bool operator <=(JulianDate lhs, JulianDate rhs)
        {
            return lhs.CompareTo(rhs) <= 0;
        }


        /// <summary>
        /// Operator "greater than or equal to" to compare two instances by value.
        /// </summary>
        /// <param name="lhs">left JulianDate</param>
        /// <param name="rhs">right JulianDate</param>
        /// <returns>true if <paramref name="lhs"/> &gt;= <paramref name="rhs"/></returns>
        public static bool operator >=(JulianDate lhs, JulianDate rhs)
        {
            return lhs.CompareTo(rhs) >= 0;
        }


        /// <summary>
        /// Operator "greater than" to compare two instances by value.
        /// </summary>
        /// <param name="lhs">left JulianDate</param>
        /// <param name="rhs">right JulianDate</param>
        /// <returns>true if <paramref name="lhs"/> &gt; <paramref name="rhs"/></returns>
        public static bool operator >(JulianDate lhs, JulianDate rhs)
        {
            return lhs.CompareTo(rhs) > 0;
        }


        /// <summary>
        /// Operator "equal to" to compare two instances by value.
        /// </summary>
        /// <param name="lhs">left JulianDate</param>
        /// <param name="rhs">right JulianDate</param>
        /// <returns>true if <paramref name="lhs"/> equals <paramref name="rhs"/></returns>
        public static bool operator ==(JulianDate lhs, JulianDate rhs)
        {
            if ((object)lhs == null)
                if ((object)rhs == null)
                    return true;
                else
                    return false;
            return lhs.CompareTo(rhs) == 0;
        }


        /// <summary>
        /// Operator "not equal to" to compare two instances by value.
        /// </summary>
        /// <param name="lhs">left JulianDate</param>
        /// <param name="rhs">right JulianDate</param>
        /// <returns>true if <paramref name="lhs"/> not equal to <paramref name="rhs"/></returns>
        public static bool operator !=(JulianDate lhs, JulianDate rhs)
        {
            return !(lhs == rhs);
        }


#if DEBUG
        /// <summary>
        /// Apportion the fields of this instance so part1 is at midnight (ends in .5)
        /// and part2 is in interval [0, 1) and the sum of the fields is unchanged.
        /// </summary>
#endif
        private void Normalize()
        {
            // After making part1 an integer and part2 a fraction, the missing
            // .5 will be restored.
            this.part1 -= .5;

            double part1_floor = Math.Floor(this.part1);
            double part2_floor = Math.Floor(this.part2);

            // remove the fractional part from part1
            double fraction = this.part1 - part1_floor;     // always positive
            this.part1 = part1_floor;

            // remove the integer part from part2
            this.part2 -= part2_floor;				        // always positive

            // put integer and fraction in their correct places
            this.part1 += part2_floor;
            this.part2 += fraction;
            if (this.part2 >= 1.0) {
                this.part2 -= 1.0;
                this.part1 += 1.0;
            }
            this.part1 += .5;
        }


#if DEBUG
        /// <summary>
        /// Re-apportion a 2-part Julian date so the sum is unchanged and <paramref name="jd1"/>
        /// is the J2000.0 base epoch.
        /// </summary>
        /// <param name="jd1">Julian date part 1</param>
        /// <param name="jd2">Julian date part 2</param>
#endif
        private static void normalizeToJ2000(ref double jd1, ref double jd2)
        {
	        double x = jd1 - 2451545.0;
	        jd1 = 2451545.0;
	        jd2 += x;
        }


#if DEBUG
        /// <summary>
        /// Initialize this instance from a calendar date. Helper method for the constructors.
        /// </summary>
        /// <param name="y">year</param>
        /// <param name="m">month (1 - 12)</param>
        /// <param name="d">day</param>
        /// <param name="fracDay">fractional days</param>
        /// <param name="julianCalendar">true if Julian calendar</param>
        /// <exception cref="ArgumentOutOfRangeException"> if <paramref name="fracDay"/> is negative
        /// </exception>
#endif
        private void calendarToJd(int y, int m, int d, double fracDay, bool julianCalendar)
        {
            if (fracDay < 0.0)
                throw new ArgumentOutOfRangeException("fracDay", "is negative");

            // Compute the day number n (which equals 1 on 1 AD Jan 1).
            int n = DateToDayNumber(y, m, d, julianCalendar);

            // Convert n to a 2-part Julian Date.
            this.part1 = (julianCalendar) ? _julianOffset : _gregorianOffset;
            this.part2 = n;

            // finish
            this.Normalize();
            this.part2 += fracDay;
            this.Normalize();
        }


        /// <summary>
        /// Is this a leap year?
        /// </summary>
        /// <param name="year">year (0 = 1 BC, -1 = 2 BC, etc.)</param>
        /// <param name="julianCalendar">true if Julian calendar</param>
        /// <returns>1 if leap year, 0 if not</returns>
        public static int IsLeapYear(int year, bool julianCalendar) {
            if (julianCalendar)
                return (year % 4 == 0) ? 1 : 0;
            else
                return ((year % 4 == 0 && year % 100 != 0) || (year % 400 == 0)) ? 1 : 0;
        }


        /// <summary>
        /// Convert date in Julian or Gregorian calendar to a day number which equals 1 on
        /// January 1, 1 AD of that calendar.
        /// </summary>
        /// <remarks>The algorithm itself has unlimited range, BC and AD. In practice, the time span
        /// is limited by integer overflow, and far exceeds any practical application of either
        /// calendar. In keeping with astronomical practice, day of month does not have to be
        /// within the normal range: it can be any integer.</remarks>
        /// <param name="y">year (0 = 1 BC, -1 = 2 BC, etc.)</param>
        /// <param name="m">month, 1 - 12</param>
        /// <param name="d">day (any integer is legal)</param>
        /// <param name="julianCalendar">true if Julian calendar</param>
        /// <returns>day number</returns>
        /// <exception cref="ArgumentOutOfRangeException">if month is out of range</exception>
        public static int DateToDayNumber(int y, int m, int d, bool julianCalendar)
        {
            int leap;   // 1 if leap year, 0 otherwise

            int[,] month_len = {
             {0,31,28,31,30,31,30,31,31,30,31,30,31},   // normal year
             {0,31,29,31,30,31,30,31,31,30,31,30,31}};  // leap year

            if (m < 1 || m > 12)
                throw new ArgumentOutOfRangeException("m", "not a legal month number");

            // Is this a leap year?
            leap = IsLeapYear(y, julianCalendar);

            // Compute the day number on Dec 31 of the preceding year.
            bool bc;
            if (y > 0) {        // AD year
                bc = false;
                y = y - 1;
            } else {
                bc = true;
                y = -y;
            }
            int n;      // day number
            if (julianCalendar)
                n = y * 365 + y / 4;
            else
                n = y * 365 + y / 4 - y / 100 + y / 400;
            if (bc)
                n = -(n + 366);     // e.g., in 1 BC, n = -366; in 2 BC, n = -731

            // add the days in the preceding months of this year
            for (int i = m - 1; i > 0; --i)
                n += month_len[leap, i];

            // finally, add day of the month
            return n + d;
        }



        #region self-test code
#if DEBUG

        /// <summary>
        /// class self-test
        /// </summary>
        public static void Test()
        {
            TimeFields tf1, tf2;
            JulianDate jd1, jd2, jd3;
#if false
            ParsedIso8601 parsed8601;
#endif
            Object obj1, obj2;
            Duration dur1, dur2;
            const double secPerDay = 86400.0;
            const double oneMicrosec = 1e-6 / secPerDay;
            const double oneNanosec = 1e-9 / secPerDay;
            double part1, part2;
            int correctDayNumber;
            int[,] monthLength = {
                { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
                { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 } };
            int y, m, d, leap;

            // MJD base epoch
            MjdBase.EqualsApprox(2400000.5, 0.0, oneNanosec);

            // J2000 base epoch
            jd1 = new JulianDate(2000, 1, 1, .5);
            J2000Base.EqualsApprox(2451544.5, 0.5, oneNanosec);

            // construct from a .NET Framework DateTime
            System.DateTime dt1 = new DateTime(2007, 6, 23, 6, 0, 0, DateTimeKind.Utc);
            jd1 = new JulianDate(dt1);
            jd1.EqualsApprox(2454274.5, .25, oneMicrosec);

            // construct from another JulianDate instance
            jd2 = new JulianDate(jd1);
            jd2.EqualsApprox(2454274.5, .25, oneNanosec);

            // construct from a Gregorian calendar date and fractional days (both forms of
            // constructor).
            jd1 = new JulianDate(2007, 6, 23, .947812, false);
            jd1.EqualsApprox(2454274.5, .947812, oneNanosec);
            jd1 = new JulianDate(2007, 6, 23, .947812);
            jd1.EqualsApprox(2454274.5, .947812, oneNanosec);

            // construct from a Gregorian calendar date and fractional days; extreme past
            jd1 = new JulianDate(-7191, 6, 13, .947812);
            jd1.EqualsApprox(-905235.5, .947812, 1e-6);

            // construct from a Julian calendar date and fractional days
            jd1 = new JulianDate(2007, 6, 10, .947812, true);   // Gregorian June 23
            jd1.EqualsApprox(2454274.5, .947812, oneNanosec);

            // construct from a Julian calendar date and fractional days; extreme past
            jd1 = new JulianDate(-7573, 1, 12, .947812, true);
            jd1.EqualsApprox(-1044969.5, .947812, 1e-6);

            // construct from a Gregorian date and sexagesimal time (both constructors)
            jd1 = new JulianDate(2007, 6, 23, 22, 44, 51.0, false);
            jd1.EqualsApprox(2454274.5, .947812, 1e-6);
            jd1 = new JulianDate(2007, 6, 23, 22, 44, 51.0);
            jd1.EqualsApprox(2454274.5, .947812, 1e-6);

            // construct from a Julian date and sexagesimal time
            jd1 = new JulianDate(2007, 6, 10, 22, 44, 51.0, true);
            jd1.EqualsApprox(2454274.5, .947812, 1e-6);

            // verify exception thrown for illegal Julian calendar date
            try {
                jd1 = new JulianDate(2007, 0, 10, .947812, true);
                Console.WriteLine("ERROR: constructor failed to detect illegal date.");
            } catch {
                ;
            }

            // verify exception is thrown if Gregorian calendar date is illegal
            try {
                jd1 = new JulianDate(2007, 0, 23, .947812);
                Console.WriteLine("ERROR: constructor failed to detect illegal date.");
            } catch {
                ;
            }

            // Verify exception if time of day is negative.
            try {
                jd1 = new JulianDate(2018, 2, 20, 13, 41, -1.0);
                Console.WriteLine("Constructor failed to detect negative time of day.");
            } catch {
                ;
            }

            // IsLeapYear(), Gregorian calendar
            for (y = -400; y <= 401; ++y) {
                leap = 0;
                if (y % 100 != 0) {      // non-century year
                    if (y % 4 == 0)
                        leap = 1;
                } else if (y % 400 == 0)
                    leap = 1;
                if (leap != IsLeapYear(y, false))
                    throw new Exception("IsLeapYear(), Gregorian.");
            }

            // IsLeapYear(), Julian calendar
            for (y = -4; y <= 5; ++y) {
                leap = 0;
                if (y % 4 == 0)
                    leap = 1;
                if (leap != IsLeapYear(y, true))
                    throw new Exception("IsLeapYear(), Julian.");
            }
                

                // test day number method with Gregorian dates, year 1 thru 401 (AD) inclusive
                correctDayNumber = 1;   // day no. at initial date (1 AD Jan 1)
            for (y = 1; y <= 401; ++y) {
                for (m = 1; m <= 12; ++m) {
                    d = 1;
                    if (DateToDayNumber(y, m, d, false) != correctDayNumber)
                        throw new Exception(String.Format("Gregorian at {0}-{1}-{2}", y, m, d));
                    d = monthLength[isLeap(y, false), m];   // last day of month
                    correctDayNumber += (d - 1);
                    if (DateToDayNumber(y, m, d, false) != correctDayNumber)
                        throw new Exception(String.Format("Gregorian at {0}-{1}-{2}", y, m, d));
                    ++correctDayNumber;     // first day of next month
                }
            }

            // test day number method with Gregorian dates, year 0 thru -400 inclusive
            correctDayNumber = 0;   // day no. at initial date (1 BC Dec 31)
            for (y = 0; y >= -400; --y) {
                for (m = 12; m >= 1; --m) {
                    d = monthLength[isLeap(y, false), m];   // last day of month
                    if (DateToDayNumber(y, m, d, false) != correctDayNumber)
                        throw new Exception(String.Format("Gregorian at {0}-{1}-{2}", y, m, d));
                    correctDayNumber -= (d - 1);
                    d = 1;
                    if (DateToDayNumber(y, m, d, false) != correctDayNumber)
                        throw new Exception(String.Format("Gregorian at {0}-{1}-{2}", y, m, d));
                    --correctDayNumber;     // last day of previous month
                }
            }

            // test day number method with Julian calendar dates, year 1 thru 5 (AD) inclusive
            correctDayNumber = 1;   // day no. at initial date (1 AD Jan 1)
            for (y = 1; y <= 5; ++y) {
                for (m = 1; m <= 12; ++m) {
                    d = 1;
                    if (DateToDayNumber(y, m, d, true) != correctDayNumber)
                        throw new Exception(String.Format("Julian at {0}-{1}-{2}", y, m, d));
                    d = monthLength[isLeap(y, true), m];   // last day of month
                    correctDayNumber += (d - 1);
                    if (DateToDayNumber(y, m, d, true) != correctDayNumber)
                        throw new Exception(String.Format("Julian at {0}-{1}-{2}", y, m, d));
                    ++correctDayNumber;     // first day of next month
                }
            }

            // test day number method with Julian calendar dates, year 0 thru -4 inclusive
            correctDayNumber = 0;   // day no. at initial date (1 BC Dec 31)
            for (y = 0; y >= -4; --y) {
                for (m = 12; m >= 1; --m) {
                    d = monthLength[isLeap(y, true), m];   // last day of month
                    if (DateToDayNumber(y, m, d, true) != correctDayNumber)
                        throw new Exception(String.Format("Julian at {0}-{1}-{2}", y, m, d));
                    correctDayNumber -= (d - 1);
                    d = 1;
                    if (DateToDayNumber(y, m, d, true) != correctDayNumber)
                        throw new Exception(String.Format("Julian at {0}-{1}-{2}", y, m, d));
                    --correctDayNumber;     // last day of previous month
                }
            }


            // construct from 2 Doubles, testing normalization too
            jd1 = new JulianDate(2545273.5, .948);      // nominal format
            jd1.EqualsApprox(2545273.5, .948, oneNanosec);
            jd1 = new JulianDate(.948, 2545273.5);      // transpose of previous
            jd1.EqualsApprox(2545273.5, .948, oneNanosec);
            jd1 = new JulianDate(2545274.448, 0.0);     // all in one argument
            jd1.EqualsApprox(2545273.5, .948, 1e-7);
            jd1 = new JulianDate(0.0, 2545274.448);     // transpose of previous
            jd1.EqualsApprox(2545273.5, .948, 1e-7);
            jd1 = new JulianDate(2400000.5, 145273.948);    // MJD
            jd1.EqualsApprox(2545273.5, .948, 1e-8);
            jd1 = new JulianDate(145273.948, 2400000.5);    // transpose of MJD
            jd1.EqualsApprox(2545273.5, .948, 1e-8);
            jd1 = new JulianDate(2600000.0, -54725.552);    // offset from 2.6e6
            jd1.EqualsApprox(2545273.5, .948, 1e-9);
            jd1 = new JulianDate(-54725.552, 2600000.0);    // transpose of previous
            jd1.EqualsApprox(2545273.5, .948, 1e-9);

            // test normalization in extreme past
            jd1 = new JulianDate(-2545273.5, .948);      // nominal format
            jd1.EqualsApprox(-2545273.5, .948, oneNanosec);
            jd1 = new JulianDate(.948, -2545273.5);      // transpose of previous
            jd1.EqualsApprox(-2545273.5, .948, oneNanosec);
            jd1 = new JulianDate(-2545273.5 + 10.0, .948 - 10.0);   // denormalized
            jd1.EqualsApprox(-2545273.5, .948, 1e-12);
            jd1 = new JulianDate(-2545273.5 - 10.0, .948 + 10.0);   // denorm.
            jd1.EqualsApprox(-2545273.5, .948, 1e-12);
            jd1 = new JulianDate(.948 + 10.0, -2545273.5 - 10.0);   // transposed, denorm.
            jd1.EqualsApprox(-2545273.5, .948, 1e-12);
            jd1 = new JulianDate(.948 - 10.0, -2545273.5 + 10.0);   // transposed, denorm.
            jd1.EqualsApprox(-2545273.5, .948, 1e-12);

#if false
            // construct from a ParsedIso8601 (implied Gregorian)
            parsed8601 = new ParsedIso8601("2007-06-23T22:44:51");
            jd1 = new JulianDate(parsed8601);
            jd1.EqualsApprox(2454274.5, .947812, 1e-6);

            // construct from a ParsedIso8601 (Julian calendar)
            jd1 = new JulianDate(new ParsedIso8601("J2007-06-10T22:44:51"));
            jd1.EqualsApprox(2454274.5, .947812, 1e-6);
#endif

            // Construct from a TimeFields and convert to TimeFields.
            tf1 = new TimeFields(2008, 11, 27, 0, 30, 45.4, 3.6e4, true);
            jd1 = new JulianDate(tf1);
            jd1.EqualsApprox(2454810.5, .021359, 1e-6);
            tf2 = jd1.ToTimeFields(3.6e4, true);
            assertTrue(tf2.Year == 2008 && tf2.Month == 11 && tf2.Day == 27
                && (int)(tf2.Hms.Hours) == 0 && (int)(tf2.Hms.Minutes) == 30
                && Math.Abs(tf2.Hms.Seconds - 45.4) < .1);

            // date and time properties
            jd1 = new JulianDate(2007, 6, 25, .910319);
            assertTrue(jd1.DatePart == 2454276.5 && Math.Abs(jd1.TimePart - .910319) < .000001);

            // ToString()
            assertTrue(jd1.ToString().Equals("2454276.5 + 0.910319"));

            // ToDouble()
            assertTrue(Math.Abs(jd1.ToDouble() - 2454277.410319) < .000001);

            // AtZeroHours(), TimeOfDay()
            jd1 = new JulianDate(2458053.5, .5);
            jd2 = jd1.AtZeroHours();
            assertTrue(jd2.DatePart == jd1.DatePart);
            assertTrue(jd2.TimePart == 0.0);
            dur1 = jd1.TimeOfDay();
            assertTrue(dur1.ToDays() == jd1.TimePart);

            // ToDateTime()
            jd1 = new JulianDate(2007, 6, 25, .910319);
            dt1 = jd1.ToDateTime();
            assertTrue(dt1.Year == 2007 && dt1.Month == 6 && dt1.Day == 25
                && dt1.Hour == 21 && dt1.Minute == 50 && dt1.Second == 51);


            // time scale conversions

            dur1 = new Duration(65.7 / 86400.0);    // delta T
            jd1 = new JulianDate(2007, 6, 24, .002350);

            jd1.Ut1ToTT(dur1).EqualsApprox(2454275.5, 0.003110, .000001);
            jd1.TTToUt1(dur1).EqualsApprox(2454275.5, 0.001590, .000001);

            // Replaced by Duration properties which give the offsets between these scales.
#if false
            jd1.TaiToTT().EqualsApprox(2454275.5, 0.002722, .000001);
            jd1.TTToTai().EqualsApprox(2454275.5, 0.001978, .000001);
            jd1.GpsToTT().EqualsApprox(2454275.5, 0.002942, .000001);
            jd1.TTToGps().EqualsApprox(2454275.5, 0.001758, .000001);
#endif

            // throw new Exception("TTToTdb() not tested");
            jd1 = new JulianDate(2015, 7, 25, 20, 36, 0);
            dur1 = new Duration(Sofa.Dtdb(JulianDate.J2000Base.ToDouble(), jd1.ToJ2000(),
                0, 0, 0, 0) / 86400.0);     // TDB-TT
            jd2 = jd1 + dur1;       // TDB
            jd1.TTToTdb().EqualsApprox(jd2.DatePart, jd2.TimePart, oneNanosec);

            // throw new Exception("TdbToTT() not tested");
            jd1.TTToTdb().TdbToTT().EqualsApprox(jd1.DatePart, jd1.TimePart, oneNanosec);

            // throw new Exception("TTToTcg() not tested");
            Sofa.Tttcg(jd1.DatePart, jd1.TimePart, out part1, out part2);
            jd2 = new JulianDate(part1, part2);     // TCG
            jd1.TTToTcg().EqualsApprox(jd2.DatePart, jd2.TimePart, oneNanosec);

            // throw new Exception("TcgToTT() not tested");
            jd1.TTToTcg().TcgToTT().EqualsApprox(jd1.DatePart, jd1.TimePart, oneNanosec);

            // throw new Exception("TTToTcb() not tested");
            Sofa.Tdbtcb(jd1.DatePart, jd1.TimePart + dur1.ToDays(), out part1, out part2);
            jd2 = new JulianDate(part1, part2);     // TCB
            jd1.TTToTcb().EqualsApprox(jd2.DatePart, jd2.TimePart, oneNanosec);


#if true
            // This code added to investigate bug in TcbToTT(). May as well be retained to get the
            // benefit of some extra checks.

            // Test TT -> TDB -> TT round trip via class methods.
            jd2 = jd1.TTToTdb();        // TDB
            jd3 = jd2.TdbToTT();        // TT
            jd3.EqualsApprox(jd1.DatePart, jd1.TimePart, oneNanosec);

            // Compare TDB-TT implied above with the SOFA value.
            dur1 = new Duration(Sofa.Dtdb(JulianDate.J2000Base.ToDouble(), jd1.ToJ2000(),
                0, 0, 0, 0) / 86400.0);     // TDB-TT
            dur2 = jd2 - jd1;               // TDB-TT
            dur1.EqualApprox(dur2.IntegerPart, dur2.FractionPart, oneNanosec);

            // Verify dur1 add and subtract are inverses.
            jd2 = jd1 + dur1;       // TDB
            jd3 = jd2 - dur1;       // TT
            jd3.EqualsApprox(jd1.DatePart, jd1.TimePart, oneNanosec);

            // TEST SOFA TDB -> TCB -> TDB ROUND TRIP
            Sofa.Tdbtcb(jd1.DatePart, jd1.TimePart, out part1, out part2);
            jd2 = new JulianDate(part1, part2);     // TCB
            Sofa.Tcbtdb(jd2.DatePart, jd2.TimePart, out part1, out part2);
            jd2 = new JulianDate(part1, part2);     // TDB
            jd1.EqualsApprox(jd2.DatePart, jd2.TimePart, oneNanosec);

            // Assume jd1 = TT. Test TT -> TCB.
            jd2 = jd1.TTToTdb();        // TDB
            Sofa.Tdbtcb(jd2.DatePart, jd2.TimePart, out part1, out part2);
            jd3 = new JulianDate(part1, part2);     // TCB
            jd1.TTToTcb().EqualsApprox(part1, part2, oneNanosec);

            // Starting with jd3 = TCB, test TCB -> TT.
            Sofa.Tcbtdb(jd3.DatePart, jd3.TimePart, out part1, out part2);
            jd2 = new JulianDate(part1, part2).TdbToTT();      // TT
            jd3.TcbToTT().EqualsApprox(jd2.DatePart, jd2.TimePart, oneNanosec);

            // Assume jd1 = TT. Round trip test.
            jd2 = jd1.TTToTcb();        // TCB
            jd3 = jd2.TcbToTT();        // TT
            jd3.EqualsApprox(jd1.DatePart, jd1.TimePart, oneNanosec);

#endif      // end debugging code


            // throw new Exception("TcbToTT() not tested");
            jd1.TTToTcb().TcbToTT().EqualsApprox(jd1.DatePart, jd1.TimePart, oneNanosec);

            // Besselian epoch
            jd1 = new JulianDate(1950, 1, 0, .923);     // B1950.0 per Almanac
            assertTrue(Math.Abs(jd1.ToBesselianEpoch() - 1950.0) < 1e-5);
            dur1 = JulianDate.FromBesselianEpoch(1950.0) - jd1;
            assertTrue(dur1.Abs() < new Duration(.001));

            // Julian epoch

            jd1 = new JulianDate(2000, 1, 1, .5);   // J2000.0
            assertTrue(Math.Abs(jd1.ToJulianEpoch() - 2000.0) < 1e-5);
            dur1 = JulianDate.FromJulianEpoch(2000.0) - jd1;
            assertTrue(dur1.Abs() < new Duration(.001));

            // Test again with data from t_sofa.c.
            jd1 = new JulianDate(2451545, -7392.5);
            assertTrue(Math.Abs(jd1.ToJulianEpoch() - 1979.760438056125941) <= 1e-12);
            dur1 = JulianDate.FromJulianEpoch(1979.760438056125941) - jd1;
            assertTrue(dur1.Abs() < new Duration(.001));

            // MJD
            jd1 = new JulianDate(2000, 1, 1, .5);   // J2000.0
            assertTrue(Math.Abs(jd1.ToMjd() - 51544.5) < 1e-6);
            dur1 = JulianDate.FromMjd(51544.5) - jd1;
            assertTrue(dur1.Abs() < new Duration(1e-6));

            // ToJ2000() and normalizeToJ2000()
            jd1 = new JulianDate(2007, 11, 17, (23.0 + (48.0 + 25.0 / 60.0) / 60.0) / 24.0);
            assertTrue(Math.Abs(jd1.ToJ2000() -  2877.491956) < 1e-6);
            part1 = jd1.DatePart;
            part2 = jd1.TimePart;
            normalizeToJ2000(ref part1, ref part2);
            assertTrue(part1 == 2451545.0);
            assertTrue(Math.Abs(part2 - 2877.491956) < 1e-6);

            // ToTimeFields tested already with the constructor that takes a TimeFields.

#if false
            // TdbMinusTT()
            assertTrue(Math.Abs(jd1.TdbMinusTT().ToDouble() * 86400.0 -
                Sofa.DTDB(d1, d2, 0.0, 0.0, 0.0, 0.0)) < .000001);
            assertTrue(jd1.TdbMinusTT() != new Duration(0.0));
#endif

            // CompareTo(JulianDate) method

            // handle a null reference
            jd1 = new JulianDate(2007, 6, 24, .1);
            jd2 = null;
            assertTrue(jd1.CompareTo(jd2) > 0);

            // different dates, same times
            jd2 = new JulianDate(2007, 6, 25, .1);
            assertTrue(jd1.CompareTo(jd1) == 0);
            assertTrue(jd1.CompareTo(jd2) < 0);
            assertTrue(jd2.CompareTo(jd1) > 0);

            // same dates, different times
            jd2 = new JulianDate(2007, 6, 24, .2);
            assertTrue(jd1.CompareTo(jd1) == 0);
            assertTrue(jd1.CompareTo(jd2) < 0);
            assertTrue(jd2.CompareTo(jd1) > 0);

            // different dates & times
            jd1 = new JulianDate(2007, 6, 23, .4);
            assertTrue(jd1.CompareTo(jd1) == 0);
            assertTrue(jd1.CompareTo(jd2) < 0);
            assertTrue(jd2.CompareTo(jd1) > 0);


            // CompareTo(object) method
            jd1 = new JulianDate(2007, 12, 4, .1);
            jd2 = new JulianDate(2007, 12, 5, .2);
            obj1 = null;
            assertTrue(jd1.CompareTo(obj1) > 0);
            obj1 = jd1;
            obj2 = jd2;
            assertTrue(jd1.CompareTo(obj1) == 0);
            assertTrue(jd1.CompareTo(obj2) < 0);
            assertTrue(jd2.CompareTo(obj1) > 0);
#if false
            try {
                Console.WriteLine("CompareTo(object) returned {0}", jd1.CompareTo(parsed8601));
                Console.WriteLine("That was an error: wrong type was not detected.");
            } catch (Exception) {
                ;
            }
#endif


            // Equals(JulianDate) method
            jd2 = jd1;
            assertTrue(jd1.Equals(jd2));
            jd2 = (JulianDate)obj2;
            assertFalse(jd1.Equals(jd2));

            // Equals(object) method
            jd1 = new JulianDate(1918, 11, 11, 11.0 / 24.0);
            jd2 = new JulianDate(2007, 6, 24, (19 + 42 / 60.0) / 24.0);
            obj1 = null;
            assertFalse(jd1.Equals(obj1));
            obj1 = jd1;
            obj2 = jd2;
#if false
            assertFalse(jd1.Equals(parsed8601));
#endif
            assertTrue(jd1.Equals(obj1));
            assertFalse(jd1.Equals(obj2));

            // hash code
            assertTrue(jd1.GetHashCode() != jd2.GetHashCode());

            // add() method
            jd1 = new JulianDate(2017, 11, 7, (15 + (21 + 36.2 / 60.0) / 60.0) / 24.0);
            dur1 = new Duration(-.617774);
            jd3 = jd1.Add(dur1);
            jd3.EqualsApprox(2458064.5, .022228, .000001);

            // + operator
            dur1 = new Duration(-5.0, -.712149);
            jd3 = jd1 + dur1;
            jd3.EqualsApprox(2458058.5, .927853, .000001);
            jd3 = dur1 + jd1;
            jd3.EqualsApprox(2458058.5, .927853, .000001);

            // subtract() method
            jd3 = jd1.Subtract(dur1);
            jd3.EqualsApprox(2458070.5, .352151, .000001);

            // - operator to subtract a Duration
            dur1 = new Duration(.502896);
            jd3 = jd1 - dur1;
            jd3.EqualsApprox(2458064.5, .137106, .000001);

            // subtract() method with 2 JulianDates
            jd2 = new JulianDate(2458058.5, .927853);
            dur2 = jd2.Subtract(jd1);
            dur2.EqualApprox(-6.0, .287851, .000001);

            // - operator with 2 JulianDates
            dur2 = jd2 - jd1;
            dur2.EqualApprox(-6.0, .287851, .000001);

            // relational operators

            jd1 = new JulianDate(2000, 2, 28, .932);
            jd2 = new JulianDate(2000, 2, 29, .778);
            jd3 = new JulianDate(2000, 2, 28, .932);

            assertTrue(jd1 < jd2);
            assertFalse(jd2 < jd1);
            assertFalse(jd1 < jd3);

            assertTrue(jd1 <= jd2);
            assertFalse(jd2 <= jd1);
            assertTrue(jd1 <= jd3);

            assertFalse(jd1 == jd2);
            assertFalse(jd2 == jd1);
            assertTrue(jd1 == jd3);

            assertFalse(jd1 >= jd2);
            assertTrue(jd2 >= jd1);
            assertTrue(jd1 >= jd3);

            assertFalse(jd1 > jd2);
            assertTrue(jd2 > jd1);
            assertFalse(jd1 > jd3);

            assertTrue(jd1 != jd2);
            assertTrue(jd2 != jd1);
            assertFalse(jd1 != jd3);

        }



        /// <summary>
        /// Throw exception if this instance isn't correct within a given tolerance.
        /// </summary>
        /// <param name="date">correct date part</param>
        /// <param name="time">correct time part</param>
        /// <param name="tolerance">time part tolerance</param>
        /// <remarks>An exception is thrown if the date parts are different, so this is not
        /// appropriate for general comparison of two instances.</remarks>
        public void EqualsApprox(double date, double time, double tolerance)
        {
#if false
            if (this.DatePart != date || Math.Abs(this.TimePart - time) > tolerance)
                throw new Exception("Julian date is out of tolerance.");
#endif
            if (this.DatePart != date)
                throw new Exception(string.Format("date is {0:f1}, should be {1:f1}",
                    this.DatePart, date));
            double error = this.TimePart - time;
            if (Math.Abs(error) > tolerance)
                throw new Exception(string.Format("time error is {0:e2} sec", error * _secPerDay));
        }


        /// <summary>
        /// Throw exception if argument is false.
        /// </summary>
        /// <param name="assertion">true = OK</param>
        private static void assertTrue(bool assertion)
        {
            if (!assertion)
                throw new Exception("Assertion was false.");
        }


        /// <summary>
        /// Throw exception if argument is true.
        /// </summary>
        /// <param name="assertion">false = OK</param>
        private static void assertFalse(bool assertion)
        {
            if (assertion)
                throw new Exception("Assertion was true.");
        }


        private static int isLeap(int year, bool julianCalendar) {
            int isLeap = 0;
            if (julianCalendar && year % 4 == 0)
                isLeap = 1;
            else
                if (year % 4 == 0 && year % 100 != 0 || year % 400 == 0)
                    isLeap = 1;
            return isLeap;
        }

#endif      // end class self-test code
        #endregion
    }
}
