// PVVector.cs
// Position/velocity vector class, based on the IAU SOFA library.
// Copyright 2018 by Paul S. Hirose

// revision history
#if false
2007-10-19 Completed rewrite to utilize .NET Framework types only. Eliminated unsafe code
and references to the Win32 Sofajpl DLL.
Many of the XML comments are from the IAU SOFA documentation, and the class uses the SOFA
library, so the SOFA software license applies to this class. There is a copy of that license
at the end of this file.

2007-12-01 Self-test code completed and runs OK.
2008-03-13 Altered the constructor that takes a Spherical, added ToSpherical(). Test OK.
2008-04-22 New EqualsApprox() with separate tolerances for position & velocity.
2012-07-19 Added an EqualsApprox overload which takes a message argument. Deleted SOFA
license.
2012-09-05 More EqualsApprox() overloads with error messages.

2015-05-10 I now compile the SOFA library from their C source, which has a different capitalization
style compared to the Fortran code, requiring corresponding change here.
2015-05-19 Properties and methods made virtual. Added a Negate() method and unary - operator.
2015-06-26 De-virtualized the class.

2017-10-26 Added a constructor which takes six parameters.
#endif


using System;


namespace HirosePS.SofaJpl
{
    /// <summary>
    /// A double vector which represents position and velocity.
    /// </summary>
    /// <remarks>
    /// <para>The vectors are stored as rectangular coordinates at double precision.</para>
    /// <para>Instances of this class are immutable.</para>
    /// </remarks>
    public class PVVector
    {
        // the representation of a PV vector
        private Vector position;
        private Vector velocity;



        /// <summary>
        /// Initialize position and velocity to 0.
        /// </summary>
        public PVVector() {
            this.position = new Vector();
            this.velocity = new Vector();
        }

        /// <summary>
        /// Initialize from another PVVector.
        /// </summary>
        /// <param name="pv">PVVector</param>
        public PVVector(PVVector pv)
        {
            this.position = pv.position;
            this.velocity = pv.velocity;
        }

        /// <summary>
        /// Initialize from separate position and velocity values.
        /// </summary>
        /// <param name="positionX">X coordinate of position</param>
        /// <param name="positionY">Y coordinate of position</param>
        /// <param name="positionZ">Z coordinate of position</param>
        /// <param name="velocityX">X component of velocity</param>
        /// <param name="velocityY">Y component of velocity</param>
        /// <param name="velocityZ">Z component of velocity</param>
        public PVVector(double positionX, double positionY, double positionZ,
            double velocityX, double velocityY, double velocityZ):
            this(new Vector(positionX, positionY, positionZ),
            new Vector(velocityX, velocityY, velocityZ))
        { }
        
        /// <summary>
        /// Initialize from a position Vector, set velocity to 0.
        /// </summary>
        /// <param name="position">Vector: position</param>
        public PVVector(Vector position)
        {
            this.position = position;
            this.velocity = new Vector();
        }
        
        /// <summary>
        /// Construct from separate position and velocity Vectors.
        /// </summary>
        /// <param name="position">Vector: position</param>
        /// <param name="velocity">Vector: velocity</param>
        public PVVector(Vector position, Vector velocity)
        {
            this.position = position;
            this.velocity = velocity;
        }

        /// <summary>
        /// Construct from an array of position and velocity.
        /// </summary>
        /// <param name="pv">Array of dimensions [2,3] with position in the top row</param>
        /// <exception cref="ArgumentException">if <paramref name="pv"/> has incorrect dimensions.
        /// </exception>
        public PVVector(double[,] pv)
        {
            if (pv.GetLength(0) != 2 || pv.GetLength(1) != 3)
                throw new ArgumentException("Constructor argument must have dimensions [2, 3].");
            this.position = new Vector(pv[0, 0], pv[0, 1], pv[0, 2]);
            this.velocity = new Vector(pv[1, 0], pv[1, 1], pv[1, 2]);
        }


#if false
        /// <summary>
        /// Construct from a Spherical with the position and another Spherical with the
        /// corresponding rates of change.
        /// </summary>
        /// <param name="position">Spherical: position</param>
        /// <param name="rate">Spherical: rate of change of position</param>
        public PVVector(Spherical position, Spherical rate)
        {
            double[,] result = new double[2, 3];
            Sofa.S2PV(position.Theta, position.Phi, position.Radius,
                rate.Theta, rate.Phi, rate.Radius, result);
            this.position = new Vector(result[0, 0], result[0, 1], result[0, 2]);
            this.velocity = new Vector(result[1, 0], result[1, 1], result[1, 2]);
        }
#endif


        /// <summary>
        /// Construct from a Spherical containing the position, and the rates of change of the
        /// spherical coordinates.
        /// </summary>
        /// <param name="position">spherical coordinates</param>
        /// <param name="thetaDot">longitude rate of change</param>
        /// <param name="phiDot">latitude rate of change</param>
        /// <param name="radiusDot">radius rate of change</param>
        public PVVector(Spherical position, double thetaDot, double phiDot, double radiusDot)
        {
            double[,] result = new double[2, 3];
            Sofa.S2pv(position.Theta, position.Phi, position.Radius,
                thetaDot, phiDot, radiusDot, result);
            this.position = new Vector(result[0, 0], result[0, 1], result[0, 2]);
            this.velocity = new Vector(result[1, 0], result[1, 1], result[1, 2]);

        }


        /// <summary>
        /// Write position x,y,z and velocity x,y,z as a string.
        /// </summary>
        /// <returns>String</returns>
        public override string ToString()
        {
            return String.Format("{0} {1}", position.ToString(), velocity.ToString());
        }


        /// <summary>
        /// Put position and velocity into a 2-dimensional array.
        /// </summary>
        /// <returns>new double[2,3] with position in the top row</returns>
        public double[,] ToArray()
        {
            return new double[,] {
            {this.position.X, this.position.Y, this.position.Z},
            {this.velocity.X, this.velocity.Y, this.velocity.Z}};
        }


        /// <summary>
        /// Convert the position part of this instance to a Spherical, and the velocity part to
        /// spherical angle rates and range rate.
        /// </summary>
        /// <param name="thetaDot">longitude angle rate</param>
        /// <param name="phiDot">latitude angle rate</param>
        /// <param name="rDot">range rate</param>
        /// <returns></returns>
        public Spherical ToSpherical(out double thetaDot, out double phiDot,
            out double rDot)
        {
            double theta, phi, r;
            Sofa.Pv2s(new double[,] {
                { this.position.X, this.position.Y, this.position.Z },
                { this.velocity.X, this.velocity.Y, this.velocity.Z } },
                out theta, out phi, out r, out thetaDot, out phiDot, out rDot);
            return new Spherical(theta, phi, r);
        }

        /// <summary>
        /// Get the position component of this instance.
        /// </summary>
        /// <value>position Vector</value>
        public Vector Position
        {
            get { return this.position; }
        }

        /// <summary>
        /// Get the velocity component of this instance.
        /// </summary>
        /// <value>new Vector with the velocity component</value>
        public Vector Velocity
        {
            get { return this.velocity; }
        }


        /// <summary>
        /// Additive inverse.
        /// </summary>
        /// <returns>negative of this instance</returns>
        public PVVector Negate() {
            return new PVVector(this.position.Negate(), this.velocity.Negate());
        }


        /// <summary>
        /// Additive inverse operator.
        /// </summary>
        /// <param name="pv">an instance</param>
        /// <returns>negative of the instance</returns>
        public static PVVector operator -(PVVector pv){
            return pv.Negate();
        }


        /// <summary>
        /// Sum of this instance and another PVVector.
        /// </summary>
        /// <param name="rhs">PVVector</param>
        /// <returns>new PVVector: this plus <paramref name="rhs"/></returns>
        public PVVector Add(PVVector rhs)
        {
            return new PVVector(this.position.Add(rhs.position), this.velocity.Add(rhs.velocity));
        }

        /// <summary>
        /// Operator: add two PVVector instances.
        /// </summary>
        /// <param name="lhs">first PVVector</param>
        /// <param name="rhs">second PVVector</param>
        /// <returns>new PVVector: <paramref name="lhs"/> plus <paramref name="rhs"/></returns>
        public static PVVector operator +(PVVector lhs, PVVector rhs)
        {
            return lhs.Add(rhs);
        }

        /// <summary>
        /// This instance minus another PVVector.
        /// </summary>
        /// <param name="rhs">the PVVector to subtract</param>
        /// <returns>new PVVector: this instance minus <paramref name="rhs"/></returns>
        public PVVector Subtract(PVVector rhs)
        {
            return new PVVector(this.position.Subtract(rhs.position),
                this.velocity.Subtract(rhs.velocity));
        }

        /// <summary>
        /// Operator: subtract one PVVector from another.
        /// </summary>
        /// <param name="lhs">the PVVector on the left</param>
        /// <param name="rhs">the PVVector on the right</param>
        /// <returns>new PVVector: <paramref name="lhs"/> minus <paramref name="rhs"/></returns>
        public static PVVector operator -(PVVector lhs, PVVector rhs)
        {
            return lhs.Subtract(rhs);
        }

        /// <summary>
        /// Multiply position and velocity of this instance by the same scalar.
        /// </summary>
        /// <param name="rhs">the scalar</param>
        /// <returns>new PVVector: this instance times <paramref name="rhs"/></returns>
        public PVVector Multiply(double rhs)
        {
            return new PVVector(this.position.Multiply(rhs), this.velocity.Multiply(rhs));
        }

        /// <summary>
        /// Operator: PVVector times a scalar.
        /// </summary>
        /// <param name="lhs">PVVector</param>
        /// <param name="rhs">scalar</param>
        /// <returns>new PVVector: <paramref name="lhs"/> times <paramref name="rhs"/></returns>
        public static PVVector operator *(PVVector lhs, double rhs)
        {
            return lhs.Multiply(rhs);
        }

        /// <summary>
        /// Operator: scalar times a PVVector.
        /// </summary>
        /// <param name="lhs">scalar</param>
        /// <param name="rhs">PVVector instance</param>
        /// <returns>new PVVector: <paramref name="lhs"/> times <paramref name="rhs"/></returns>
        public static PVVector operator *(double lhs, PVVector rhs)
        {
            return rhs.Multiply(lhs);
        }

        /// <summary>
        /// Apply different scalar multipliers to the position and velocity components
        /// of this instance.
        /// </summary>
        /// <param name="posMult">position multiplier</param>
        /// <param name="velMult">velocity multiplier</param>
        /// <returns>new PVVector: this instance times <paramref name="posMult"/> and
        /// <paramref name="velMult"/> respectively</returns>
        public PVVector Multiply(double posMult, double velMult)
        {
            return new PVVector(
                new Vector(this.position.Multiply(posMult)),
                new Vector(this.velocity.Multiply(velMult)));
        }

        /// <summary>
        /// Divide this instance by a scalar.
        /// </summary>
        /// <param name="rhs">scalar divisor</param>
        /// <returns>new PVVector: this instance divided by <paramref name="rhs"/></returns>
        /// <exception cref="ArgumentOutOfRangeException">if rhs is 0.</exception>
        public PVVector Divide(double rhs)
        {
            return new PVVector(this.position.Divide(rhs), this.velocity.Divide(rhs));
        }

        /// <summary>
        /// Operator: divide this instance by a scalar.
        /// </summary>
        /// <param name="lhs">PVVector instance</param>
        /// <param name="rhs">scalar</param>
        /// <returns>new PVVector: <paramref name="lhs"/> divided by <paramref name="rhs"/></returns>
        public static PVVector operator /(PVVector lhs, double rhs)
        {
            return lhs.Divide(rhs);
        }

        /// <summary>
        /// Divide this instance by separate scalars for the position and velocity components.
        /// </summary>
        /// <param name="positionDiv">position divisor</param>
        /// <param name="velocityDiv">velocity divisor</param>
        /// <returns>new PVVector: this instance divided by the respective divisors</returns>
        public PVVector Divide(double positionDiv, double velocityDiv)
        {
            return new PVVector(
                this.position.Divide(positionDiv),
                this.velocity.Divide(velocityDiv));
        }

        /// <summary>
        /// Update the position component of this instance to a new epoch
        /// <paramref name="deltaTime"/>
        /// different from the present epoch. Velocity component is not changed.
        /// </summary>
        /// <remarks>It is the user's responsibility to make sure the <paramref name="deltaTime"/>
        /// is expressed in units compatible with the velocity component of this instance.
        /// </remarks>
        /// <param name="deltaTime">elapsed time</param>
        /// <returns>new PVVector (updated)</returns>
        public PVVector Update(double deltaTime)
        {
            double[,] result = this.ToArray();
            Sofa.Pvu(deltaTime, result, result);
            return new PVVector(result);
        }

#if DEBUG
        /// <summary>
        /// Inner (dot) product of this instance and another PVVector.
        /// </summary>
        /// <remarks>If the position and velocity components of the two PVVectors are
        /// ( Ap, Av ) and ( Bp, Bv ), the result, A . B, is the pair of
        /// numbers ( Ap . Bp , Ap . Bv + Av . Bp ).  The two numbers are the
        /// dot product of the two PVVectors and its derivative.
        /// </remarks>
        /// <param name="rhs">the other PVVector</param>
        /// <param name="dotProd">dot product</param>
        /// <param name="deriv">dot product derivative</param>
#endif
        private void dotBoth(PVVector rhs, out double dotProd, out double deriv)
        {
            double[] result = new double[2];
            Sofa.Pvdpv(this.ToArray(), rhs.ToArray(), result);
            dotProd = result[0];
            deriv = result[1];
        }

        /// <summary>
        /// The dot (inner) product of the position components of this PVVector and another.
        /// The velocity components have no effect on the result.
        /// </summary>
        /// <param name="rhs">the other PVVector</param>
        /// <returns>dot product of the position components</returns>
        public double Dot(PVVector rhs)
        {
            double dotProd, deriv;
            dotBoth(rhs, out dotProd, out deriv);
            return dotProd;
        }

        /// <summary>
        /// The dot product derivative of this PVVector and another.
        /// </summary>
        /// <remarks>If the position and velocity components of the two PVVectors are
        /// ( Ap, Av ) and ( Bp, Bv ), the result is the number ( Ap . Bv + Av . Bp ).
        /// </remarks>
        /// <param name="rhs">the other PVVector</param>
        /// <returns>dot product derivative</returns>
        public double DotDerivative(PVVector rhs)
        {
            double dotProd, deriv;
            dotBoth(rhs, out dotProd, out deriv);
            return deriv;
        }

        /// <summary>
        /// Outer (cross) product of this instance and another PVVector.
        /// </summary>
        /// <remarks>If the position and velocity components of the two PVVectors are
        /// ( Ap, Av ) and ( Bp, Bv ), the result, A x B, is the pair of
        /// vectors ( Ap x Bp, Ap x Bv + Av x Bp ).  The two vectors are the
        /// cross product of the two PVVectors and its derivative.
        /// </remarks>
        /// <param name="rhs">the other PVVector</param>
        /// <returns>new PVVector: cross product</returns>
        public PVVector Cross(PVVector rhs)
        {
            double[,] result = this.ToArray();
            Sofa.Pvxpv(result, rhs.ToArray(), result);
            return new PVVector(result);
        }




        #region self-test code
#if DEBUG


        /// <summary>
        /// Self-test function. Runs silently if all is well.
        /// </summary>
        public static void Test()
        {
            // Initialize position and velocity to 0.
            PVVector pvvec1 = new PVVector();
            pvvec1.EqualsApprox(0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0);

            // Initialize from a position vector
            Vector vec1 = new Vector(7.0, 8.0, 9.0);
            pvvec1 = new PVVector(vec1);
            pvvec1.EqualsApprox(7.0, 8.0, 9.0, 0.0, 0.0, 0.0, .001);

            // Initialize from position and velocity vectors
            vec1 = new Vector(7.0, 8.0, 9.0);
            Vector vec2 = new Vector(.1, .2, .3);
            pvvec1 = new PVVector(vec1, vec2);
            pvvec1.EqualsApprox(7.0, 8.0, 9.0, .1, .2, .3, .001);

            // Initialize from another PVVector
            PVVector pvvec2 = new PVVector(pvvec1);
            pvvec2.EqualsApprox(7.0, 8.0, 9.0, .1, .2, .3, .001);

            // Initialize from separate position and velocity components.
            pvvec1 = new PVVector(11, 12, 13, 21, 22, 23);
            pvvec1.EqualsApprox(11, 12, 13, 21, 22, 23, 0);

            // Construct from an array of position and velocity.
            double[,] pvArray1 = { { 5.0, 6.0, 7.0 }, { .2, .3, .4 } };
            pvvec1 = new PVVector(pvArray1);
            pvvec1.EqualsApprox(5.0, 6.0, 7.0, .2, .3, .4, .001);
#if false
            // Initialize from 2 Sphericals
            {
                Spherical sph1 = new Spherical(vec1);
                Spherical sph2 = new Spherical(vec2);
                pvvec2 = new PVVector(sph1, sph2);
                double zSep = Math.Cos(sph1.Phi) * sph1.Radius;     // distance from Z axis
                double sinTheta = Math.Sin(sph1.Theta);
                double cosTheta = Math.Cos(sph1.Theta);
                double sinPhi = Math.Sin(sph1.Phi);
                double cosPhi = Math.Cos(sph1.Phi);
                double xVel = -sph2.Theta * sinTheta * zSep
                    - sph2.Phi * sinPhi * cosTheta * sph1.Radius
                    + sph2.Radius * cosPhi * cosTheta;
                double yVel = sph2.Theta * cosTheta * zSep
                    - sph2.Phi * sinPhi * sinTheta * sph1.Radius
                    + sph2.Radius * cosPhi * sinTheta;
                double zVel = sph2.Phi * cosPhi * sph1.Radius + sph2.Radius * sinPhi;
                pvvec2.EqualsApprox(7.0, 8.0, 9.0, xVel, yVel, zVel, .001);
            }
#endif

            // Initialize from a Spherical and spherical angle rates
            {
                Spherical sph1 = new Spherical(vec1);
                double thetaDot = .01, phiDot = .02, rDot = .03;
                pvvec2 = new PVVector(sph1, thetaDot, phiDot, rDot);
                double zSep = Math.Cos(sph1.Phi) * sph1.Radius;     // distance from Z axis
                double sinTheta = Math.Sin(sph1.Theta);
                double cosTheta = Math.Cos(sph1.Theta);
                double sinPhi = Math.Sin(sph1.Phi);
                double cosPhi = Math.Cos(sph1.Phi);
                double xVel = -thetaDot * sinTheta * zSep
                    - phiDot * sinPhi * cosTheta * sph1.Radius
                    + rDot * cosPhi * cosTheta;
                double yVel = thetaDot * cosTheta * zSep
                    - phiDot * sinPhi * sinTheta * sph1.Radius
                    + rDot * cosPhi * sinTheta;
                double zVel = phiDot * cosPhi * sph1.Radius + rDot * sinPhi;
                pvvec2.EqualsApprox(7.0, 8.0, 9.0, xVel, yVel, zVel, .001);

                // test ToSpherical()
                double thetaDot2, phiDot2, rDot2;
                Spherical sph2 = pvvec2.ToSpherical(out thetaDot2, out phiDot2, out rDot2);
                assertTrue(
                    Math.Abs(sph1.LonEast - sph2.LonEast) < 1e-6 &&
                    Math.Abs(sph1.Lat - sph2.Lat) < 1e-6 &&
                    Math.Abs(sph1.Radius - sph2.Radius) < 1e-6 &&
                    Math.Abs(thetaDot - thetaDot2) < 1e-6 &&
                    Math.Abs(phiDot - phiDot2) < 1e-6 &&
                    Math.Abs(rDot - rDot2) < 1e-6);
            }

            
            // ToString()
            assertTrue(pvvec1.ToString().Equals("5 6 7 0.2 0.3 0.4"));

            // ToArray()
            double[,] pvArray2 = pvvec1.ToArray();
            pvvec2 = new PVVector(pvArray2);
            pvvec2.EqualsApprox(5.0, 6.0, 7.0, .2, .3, .4, .0001);

            // position and velocity components
            pvvec1.position.EqualsApprox(5.0, 6.0, 7.0, .001);
            pvvec1.velocity.EqualsApprox(.2, .3, .4, .0001);

            Vector pos1 = new Vector(-6.26, -2.65, -.23);
            Vector pos2 = new Vector(-2.33, 1.32, -3.79);
            Vector vel1 = new Vector(.41, .93, -.18);
            Vector vel2 = new Vector(.31, -.76, -.79);

            // additive inverse
            pvvec1 = new PVVector(new double[,] {{-1.1, -1.2, -1.3}, {-2.1, -2.2,- 2.3}});
            pvvec1.Negate().EqualsApprox(1.1, 1.2, 1.3, 2.1, 2.2, 2.3, 0.0, "Negate()");
            pvvec2 = -pvvec1;
            pvvec2.EqualsApprox(1.1, 1.2, 1.3, 2.1, 2.2, 2.3, 0.0, "unary -");

            // Add method and operator
            pvvec1 = new PVVector(pos1, vel1);
            pvvec2 = new PVVector(pos2, vel2);
            PVVector pvvec3 = pvvec1.Add(pvvec2);
            pvvec3.equalsApprox(pos1 + pos2, vel1 + vel2, .0001);
            pvvec3 = pvvec1 + pvvec2;
            pvvec3.equalsApprox(pos1 + pos2, vel1 + vel2, .0001);

            // Subtract method and operator
            pvvec3 = pvvec1.Subtract(pvvec2);
            pvvec3.equalsApprox(pos1 - pos2, vel1 - vel2, .0001);
            pvvec3 = pvvec1 - pvvec2;
            pvvec3.equalsApprox(pos1 - pos2, vel1 - vel2, .0001);

            // Multiply by one scalar
            double scalar1 = -.43;
            pvvec1.Multiply(scalar1).equalsApprox(pos1 * scalar1, vel1 * scalar1, .0001);
            pvvec3 = pvvec1 * scalar1;
            pvvec1.Multiply(scalar1).equalsApprox(pos1 * scalar1, vel1 * scalar1, .0001);
            pvvec3 = scalar1 * pvvec1;
            pvvec1.Multiply(scalar1).equalsApprox(pos1 * scalar1, vel1 * scalar1, .0001);

            // Multiply by two scalars
            double scalar2 = .70;
            pvvec1.Multiply(scalar1, scalar2).equalsApprox(pos1 * scalar1, vel1 * scalar2, .0001);

            // Divide by one scalar
            pvvec1.Divide(scalar1).equalsApprox(pos1 / scalar1, vel1 / scalar1, .00001);
            pvvec3 = pvvec1 / scalar1;
            pvvec3.equalsApprox(pos1 / scalar1, vel1 / scalar1, .00001);

            // Divide by two scalars
            pvvec1.Divide(scalar1, scalar2).equalsApprox(pos1 / scalar1, vel1 / scalar2, .0001);

            // update a PV vector
            double deltaTime = .34;
            pvvec1.Update(deltaTime).equalsApprox(pos1 + vel1 * deltaTime, vel1, .0001);

            // dot product
            assertTrue(Math.Abs(pvvec1.Dot(pvvec2) - pvvec1.Position.Dot(pvvec2.Position))
                < .0001);

            // dot product derivative
            assertTrue(Math.Abs(pvvec1.DotDerivative(pvvec2) -
                (pvvec1.Position.Dot(pvvec2.Velocity) + pvvec1.Velocity.Dot(pvvec2.Position)))
                < .0001);

            // cross product
            pvvec1.Cross(pvvec2).equalsApprox(pvvec1.Position.Cross(pvvec2.Position),
                pvvec1.Position.Cross(pvvec2.Velocity) + pvvec1.Velocity.Cross(pvvec2.Position),
                .0001);
        }


        /// <summary>
        /// Throws Exception if the instance is out of tolerance.
        /// </summary>
        /// <param name="px">expected x position</param>
        /// <param name="py">expected y position</param>
        /// <param name="pz">expected z position</param>
        /// <param name="vx">expected x velocity</param>
        /// <param name="vy">expected y velocity</param>
        /// <param name="vz">expected z velocity</param>
        /// <param name="ptolerance">max allowable position discrepancy</param>
        /// <param name="vtolerance">max allowable velocity discrepancy</param>
        /// <exception cref="Exception">if any value in instance exceeds tolerance.</exception>
        /// <exception cref="ArgumentException"> if a NaN is in the instance or any parameter.
        /// </exception>
        public void EqualsApprox(
            double px, double py, double pz,
            double vx, double vy, double vz,
            double ptolerance, double vtolerance)
        {
            if (double.IsNaN(this.Position.X) ||
                double.IsNaN(this.Position.Y) ||
                double.IsNaN(this.Position.Z) ||
                double.IsNaN(this.Velocity.X) ||
                double.IsNaN(this.Velocity.Y) ||
                double.IsNaN(this.Velocity.Z))
                throw new ArgumentException("NaN in PVVector.");

            if (double.IsNaN(px) ||
                double.IsNaN(py) ||
                double.IsNaN(pz) ||
                double.IsNaN(vx) ||
                double.IsNaN(vy) ||
                double.IsNaN(vz))
                throw new ArgumentException("One of the expected values was a NaN.");

            if (double.IsNaN(ptolerance) || double.IsNaN(vtolerance))
                throw new ArgumentException("Tolerance was a NaN.");

            if (Math.Abs(this.position.X - px) > ptolerance ||
                Math.Abs(this.position.Y - py) > ptolerance ||
                Math.Abs(this.position.Z - pz) > ptolerance ||
                Math.Abs(this.velocity.X - vx) > vtolerance ||
                Math.Abs(this.velocity.Y - vy) > vtolerance ||
                Math.Abs(this.velocity.Z - vz) > vtolerance)
                throw new Exception("PVVector was out of tolerance.");
        }


        /// <summary>
        /// Throws Exception if the instance is out of tolerance.
        /// </summary>
        /// <param name="px">expected x position</param>
        /// <param name="py">expected y position</param>
        /// <param name="pz">expected z position</param>
        /// <param name="vx">expected x velocity</param>
        /// <param name="vy">expected y velocity</param>
        /// <param name="vz">expected z velocity</param>
        /// <param name="tolerance">max allowable discrepancy in position or velocity</param>
        /// <exception cref="Exception">if any value in instance exceeds tolerance.</exception>
        /// <exception cref="ArgumentException"> if a NaN is in the instance or any parameter.
        /// </exception>
        public void EqualsApprox(
            double px, double py, double pz,
            double vx, double vy, double vz, double tolerance)
        {
            this.EqualsApprox(px, py, pz, vx, vy, vz, tolerance, tolerance);
        }


        /// <summary>
        /// Throws Exception if the instance is out of tolerance.
        /// </summary>
        /// <param name="px">expected x position</param>
        /// <param name="py">expected y position</param>
        /// <param name="pz">expected z position</param>
        /// <param name="vx">expected x velocity</param>
        /// <param name="vy">expected y velocity</param>
        /// <param name="vz">expected z velocity</param>
        /// <param name="tolerance">max allowable discrepancy</param>
        /// <param name="message">error message</param>
        /// <exception cref="Exception">if any value in instance exceeds tolerance.
        /// </exception>
        public void EqualsApprox(
            double px, double py, double pz,
            double vx, double vy, double vz, double tolerance, string message)
        {
            try {
                EqualsApprox(px, py, pz, vx, vy, vz, tolerance);
            } catch {
                throw new Exception(message);
            }
        }


        /// <summary>
        /// Throws Exception if the instance is out of tolerance.
        /// </summary>
        /// <param name="px">expected x position</param>
        /// <param name="py">expected y position</param>
        /// <param name="pz">expected z position</param>
        /// <param name="vx">expected x velocity</param>
        /// <param name="vy">expected y velocity</param>
        /// <param name="vz">expected z velocity</param>
        /// <param name="ptolerance">max allowable position discrepancy</param>
        /// <param name="vtolerance">max allowable velocity discrepancy</param>
        /// <param name="message">error message</param>
        /// <exception cref="Exception">if any value in instance exceeds tolerance.
        /// </exception>
        public void EqualsApprox(
            double px, double py, double pz,
            double vx, double vy, double vz,
            double ptolerance, double vtolerance, string message)
        {
            try {
                EqualsApprox(px, py, pz, vx, vy, vz, ptolerance, vtolerance);
            } catch {
                throw new Exception(message);
            }
        }


        /// <summary>
        /// Throw exception if any component of the position or velocity differs by more
        /// than the allowable tolerance from another instance.
        /// </summary>
        /// <param name="pv">instance to compare to</param>
        /// <param name="posTolerance">position tolerance</param>
        /// <param name="velTolerance">velocity tolerance</param>
        public void EqualsApprox(PVVector pv, double posTolerance, double velTolerance)
        {
            try {
                this.position.EqualsApprox(pv.position, posTolerance);
            } catch {
                throw new Exception("PV vector position out of tolerance.");
            }
            try {
                this.velocity.EqualsApprox(pv.velocity, velTolerance);
            } catch {
                throw new Exception("PV vector velocity out of tolerance.");
            }
        }


        private void equalsApprox(Vector pos, Vector vel, double tolerance)
        {
            try {
                this.Position.EqualsApprox(pos.X, pos.Y, pos.Z, tolerance);
                this.Velocity.EqualsApprox(vel.X, vel.Y, vel.Z, tolerance);
            } catch {
                throw new Exception("PVVector was out of tolerance.");
            }
        }

        /// <summary>
        /// Throw exception if any component of the position or velocity differs by more
        /// than the allowable tolerance from another instance.
        /// </summary>
        /// <param name="pv">instance to compare to</param>
        /// <param name="positionTolerance">position tolerance</param>
        /// <param name="velocityTolerance">velocity tolerance</param>
        /// <param name="message">error message</param>
        public void EqualsApprox(PVVector pv, double positionTolerance,
            double velocityTolerance, string message)
        {
            try {
                position.EqualsApprox(pv.position, positionTolerance, "(position) ");
                velocity.EqualsApprox(pv.velocity, velocityTolerance, "(velocity) ");
            } catch (Exception e) {
                throw new Exception(message + e.Message);
            }
        }


        private static void assertTrue(bool assertion)
        {
            if (!assertion)
                throw new Exception("Assertion was false.");
        }


        private static void assert(bool assertion, string message)
        {
            if (!assertion)
                throw new Exception(message);
        }

#endif
        #endregion
    }
}
