// RMatrix.cs
// Rotation matrix class.
// Copyright 2018 by Paul S. Hirose.

// revision history:
#if false

2007-10-19 Completed rewrite to utilize .NET Framework types only. Eliminated unsafe code
and references to the Win32 Sofajpl DLL.
Many of the XML comments are from the IAU SOFA documentation, and the class uses the SOFA
library, so the SOFA software license applies to this class.

2007-12-02 Completed self-test code.
2008-01-19 Changed the x-, y-, and z-rotation methods to return a new instance instead of
modifying the original instance.
2008-02-11 New constructor which takes a 3x3 array.
2008-03-17 Changed the constructor that takes an array so the matrix elements are supplied
in normal left to right, top to bottom order and are automatically transposed to Fortran
order. The ToString() method likewise transposes the elements for display.
2008-03-19 Modified ToArray() so it's consistent with the constructor that takes an array.
2008-04-06 The previous changes to the constructor that takes an array, and ToArray(),
were more trouble than they were worth. Changed them back. Added a constructor that takes
individual values.
2008-06-19 TrsToHor() added + 1 overload; test OK.
2008-06-29 New CRS to intermediate and CRS to true methods test OK.
2008-07-01 ICRS to ecliptical, mean, intermediate, true systems test OK.
2010-11-28 Minor correction to comment describing the order of the matrix elements.
2012-03-30 Added constructor taking 3 Vectors, and rotation methods taking cos and sin.
Self-test good.
2012-04-14 Added CrsToMeanLongTerm() and CrsToEclipMeanLongTerm(), based on the Vondrak
et al long term precession model. Self-Tests not written yet.
2012-04-16 Added mean obliquity 1980, 2000, and 2006 methods. Working on filling out all
the combinations of PN model and reference system. Do IAU 2000 mean ecliptical next.
2012-04-28 Added methods to extract a specified row or column as a Vector.
2012-05-06 Decided my goal of filling out all combinations of PN model and ref system
combinations is unnecessary and will make too much clutter. Instead, will support the
IAU 2006/2000A model in all reference systems, and provide precession, nutation, and
obliquity methods for the other IAU models so any required transformation will be
possible without much trouble.
2012-05-12 Added nutation angle methods for all IAU models. True obliquity is simplified
to a single method taking delegates which supply mean obliquity and nutation angles.
2012-05-15 Methods for s.
2012-05-25 Removed TrueObliquity(), as the calculation is trivial.
2012-05-30 Added CRS to ITRS method.
2012-06-15 Renamed GetCipXY() to CipXY(). Edited XML comments.
2012-08-16 Removed MICA functions.
2012-08-19 Added remark that RMatrix objects are immutable (was already true).
Is true obliquity really trivial to compute? Maybe ought to re-think that.
2012-08-20 Put TrueObliquity() back in, plus a couple delegates to support it.
2012-08-21 CrsToEclip() is inefficient if the true equinox is wanted because you end up
calling the nutation routine twice. And would it be helpful to add frame bias to 1976
precession to make it like the other precession methods?
Wrote separate CrsToEclipMean() and CrsToEclipTrue(), with crsToEclip() retained as a
helper method. The 1976 precession method includes frame bias now. Corrected the X, Y,
and Z rotation methods so they're immutable.
2012-08-25 Declared the instance and static fields readonly to guarantee immutability.
Added a constructor which takes a 3x3 array and argument which indicates whether or not
the array is immutable. If true, an array copy is eliminated. Rewrote rotation and
multiplication methods to take advantage of this efficiency.
2012-08-26 Changed CipXY() to a non-static method. Renamed CrsToMean methods to
Precession. Renamed MeanToTrue methods to Nutation. Renamed delegates to put "delegate"
in their names.
2012-08-27 Renamed Precession() and Nutation() methods to Precess() and Nutate().
2012-08-28 Added a CrsToItrs() overload which takes a CRS to true matrix. Reduced the
EqualsApprox() methods to only two. The one that takes an array automatically
transposes to match the RMatrix storage order. Began revising self test.
2012-09-06 Self test good through GCRS to true.
2012-09-30 Added MeanObliqLongTerm(). Improved the GCRS to true tests. All tests good
thru GCRS to intermediate.
2012-10-01 Renamed methods to eliminate the confusing "Trs" from their names. Now it's
either Tirs or Itrs. Also renamed "Obliquity" to "Obliq" in all applicable names.
Renamed Crs to Gcrs. Added another GcrsToItrs() overload.
2012-10-03 Reduced the GcrsToItrs() methods to three.
2012-10-05 ItrsToHor uses a more accurate vector solution for deflection of the vertical.
Self test OK through the GcrsToItrs methods.
2012-10-07 Self test OK through GcrsToHor methods. Thinking about eliminating the
delegates from method argument lists to make calls more flexible. I.e., the user could
use either a variable or (usually) a method call as a argument.
2012-10-22 Updated the self test to the new Topocenter constructor, which takes degrees
instead of radians. Runs OK. Renamed CrsToJ2000() to IcrsToJ2000().
2012-10-23 Rewrote TrueObliquity() to take angle arguments, not delegates. Deleted all
GcrsToItrs() overloads except the one that takes the GCRS to CIRS matrix. However, one
should be added, to assume no polar motion.
2012-10-24 With no polar motion, GcrsToItrs() is actually GcrsToTirs(), so named it thus.

2015-05-10 I changed to C source for the SOFA library. Its different capitalization style compared
to the Fortran code required corresponding changes here. Storage order for matrices is reversed to
match the C convention.
    In tests I've been impressed with the IAU 2000B nutation, so it replaces 2000A in the simple
methods that provide default precession and nutation models.
    Verified Test() checks everything. Made all methods virtual. No longer use any delegates.
    Long term precession still needs to be translated to C#.

2015-05-16 Complete except for long term precession by Vondrak. Self test OK.
2015-05-18 Improved EqualsApprox() to report the amount of error and the matrix location where it
exceeded tolerance.
    Translation of the long term precession model of Vondrak et al to C# is finished. This class has
calls to the mean obliquity and precession routines. Self test OK.
2015-06-26 De-virtualized the class.

2015-08-25 ItrsToHor() is rewritten to orient the azimuth of the horizontal system to the CIP.

2018-08-11 Improved the XML comments.
#endif


using System;


namespace HirosePS.SofaJpl
{
    /// <summary>
    /// A rotation matrix to transform the coordinate system of a vector, for example, from the ICRS
    /// to the equator and equinox of date.
    /// </summary>
    /// <remarks>
    /// <para> The multiplication (*) operator applies a rotation matrix to a vector, or combines
    /// two rotation matrices into one, as in the textbook formulas. The matrix is represented
    /// internally as a 3x3 array of double precision values.</para>
    /// <para>Routines in the class generate rotation matrices to convert from the ICRS/GRCS to
    /// systems such as the true equatorial, ecliptical, horizontal, etc. By transposing the matrix
    /// with the provided routine, it performs the opposite transformation.</para>
    /// <para>Instances of this class are immutable.</para>
    /// </remarks>
    public class RMatrix
    {
        // The representation of the rotation matrix.
        readonly double[,] mat = new double[3, 3];

        // Frame bias matrix.
        static readonly RMatrix _icrsToJ2000;


        static RMatrix()
        {
            // Initialize the frame bias matrix.
            double[,] rb = new double[3, 3];
            double[,] rp = new double[3, 3];
            double[,] rbp = new double[3, 3];
            Sofa.Bp00(JulianDate.J2000Base.ToDouble(), 0.0, rb, rp, rbp);
            _icrsToJ2000 = new RMatrix(rb, true);
        }

        
        /// <summary>
        /// Initialize to an identity matrix
        /// </summary>
        public RMatrix()
        {
            Sofa.Ir(this.mat);
        }


        /// <summary>
        /// Construct from individual values.
        /// </summary>
        /// <param name="a11">top left</param>
        /// <param name="a12">top center</param>
        /// <param name="a13">top right</param>
        /// <param name="a21">center left</param>
        /// <param name="a22">center</param>
        /// <param name="a23">center right</param>
        /// <param name="a31">bottom left</param>
        /// <param name="a32">bottom center</param>
        /// <param name="a33">bottom right</param>
        public RMatrix(
            double a11, double a12, double a13,
            double a21, double a22, double a23,
            double a31, double a32, double a33)
        {
            this.mat[0, 0] = a11; this.mat[0, 1] = a12; this.mat[0, 2] = a13;
            this.mat[1, 0] = a21; this.mat[1, 1] = a22; this.mat[1, 2] = a23;
            this.mat[2, 0] = a31; this.mat[2, 1] = a32; this.mat[2, 2] = a33;
        }


        /// <summary>
        /// Construct from a 3x3 array.
        /// </summary>
        /// <remarks>
        /// <para>The top row of <paramref name="source"/> becomes the top row of the matrix.</para>
        /// <para>If the source array will never change, make <paramref name="immutable"/>
        /// true to eliminate an unnecessary array copy.</para>
        /// </remarks>
        /// <param name="source">3x3 source array</param>
        /// <param name="immutable">true if source array is guaranteed to never change
        /// </param>
        /// <exception cref="ArgumentException">if source array has wrong dimensions.
        /// </exception>
        public RMatrix(double[,] source, bool immutable)
        {
            if (source.GetLength(1) != 3 || source.GetLength(0) != 3)
                throw new ArgumentException("Source array must have dimensions [3,3].");
            if (immutable)
                this.mat = source;
            else
                Array.Copy(source, this.mat, 9);
        }


        /// <summary>
        /// Construct from a 3x3 array.
        /// </summary>
        /// <remarks>
        /// <para>The top row of <paramref name="source"/> becomes the top row of the matrix.</para>
        /// <para>This overload assumes the source array can change, so to be safe it makes a copy.
        /// </para>
        /// </remarks>
        /// <param name="source">3x3 source array</param>
        /// <exception cref="ArgumentException">if source array has wrong dimensions.
        /// </exception>
        public RMatrix(double[,] source)
            : this(source, false) { }


        /// <summary>
        /// Initialize from another rotation matrix.
        /// </summary>
        /// <param name="source">The other RMatrix</param>
        public RMatrix(RMatrix source)
        {
            // Take advantage of the fact that instances are immutable.
            this.mat = source.mat;
        }


        /// <summary>
        /// Construct a rotation matrix from a rotation vector.
        /// </summary>
        /// <remarks> The reference frame rotates clockwise as seen looking along the
        /// rotation vector from the origin. If the rotation vector has zero length, an
        /// identity matrix is constructed.</remarks>
        /// <param name="rvec">rotation vector</param>
        public RMatrix(Vector rvec)
        {
            Sofa.Rv2m(rvec.ToArray(), this.mat);
        }


        /// <summary>
        /// Construct a rotation matrix from unit vectors representing the rows.
        /// </summary>
        /// <param name="row1">top row</param>
        /// <param name="row2">center row</param>
        /// <param name="row3">bottom row</param>
        public RMatrix(Vector row1, Vector row2, Vector row3)
        {
            this.mat[0, 0] = row1.X;
            this.mat[0, 1] = row1.Y;
            this.mat[0, 2] = row1.Z;
            this.mat[1, 0] = row2.X;
            this.mat[1, 1] = row2.Y;
            this.mat[1, 2] = row2.Z;
            this.mat[2, 0] = row3.X;
            this.mat[2, 1] = row3.Y;
            this.mat[2, 2] = row3.Z;
        }


        /// <summary>
        /// Get a Vector representing a specified row of the rotation matrix.
        /// </summary>
        /// <param name="r">1 (top row), 2, or 3</param>
        /// <returns>the specified row, as a Vector</returns>
        /// <exception cref="ArgumentException"> if <paramref name="r"/> is not 1, 2, or 3
        /// </exception>
        public Vector Row(int r)
        {
            if (r < 1 || r > 3)
                throw new ArgumentException("Row must be 1, 2, or 3.");
            r--;
            return new Vector(this.mat[r, 0], this.mat[r, 1], this.mat[r, 2]);
        }


        /// <summary>
        /// Get a Vector representing a specified column of the rotation matrix.
        /// </summary>
        /// <param name="c">1 (left column), 2, or 3</param>
        /// <returns>the specified column, as a Vector</returns>
        /// <exception cref="ArgumentException">if <paramref name="c"/> is not 1, 2, or 3
        /// </exception>
        public Vector Column(int c)
        {
            if (c < 1 || c > 3)
                throw new ArgumentException("Column must be 1, 2, or 3.");
            c--;
            return new Vector(this.mat[0, c], this.mat[1, c], this.mat[2, c]);
        }


        /// <summary>
        /// Frame bias matrix to transform a vector from the ICRS to the J2000 system.
        /// </summary>
        public static RMatrix IcrsToJ2000 { get { return _icrsToJ2000; } }


        /// <summary>
        /// Output the elements as a string with 3 rows of values in f6 format.
        /// </summary>
        /// <returns>String</returns>
        public override string ToString()
        {
            return string.Format("{0:f6} {1:f6} {2:f6}\n{3:f6} {4:f6} {5:f6}\n{6:f6} {7:f6} {8:f6}",
                mat[0, 0], mat[0, 1], mat[0, 2],
                mat[1, 0], mat[1, 1], mat[1, 2],
                mat[2, 0], mat[2, 1], mat[2, 2]);
        }


        /// <summary>
        /// Output the values from this instance as a 3x3 Array.
        /// </summary>
        /// <remarks>Storage order is compatible with the C language and the routines in the Sofa
        /// class in this DLL. The first three elements are the top row of the matrix.</remarks>
        /// <returns>new 3x3 Array</returns>
        public double[,] ToArray()
        {
            double[,] a = new double[3, 3];
            Array.Copy(this.mat, a, 9);
            return a;
        }


        /// <summary>
        /// Apply an x axis rotation to the rotation matrix, given the angle. The original matrix
        /// is not affected.
        /// </summary>
        /// <remarks>Calling with a positive angle applies an additional rotation about
        /// the x-axis, anticlockwise as seen looking towards the origin from positive x.
        /// </remarks>
        /// <param name="angle">x rotation angle (radians)</param>
        /// <returns>new RMatrix instance with an additional X rotation</returns>
        public RMatrix XRotate(double angle)
        {
            double[,] rmat = new double[3, 3];
            Array.Copy(this.mat, rmat, 9);
            Sofa.Rx(angle, rmat);
            return new RMatrix(rmat, true);
        }


        /// <summary>
        /// Apply an x axis rotation to the rotation matrix, given the cosine and
        /// sine of the angle. The original matrix is not affected.
        /// </summary>
        /// <remarks>The arguments must be proportional to the cosine and sine of the
        /// rotation angle. If both are zero, the rotation angle is zero.</remarks>
        /// <param name="cosx">quasi cosine of rotation angle</param>
        /// <param name="sinx">quasi sine of rotation angle</param>
        /// <returns>new RMatrix instance with an additional X rotation</returns>
        public RMatrix XRotate(double cosx, double sinx)
        {
            if (sinx == 0.0 && cosx == 0.0)
                return new RMatrix(this);
            double modulus = Math.Sqrt(sinx * sinx + cosx * cosx);
            return new RMatrix(1, 0, 0,
                0, cosx / modulus, sinx / modulus,
                0, -sinx / modulus, cosx / modulus) * this;
        }
        

        /// <summary>
        /// Apply a y axis rotation to the rotation matrix, given the angle. The original matrix
        /// is not affected.
        /// </summary>
        /// <remarks>Calling with a positive angle applies an additional rotation about
        /// the y-axis, anticlockwise as seen looking towards the origin from positive y.
        /// </remarks>
        /// <param name="angle">y rotation angle (radians)</param>
        /// <returns>new RMatrix instance with an additional Y rotation</returns>
        public RMatrix YRotate(double angle)
        {
            double[,] rmat = new double[3, 3];
            Array.Copy(this.mat, rmat, 9);
            Sofa.Ry(angle, rmat);
            return new RMatrix(rmat, true);
        }


        /// <summary>
        /// Apply a y axis rotation to the rotation matrix, given the cosine and sine of
        /// the angle. The original matrix is not affected.
        /// </summary>
        /// <remarks>The arguments must be proportional to the cosine and sine of the
        /// rotation angle. If both are zero, the rotation angle is zero.</remarks>
        /// <param name="cosx">quasi cosine of rotation angle</param>
        /// <param name="sinx">quasi sine of rotation angle</param>
        /// <returns>new RMatrix instance with an additional Y rotation</returns>
        public RMatrix YRotate(double cosx, double sinx)
        {
            if (sinx == 0.0 && cosx == 0.0)
                return new RMatrix(this);
            double modulus = Math.Sqrt(sinx * sinx + cosx * cosx);
            return new RMatrix(cosx / modulus, 0, -sinx / modulus,
                0, 1, 0,
                sinx / modulus, 0, cosx / modulus) * this;
        }
         

        /// <summary>
        /// Apply a z axis rotation to the rotation matrix, given the angle. The original matrix
        /// is not affected.
        /// </summary>
        /// <remarks>Calling with a positive angle applies an additional rotation about
        /// the z-axis, anticlockwise as seen looking towards the origin from positive z.
        /// </remarks>
        /// <param name="angle">z rotation angle (radians)</param>
        /// <returns>new RMatrix instance with an additional Z rotation</returns>
        public RMatrix ZRotate(double angle)
        {
            double[,] rmat = new double[3, 3];
            Array.Copy(this.mat, rmat, 9);
            Sofa.Rz(angle, rmat);
            return new RMatrix(rmat, true);
        }


        /// <summary>
        /// Apply a z axis rotation to the rotation matrix, given the cosine and sine of
        /// the angle. The original matrix is not affected.
        /// </summary>
        /// <remarks>The arguments must be proportional to the cosine and sine of the
        /// rotation angle. If both are zero, the rotation angle is zero.</remarks>
        /// <param name="cosx">quasi cosine of rotation angle</param>
        /// <param name="sinx">quasi sine of rotation angle</param>
        /// <returns>new RMatrix instance with an additional Z rotation</returns>
        public RMatrix ZRotate(double cosx, double sinx)
        {
            if (sinx == 0.0 && cosx == 0.0)
                return new RMatrix(this);
            double modulus = Math.Sqrt(sinx * sinx + cosx * cosx);
            return new RMatrix(
                cosx / modulus, sinx / modulus, 0,
                -sinx / modulus, cosx / modulus, 0,
                0, 0, 1) * this;
        }



        /// <summary>
        /// Multiply this RMatrix instance by another.
        /// </summary>
        /// <param name="rhs">the other RMatrix</param>
        /// <returns>new RMatrix: this RMatrix times <paramref name="rhs"/></returns>
        public RMatrix Multiply(RMatrix rhs)
        {
            double[,] result = new double[3, 3];
            Sofa.Rxr(this.mat, rhs.mat, result);
            return new RMatrix(result, true);
        }
        
        /// <summary>
        /// Operator to multiply two RMatrix instances.
        /// </summary>
        /// <param name="lhs">left RMatrix</param>
        /// <param name="rhs">right RMatrix</param>
        /// <returns>new RMatrix: <paramref name="lhs"/> times <paramref name="rhs"/>
        /// </returns>
        public static RMatrix operator *(RMatrix lhs, RMatrix rhs)
        {
            return lhs.Multiply(rhs);
        }

        /// <summary>
        /// Multiply this RMatrix and a Vector.
        /// </summary>
        /// <param name="rhs">Vector</param>
        /// <returns>new Vector: this RMatrix times <paramref name="rhs"/></returns>
        public Vector Multiply(Vector rhs)
        {
            double[] result = new double[3];
            Sofa.Rxp(this.mat, rhs.ToArray(), result);
            return new Vector(result);
        }

        /// <summary>
        /// Operator to multiply an RMatrix and a Vector.
        /// </summary>
        /// <param name="lhs">RMatrix</param>
        /// <param name="rhs">Vector</param>
        /// <returns>new Vector: <paramref name="lhs"/> times <paramref name="rhs"/>
        /// </returns>
        public static Vector operator *(RMatrix lhs, Vector rhs)
        {
            return lhs.Multiply(rhs);
        }

        /// <summary>
        /// This RMatrix times a PVVector.
        /// </summary>
        /// <param name="rhs">PVVector</param>
        /// <returns>new PVVector: this RMatrix times <paramref name="rhs"/></returns>
        public PVVector Multiply(PVVector rhs)
        {
            return new PVVector(
                this.Multiply(rhs.Position),
                this.Multiply(rhs.Velocity));
        }

        /// <summary>
        /// Operator to multiply an RMatrix and a PVVector.
        /// </summary>
        /// <param name="lhs">RMatrix</param>
        /// <param name="rhs">PVVector</param>
        /// <returns>new PVVector: <paramref name="lhs"/> times <paramref name="rhs"/>
        /// </returns>
        public static PVVector operator *(RMatrix lhs, PVVector rhs)
        {
            return lhs.Multiply(rhs);
        }

        /// <summary>
        /// Transpose this instance. (For a valid rotation matrix, the transpose
        /// is identical to the inverse.) The original matrix is not affected.
        /// </summary>
        /// <returns>new RMatrix: transpose of the original instance</returns>
        public RMatrix Transpose()
        {
            double[,] result = new double[3, 3];
            Sofa.Tr(this.mat, result);
            return new RMatrix(result, true);
        }


        /// <summary>
        /// Mean obliquity of the ecliptic, IAU 1980 model.
        /// </summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>mean obliquity (radians)</returns>
        public static double MeanObliq80(JulianDate tt)
        {
            return Sofa.Obl80(JulianDate.J2000Base.ToDouble(), tt.ToJ2000());
        }


        /// <summary>
        /// Mean obliquity of the ecliptic, IAU 2000 model.
        /// </summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>mean obliquity (radians)</returns>
        public static double MeanObliq00(JulianDate tt)
        {
            // IAU 2000 precession-rate adjustments.
            double DPSIPR, DEPSPR;
            Sofa.Pr00(JulianDate.J2000Base.ToDouble(), tt.ToJ2000(), out DPSIPR, out DEPSPR);

            // Mean obliquity, consistent with IAU 2000 precession-nutation.
            return Sofa.Obl80(JulianDate.J2000Base.ToDouble(), tt.ToJ2000()) + DEPSPR;
        }


        /// <summary>
        /// Mean obliquity of the ecliptic, IAU 2006 precession model.
        /// </summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>mean obliquity (radians)</returns>
        public static double MeanObliq06(JulianDate tt)
        {
            return Sofa.Obl06(JulianDate.J2000Base.ToDouble(), tt.ToJ2000());
        }


        /// <summary>
        /// Mean obliquity of the ecliptic (long term precession model of Vondrak,
        /// Capitaine, Wallace, 2011).
        /// </summary>
        /// <remarks>This is more computationally expensive than a typical obliquity
        /// routine, as it must calculate the poles of the ecliptic and mean equator.
        /// </remarks>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>mean obliquity (radians)</returns>
        public static double MeanObliqLongTerm(JulianDate tt)
        {
            return LongTermPrecession.Obl(tt);
        }


        /// <summary>
        /// Nutation angles (IAU 1980 model).
        /// </summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <param name="deltaPsi">nutation in longitude (radians)</param>
        /// <param name="deltaEpsilon">nutation in obliquity (radians)</param>
        public static void NutationAngles80(JulianDate tt,
            out double deltaPsi, out double deltaEpsilon)
        {
            Sofa.Nut80(JulianDate.J2000Base.ToDouble(), tt.ToJ2000(),
                out deltaPsi, out deltaEpsilon);
        }


        /// <summary>
        /// Nutation angles (IAU 2000A model).
        /// </summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <param name="deltaPsi">nutation in longitude (radians)</param>
        /// <param name="deltaEpsilon">nutation in obliquity (radians)</param>
        public static void NutationAngles00a(JulianDate tt,
            out double deltaPsi, out double deltaEpsilon)
        {
            Sofa.Nut00a(JulianDate.J2000Base.ToDouble(), tt.ToJ2000(),
                out deltaPsi, out deltaEpsilon);
        }


        /// <summary>
        /// Nutation angles (IAU 2000B model).
        /// </summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <param name="deltaPsi">nutation in longitude (radians)</param>
        /// <param name="deltaEpsilon">nutation in obliquity (radians)</param>
        public static void NutationAngles00b(JulianDate tt,
            out double deltaPsi, out double deltaEpsilon)
        {
            Sofa.Nut00b(JulianDate.J2000Base.ToDouble(), tt.ToJ2000(),
                out deltaPsi, out deltaEpsilon);
        }


        /// <summary>
        /// Nutation angles (IAU 2000A model with adjustments to match IAU 2006
        /// precession).
        /// </summary>
        /// <remarks>Output includes the small nutation adjustments that are formally
        /// necessary when IAU 2006 precession and 2000A nutation are used together. These
        /// adjustments may be neglected with IAU 2000B nutation.</remarks>
        /// <param name="tt">Terrestrial Time</param>
        /// <param name="deltaPsi">nutation in longitude (radians)</param>
        /// <param name="deltaEpsilon">nutation in obliquity (radians)</param>
        public static void NutationAngles06a(JulianDate tt,
            out double deltaPsi, out double deltaEpsilon)
        {
            Sofa.Nut06a(JulianDate.J2000Base.ToDouble(), tt.ToJ2000(),
                out deltaPsi, out deltaEpsilon);
        }


        /// <summary>
        /// True obliquity of the ecliptic, given mean obliquity and nutation in
        /// obliquity.
        /// </summary>
        /// <param name="meanObliq">mean obliquity (radians)</param>
        /// <param name="deltaEpsilon">nutation in obliquity (radians)</param>
        /// <returns>true obliquity (radians)</returns>
        public static double TrueObliq(double meanObliq, double deltaEpsilon)
        {
            return meanObliq + deltaEpsilon;
        }

        
        /// <summary>
        /// True obliquity of the ecliptic (IAU 2006/2000B).
        /// </summary>
        /// <remarks>Consistent with IAU 2006 precession and 2000B nutation models.
        /// </remarks>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>true obliquity (radians)</returns>
        public static double TrueObliq06b(JulianDate tt) {
            double deltaPsi, deltaEpsilon;
            NutationAngles00b(tt, out deltaPsi, out deltaEpsilon);
            return TrueObliq(MeanObliq06(tt), deltaEpsilon);
        }


        /// <summary>
        /// GCRS x and y of the CIP, given the RMatrix that transforms from the GCRS to
        /// the intermediate system or true equatorial system.
        /// </summary>
        /// <param name="cipX">GCRS x component of a unit vector to the CIP</param>
        /// <param name="cipY">GCRS y component of a unit vector to the CIP</param>
        public void CipXY(out double cipX, out double cipY)
        {
            cipX = this.mat[2, 0];
            cipY = this.mat[2, 1];
        }


        /// <summary>
        /// The CIO locator s, given the GCRS xy of the CIP. Compatible with IAU 2000A or
        /// 2000B precession and nutation.
        /// </summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <param name="cipX">GCRS x component of a unit vector to the CIP</param>
        /// <param name="cipY">GCRS y component of a unit vector to the CIP</param>
        /// <returns>CIO locator s (radians)</returns>
        public static double S00(JulianDate tt, double cipX, double cipY)
        {
            return Sofa.S00(JulianDate.J2000Base.ToDouble(), tt.ToJ2000(), cipX, cipY);
        }


        /// <summary>
        /// The CIO locator s, given the GCRS xy of the CIP. Compatible with IAU 2006
        /// precession and 2000A or 2000B nutation.
        /// </summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <param name="cipX">GCRS x component of a unit vector to the CIP</param>
        /// <param name="cipY">GCRS y component of a unit vector to the CIP</param>
        /// <returns>CIO locator s (radians)</returns>
        public static double S06(JulianDate tt, double cipX, double cipY)
        {
            return Sofa.S06(JulianDate.J2000Base.ToDouble(), tt.ToJ2000(), cipX, cipY);
        }


        /// <summary>
        /// The CIO locator s, given TT, using IAU 2006 precession and 2006A nutation.
        /// </summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>CIO locator s (radians)</returns>
        public static double S06a(JulianDate tt)
        {
            return Sofa.S06a(JulianDate.J2000Base.ToDouble(), tt.ToJ2000());
        }


        /// <summary>
        /// Precession matrix (IAU 1976 precession model), GCRS to mean equator and
        /// equinox system of date, including frame bias.
        /// </summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>precession matrix (GCRS to mean equator system)</returns>
        public static RMatrix Precess76(JulianDate tt)
        {
            double[,] rp = new double[3, 3];
            Sofa.Pmat76(JulianDate.J2000Base.ToDouble(), tt.ToJ2000(), rp);
            return new RMatrix(rp, true) * IcrsToJ2000;
        }


        /// <summary>
        /// Precession matrix (IAU 2000 precession model), GCRS to mean equator and
        /// equinox system of date, including frame bias.</summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>precession matrix (GCRS to mean equator system)</returns>
        public static RMatrix Precess00(JulianDate tt)
        {
            double[,] rbp = new double[3, 3];
            Sofa.Pmat00(JulianDate.J2000Base.ToDouble(), tt.ToJ2000(), rbp);
            return new RMatrix(rbp, true);
        }


        /// <summary>
        /// Precession matrix (IAU 2006 precession model), GCRS to mean equator and
        /// equinox system of date, including frame bias.</summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>precession matrix (GCRS to mean equator system)</returns>
        public static RMatrix Precess06(JulianDate tt)
        {
            double[,] result = new double[3, 3];
            Sofa.Pmat06(JulianDate.J2000Base.ToDouble(), tt.ToJ2000(), result);
            return new RMatrix(result, true);
        }


        /// <summary>
        /// Precession matrix (long term precession model of Vondrak, Capitaine, Wallace,
        /// 2011), GCRS to mean equator and equinox system of date, including frame bias.
        /// </summary>
        /// <remarks>Citation: A&amp;A 534, A22 (2011), doi:10.1051/0004-6361/201117274
        /// and correction doi:10.1051/0004-6361/201117274e.
        /// </remarks>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>precession matrix (GCRS to mean equator system)</returns>
        public static RMatrix PrecessLongTerm(JulianDate tt)
        {
            return LongTermPrecession.Pmat(tt) * IcrsToJ2000;
        }


        /// <summary>
        /// Nutation matrix, given mean obliquity and nutation angles.
        /// </summary>
        /// <param name="meanObliquity">mean obliquity of the ecliptic (radians)</param>
        /// <param name="deltaPsi">nutation in longitude (radians)</param>
        /// <param name="deltaEpsilon">nutation in obliquity (radians)</param>
        /// <returns>mean to true equator and equinox matrix</returns>
        public static RMatrix Nutate(double meanObliquity, double deltaPsi,
            double deltaEpsilon)
        {
            double[,] rn = new double[3, 3];
            Sofa.Numat(meanObliquity, deltaPsi, deltaEpsilon, rn);
            return new RMatrix(rn, true);
        }


        /// <summary>
        /// GCRS to true equator and equinox matrix, given the precession matrix, mean
        /// obliquity, and the nutation angles.
        /// </summary>
        /// <param name="precession">GCRS to mean matrix</param>
        /// <param name="meanObliq">mean obliquity (radians)</param>
        /// <param name="deltaPsi">nutation in longitude (radians)</param>
        /// <param name="deltaEpsilon">nutation in obliquity (radians)</param>
        /// <returns>GCRS to true matrix</returns>
        public static RMatrix GcrsToTrue(RMatrix precession, double meanObliq,
            double deltaPsi, double deltaEpsilon)
        {
            RMatrix nutationMatrix = Nutate(meanObliq, deltaPsi, deltaEpsilon);
            return nutationMatrix * precession;
        }


        /// <summary>
        /// GCRS to true equator and equinox matrix (IAU 2006 precession, 2000B nutation).
        /// </summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>GCRS to true equator and equinox matrix</returns>
        public static RMatrix GcrsToTrue06b(JulianDate tt) {
            double deltaPsi, deltaEpsilon;
            NutationAngles00b(tt, out deltaPsi, out deltaEpsilon);
            return Nutate(MeanObliq06(tt), deltaPsi, deltaEpsilon) * Precess06(tt);
        }


        /// <summary>
        /// GCRS to celestial intermediate matrix, given the xy coordinates of the CIP
        /// and the CIO locator s.
        /// </summary>
        /// <param name="cipX">GCRS x component of a unit vector to the CIP</param>
        /// <param name="cipY">GCRS y component of a unit vector to the CIP</param>
        /// <param name="s">CIO locator s (radians)</param>
        /// <returns>GCRS to celestial intermediate matrix</returns>
        public static RMatrix GcrsToCirs(double cipX, double cipY, double s)
        {
            double[,] rc2i = new double[3, 3];
            Sofa.C2ixys(cipX, cipY, s, rc2i);
            return new RMatrix(rc2i, true);
        }


        /// <summary>
        /// GCRS to celestial intermediate matrix, given the GCRS to true equator and
        /// equinox matrix and the CIO locator s.
        /// </summary>
        /// <param name="gcrsToTrue">GCRS to true matrix</param>
        /// <param name="s">CIO locator s (radians)</param>
        /// <returns>GCRS to celestial intermediate matrix</returns>
        public static RMatrix GcrsToCirs(RMatrix gcrsToTrue, double s)
        {
            double cipX, cipY;
            gcrsToTrue.CipXY(out cipX, out cipY);
            return GcrsToCirs(cipX, cipY, s);
        }


        /// <summary>
        /// GCRS to celestial intermediate matrix, using IAU 2006 precession and IAU
        /// 2006B nutation.
        /// </summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>GCRS to celestial intermediate matrix</returns>
        public static RMatrix GcrsToCirs06b(JulianDate tt) {
            double cipX, cipY;
            RMatrix gcrsToTrue = GcrsToTrue06b(tt);
            gcrsToTrue.CipXY(out cipX, out cipY);
            return GcrsToCirs(cipX, cipY, S06(tt, cipX, cipY));
        }


        /// <summary>
        /// GCRS to terrestrial intermediate matrix, given the GCRS to celestial
        /// intermediate matrix and Earth rotation angle.
        /// </summary>
        /// <remarks>The equinox based components (GCRS to true matrix and Greenwich
        /// apparent sidereal time) will also produce a correct result.</remarks>
        /// <param name="gcrsToCirs">GCRS to celestial intermediate matrix</param>
        /// <param name="era">Earth rotation angle (radians)</param>
        /// <returns>GCRS to TIRS matrix</returns>
        public static RMatrix GcrsToTirs(RMatrix gcrsToCirs, double era)
        {
            return gcrsToCirs.ZRotate(era);
        }


        /// <summary>
        /// Polar motion matrix (transforms from the Terrestrial Intermediate Reference
        /// System to the International Terrestrial Reference System), given TT and the
        /// polar motion angles.
        /// </summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <param name="poleX">polar motion x (radians)</param>
        /// <param name="poleY">polar motion y (radians)</param>
        /// <returns>TIRS to ITRS matrix</returns>
        public static RMatrix TirsToItrs(JulianDate tt, double poleX, double poleY)
        {
            double[,] rpom = new double[3, 3];
            Sofa.Pom00(poleX, poleY, Sofa.Sp00(JulianDate.J2000Base.ToDouble(), tt.ToJ2000()),
                rpom);
            return new RMatrix(rpom, true);
        }


        /// <summary>
        /// GCRS to ITRS matrix, given the GCRS to celestial intermediate matrix, Earth
        /// rotation angle, and the polar motion matrix.
        /// </summary>
        /// <remarks>The equinox based components (GCRS to true matrix and Greenwich
        /// apparent sidereal time) will also produce a correct result.</remarks>
        /// <param name="gcrsToCirs">GCRS to celestial intermediate matrix</param>
        /// <param name="era">Earth rotation angle (radians)</param>
        /// <param name="tirsToItrs">polar motion matrix (terrestrial intermediate to
        /// ITRS)</param>
        /// <returns>GCRS to ITRS matrix</returns>
        public static RMatrix GcrsToItrs(RMatrix gcrsToCirs, double era,
            RMatrix tirsToItrs)
        {
            return tirsToItrs * GcrsToTirs(gcrsToCirs, era);
        }


        /// <summary>
        /// GCRS to ITRS matrix, given TT, UT1, and polar motion angles. (IAU 2006
        /// precession and IAU 2006B nutation.)
        /// </summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <param name="ut1">UT1</param>
        /// <param name="xPole">polar motion x (radians)</param>
        /// <param name="yPole">polar motion y (radians)</param>
        /// <returns>rotation matrix, GCRS to ITRS</returns>
        public static RMatrix GcrsToItrs06b(JulianDate tt, JulianDate ut1, double xPole,
            double yPole) {
            return GcrsToItrs(GcrsToCirs06b(tt), Angle.Era00(ut1), TirsToItrs(tt, xPole, yPole));
        }



        /// <summary>
        /// ITRS to horizontal (east, north, zenith) matrix, given latitude, longitude, deflection
        /// of the vertical, and the polar motion matrix.
        /// </summary>
        /// <param name="lat">geodetic north latitude (radians)</param>
        /// <param name="lon">geodetic east longitude (radians)</param>
        /// <param name="xi">north deflection of the vertical (radians)</param>
        /// <param name="eta">east deflection of the vertical (radians)</param>
        /// <param name="tirsToItrs">polar motion matrix</param>
        /// <returns>ITRS to horizontal matrix</returns>
        /// <remarks>
        /// <para><paramref name="xi"/> and <paramref name="eta"/> are positive if the plumb line
        /// intersects the celestial sphere north and east of the ellipsoid normal.</para>
        /// <para>If <paramref name="xi"/> and <paramref name="eta"/> are both zero, the polar
        /// motion matrix <paramref name="tirsToItrs"/> has no effect.</para>
        /// <para>To ignore polar motion, set <paramref name="tirsToItrs"/> to the identity matrix
        /// or a null reference. The latter is more efficient.</para>
        /// </remarks>
        public static RMatrix ItrsToHor(double lat, double lon, double xi, double eta,
            RMatrix tirsToItrs) {

            // Rotation matrix, ITRS to the ellipsoid horizontal system.
            RMatrix itrsToHor = new RMatrix().
                ZRotate(lon + Angle.HalfPi).
                XRotate(Angle.Complement(lat));

            if (xi != 0.0 || eta != 0.0) {
                // Level and plumb line unit vectors with respect to the ellipsoid horizontal.
                Vector east = new Vector(Math.Cos(eta), 0, -Math.Sin(eta));
                Vector north = new Vector(0, Math.Cos(xi), -Math.Sin(xi));
                Vector zenith = east.Cross(north).Unit();

                if (tirsToItrs != null) {
                    // Unit vector to the CIP, with respect to the ITRS.
                    Vector cip = tirsToItrs.Column(3);

                    // Calculate the basis vectors of the astronomic horizontal system with respect
                    // to the ellipsoid horizontal system. First, put the CIP vector in the latter
                    // system.
                    cip = itrsToHor * cip;

                    // Then get a vector to astronomic east. It's perpendicular to the CIP and
                    // plumb line. If the topocenter is extremely close to the pole it may be
                    // necessary to negate the cross product because it points west instead of east.
                    east = cip.Cross(zenith);
                    if (east.X < 0.0)
                        east = -east;

                    // Set astronomical north perpendicular to the plumb line and east vectors.
                    if (east.Modulus() != 0.0) {
                        east = east.Unit();     // set modulus to 1
                        north = zenith.Cross(east);
                    }
                    else {
                        // The topocenter is at the CIP. Therefore, astronomic azimuth is
                        // indeterminate. So use the existing north vector (which includes N-S
                        // deflection of the vertical) as astronomic north, then set the east vector
                        // perpendicular to the north vector and plumb line.
                        east = north.Cross(zenith);
                    }
                }
                else
                    // The caller has not supplied a polar motion matrix, so adopt the existing
                    // north vector as astronomic north, and set the east vector perpendicular to
                    // the north vector and the plumb line.
                    east = north.Cross(zenith);

                // Construct a deflection of the vertical matrix and apply it to the ellipsoid
                // horizontal system.
                itrsToHor = new RMatrix(east, north, zenith) * itrsToHor;
            }       // end code that executes if deflection of the vertical is nonzero

            return itrsToHor;
        }


        /// <summary>
        /// ITRS to horizontal (east, north, zenith) matrix, given a Topocenter, deflection
        /// of the vertical, and the polar motion matrix.
        /// </summary>
        /// <param name="obs">topocenter</param>
        /// <param name="xi">north deflection of the vertical (radians)</param>
        /// <param name="eta">east deflection of the vertical (radians)</param>
        /// <param name="tirsToItrs">polar motion matrix</param>
        /// <returns>ITRS to horizontal matrix</returns>
        /// <remarks>
        /// <para><paramref name="xi"/> and <paramref name="eta"/> are positive if the plumb line
        /// intersects the celestial sphere north and east of the ellipsoid normal.</para>
        /// <para>If <paramref name="xi"/> and <paramref name="eta"/> are both zero, the polar
        /// motion matrix <paramref name="tirsToItrs"/> has no effect.</para>
        /// <para>To ignore polar motion, set <paramref name="tirsToItrs"/> to the identity matrix
        /// or a null reference. The latter is more efficient.</para>
        /// </remarks>
        public static RMatrix ItrsToHor(Topocenter obs, double xi, double eta, RMatrix tirsToItrs) {
            return ItrsToHor(obs.Lat, obs.Lon, xi, eta, tirsToItrs);
        }


        /// <summary>
        /// ITRS to horizontal (east, north, zenith) matrix, given latitude and longitude.
        /// </summary>
        /// <param name="lat">north latitude (radians)</param>
        /// <param name="lon">east longitude (radians)</param>
        /// <returns>ITRS to horizontal matrix</returns>
        public static RMatrix ItrsToHor(double lat, double lon)
        {
#if false
            return new RMatrix().
                ZRotate(lon + Angle.HalfPi).
                XRotate(Angle.Complement(lat));
#endif
            return ItrsToHor(lat, lon, 0, 0, null);
        }


        /// <summary>
        /// ITRS to horizontal (east, north, zenith) matrix, given a Topocenter object.
        /// </summary>
        /// <param name="position">geographic position</param>
        /// <returns>ITRS to horizontal matrix</returns>
        public static RMatrix ItrsToHor(Topocenter position)
        {
#if false
            return ItrsToHor(position.Lat, position.Lon);
#endif
            return ItrsToHor(position, 0, 0, null);
        }


        /// <summary>
        /// ITRS to horizontal (east, north, zenith) matrix, given latitude, longitude,
        /// and deflection of the vertical.
        /// </summary>
        /// <remarks><paramref name="xi"/> and <paramref name="eta"/> are positive if the
        /// plumb line intersects the celestial sphere north and east of the ellipsoid
        /// normal.</remarks>
        /// <param name="lat">north latitude (radians)</param>
        /// <param name="lon">east longitude (radians)</param>
        /// <param name="xi">north deflection of the vertical (radians)</param>
        /// <param name="eta">east deflection of the vertical (radians)</param>
        /// <returns>ITRS to horizontal matrix</returns>
        public static RMatrix ItrsToHor(double lat, double lon, double xi, double eta)
        {
#if false
            Vector vNorth = new Vector(new Spherical(Angle.HalfPi, -xi));
            Vector vEast = new Vector(new Spherical(0, -eta));
            Vector vZenith = vEast.Cross(vNorth);
            Spherical sph = new Spherical(vZenith);

            // The problem is similar to computing the GCRS to celestial intermediate matrix from
            // the CIP x and y coordinates. See the SOFA C2IXYS routine.
            // I use the same sequence of rotations, but ignore the tiny angle s.
            return ItrsToHor(lat, lon).ZRotate(sph.Theta)
                .YRotate(sph.PolarAngle).ZRotate(-sph.Theta);
#endif
            return ItrsToHor(lat, lon, xi, eta, null);
        }


        /// <summary>
        /// ITRS to horizontal (east, north, zenith) matrix, given a Topocenter object
        /// and deflection of the vertical.
        /// </summary>
        /// <remarks><paramref name="xi"/> and <paramref name="eta"/> are positive if the
        /// plumb line intersects the celestial sphere north and east of the ellipsoid
        /// normal.</remarks>
        /// <param name="position">geographic position</param>
        /// <param name="xi">north deflection of the vertical (radians)</param>
        /// <param name="eta">east deflection of the vertical (radians)</param>
        /// <returns>ITRS to horizontal matrix</returns>
        public static RMatrix ItrsToHor(Topocenter position, double xi, double eta)
        {
#if false
            return ItrsToHor(position.Lat, position.Lon, xi, eta);
#endif
            return ItrsToHor(position, xi, eta, null);
        }



        /// <summary>
        /// GCRS to ecliptic and mean (or true) equinox matrix, given the GCRS to mean
        /// (or true) matrix and the mean (or true) obliquity.
        /// </summary>
        /// <remarks>The parameters must be both mean or both true</remarks>
        /// <param name="gcrsToEquator">GCRS to mean (or true) equator and equinox matrix
        /// </param>
        /// <param name="obliquity">mean or true obliquity (radians)</param>
        /// <returns>GCRS to ecliptic matrix</returns>
        public static RMatrix GcrsToEclip(RMatrix gcrsToEquator, double obliquity)
        {
            return gcrsToEquator.XRotate(obliquity);
        }


        /// <summary>
        /// GCRS to ecliptic and mean equinox matrix (IAU 2006 precession).
        /// </summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>GCRS to mean ecliptical matrix</returns>
        public static RMatrix GcrsToEclipMean06(JulianDate tt)
        {
            return GcrsToEclip(Precess06(tt), MeanObliq06(tt));
        }


        /// <summary>
        /// GCRS to ecliptic and true equinox system, IAU 2006 and 2000B nutation.
        /// </summary>
        /// <param name="tt">Terrestrial Time</param>
        /// <returns>rotation matrix, GCRS to the true ecliptic system</returns>
        public static RMatrix GcrsToEclipTrue06b(JulianDate tt) {
            return GcrsToEclip(GcrsToTrue06b(tt), TrueObliq06b(tt));
        }



        #region self-test code
#if DEBUG
        /// <summary>
        /// self-test the class
        /// </summary>
        public static void Test()
        {
            RMatrix rmat1, rmat2, gcrsToCirs, tirsToItrs, itrsToHor, gcrsToItrs, deflect;
            double[,] arr1, arr2, rb, rp, rbp, rn, rbpn;
            Vector vec1, vec2, vec3, northVec, eastVec, plumbVec, cipVec;
            PVVector pvvec1, pvvec2;
            bool error = false;
            JulianDate tt, ut1;
            Duration deltaT;
            double deltaPsi, deltaEpsilon, epsilon, era;
            double tolerance, sepAngle;
            double x, y, s;
            double lat, lon, xi, eta, degLat, degLon, poleX, poleY, cipAzimuth;
            Topocenter top1;


            // Test the test methods.
            arr1 = new double[,] {
                { 1.1, 1.2, 1.3 }, { 2.1, 2.2, 2.3 }, { 3.1, 3.2, 3.3 } };
            arr2 = new double[,] {
                { 1.1, 1.2, 1.3 }, { 2.1, 2.2, 2.3 }, { 3.1, 3.2, 0.0 } };  // note last number
            rmat1 = new RMatrix(arr1);
            rmat2 = new RMatrix(arr1);
            rmat1.EqualsApprox(rmat2, 0.0, "EqualsApprox() failed");
            rmat1 = new RMatrix(arr2);
            // Should throw exception.
#if false
            rmat1.EqualsApprox(rmat2, 0.0, "EqualsApprox() correct exception");
#endif

            // default constructor
            rmat1 = new RMatrix();
            arr1 = new double[,]
                { { 1.0, 0.0, 0.0 }, { 0.0, 1.0, 0.0 }, { 0.0, 0.0, 1.0 } };
            rmat1.EqualsApprox(arr1, 0.0, "default constructor");

            // constructor that takes 9 doubles
            rmat1 = new RMatrix(1.1, 1.2, 1.3, 2.1, 2.2, 2.3, 3.1, 3.2, 3.3);
            arr1 = new double[,]
                {{1.1, 1.2, 1.3}, {2.1, 2.2, 2.3}, {3.1, 3.2, 3.3}};
            rmat1.EqualsApprox(arr1, 0.0, "construct from 9 doubles");

            // Test constructor that takes an array of doubles. Re-use the above array.
            rmat1 = new RMatrix(arr1);
            arr2 = new double[,]
                { { 1.1, 1.2, 1.3 }, { 2.1, 2.2, 2.3 }, { 3.1, 3.2, 3.3 } };
            rmat1.EqualsApprox(arr2, .0, "construct from array of doubles");

            // Verify it makes a copy of the source array.
            assertTrue(rmat1.mat != arr1, "Array constructor didn't copy source.");

            // Constructor that takes an array of doubles, plus a bool to indicate
            // array is immutable.
            rmat1 = new RMatrix(arr1, false);
            rmat1.EqualsApprox(arr2, 0.0, "Construct from double[,] and false bool.");
            assertTrue(rmat1.mat != arr1,
                "Construct from double[,] and false bool; references equal.");
            rmat1 = new RMatrix(arr1, true);
            rmat1.EqualsApprox(arr2, 0.0, "Construct from double[,], and true bool.");
            assertTrue(rmat1.mat == arr1,
                "Construct from double[,] & true bool; references not equal.");
            
            // construct from an RMatrix
            rmat2 = new RMatrix(rmat1);
            rmat2.EqualsApprox(rmat1, .0, "Construct from RMatrix.");

            // Construct from a rotation vector.
            // First, construct a rotation vector to x-rotate the system by +pi/2.
            vec1 = new Vector(Angle.HalfPi, 0.0, 0.0);
            // Use that to construct a rotation matrix, then test it on a vector.
            rmat1 = new RMatrix(vec1);
            vec2 = rmat1 * new Vector(0.0, 1.0, 0.0);
            vec2.EqualsApprox(0.0, 0.0, -1.0, 1e-9);

            // Construct from 3 basis vectors such that a 90 degree Z rotation is applied.
            vec1 = new Vector(0, 1, 0);         // X basis vector
            vec2 = new Vector(-1, 0, 0);        // Y basis vector
            vec3 = new Vector(0, 0, 1);         // Z basis vector
            vec1 = new RMatrix(vec1, vec2, vec3) * new Vector(1, 0, 0);
            vec1.EqualsApprox(0, -1, 0, 1e-9);

            // Get the specified row or column.
            arr1 = new double[,] { { 11, 12, 13 }, { 21, 22, 23 }, { 31, 32, 33 } };
            rmat1 = new RMatrix(arr1, true);
            vec1 = rmat1.Row(1);
            vec1.EqualsApprox(11, 12, 13, 0.0, "RMatrix Row(1)");
            error = false;
            try {
                vec1 = rmat1.Row(0);
                error = true;
            } catch {
                if (error)
                    throw new Exception("RMatrix.Row(0)");
            }

            vec1 = rmat1.Column(1);
            vec1.EqualsApprox(11, 21, 31, 0.0, "RMatrix.Column(1)");
            try {
                vec1 = rmat1.Column(0);
                error = true;
            } catch {
                if (error)
                    throw new Exception("RMatrix.Column(0)");
            }

            // Frame bias property. Correct bias matrix from Almanac 2012 p. B51.
            arr1 = new double[,] {
                {.9999999999999942, -.000000071, .00000008056},
                {.000000071, .9999999999999969, .00000003306},
                {-.00000008056, -.00000003306, .9999999999999962}};
            IcrsToJ2000.EqualsApprox(arr1, 1e-9, "frame bias matrix");

            // ToString()
            arr1 =
                new double[,] { { 11, 12, 13 }, { 21, 22, 23 }, { 31, 32, 33 } };
            rmat1 = new RMatrix(arr1);
            assertTrue(rmat1.ToString().Equals(
                "11.000000 12.000000 13.000000\n" +
                "21.000000 22.000000 23.000000\n" +
                "31.000000 32.000000 33.000000"), "ToString()");

            // ToArray()
            arr1 = rmat1.ToArray();
            rmat1.EqualsApprox(arr1, 0.0, "ToArray()");
            assertTrue(arr1 != rmat1.mat, "ToArray() didn't create new array.");


            // x-rotation and RMatrix * Vector operator
            rmat1 = new RMatrix();
            rmat2 = rmat1.XRotate(-Angle.HalfPi);
            assertTrue(rmat1.mat != rmat2.mat,
                "XRotate(angle) didn't create new array.");
            vec1 = new Vector(0.0, 1.0, 0.0);
            vec2 = rmat2 * vec1;
            vec2.EqualsApprox(0.0, 0.0, 1.0, 1e-9, "Rmatrix.XRotate(angle) * Vector.");

            // same x-rotation, given cosine and sine (which need scaling)
            rmat2 = rmat1.XRotate(0, -2.0);
            assertTrue(rmat1.mat != rmat2.mat,
                "XRotate(cos,sin) didn't create new array.");
            vec2 = rmat2 * vec1;
            vec2.EqualsApprox(0.0, 0.0, 1.0, 1e-9, "RMatrix.XRotate(cos, sin) * Vector.");

            // y-rotation and RMatrix.Multiply(Vector) method
            rmat2 = rmat1.YRotate(-Angle.Pi);
            assertTrue(rmat1.mat != rmat2.mat,
                "YRotate(angle) didn't create new array.");
            vec1 = new Vector(1.0, 0.0, 0.0);
            vec2 = rmat2.Multiply(vec1);
            vec2.EqualsApprox(-1.0, 0.0, 0.0, 1e-9,
                "RMatrix.YRotate(angle).Multiply(Vector).");

            // same y-rotation, given cosine and sine
            rmat2 = rmat1.YRotate(-1.0, 0.0);
            assertTrue(rmat1.mat != rmat2.mat,
                "YRotate(cos,sin) didn't create new array.");
            vec2 = rmat2.Multiply(vec1);
            vec2.EqualsApprox(-1.0, 0.0, 0.0, 1e-9, "YRotate(cos,sin).Multiply(Vector)");

            // z-rotation
            rmat2 = rmat1.ZRotate(Angle.HalfPi);
            assertTrue(rmat1.mat != rmat2.mat,
                "RMatrix.ZRotate(angle) didn't create new array.");
            vec1 = new Vector(0.0, 1.0, 0.0);
            vec2 = rmat2 * vec1;
            vec2.EqualsApprox(1.0, 0.0, 0.0, 1e-9, "RMatrix.ZRotate(angle) * Vector");

            // same z-rotation, given cosine and sine (which need scaling)
            rmat2 = rmat1.ZRotate(0.0, .5);
            assertTrue(rmat1.mat != rmat2.mat,
                "ZRotate(cos,sin) didn't create new array.");
            vec2 = rmat2 * vec1;
            vec2.EqualsApprox(1.0, 0.0, 0.0, 1e-9, "RMatrix.ZRotate(cos,sin) * Vector");

            // RMatrix.Multiply(RMatrix) method
            rmat1 = new RMatrix().XRotate(Angle.HalfPi);
            rmat2 = new RMatrix().YRotate(Angle.HalfPi);
            vec1 = new Vector(0.0, 1.0, 0.0);
            vec2 = rmat2.Multiply(rmat1) * vec1;
            vec2.EqualsApprox(1.0, 0.0, 0.0, 1e-9);

            // RMatrix * RMatrix operator
            vec2 = rmat2 * rmat1 * vec1;
            vec2.EqualsApprox(1.0, 0.0, 0.0, 1e-9);

            // RMatrix.Multiply(PVVector) method
            rmat1 = new RMatrix().ZRotate(-Angle.HalfPi);
            pvvec1 = new PVVector(new Vector(1.0, 0.0, 0.0),
                new Vector(1.0, 0.0, 0.0));
            rmat1.Multiply(pvvec1).EqualsApprox(0.0, 1.0, 0.0, 0.0, 1.0, 0.0, 1e-9,
                "RMatrix.Multiply(PVVector)");

            // RMatrix * PVVector operator
            pvvec2 = rmat1 * pvvec1;
            pvvec2.EqualsApprox(0.0, 1.0, 0.0, 0.0, 1.0, 0.0, 1e-9, "RMatrix * PVVector");

            // Transpose() method. Create an RMatrix to Z rotate +pi/2, transpose it to
            // do the inverse rotation.
            rmat1 = new RMatrix().ZRotate(Angle.HalfPi);
            rmat2 = rmat1.Transpose();
            assertTrue(rmat1.mat != rmat2.mat, "Transpose() didn't create new matrix.");
            vec1 = new Vector(1, 0, 0);
            vec2 = rmat2 * vec1;
            vec2.EqualsApprox(0.0, 1.0, 0.0, 1e-9, "Transpose()");

            // obliquity & nutation

            // IAU 1980 model. (SOFA test)
            equalsApprox(MeanObliq80(new JulianDate(2400000.5, 54388.0)),
                0.4090751347643816218, 1e-14, "MeanObliq80()");
            NutationAngles80(new JulianDate(2400000.5, 53736.0),
                out deltaPsi, out deltaEpsilon);
            equalsApprox(deltaPsi, -0.9643658353226563966e-5, 1e-13,
                "NutationAngles80() deltaPsi");
            equalsApprox(deltaEpsilon, 0.4060051006879713322e-4, 1e-13,
                "NutationAngles80() deltaEpsilon");

            // IAU 2000 model. (SOFA test, except for mean obliqutiy)
            tt = new JulianDate(2400000.5, 53736.0);
            rb = new double[3, 3];
            rp = new double[3, 3];
            rbp = new double[3, 3];
            rn = new double[3, 3];
            rbpn = new double[3, 3];
            Sofa.Pn00b(JulianDate.J2000Base.ToDouble(), tt.ToJ2000(), out deltaPsi,
                out deltaEpsilon, out epsilon, rb, rp, rbp, rn, rbpn);
            equalsApprox(MeanObliq00(tt), epsilon, 1e-13, "MeanObliquity00()");
            NutationAngles00b(tt, out deltaPsi, out deltaEpsilon);
            equalsApprox(deltaPsi, -0.9632552291148362783e-5, 1e-13,
                "NutationAngles00b() deltaPsi");
            equalsApprox(deltaEpsilon, 0.4063197106621159367e-4, 1e-13,
                "NutationAngles00b() deltaEpsilon");
            NutationAngles00a(tt, out deltaPsi, out deltaEpsilon);
            equalsApprox(deltaPsi, -0.9630909107115518431e-5, 1e-13,
                "NutationAngles00a() deltaPsi");
            equalsApprox(deltaEpsilon, 0.4063239174001678710e-4, 1e-13,
                "NutationAngles00a() deltaEpsilon");

            // IAU 2006 model. (SOFA tests)
            tt = new JulianDate(2400000.5, 54388.0);
            equalsApprox(MeanObliq06(tt), 0.4090749229387258204, 1e-14, "MeanObliq06()");
            tt = new JulianDate(2400000.5, 53736.0);
            NutationAngles00b(tt, out deltaPsi, out deltaEpsilon);
            equalsApprox(deltaPsi, -0.9632552291148362783e-5, 1e-13,
                "NutationAngles06a() deltaPsi");
            equalsApprox(deltaEpsilon, 0.4063197106621159367e-4, 1e-13,
                "NutationAngles06a() deltaEpsilon");
            equalsApprox(TrueObliq06b(tt), MeanObliq06(tt) + deltaEpsilon, 1e-12,
                "TrueObliquity06b()");

#if true
            // Long term model's mean obliquity. Compare to IAU 2006 value. Not a
            // truly satisfactory test, but can't think of anything better.
            equalsApprox(MeanObliqLongTerm(tt), MeanObliq06(tt), Angle.DmsToRad(0,0,.1),
                "MeanObliqLongTerm()");
#endif

            // TrueObliq() and TrueObliq06b(). Values from Almanac.
            tt = new JulianDate(2012, 4, 1, 0.0);
            RMatrix.NutationAngles06a(tt, out deltaPsi, out deltaEpsilon);
            equalsApprox(TrueObliq(MeanObliq06(tt), deltaEpsilon),
                Angle.DmsToRad(23, 26, 12.8152), Angle.DmsToRad(0, 0, .0001),
                "TrueObliq()");
            equalsApprox(TrueObliq06b(tt),
                Angle.DmsToRad(23, 26, 12.8152), Angle.DmsToRad(0, 0, .001),
                "TrueObliq06b()");
            

            // CIP XY and s (2012 Almanac).

            tt = new JulianDate(2012, 4, 1, 0.0);
            GcrsToTrue06b(tt).CipXY(out x, out y);
            equalsApprox(x, +12205202e-10, 1e-8, "CipXY() x");
            equalsApprox(y, -156010e-10, 1e-8, "CipXY() y");

            s = Angle.DmsToRad(0, 0, .0048);
            tolerance = Angle.DmsToRad(0, 0, .001);
            equalsApprox(S00(tt, x, y), s, tolerance, "S00()");
            equalsApprox(S06(tt, x, y), s, tolerance, "S06()");
            equalsApprox(S06a(tt), s, tolerance, "S06a()");


            // Long term precession model by Vondrak et al. The correct result is from the A&A
            // article test example (including the erratum). However, the original code precesses
            // from the J2000 frame while my PrecessLongTerm() precesses from the ICRS. Thus the
            // frame bias correction is removed when computing rmat1.
            tt = new JulianDate(-1374, 5, 3, 13, 52, 19.2);
            rmat1 = PrecessLongTerm(tt) * IcrsToJ2000.Transpose();
            rmat1.EqualsApprox(new double[,] {
                {0.684733909271267, 0.666477936491748, 0.294867145785675},
                {-0.666694822433781, 0.736256364537221, -0.115950762905741},
                {-0.29437643797369, -0.117190980233703, 0.948477088240821}},
                1e-14, "LongTermPrecession.Pmat()");

            // GCRS to true. This is a combined test of precession and nutation.

            tt = new JulianDate(2012, 11, 16, 0.0);
            rmat1 = new RMatrix(1.0 - 51278e-10, -29371971e-10, -12760907e-10,
                29372305e-10, 1.0 - 43140e-10, 242916e-10,
                12760139e-10, -280397e-10, 1.0 - 8145e-10);

            // GCRS to true, 2006/2000B
            tolerance = Angle.DmsToRad(0, 0, .001);
            rmat2 = Precess06(tt);
            epsilon = MeanObliq06(tt);
            NutationAngles00b(tt, out deltaPsi, out deltaEpsilon);
            GcrsToTrue(rmat2, epsilon, deltaPsi, deltaEpsilon)
                .EqualsApprox(rmat1, tolerance, "GcrsToTrue() with 2006/00B angles");
            GcrsToTrue06b(tt).EqualsApprox(rmat1, tolerance, "GcrsToTrue06b()");

            // Test Nutate() with the values above.
            (Nutate(epsilon, deltaPsi, deltaEpsilon) * rmat2).EqualsApprox(rmat1, tolerance,
                "Nutate()");

            // GCRS to true, 2000/2000A and B
            tolerance = Angle.DmsToRad(0, 0, .01);
            rmat2 = Precess00(tt);
            epsilon = MeanObliq00(tt);
            NutationAngles00a(tt, out deltaPsi, out deltaEpsilon);
            GcrsToTrue(rmat2, epsilon, deltaPsi, deltaEpsilon).EqualsApprox(
                rmat1, tolerance, "GCRS to true, 2000/2000A");
            NutationAngles00b(tt, out deltaPsi, out deltaEpsilon);
            GcrsToTrue(rmat2, epsilon, deltaPsi, deltaEpsilon).EqualsApprox(
                rmat1, tolerance, "GCRS to true, 2000/2000B");

#if true
            // GCRS to true, long term precession model + 2000B nutation angles above.
            GcrsToTrue(PrecessLongTerm(tt), MeanObliqLongTerm(tt), deltaPsi, deltaEpsilon)
                .EqualsApprox(rmat1, tolerance, "GCRS to true, long term");
#endif

            // GCRS to true, 1976/1980. Relax the tolerance a little.
            tolerance = Angle.DmsToRad(0, 0, .1);
            NutationAngles80(tt, out deltaPsi, out deltaEpsilon);
            GcrsToTrue(Precess76(tt), MeanObliq80(tt), deltaPsi, deltaEpsilon)
                .EqualsApprox(rmat1, tolerance, "GCRS to true, 1976/1980");


            // GCRS to celestial intermediate.

            // Correct rotation matrix from Almanac.
            tt = new JulianDate(2012, 11, 16, 0.0);
            rmat1 = new RMatrix(1.0 - 8141e-10, -124e-10, -12760139e-10,
                482e-10, 1.0 - 4e-10, 280396e-10,
                12760139e-10, -280397e-10, 1.0 - 8145e-10);

            // Compute the CRS to true matrix, CIP xy, and s, with the 2006/2000B model.
            rmat2 = GcrsToTrue06b(tt);
            rmat2.CipXY(out x, out y);
            s = S06a(tt);

            // Test the GCRS to celestial intermediate methods.
            tolerance = Angle.DmsToRad(0, 0, .001);
            GcrsToCirs(x, y, s).EqualsApprox(rmat1, tolerance, "CrsToCirs(x, y, s)");
            GcrsToCirs(rmat2, s).EqualsApprox(rmat1, tolerance, "CrsToCirs(rmat2, s)");
            GcrsToCirs06b(tt).EqualsApprox(rmat1, tolerance, "CrsToCirs06b(tt)");

            // polar motion matrix

            // Wallace example http://syrte.obspm.fr/iauWGnfa/ExPW04.html
            tolerance = 1e-9;
            tt = new JulianDate(2003, 08, 26, 00, 37, 38.973810);
            rmat1 = new RMatrix(
               +0.99999999999920940000, -0.00000000000831300878, +0.00000125746609283028,
               +0.00000000001084649458, +0.99999999999797040000, -0.00000201475475899438,
               -0.00000125746609281098, +0.00000201475475900642, +0.99999999999717980000);
            TirsToItrs(tt, Angle.DmsToRad(0, 0, +0.259371),
                Angle.DmsToRad(0, 0, +0.415573)).EqualsApprox(rmat1, tolerance,
                "TirsToItrs()");


            // GCRS to TIRS and ITRS

            // Since the GCRS to CIRS and the TIRS to ITRS methods have been proven,
            // use them to compute the expected GCRS to ITRS matrix, then test
            // GcrsToTirs(), GcrsToItrs() and GcrsToItrs06b().
            tolerance = 1e-9;
            tt = new JulianDate(2012, 11, 16, 0.0);
            deltaT = new Duration(0, 0, 67.0);
            ut1 = tt - deltaT;
            x = Angle.DmsToRad(0, 0, .2);   // polar motion
            y = Angle.DmsToRad(0, 0, .4);
            gcrsToCirs = GcrsToCirs06b(tt);
            era = Angle.Era00(ut1);
            tirsToItrs = TirsToItrs(tt, x, y);
            gcrsToItrs = tirsToItrs * gcrsToCirs.ZRotate(era);

            // new RMatrix(TirsToItrs(tt, x, y) * GcrsToTirs(GcrsToCirs06b(tt),
            //     Angle.Era00(ut1))).EqualsApprox(rmat1, tolerance, "GcrsToTirs(");
            rmat1 = tirsToItrs * GcrsToTirs(gcrsToCirs, era);
            rmat1.EqualsApprox(gcrsToItrs, tolerance, "GcrsToTirs(");

            // GcrsToItrs(GcrsToCirs06b(tt), Angle.Era00(ut1), TirsToItrs(tt, x, y))
            //    .EqualsApprox(rmat1, tolerance, "GcrsToItrs()");
            rmat1 = GcrsToItrs(gcrsToCirs, era, tirsToItrs);
            rmat1.EqualsApprox(gcrsToItrs, tolerance, "GcrsToItrs()");

            rmat1 = GcrsToItrs06b(tt, ut1, x, y);
            rmat1.EqualsApprox(gcrsToItrs, tolerance, "GcrsToItrs06b()");


            // ITRS to horizontal

            // observer at equator & prime meridian
            rmat1 = ItrsToHor(0, 0);

            // Check the axes of the horizontal coordinate system, in terms of the ITRS.
            // The +x axis should be directed to the ITRS +y, and the +y to the ITRS +z.
            vec1 = new Vector(0, 1, 0);
            vec2 = new Vector(0, 0, 1);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor(0,0) x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor(0,0) y axis");

            // Same check on the other overloads.

            rmat1 = ItrsToHor(0, 0, 0, 0);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor(0,0,0,0) x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor(0,0,0,0) y axis");

            top1 = new Topocenter(0, 0, 0);
            rmat1 = ItrsToHor(top1);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor(Topocenter(0,0)), x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor(Topocenter(0,0)), y axis");

            rmat1 = ItrsToHor(top1, 0, 0);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor(Topocenter(0,0)), 0, 0), x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor(Topocenter(0,0)), 0, 0), y axis");

            // the master method
            rmat1 = ItrsToHor(0, 0, 0, 0, null);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor() master method, x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor() master method, y axis");


            // Now at latitude 10 north.

            degLat = 10.0;
            degLon = 0.0;
            lat = Angle.DegreesToRadians(degLat);
            lon = 0.0;
            vec2 = new Vector(-Math.Sin(lat), 0, Math.Cos(lat));

            rmat1 = ItrsToHor(lat, lon);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor(10N,0), x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor(10N,0), y axis");

            rmat1 = ItrsToHor(lat, 0, 0, 0);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor(10N,0,0,0), x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor(10N,0,0,0), y axis");

            top1 = new Topocenter(lat, lon, 0);
            rmat1 = ItrsToHor(top1);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor(Topocenter(10,0)), x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor(Topocenter(10,0)), y axis");

            rmat1 = ItrsToHor(top1, 0, 0);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor(Topocenter(10,0)), x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor(Topocenter(10,0)), y axis");

            // master method
            rmat1 = ItrsToHor(lat, lon, 0, 0, null);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor() master method, x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor() master method, y axis");


            // latitude 0, longitude +20
            degLat = 0.0;
            degLon = 20.0;
            lat = 0.0;
            lon = Angle.DegreesToRadians(degLon);
            vec1 = new Vector(-Math.Sin(lon), Math.Cos(lon), 0);
            vec2 = new Vector(0, 0, 1);

            rmat1 = ItrsToHor(lat, lon);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor(0,20), x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor(0,20), y axis");

            rmat1 = ItrsToHor(lat, lon, 0, 0);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor(0,20,0,0), x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor(0,20,0,0), y axis");

            top1 = new Topocenter(lat, lon, 0);
            rmat1 = ItrsToHor(top1);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor(Topocenter(0,20)), x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor(Topocenter(0,20)), y axis");

            rmat1 = ItrsToHor(top1, 0, 0);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor(Topocenter(0,20),0,0), x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor(Topocenter(0,20),0,0), y axis");

            // master method
            rmat1 = ItrsToHor(lat, lon, 0, 0, null);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor() master method, x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor() master method, y axis");

            // Return to lat/lon 0, and deflect the vertical north.

            degLat = degLon = 0.0;
            lat = lon = 0.0;
            xi = Angle.DmsToRad(0, 1, 0);
            eta = 0.0;
            vec1 = new Vector(0, 1, 0);
            vec2 = new Vector(-Math.Sin(xi), 0, Math.Cos(xi));

            rmat1 = ItrsToHor(lat, lon, xi, eta);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor(0,0,1',0) x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor(0,0,1',0) y axis");

            top1 = new Topocenter(0, 0, 0);
            rmat1 = ItrsToHor(top1, xi, eta);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor(Topocenter(0,0),1',0) x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor(Topocenter(0,0),1',0) y axis");

            // master method
            rmat1 = ItrsToHor(lat, lon, xi, eta, null);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor() master method, x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor() master method, y axis");

            // Still at lat/lon 0, return xi to 0 and deflect eta east.

            xi = 0.0;
            eta = Angle.DmsToRad(0, 1, 0);
            vec1 = new Vector(-Math.Sin(eta), Math.Cos(eta), 0);
            vec2 = new Vector(0, 0, 1);

            rmat1 = ItrsToHor(lat, lon, xi, eta);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor(0,0,0,1') x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor(0,0,0,1') y axis");

            rmat1 = ItrsToHor(top1, xi, eta);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor(Topocenter(0,0),0,1') x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor(Topocenter(0,0),0,1') y axis");

            // master method
            rmat1 = ItrsToHor(lat, lon, xi, eta, null);
            assertTrue(rmat1.Row(1).SeparationAngle(vec1) < 1e-9,
                "ItrsToHor() master method, x axis");
            assertTrue(rmat1.Row(2).SeparationAngle(vec2) < 1e-9,
                "ItrsToHor() master method, y axis");

            // Test the master method with deflection of the vertical combined with polar motion.
            
            lat = Angle.DegreesToRadians(50);
            lon = 0;
            top1 = new Topocenter(lat, lon, 0);
            xi = eta = Angle.DmsToRad(0, 1, 0);
            poleX = poleY = Angle.DmsToRad(0, 0, 1);
            // rotation matrix, TIRS to ITRS
            tirsToItrs = TirsToItrs(new JulianDate(2015, 8, 25, 0), poleX, poleY);
            // rotation matrix, ITRS to horizontal, sans deflection & polar motion
            itrsToHor = ItrsToHor(top1);
            // vector to the CIP wrt ellipsoid horizontal
            cipVec = itrsToHor * tirsToItrs.Column(3);
            // deflected vectors wrt ellipsoid horizontal
            eastVec = new RMatrix().YRotate(eta) * new Vector(1, 0, 0);
            northVec = new RMatrix().XRotate(-xi) * new Vector(0, 1, 0);
            plumbVec = eastVec.Cross(northVec).Unit();
            // Set east vector perpendicular to plumb line and north vector.
            eastVec = northVec.Cross(plumbVec);
            // rotation matrix to apply deflection of the vertical
            deflect = new RMatrix(eastVec, northVec, plumbVec);
            // Z-rotate the deflection matrix so the CIP azimuth is exactly north.
            cipAzimuth = new Spherical(cipVec).Az;
            deflect = new RMatrix().ZRotate(-cipAzimuth) * deflect;
            // rotation matrix, ITRS to astronomic
            itrsToHor = deflect * itrsToHor;
            // Compare to ItrsToHor() master method.
            rmat1 = ItrsToHor(top1, xi, eta, tirsToItrs);
            assertTrue(rmat1.Row(1).SeparationAngle(itrsToHor.Row(1)) < 1e9,
                "ItrsToHor() master method.");
            assertTrue(rmat1.Row(2).SeparationAngle(itrsToHor.Row(2)) < 1e9,
                "ItrsToHor() master method.");


            // GCRS to ecliptical

            // Heliocentric Earth coords at 2015-07-01 0 h TT, J2000 mean equator and equinox,
            // from MICA.
            tt = JulianDate.J2000Base;
            vec1 = new Vector(0.153140622, -0.922100205, -0.399746302);
            // Convert to ICRS.
            rmat1 = RMatrix.IcrsToJ2000.Transpose();
            vec1 = rmat1 * vec1;

            // Set error tolerance to .2 arcsec.
            tolerance = Angle.DmsToRad(0, 0, .2);

            // Convert from ICRS to ecliptic and mean equinox, compare to MICA values.
            vec2 = RMatrix.GcrsToEclipMean06(tt) * vec1;
            vec3 = new Vector(new Spherical(Angle.DmsToRad(278, 39, 49.8),
                Angle.DmsToRad(0, 00, 06.2)));
            assertTrue(vec2.SeparationAngle(vec3) <= tolerance, "GcrsToEclipMean06()");

            // Convert from ICRS to true equatorial, then true ecliptical. (MICA uses 1976/80
            // precession nutation.)
            epsilon = RMatrix.MeanObliq80(tt);
            RMatrix.NutationAngles80(tt, out deltaPsi, out deltaEpsilon);
            rmat1 = RMatrix.GcrsToTrue(RMatrix.Precess76(tt), epsilon, deltaPsi, deltaEpsilon);
            // GcrsToEclip() was indirectly tested in the GcrsToEclipMean06(tt) test above.
            vec3 = RMatrix.GcrsToEclip(rmat1, epsilon + deltaEpsilon) * vec1;

            // Verify GcrsToEclipTrue06b() gives same result.
            vec2 = RMatrix.GcrsToEclipTrue06b(tt) * vec1;
            sepAngle = Angle.RadiansToDegrees(vec2.SeparationAngle(vec3)) * 3600.0;
            assertTrue(vec2.SeparationAngle(vec3) <= tolerance, "GcrsToEclipTrue06b()");
        }


        /// <summary>
        /// Compare to an array, throw exception if any element is out of tolerance.
        /// </summary>
        /// <param name="a">expected result</param>
        /// <param name="tolerance">max allowable discrepancy</param>
        /// <param name="message">message to display if error</param>
        /// <exception cref="Exception">if any element exceeds tolerance.</exception>
        public void EqualsApprox(double[,] a, double tolerance, string message)
        {
            double error;
            for (int i = 0; i < 3; ++i)
                for (int j = 0; j < 3; ++j) {
                    error = this.mat[i, j] - a[i, j];
                    if (Math.Abs(error) > tolerance)
                        throw new Exception(message + string.Format(
                            ": error at [{0:d}, {1:d}] = {2:e2}", i, j, error));
                }
        }


        /// <summary>
        /// Compare to another RMatrix, throw exception if any element is out of tolerance.
        /// </summary>
        /// <param name="rm">expected result</param>
        /// <param name="tolerance">max allowable discrepancy</param>
        /// <param name="message">message to display if error</param>
        public void EqualsApprox(RMatrix rm, double tolerance, string message)
        {
            EqualsApprox(rm.mat, tolerance, message);
        }


        private static void assertTrue(bool assertion, string message)
        {
            if (!assertion)
                throw new Exception(message);
        }

        private static void equalsApprox(double a, double b, double tolerance,
            string message)
        {
            if (Math.Abs(a - b) > tolerance)
                throw new Exception(message + string.Format(": error = {0:e2}", a - b));
        }
#endif
        #endregion
    }
}
