﻿// Sexagesimal.cs
// A class to represent and format sexagesimal quantities.
// Copyright 2018 by Paul S. Hirose

// revision history
#if false
2007-11-11 Minor change to argument name, improved XML comments.
2007-12-03 Finished self-test code.
2007-12-04 CompareTo(Sexagesimal) and related methods have been made null safe.
2008-11-07 Fixed bug in ToDouble() when seconds exceed 60 due to a step adjustment.
2012-03-31 ToString() modified so padding space will precede any sign.
2014-10-07 Minor improvement to exception message.
2014-10-22 All properties and methods are virtual now.
2014-11-01 Rewrote XML comments. Self test checks Unicode labels. OK.
2015-06-26 De-virtualized the class.

2017-02-13 Verified bug when angle accuracy is .001f arcsec and epoch is 2015-08-28s 0300 UTC.
The time formats as 2015-08-28T02:59:60.0000.

2017-11-30 Rewrote the constructor to eliminate the anomaly noted on 2017-02-13. Not yet tested.
2017-12-21 Self test OK, and the "seconds = 60" bugs are fixed.
2018-04-19 Bug fix. When beyond the midpoint of a leap second, and precision is 3600, time should
be 00:00:00, not 00:00:01. I.e., when roundoff takes you out of the step adjustment, the step
seconds parameter must be disregarded. Test OK.
#endif


using System;
using System.Text;
using System.Text.RegularExpressions;


namespace HirosePS.SofaJpl
{
    /// <summary>
    /// A Sexagesimal object formats a decimal quantity (degrees or hours) into a sexagesimal string
    /// for display.</summary>
    /// <remarks>
    /// <para>The DMS.s, DM.m, D.d, MS.s, M.m, and S.s formats are supported. The last three
    /// overflow to a different format if necessary. For instance, S.s overflows to MS.s or
    /// DMS.s if the value is too big. Precision of all formats is determined by a parameter to
    /// the object's constructor.</para>
    /// <para>Instances of this class are immutable.</para>
    /// </remarks>
    public class Sexagesimal : IComparable<Sexagesimal>, IComparable,
        IEquatable<Sexagesimal>, IFormattable
    {
        private char sign;
        private double degrees, minutes, seconds;

        // .1 = 10 deg resolution, 1 = 1 deg resolution, 60.0 = 1 minute resolution, etc.
        private double resolution;

        private int modulus;

        // the number of multiples (e.g., of 24 or 360) removed during a modulo operation
        private int revolutions;

        const double minPerDeg = 60.0;
        const double secPerMin = 60.0;
        const double secPerDeg = 3600.0;


        // constructors


        /// <summary>
        /// Construct from time in hours or angle in degrees, plus specified roundoff resolution,
        /// and modulus, with provision for leap seconds.
        /// </summary>
        /// <param name="degrees">Degrees or hours. Negative OK.</param>
        /// <param name="resolution">Resolution units per degree (or hour): 0.1 = 10 degree (or 10
        /// hour) resolution, 60 = 1 minute, 36000 = .1 second resolution. Must be &gt; 0.</param>
        /// <param name="modulus">A non-negative modulus, typically 0 (= no modulo adjustment), 24,
        /// or 360.</param>
        /// <param name="stepSeconds">Use 1.0 if time is inside a leap second, otherwise 0. A value
        /// less than 1 is legal to accommodate pre-1972 UTC. Ignored if negative.</param>
        /// <exception cref="ArgumentOutOfRangeException"> if <paramref name="modulus"/>
        /// &lt; 0, or <paramref name="resolution"/> &lt;= 0.
        /// </exception>
        /// <remarks>
        /// <para>The precision of the <paramref name="degrees"/> parameter limits the maximum
        /// practical resolution. A request for more is accepted, but the extra digits are not
        /// meaningful.</para>
        /// <para>The <paramref name="resolution"/> parameter is normalized before the object is
        /// constructed. If greater than 1, any fractional part is truncated. If less than 1, it's
        /// converted to the nearest reciprocal of a whole number. E.g. .19 is
        /// normalized to 1/5 (as near as can be represented in floating point), and thus the angle
        /// will be displayed rounded to the nearest multiple of 5 degrees.</para>
        /// <para>The <paramref name="modulus"/>, if nonzero, puts the result in
        /// the interval (-<paramref name="modulus"/>, +<paramref name="modulus"/>).</para>
        /// <para>The <paramref name="stepSeconds"/> parameter is to allow seconds to
        /// equal or exceed 60 during a UTC negative step adjustment. It adds directly to
        /// seconds. For example, if seconds would normally be 59.5, setting this
        /// parameter to 1.0 will make seconds 60.5. Note that the value is first rounded
        /// to the nearest multiple of the resolution unit, so at 1 minute resolution it
        /// has no effect.</para>
        /// </remarks>
        public Sexagesimal(
            double degrees,
            double resolution,
            int modulus,
            double stepSeconds)
        {
            if (resolution <= 0.0)
                throw new ArgumentOutOfRangeException(
                    "Sexagesimal constructor resolution argument must be > 0.");
            if (modulus < 0)
                throw new ArgumentOutOfRangeException(
                    "Sexagesimal constructor modulus argument must be >= 0.");

            if (degrees < 0.0) {
                this.sign = '-';
                degrees = -degrees;
            } else
                this.sign = '+';

            this.resolution = resolution;
            this.modulus = modulus;


            // old (buggy) algorithm
#if false
            // Round the angle to nearest integral number of resolution units.
            double units = Math.Floor(degrees * resolution + .5);

            // floating point degrees, quantized to the nearest resolution unit
            this.degrees = units / resolution;
            // number of resolution units in 1 second and 1 minute
            double unitsPerMinute = resolution / 60.0;
            double unitsPerSecond = resolution / 3600.0;

            // Get the fractional degrees. If nonzero, convert it to minutes and seconds.
            units -= Math.Floor(this.degrees) * resolution;
            if (units != 0.0) {
                // floating point minutes
                this.minutes = units / unitsPerMinute;

                // Get the fractional minutes.
                units -= Math.Floor(this.minutes) * unitsPerMinute;

                // Convert stepSeconds into resolution units. This is only applicable
                // during a negative UTC step adjustment, i.e., when stepSeconds is
                // positive.
                double stepUnits = 0.0;
                if (stepSeconds > 0.0)
                    stepUnits = Math.Floor(stepSeconds * unitsPerSecond + .5);

                // Get floating point seconds. The stepUnits are added here so seconds
                // can be >= 60 during a UTC negative step adjustment.
                this.seconds = (units + stepUnits) / unitsPerSecond;
            }
#endif

            // new algorithm

            // Make the resolution unit either an integer fraction of 1 degree, or
            // an integer number of degrees.
            if (resolution < 1.0)
                resolution = 1.0 / (Math.Floor(1.0 / resolution + .5));
            else
                resolution = Math.Floor(resolution + .5);

            // resolution units per minute and second
            double unitsPerMin = resolution / minPerDeg;
            double unitsPerSec = resolution / secPerDeg;

            // Convert the angle to the nearest whole number of resolution units.
            double units = Math.Floor(degrees * resolution + .5);

            // Convert stepSeconds to the nearest whole number of resolution units. This is only
            // applicable during a negative UTC step adjustment.
            double stepUnits = 0.0;
            if (stepSeconds > 0.0)
                stepUnits = Math.Floor(stepSeconds * unitsPerSec + .5);

            // degrees (or hours), minutes, seconds, all in multiples of the resolution unit
            this.degrees = units / resolution;
            units -= Math.Floor(this.degrees) * resolution;     // remove whole degrees
            this.minutes = units / unitsPerMin;
            units -= Math.Floor(this.minutes) * unitsPerMin;    // remove whole minutes

            // Do a modulo operation on degrees, if applicable.
            if (modulus > 0) {
                this.revolutions = (int)this.degrees / modulus * (sign == '+' ? 1 : -1);
                this.degrees %= modulus;
            }

            // If time has rounded up to the next day, it is outside the leap second, so the
            // time step must be disregarded.
            if (revolutions > 0.0)
                stepUnits = 0.0;
            this.seconds = (units + stepUnits) / unitsPerSec;
        }


        /// <summary>
        /// Construct from time in hours or angle in degrees, plus specified roundoff resolution,
        /// and modulus.
        /// </summary>
        /// <param name="degrees">Degrees or hours. Negative OK.</param>
        /// <param name="resolution">Resolution units per degree (or hour): 0.1 = 10 degree (or 10
        /// hour) resolution, 60 = 1 minute, 36000 = .1 second resolution. Must be &gt; 0.</param>
        /// <param name="modulus">A non-negative modulus, typically 0 (= no modulo adjustment), 24,
        /// or 360.</param>
        /// <exception cref="ArgumentOutOfRangeException"> if <paramref name="modulus"/>
        /// &lt; 0, or <paramref name="resolution"/> &lt;= 0.
        /// </exception>
        /// <remarks>
        /// <para>The precision of the <paramref name="degrees"/> parameter limits the maximum
        /// practical resolution. A request for more is accepted, but the extra digits are not
        /// meaningful.</para>
        /// <para>The <paramref name="resolution"/> parameter is normalized before the object is
        /// constructed. If greater than 1, any fractional part is truncated. If less than 1, it's
        /// converted to the nearest reciprocal of a whole number. E.g. .19 is
        /// normalized to 1/5 (as near as can be represented in floating point), and thus the angle
        /// will be displayed rounded to the nearest multiple of 5 degrees.</para>
        /// <para>The <paramref name="modulus"/>, if nonzero, puts the result in
        /// the interval (-<paramref name="modulus"/>, +<paramref name="modulus"/>).</para>
        /// </remarks>
        public Sexagesimal(
            double degrees,
            double resolution,
            int modulus)
            : this(degrees, resolution, modulus, 0.0)
        { }


        /// <summary>
        /// Construct from time in hours or angle in degrees, plus specified roundoff resolution.
        /// </summary>
        /// <param name="degrees">Degrees or hours. Negative OK.</param>
        /// <param name="resolution">Resolution units per degree (or hour): 0.1 = 10 degree (or 10
        /// hour) resolution, 60 = 1 minute, 36000 = .1 second resolution. Must be &gt; 0.</param>
        /// <exception cref="ArgumentOutOfRangeException"> if <paramref name="resolution"/> &lt;= 0.
        /// </exception>
        /// <remarks>
        /// <para>The precision of the <paramref name="degrees"/> parameter limits the maximum
        /// practical resolution. A request for more is accepted, but the extra digits are not
        /// meaningful.</para>
        /// <para>The <paramref name="resolution"/> parameter is normalized before the object is
        /// constructed. If greater than 1, any fractional part is truncated. If less than 1, it's
        /// converted to the nearest reciprocal of a whole number. E.g. .19 is
        /// normalized to 1/5 (as near as can be represented in floating point), and thus the angle
        /// will be displayed rounded to the nearest multiple of 5 degrees.</para>
        /// </remarks>
        public Sexagesimal(
            double degrees,
            double resolution)
            : this(degrees, resolution, 0, 0.0)
        { }




        // Properties.


        /// <summary>
        /// Get the sign of this instance.
        /// </summary>
        /// <value>'+' or '-'</value>
        public char Sign { get { return this.sign; } }

        /// <summary>
        /// Get the floating point degrees or hours of this instance.
        /// </summary>
        /// <value>degrees or hours, always positive</value>
        public double Degrees { get { return this.degrees; } }

        /// <summary>
        /// Get the floating point degrees or hours of this instance. Identical to
        /// Degrees; provided for convenience.
        /// </summary>
        /// <value>degrees or hours, always positive</value>
        public double Hours { get { return Degrees; } }

        /// <summary>
        /// Get the floating point minutes of this instance, i.e., the fractional part of
        /// degrees or hours, times 60.
        /// </summary>
        /// <value>minutes, always positive</value>
        public double Minutes { get { return this.minutes; } }

        /// <summary>
        /// Get the floating point seconds of this instance, i.e., the fractional part of
        /// minutes, times 60.
        /// </summary>
        /// <value>seconds, always positive</value>
        public double Seconds { get { return this.seconds; } }

        /// <summary>
        /// Get the resolution of this instance: 60.0 = 1 minute resolution,
        /// 3600.0 = 1 second, etc.
        /// </summary>
        /// <value>resolution</value>
        public double Resolution { get { return this.resolution; } }

        /// <summary>
        /// Get the whole multiples (e.g., of 24 or 360) removed during a modulo
        /// operation. Zero if no modulus was passed to the constructor. Can be negative,
        /// if the value of this instance is negative.
        /// </summary>
        /// <value>integer (can be negative) number of multiples</value>
        public int Revolutions { get { return this.revolutions; } }




        // methods


        #region IComparable<Sexagesimal> Members

        /// <summary>
        /// Compare the value of this instance (from ToDouble()) to
        /// <paramref name="other"/>. Whole multiples of the modulus are ignored.
        /// </summary>
        /// <param name="other">the other instance to compare to</param>
        /// <returns>a value less than, equal to, or greater than 1, if value of this
        /// instance is less than, equal to, or greater than <paramref name="other"/>,
        /// respectively</returns>
        public int CompareTo(Sexagesimal other)
        {
            if ((object)other == null)
                return 1;
            return ToDouble().CompareTo(other.ToDouble());
        }

        #endregion


        #region IComparable Members

        /// <summary>
        /// Compare this instance and <paramref name="obj"/> by value, according to the
        /// value from ToDouble(). Whole multiples of the modulus (e.g., 360 degrees) are
        /// ignored.
        /// </summary>
        /// <param name="obj">the other object to compare to</param>
        /// <returns>integer which is &lt;0, 0, or &gt;0, if this instance is less than,
        /// equal to, or greater than <paramref name="obj"/>, respectively.
        /// </returns>
        /// <exception cref="ArgumentException"> if <paramref name="obj"/> is the wrong
        /// type.</exception>
        public int CompareTo(object obj)
        {
            if (obj == null)
                return 1;
            if (this.GetType() != obj.GetType())
                throw new ArgumentException("Sexagesimal CompareTo(obj): type mismatch.");
            return this.CompareTo((Sexagesimal)obj);
        }

        #endregion


        #region IEquatable<Sexagesimal> Members

        /// <summary>
        /// Test this instance and <paramref name="other"/> for value equality, based on
        /// ToDouble(). Whole multiples of the modulus (e.g., 360 degrees) are ignored.
        /// </summary>
        /// <param name="other">A Sexagesimal instance.</param>
        /// <returns>true if the values are equal</returns>
        public bool Equals(Sexagesimal other)
        {
            return this.CompareTo(other) == 0;
        }

        #endregion


        /// <summary>
        /// Override of the Equals(object) method. Compares this instance to
        /// <paramref name="obj"/> by value, using ToDouble().
        /// </summary>
        /// <param name="obj">Must be castable to Sexagesimal.</param>
        /// <returns>true if the values are equal</returns>
        public override bool Equals(object obj)
        {
            if (obj == null || GetType() != obj.GetType())
                return false;
            return this.CompareTo((Sexagesimal)obj) == 0;
        }


        /// <summary>
        /// Get a hash code based on the value of this instance.
        /// </summary>
        /// <returns>hash of the ToDouble() value</returns>
        public override int GetHashCode()
        {
            return this.ToDouble().GetHashCode();
        }

        
        /// <summary>
        /// Convert the value of this instance to a string.
        /// </summary>
        /// <returns>String in "+g " format</returns>
        public override string ToString()
        {
            return string.Format("{0:+g }", this);
        }

        /// <summary>
        /// Convert the value of this instance (rounded to the resolution specified to the
        /// constructor) to floating point. Whole multiples of the modulus are ignored.
        /// </summary>
        /// <returns>degrees or hours</returns>
        public double ToDouble()
        {
            return (Math.Floor(this.Hours)
                + Math.Floor(this.Minutes) / 60.0
                + this.Seconds / 3600.0)
                * (sign == '+' ? 1.0 : -1.0);
        }


        // old versions
#if false
        /// <summary>
        /// Returns the number of decimal places that should be displayed in floating
        /// point degrees, appropriate to the resolution of this instance. If resolution
        /// &lt; 2, this method returns 0. If 2 &lt;= resolution &lt; 20, it returns 1.
        /// If 20 &lt;= resolution &lt; 200, it returns 2, etc.
        /// </summary>
        /// <returns>integer &gt;= 0</returns>
        public int DegreeDecimals()
        {
            int i = (int)(Math.Log10(Resolution) + .8);
            return (i > 0) ? i : 0;
        }


        /// <summary>
        /// Returns the number of decimal places that should be displayed in floating
        /// point minutes, appropriate to the resolution of this instance. If resolution
        /// &lt; 100, this method returns 0. If 100 &lt;= resolution &lt; 1000, it
        /// returns 1. If 1000 &lt;= resolution &lt; 10000, it returns 2, etc.
        /// </summary>
        /// <returns>integer &gt;= 0</returns>
        public int MinuteDecimals()
        {
            int i = (int)Math.Log10(Resolution) - 1;
            return (i > 0) ? i : 0;
        }


        /// <summary>
        /// Returns the number of decimal places that should be displayed in floating
        /// point seconds, appropriate to the resolution of this instance. If resolution
        /// &lt; 6300 (approx.), this method returns 0. (The value 6300 corresponds to a
        /// resolution a bit more coarse than .5 second.) If 6300 &lt;= resolution &lt;
        /// 63000, it returns 1. If 63000 &lt;= resolution &lt; 630000, it returns 2, etc.
        /// </summary>
        /// <returns>integer &gt;= 0</returns>
        public int SecondDecimals()
        {
            int i = (int)(Math.Log10(Resolution) - 2.8);
            return (i > 0) ? i : 0;
        }
#endif  // end old DegreeDecimals(), MinuteDecimals(), SecondDecimals()


        // new versions
#if true

        /// <summary>
        /// Returns the number of decimal places that should be displayed in floating
        /// point degrees, appropriate to the resolution of this instance: 0 if resolution
        /// less than 1.26, 1 if resolution is 1.26 to 12.6, etc.
        /// </summary>
        /// <returns>integer &gt;= 0</returns>
        public int DegreeDecimals() {
            // If the constant were 1.0 instead of .9, this method would call for degrees to be
            // desplayed with one decimal place if resolution were 1.0 (i.e., 1-degree resolution).
            // That's excessive, so I adjust the constant by a tenth order of magnitude.
            int i = (int)(Math.Log10(Resolution) + .9);
            return (i > 0) ? i : 0;
        }


        /// <summary>
        /// Returns the number of decimal places that should be displayed in floating
        /// point minutes, appropriate to the resolution of this instance: 0 if resolution
        /// &lt; 75.6, 1 if resolution is 75.6 to 756, etc.
        /// </summary>
        /// <returns>integer &gt;= 0</returns>
        public int MinuteDecimals() {
            // If the constant were about -.778 (= 1 - log 60) instead of .878, this method would
            // call for one decimal place on minutes if resolution were 60.0 (i.e., 1-minute
            // resolution). To prevent that, I adjust the constant by a tenth order of magnitude.
            int i = (int)(Math.Log10(Resolution) - .878);
            return (i > 0) ? i : 0;
        }


        /// <summary>
        /// Returns the number of decimal places that should be displayed in floating
        /// point seconds, appropriate to the resolution of this instance: 0 if resolution
        /// &lt; 4530 (approx.), 1 if resolution is 4530 to 45300, etc.
        /// </summary>
        /// <returns>integer &gt;= 0</returns>
        public int SecondDecimals() {
            // If the constant were about -2.556 (= 1 - log 3600), this method would
            // call for one decimal place on seconds if resolution were 3600 (i.e., 1-second
            // resolution). To prevent that, I adjust the constant by a tenth order of magnitude.
            int i = (int)(Math.Log10(Resolution) - 2.656);
            return (i > 0) ? i : 0;
        }


#endif  // end new DegreeDecimals(), MinuteDecimals(), SecondDecimals()


        #region IFormattable Members

        /// <summary>
        /// Convert this instance to a string for display in a specified format.
        /// </summary>
        /// <param name="format">format control string</param>
        /// <param name="formatProvider">ignored</param>
        /// <returns>formatted String</returns>
        /// <exception cref="ArgumentException"> if the format control string is malformed.
        /// </exception>
        /// <remarks>
        /// <para>The format control string begins with 0, 1, or 2 of the flag characters ' ',
        /// '+', and '0'.</para>
        /// <para>The ' ' flag or '+' flag puts a leading ' ' or '+' on positive numbers. It is an
        /// error to use both.</para>
        /// <para>The '0' (zero) flag changes the padding character from ' ' to '0' when left
        /// padding is necessary.</para>
        /// <para>Flag characters may be in any order, e.g., "0 " or " 0".</para>
        /// <para>An optional number (one or more digits) after the flag characters (if any) is the
        /// minimum width of the leftmost field's integer part. For example, if the format is
        /// decimal degrees, a width of 2 ensures integer degrees occupy at least 2 characters.
        /// Any padding is on the left, and is either ' ' or '0', depending on the '0' flag.</para>
        /// <para>The minimum field width affects only the leftmost field. The fields to the
        /// right (e.g., minutes and seconds) are always padded with leading zeros to 2
        /// digits in the integer part.</para>
        /// <para>Next is one letter (not case sensitive) to specify the format. It's the only
        /// mandatory part of the format string.</para>
        /// <para>Format 'a' is degrees (or hours), minutes, and decimal seconds. This
        /// format, like the others, automatically adjusts to the resolution of the
        /// instance. E.g., if resolution is 60 (1 minute), seconds do not appear.</para>
        /// <para>Format 'b' is degrees and decimal minutes.</para>
        /// <para>Format 'c' is decimal degrees.</para>
        /// <para>Format 'd' is minutes and decimal seconds. It overflows to format 'a' if the value
        /// is too large.</para>
        /// <para>Format 'e' is decimal minutes. It overflows to format 'b'.</para>
        /// <para>Format 'f' is decimal seconds. It overflows to format 'd', which may overflow
        /// to format 'a'</para>
        /// <para>Format 'g' is identical to 'a'.</para>
        /// <para>To prevent display of meaningless digits, precision (set when the instance is
        /// constructed) has precedence over format. For example, if you request format 'f' but
        /// precision is 60 (1 minute), format 'd' is used.</para>
        /// <para>After the format letter, there may be either 0 or 3 (no other number) label
        /// characters for the fields. The first label appears after degrees or hours, the second
        /// label after minutes, the third after seconds. Even if some labels are unused due to the
        /// precision or format of the instance, they must still be present.</para>
        /// <para>Finally, one optional separator character (typically ' ' or ':') may be present.
        /// This character separates the degrees (hours), minutes, and seconds fields.</para>
        /// <para>If labels and a separator are both supplied, the separator must be last. If labels
        /// and a separator are both omitted, there will be nothing between the fields ("packed"
        /// format). The labels and separator may be any Unicode characters.</para>
        /// <para>The result string has the resoluton specified when the instance was constructed.
        /// This cannot be changed.</para>
        /// </remarks>
        public string ToString(string format, IFormatProvider formatProvider)
        {
            if (format == "" || format == null)
                format = "a ";

            // Use a regular expression to decode the format string.
            string pattern = @"^([ +0]){0,2}([1-9][0-9]*)?([a-zA-Z])(.{0,4})$";
            Match m = Regex.Match(format, pattern);
            if (m.Success == false)
                throw new ArgumentException(
                    "Sexagesimal.ToString() received invalid format string.");
            GroupCollection gc = m.Groups;

            // process the flag characters
            if (gc[1].Captures.Count > 2)
                throw new ArgumentException(
                    "Sexagesimal format string has too many flag characters.");
            bool spaceFlag = false;
            bool plusFlag = false;
            bool zeroFlag = false;
            for (int i = 0; i < gc[1].Captures.Count; ++i) {
                switch (gc[1].Captures[i].Value) {
                case " ":
                    spaceFlag = true;
                    break;
                case "+":
                    plusFlag = true;
                    break;
                case "0":
                    zeroFlag = true;
                    break;
                default:
                    throw new ArgumentException(
                        "Sexagesimal.ToString() control string has invalid flag character.");
                }
            }
            if (spaceFlag && plusFlag)
                throw new ArgumentException(
                    "Sexagesimal.ToString() control string invalid: ' ' and '+' are both present.");

            // get the minimum width of the first value
            int width = 0;
            if (gc[2].Success)
                width = int.Parse(gc[2].Value);

            // labels for the values, and the separator character
            string degreeLabel = "";
            string minuteLabel = "";
            string secondLabel = "";
            string separator = "";
            if (gc[4].Success) {
                string labels = gc[4].Value;
                if (labels.Length == 2 || labels.Length > 4)
                    throw new ArgumentException(
                            "Sexagesimal ToString(): 0, 1, 3, or 4 characters"
                            + " must follow the format letter.");
                if (labels.Length >= 3) {
                    degreeLabel = labels[0].ToString();
                    minuteLabel = labels[1].ToString();
                    secondLabel = labels[2].ToString();
                }
                if (labels.Length == 1)
                    separator = labels[0].ToString();
                if (labels.Length == 4)
                    separator = labels[3].ToString();
            }

            // Start building the result, beginning with sign.
            StringBuilder sb = new StringBuilder(20);
            sb.Append(formatSign(spaceFlag, plusFlag));

            switch (char.ToLower(gc[3].Value[0])) {
            case 'a':
            case 'g':
                // degrees (or hours), minutes, seconds
                sb.Append(FormatDegrees(zeroFlag, width, false));
                sb.Append(degreeLabel);
                if (Resolution >= 2) {       // .5 degree or .5 hour
                    sb.Append(separator);
                    sb.Append(FormatMinutes(true, 2, false));
                    sb.Append(minuteLabel);
                    if (Resolution >= 120) {     // .5 minute
                        sb.Append(separator);
                        sb.Append(FormatSeconds(true, 2));
                        sb.Append(secondLabel);
                    }
                }
                break;
            case 'b':
                // degrees or hours, and decimal minutes
                sb.Append(FormatDegrees(zeroFlag, width, false));
                sb.Append(degreeLabel);
                if (Resolution >= 2) {       // .5 degree or .5 hour
                    sb.Append(separator);
                    sb.Append(FormatMinutes(true, 2, true));
                    sb.Append(minuteLabel);
                }
                break;
            case 'c':
                // decimal degrees or hours
                sb.Append(FormatDegrees(zeroFlag, width, true));
                sb.Append(degreeLabel);
                break;
            case 'd':
                // Minutes and decimal seconds. Omit degrees if possible.
                if ((int)degrees != 0 || resolution < 2.0)
                    goto case 'a';      // use deg, min, sec format
                sb.Append(FormatMinutes(zeroFlag, width, false));
                sb.Append(minuteLabel);
                if (Resolution >= 120) {     // .5 minute
                    sb.Append(separator);
                    sb.Append(FormatSeconds(true, 2));
                    sb.Append(secondLabel);
                }
                break;
            case 'e':
                // Decimal minutes. Omit degrees if possible.
                if ((int)degrees != 0 || resolution < 2.0)
                    goto case 'b';      // use deg, decimal min format
                sb.Append(FormatMinutes(zeroFlag, width, true));
                sb.Append(minuteLabel);
                break;
            case 'f':
                // Decimal seconds. Omit degrees and minutes if possible.
                if ((int)degrees != 0 || (int)minutes != 0 || resolution < 120.0)
                    goto case 'd';      // minutes and decimal seconds format
                sb.Append(FormatSeconds(zeroFlag, width));
                sb.Append(secondLabel);
                break;
            default:
                throw new ArgumentException(
                    "Sexagesimal.ToString(): requested format letter not implemented.");
            }

            // Move the sign so it follows any padding.
            if (sb.Length >= 3 && (sb[0] == '-' || sb[0] == '+') && sb[1] == ' ') {
                int i;
                for (i = 2; sb[i] == ' '; ++i)
                    ;
                --i;
                sb[i] = sb[0];    // replace the rightmost ' ' with the sign
                sb[0] = ' ';
            }
            return sb.ToString();
        }

        #endregion


#if DEBUG
        /// <summary>
        /// Convert the sign of this instance to a string.
        /// </summary>
        /// <remarks>If both <paramref name="spaceFlag"/> and <paramref name="plusFlag"/>
        /// are true, <paramref name="plusFlag"/> has precedence.
        /// </remarks>
        /// <param name="spaceFlag">If true, output " " or "-", depending on sign.</param>
        /// <param name="plusFlag">If true, output "+" or "-", depending on sign.</param>
        /// <returns>a new String</returns>
#endif
        private string formatSign(bool spaceFlag, bool plusFlag)
        {
            string s = "";
            if (plusFlag)
                s = Sign.ToString();
            else if (spaceFlag)
                if (Sign == '-')
                    s = "-";
                else
                    s = " ";
            else if (Sign == '-')
                s = "-";
            return s;
        }


#if DEBUG
        /// <summary>
        /// Convert the degrees part of this instance to string. Pad on the left to
        /// make the integer part at least <paramref name="width"/> wide. Only the integer
        /// part is output, unless <paramref name="decimalFormat"/> is true. In that case,
        /// the number of decimal places will be appropriate to the resolution of this
        /// instance.
        /// </summary>
        /// <param name="zeroFlag">If true, use '0' (zero) for padding instead of space.
        /// </param>
        /// <param name="width">the minimum width of the integer part. More width
        /// is used if necessary.</param>
        /// <param name="decimalFormat">If true, fractional degrees are included.</param>
        /// <returns>String</returns>
#endif
        private string FormatDegrees(bool zeroFlag, int width, bool decimalFormat)
        {
            string s;
            int decimalPlaces = 0;   // number of decimal places

            if (decimalFormat) {
                decimalPlaces = DegreeDecimals();
                s = string.Format("{0:f" + decimalPlaces.ToString() + "}", Degrees);
            }  else
                s = ((int)Degrees).ToString();

            if (width != 0) {     // degrees field width was given
                // adjust width for the decimal fraction
                width += decimalPlaces + (decimalPlaces > 0 ? 1 : 0);
                if (zeroFlag)
                    s = s.PadLeft(width, '0');    // pad with '0'
                else
                    s = s.PadLeft(width);         // pad with ' '
            }
            return s;
        }


#if DEBUG
        /// <summary>
        /// Convert the minutes part of this instance to string. Pad on the left to
        /// make the integer part at least <paramref name="width"/> wide. Only the integer
        /// part is output, unless <paramref name="decimalFormat"/> is true. In that case,
        /// the number of decimal places is appropriate to the resolution of this
        /// instance.
        /// </summary>
        /// <param name="zeroFlag">If true, use '0' (zero) for padding instead of space.
        /// </param>
        /// <param name="width">the minimum width of the integer part. More width is
        /// used if necessary.</param>
        /// <param name="decimalFormat">If true, fractional minutes are included.</param>
        /// <returns>String</returns>
#endif
        private string FormatMinutes(bool zeroFlag, int width, bool decimalFormat)
        {
            string s;
            int decimalPlaces = 0;   // number of decimal places

            if (decimalFormat) {
                decimalPlaces = MinuteDecimals();
                s = string.Format("{0:f" + decimalPlaces.ToString() + "}", Minutes);
            } else
                s = ((int)Minutes).ToString();

            if (width != 0) {     // minimum field width was given
                // adjust width for the decimal fraction
                width += decimalPlaces + (decimalPlaces > 0 ? 1 : 0);
                if (zeroFlag)
                    s = s.PadLeft(width, '0');  // pad with '0'
                else
                    s = s.PadLeft(width);       // pad with ' '
            }
            return s;
        }


#if DEBUG
        /// <summary>
        /// Convert the seconds part of this instance to string. Pad on the left to make
        /// the integer part at least <paramref name="width"/> wide.
        /// A decimal part is present if appropriate for the resolution of this instance.
        /// </summary>
        /// <param name="zeroFlag">If true, use '0' (zero) for padding instead of space.
        /// </param>
        /// <param name="width">the minimum width of the integer part. More width is
        /// used if necessary.</param>
        /// <returns>String</returns>
#endif
        private string FormatSeconds(bool zeroFlag, int width)
        {
            int decimalPlaces = SecondDecimals();   // number of decimal places

            string s = string.Format("{0:f" + decimalPlaces.ToString() + "}", Seconds);

            if (width != 0) {     // a minimum width was specified
                // adjust width for the decimal fraction
                width += decimalPlaces + (decimalPlaces > 0 ? 1 : 0);
                if (zeroFlag)
                    s = s.PadLeft(width, '0');      // pad with '0'
                else
                    s = s.PadLeft(width);           // pad with ' '
            }
            return s;
        }




        #region self-test code
#if DEBUG


        /// <summary>
        /// Test the class. Runs silently if OK.
        /// </summary>
        /// <exception cref="Exception">if an error is detected.</exception>
        public static void Test()
        {
            Sexagesimal sex1, sex2;

            // TEST CODE TO INVESTIGATE TIME FORMATTING BUGS.
#if true
            // (60 SEC INSTEAD OF 00 IN NEXT DAY)

            // mathematical constants
            const double arcsecPerDeg = 3600.0;
            const double degPerHour = 15.0;

            // desired angle accuracy (arcsec)
            float arcsecAccuracy = .01f;

            // From the above value, derive the corresponding accuracy in radians, and resolution in
            // time and angle to control the formatting of the output.
            double angleAccuracy = SofaJpl.Angle.DmsToRad(0, 0, arcsecAccuracy);
            double angleResolution = arcsecPerDeg / arcsecAccuracy;     // resolution units per degree
            double timeResolution = angleResolution * degPerHour;       // resolution units per hour

            // epoch (time scale depends on preprocessor symbols)
            int year = 2015;
            int month = 8;
            int day = 28;
            int hour = 3;
            int minute = 0;
            double second = 0.0;

            SofaJpl.UtcTable utcTable = new SofaJpl.UtcTable(new System.IO.StreamReader(
                @"C:\Users\Stan\Documents\astro\IERS\leapSecTable.txt"));
            double fracday = SofaJpl.Angle.HmsToDays(hour, minute, second);
            SofaJpl.Utc utc = new SofaJpl.Utc(year, month, day, fracday, false, utcTable);

            SofaJpl.TimeFields tf1 = utc.ToTimeFields(timeResolution);
            // Console.WriteLine("{0} UTC", tf1);
            assertTrue(tf1.ToString().Equals("2015-08-28T03:00:00.0000"));


            // SEXAGESIMAL BUG FROM THE DEMO PROGRAM
            //const double angleResolution = 360000.00000000006;
            angleResolution = 360000.0000000000;
            const double lonEast = 3.1415926734528954;
            sex1 = new Sexagesimal(Angle.RadiansToDegrees(lonEast), angleResolution);
            // Console.WriteLine("{0}", sex1);
            // Console.WriteLine(sex1.ToString());
            assertTrue(sex1.ToString().Equals("+180 00 00.00"));
#endif


            // the master constructor

            sex1 = new Sexagesimal(0.0, 3600.0, 0, 0.0);
            assertTrue(sex1.ToString().Equals("+0 00 00"));

            // 86 09 46.51
            sex1 = new Sexagesimal(86.16292, 36000.0, 0, 0.0);
            assertTrue(sex1.ToString().Equals("+86 09 46.5"));

            // -112 51 02.30
            sex1 = new Sexagesimal(-112.85064, 3600.0, 0, 0.0);
            assertTrue(sex1.ToString().Equals("-112 51 02"));

            // -347 05 19.03 aka -347 05.317
            sex1 = new Sexagesimal(-347.08862, 600.0, 0, 0.0);
            assertTrue(sex1.ToString().Equals("-347 05 18"));

            // -63 17.34
            sex1 = new Sexagesimal(-63.29298, 60.0, 0, 0.0);
            assertTrue(sex1.ToString().Equals("-63 18"));

            sex1 = new Sexagesimal(-150.97099, .1, 0, 0.0);
            assertTrue(sex1.ToString().Equals("-150"));

            sex1 = new Sexagesimal(47.07338, 3600.0, 24, 0.0);
            assertTrue(sex1.ToString().Equals("+23 04 24") && sex1.Revolutions == 1);

            sex1 = new Sexagesimal(-637.60346, 36000.0, 360, 0.0);
            assertTrue(sex1.ToString().Equals("-277 36 12.5") && sex1.Revolutions == -1);

            sex1 = new Sexagesimal(359.0, 600.0, 360, 0.0);
            assertTrue(sex1.ToString().Equals("+359 00 00"));

            // -23 59 58.999
            sex1 = new Sexagesimal(-23.999722, 3600.0, 24, 0.0);
            assertTrue(sex1.ToString().Equals("-23 59 59") && sex1.Revolutions == 0);

            sex1 = new Sexagesimal(-24.000278, 3600.0, 24, 0.0);
            assertTrue(sex1.ToString().Equals("-0 00 01") && sex1.Revolutions == -1);

            sex1 = new Sexagesimal(24.00 + .6 / 86400, 3600.0, 24, 1.0);
            assertTrue(sex1.ToString().Equals("+0 00 00") && sex1.Revolutions == 1);

            sex1 = new Sexagesimal(23.999861, 36000.0, 24, 1.0);
            assertTrue(sex1.ToString().Equals("+23 59 60.5"));

            // 23 59 59.701
            sex1 = new Sexagesimal(23.999917, 3600.0, 24, 0.0);
            assertTrue(sex1.ToString().Equals("+0 00 00") && sex1.Revolutions == 1);


            // test the simplified constructors

            sex1 = new Sexagesimal(-450.0, 36000.0, 360);
            assertTrue(sex1.ToString().Equals("-90 00 00.0") && sex1.Revolutions == -1);

            sex1 = new Sexagesimal(450.0, 1.0);
            assertTrue(sex1.ToString().Equals("+450") && sex1.Revolutions == 0);


            // properties
            sex1 = new Sexagesimal(-12.567670, 36000.0);
            assertTrue(sex1.Sign == '-' &&
                Math.Abs(sex1.Degrees - 12.567667) < .000001 &&
                sex1.Degrees == sex1.Hours &&
                Math.Abs(sex1.Minutes - 34.0600) < .0001 &&
                Math.Abs(sex1.Seconds - 3.6) < .01 &&
                sex1.Revolutions == 0 &&
                sex1.Resolution == 36000.0);


            // methods

            // CompareTo(Sexagesimal)
            sex1 = new Sexagesimal(1.0, 1.0);
            sex2 = null;
            assertTrue(sex1.CompareTo(sex2) > 0);
            sex2 = sex1;
            assertTrue(sex1.CompareTo(sex2) == 0);
            sex2 = new Sexagesimal(2.0, 1.0);
            assertTrue(sex1.CompareTo(sex2) < 0 && sex2.CompareTo(sex1) > 0);

            // CompareTo(object)
            object obj1 = sex1;
            object obj2 = sex2;
            assertTrue(sex1.CompareTo(obj2) < 0 && sex2.CompareTo(obj1) > 0 &&
                sex2.CompareTo(obj2) == 0);
            JulianDate jd1 = new JulianDate(0, 0);
            try {
                assertTrue(sex1.CompareTo(jd1) != 0);
                Console.WriteLine("CompareTo(object) failed to detect incompatible object.");
            } catch {
                ;
            }

            // Equals(Sexagesimal)
            sex2 = sex1;
            assertTrue(sex2.Equals(sex1));
            sex2 = (Sexagesimal)obj2;
            assertTrue(!sex1.Equals(sex2) && !sex1.Equals(jd1));

            // equals(object)
            obj1 = null;
            assertTrue(!sex1.Equals(obj1));
            obj1 = sex1;
            assertTrue(sex1.Equals(obj1));
            assertTrue(!sex1.Equals(obj2));
            obj1 = new JulianDate(0, 0);
            assertTrue(!sex1.Equals(obj1));

            // hash code
            sex2 = new Sexagesimal(12.5731, 3600.0);
            assertTrue(sex1.GetHashCode() != sex2.GetHashCode());

            // ToString()
            sex1 = new Sexagesimal(-127.6479151, 3600000.0);
            assertTrue(sex1.ToString().Equals("-127 38 52.494"));

            // ToDouble()
            assertTrue(Math.Abs(sex2.ToDouble() - 12.5731) < 1.0 / 3600.0);
            sex1 = new Sexagesimal(23.0 + (59.0 + 59.0 / 60.0) / 60.0, 3.6e4, 24, 1.0);
            assertTrue(Math.Abs(sex1.ToDouble() - 24.0) < .1 / 3600.0);
            sex1 = new Sexagesimal(23.0 + (59.0 + 59.5 / 60.0) / 60.0, 3.6e4, 24, 1.0);
            assertTrue(Math.Abs(sex1.ToDouble() - 24.000139) < .1 / 3600.0);

            // DegreeDecimals(), MinuteDecimals(), SecondDecimals()
            sex1 = new Sexagesimal(-259.32, .1);
            sex2 = new Sexagesimal(-259.32, 36000.0);
            assertTrue(sex1.DegreeDecimals() == 0);
            assertTrue(sex2.DegreeDecimals() == 5);
            assertTrue(sex1.MinuteDecimals() == 0);
            assertTrue(sex2.MinuteDecimals() == 3);
            assertTrue(sex1.SecondDecimals() == 0);
            assertTrue(sex2.SecondDecimals() == 1);


            // string formatting

            // 'a' and 'g' formats

            sex1 = new Sexagesimal(-191.004688, 3600.0);
            assertTrue(string.Format("{0:a }", sex1).Equals("-191 00 17"));
            assertTrue(string.Format("{0:a}", sex1).Equals("-1910017"));
            assertTrue(string.Format("{0:a°'\"}", sex1).Equals("-191°00'17\""));
            // Unicode prime and double prime
            assertTrue(string.Format("{0:a°′″}", sex1).Equals("-191°00′17″"));
            
            sex1 = new Sexagesimal(7.401644, 36000.0);
            assertTrue(string.Format("{0: 3a }", sex1).Equals("   7 24 05.9"));
            assertTrue(string.Format("{0:3g }", sex1).Equals("  7 24 05.9"));
            assertTrue(string.Format("{0:03a }", sex1).Equals("007 24 05.9"));
            assertTrue(string.Format("{0:+03g }", sex1).Equals("+007 24 05.9"));
            assertTrue(string.Format("{0:+3a }", sex1).Equals("  +7 24 05.9"));

            sex1 = new Sexagesimal(-1.750268, 3600.0);
            assertTrue(string.Format("{0:02a:}", sex1).Equals("-01:45:01"));

            // format 'b'

            sex1 = new Sexagesimal(351.9971, 600.0);
            assertTrue(string.Format("{0: 3b°'\"}", sex1).Equals(" 351°59.8'"));

            sex1 = new Sexagesimal(298.0, 1.0);
            assertTrue(string.Format("{0:+b°'\"}", sex1).Equals("+298°"));

            sex1 = new Sexagesimal(-3.2996, 6000.0);
            assertTrue(string.Format("{0: 0b }", sex1).Equals("-3 17.98"));

            // format 'c'

            sex1 = new Sexagesimal(15.265, 10.0);
            assertTrue(string.Format("{0:c}", sex1).Equals("15.3"));

            sex1 = new Sexagesimal(-60.817, .1);
            assertTrue(string.Format("{0: 03c°'\"}", sex1).Equals("-060°"));

            sex1 = new Sexagesimal(-303.0095, 100.0);
            assertTrue(string.Format("{0:+c°'\"}", sex1).Equals("-303.01°"));

            // format 'd'

            sex1 = new Sexagesimal(-.80627, 6000.0);
            assertTrue(string.Format("{0:d}", sex1).Equals("-4822.8"));

            sex1 = new Sexagesimal(.05499, 3600.0);
            assertTrue(string.Format("{0:+02d°'\"}", sex1).Equals("+03'18\""));

            sex1 = new Sexagesimal(218.67193, 360.0);
            assertTrue(string.Format("{0: d°'\" }", sex1).Equals(" 218° 40' 20\""));

            // format 'e'

            sex1 = new Sexagesimal(-.009472, 6000.0);
            assertTrue(string.Format("{0:02e°'\" }", sex1).Equals("-00.57'"));

            sex1 = new Sexagesimal(.11193, 600.0);
            assertTrue(string.Format("{0: e}", sex1).Equals(" 6.7"));

            sex1 = new Sexagesimal(-337.41097, 60.0);
            assertTrue(string.Format("{0: e }", sex1).Equals("-337 25"));

            sex1 = new Sexagesimal(0.0, 1.0);
            assertTrue(string.Format("{0: e°'\"}", sex1).Equals(" 0°"));

            // format 'f'

            sex1 = new Sexagesimal(.000277, 36000.0);
            assertTrue(string.Format("{0:f}", sex1).Equals("1.0"));

            sex1 = new Sexagesimal(-.003593, 360000.0);
            assertTrue(string.Format("{0:+f°'\"}", sex1).Equals("-12.93\""));

            sex1 = new Sexagesimal(.009000, 60.0);
            assertTrue(string.Format("{0:+f°'\"}", sex1).Equals("+1'"));

            sex1 = new Sexagesimal(.600237, 3600.0);
            assertTrue(string.Format("{0:+f }", sex1).Equals("+36 01"));

            sex1 = new Sexagesimal(129.7984, 360.0);
            assertTrue(string.Format("{0:+f°'\" }", sex1).Equals("+129° 47' 50\""));


            // verify that illegal formats are rejected

            sex1 = new Sexagesimal(129.7984, 360.0);

            try {
                Console.WriteLine("\n{0:m}", sex1);
                Console.WriteLine("FAILED TO CATCH ILLEGAL FORMAT");
            } catch {
                ;
            }
            try {
                Console.WriteLine("\n{0: +02a}", sex1);
                Console.WriteLine("FAILED TO CATCH ILLEGAL FORMAT");
            } catch {
                ;
            }
            try {
                Console.WriteLine("\n{0:-03a}", sex1);
                Console.WriteLine("FAILED TO CATCH ILLEGAL FORMAT");
            } catch {
                ;
            }
            try {
                Console.WriteLine("\n{0:a'\"}", sex1);
                Console.WriteLine("FAILED TO CATCH ILLEGAL FORMAT");
            } catch {
                ;
            }
            try {
                Console.WriteLine("\n{0:gdmsf }", sex1);
                Console.WriteLine("FAILED TO CATCH ILLEGAL FORMAT");
            } catch {
                ;
            }
        }


        private static void assertTrue(bool assertion)
        {
            if (!assertion)
                throw new Exception("Assertion failed.");
        }


#endif
        #endregion
    }
}
