﻿// SofaBody.cs
// Class to represent a body in the SOFA EPV00 and PLAN96 ephemeris routines.
// Copyright 2018 by Paul S. Hirose

// revision history
#if false

2015-06-29  Created
2015-07-01  Self test OK.
2015-09-05  Minor modification to self test because TopocentricGeometric() now returns a PVVector.

#endif


using System;


namespace HirosePS.SofaJpl {

/// <summary>
/// A class to represent a body in the SOFA EPV00 and PLAN96 ephemeris routines.
/// </summary>
public class SofaBody : SolarSystemBody {

    private SofaEphemeris.Body _body;



    /// <summary>
    /// Create a SOFA ephemeris body
    /// </summary>
    /// <param name="body">identity of body</param>
    /// <param name="ephemeris">a SOFA ephemeris</param>
    /// <param name="accuracy">light time correction accuracy (radians)</param>
    public SofaBody(SofaEphemeris.Body body, SofaEphemeris ephemeris, double accuracy) {
        Name = body.ToString();
        _body = body;
        Ephemeris = ephemeris;

        IsSun = body == SofaEphemeris.Body.Sun;
        IsSolarSystemBody = true;
        IsEarth = body == SofaEphemeris.Body.Earth;
        Radius = SolarSystemBody.BodyRadius(Name);

        Accuracy = accuracy;
    }


    /// <summary>
    /// Barycentric geometric position and velocity.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>barycentric ICRS position (km) and velocity (km/day)</returns>
    public override PVVector Barycentric(JulianDate tdb) {
        return Ephemeris.Barycentric(tdb, (int)_body);
    }

    /// <summary>
    /// Heliocentric geometric position and velocity.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>heliocentric geometric ICRS position (km) and velocity (km/day)</returns>
    public override PVVector Heliocentric(JulianDate tdb) {
        return Ephemeris.Heliocentric(tdb, (int)_body);
    }

    /// <summary>
    /// Geocentric geometric position.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>geocentric geometric ICRS position (km) and velocity (km/day)</returns>
    public override PVVector GeocentricGeometric(JulianDate tdb) {
        return Ephemeris.Geocentric(tdb, (int)_body);
    }


    // self test code
#if DEBUG

    /// <summary>
    /// Run the class self test.
    /// </summary>
    public static void Test() {
        // angle tolerance on vectors, radians
        double tolerance = Angle.DmsToRad(0, 0, 10);

        // time (TDB)
        JulianDate tdb = new JulianDate(2008, 6, 24, 4.0 / 24.0);

        // Open SOFA and JPL ephemerides, the latter as the standard of comparison.
        SofaEphemeris sofaEph = new SofaEphemeris();
        JplEphemeris jplEph = new JplEphemeris("binp2000_2099.406");

        // desired light time accuracy for both ephemerides
        double accuracy = Angle.DegreesToRadians(1.0 / 3600.0);

        // the ephemeris objects are Venus initially
        JplBody jplBody = new JplBody(JplEphemeris.Body.Venus, jplEph, accuracy);
        SofaBody sofaBody = new SofaBody(SofaEphemeris.Body.Venus, sofaEph, accuracy);

        // Verify name is OK.
        if (sofaBody.Name != "Venus")
            throw new Exception("SofaEphBody body name = " + sofaBody.Name);

        // barycentric
        PVVector pvJpl = jplBody.Barycentric(tdb);
        PVVector pvSofa = sofaBody.Barycentric(tdb);
        equalsApprox(pvSofa.Position, pvJpl.Position, tolerance *
            pvSofa.Position.Modulus(), "SofaEphBody barycentric position");
        equalsApprox(pvSofa.Velocity, pvJpl.Velocity, tolerance *
            pvSofa.Velocity.Modulus(), "SofaEphBody barycentric velocity");

        // heliocentric
        pvJpl = jplBody.Heliocentric(tdb);
        pvSofa = sofaBody.Heliocentric(tdb);
        equalsApprox(pvSofa.Position, pvJpl.Position, tolerance *
            pvSofa.Position.Modulus(), "SofaEphBody heliocentric position");
        equalsApprox(pvSofa.Velocity, pvJpl.Velocity, tolerance *
            pvSofa.Velocity.Modulus(), "SofaEphBody heliocentric velocity");

        // geocentric apparent
        Vector pJpl = jplBody.GeocentricApparent(tdb);
        Vector pSofa = sofaBody.GeocentricApparent(tdb);
        double sepAngle = pJpl.SeparationAngle(pSofa);
        assert(pJpl.SeparationAngle(pSofa) < tolerance, "geocentric apparent");

        // Sun barycentric
        jplBody = new JplBody(JplEphemeris.Body.Sun, jplEph, accuracy);
        sofaBody = new SofaBody(SofaEphemeris.Body.Sun, sofaEph, accuracy);
        pvJpl = jplBody.Barycentric(tdb);
        pvSofa = sofaBody.Barycentric(tdb);
        assert(Angle.RadiansToDegrees(pvJpl.Position.SeparationAngle(pvSofa.Position))
            * 3600.0 < .5);
        assert(new Vector(pvSofa.Velocity - pvJpl.Velocity).Modulus() /
            pvJpl.Velocity.Modulus() < 2e-5);

        // SunBaryPV() should exactly match Barycentric() position and velocity.
        sofaEph.SunBaryPV(tdb).EqualsApprox(pvSofa, 0, 0, "SunBaryPV()");

        // Sun geocentric geometric
        pJpl = jplBody.GeocentricGeometric(tdb).Position;
        pSofa = sofaBody.GeocentricGeometric(tdb).Position;
        assert(Angle.RadiansToDegrees(pJpl.SeparationAngle(pSofa)) * 3600.0
            < .002);

        // Earth barycentric
        jplBody = new JplBody(JplEphemeris.Body.Earth, jplEph, accuracy);
        sofaBody = new SofaBody(SofaEphemeris.Body.Earth, sofaEph, accuracy);
        pvJpl = jplBody.Barycentric(tdb);
        pvSofa = sofaBody.Barycentric(tdb);
        assert(Angle.RadiansToDegrees(pvJpl.Position.SeparationAngle(pvSofa.Position))
            * 3600.0 < .002);
        assert(new Vector(pvSofa.Velocity - pvJpl.Velocity).Modulus() /
            pvJpl.Velocity.Modulus() < 3e-8);

        // EarthBaryPV() should exactly match Barycentric().
        sofaEph.EarthBaryPV(tdb).EqualsApprox(pvSofa, 0, 0, "EarthBaryPV()");


        // Verify Status property warns of reduced accuracy.

        sofaBody = new SofaBody(SofaEphemeris.Body.Mars, sofaEph, accuracy);
        JulianDate goodTdb = new JulianDate(2000, 1, 1, 0);
        JulianDate badTdb = new JulianDate(900, 1, 1, 0);
        RMatrix crsToTrs = new RMatrix();   // doesn't matter for this test
        Topocenter observer = new Topocenter(40, -95, 0);

        sofaEph.Status = 0;     // clear the warning/error status
        pvSofa = sofaBody.Barycentric(goodTdb);
        assert(sofaEph.Status == 0,
            "SofaEphBody Barycentric status should be 0.");
        pvSofa = sofaBody.Barycentric(badTdb);
        assert(sofaEph.Status != 0,
            "SofaEphBody Barycentric status should not be 0.");

        sofaEph.Status = 0;     // clear the warning/error status
        pvSofa = sofaBody.Heliocentric(goodTdb);
        assert(sofaEph.Status == 0,
            "SofaEphBody Heliocentric status should be 0.");
        // Result is null vector, valid regardless of epoch.
        pvSofa = sofaBody.Heliocentric(badTdb);
        assert(sofaEph.Status != 0,
            "SofaEphBody Heliocentric status should not be 0.");

        sofaEph.Status = 0;
        pSofa = sofaBody.GeocentricGeometric(goodTdb).Position;
        assert(sofaEph.Status == 0,
            "SofaEphBody GeocentricGeometric status should be 0.");
        pSofa = sofaBody.GeocentricGeometric(badTdb).Position;
        assert(sofaEph.Status != 0,
            "SofaEphBody GeocentricGeometric status should not be 0.");

        sofaEph.Status = 0;
        pSofa = sofaBody.GeocentricAstrometric(goodTdb);
        assert(sofaEph.Status == 0,
            "SofaEphBody GeocentricAstrometric status should be 0.");
        pSofa = sofaBody.GeocentricAstrometric(badTdb);
        assert(sofaEph.Status != 0,
            "SofaEphBody GeocentricAstrometric status should not be 0.");

        sofaEph.Status = 0;
        pSofa = sofaBody.GeocentricApparent(goodTdb);
        assert(sofaEph.Status == 0,
            "SofaEphBody GeocentricApparent status should be 0.");
        pSofa = sofaBody.GeocentricApparent(badTdb);
        assert(sofaEph.Status != 0,
            "SofaEphBody GeocentricApparent status should not be 0.");

        PVVector observerPV = observer.ToPVVector(crsToTrs);

        sofaEph.Status = 0;
        pSofa = sofaBody.TopocentricGeometric(goodTdb, observerPV).Position;
        assert(sofaEph.Status == 0,
            "SofaEphBody TopocentricGeometric status should be 0.");
        pSofa = sofaBody.TopocentricGeometric(badTdb, observerPV).Position;
        assert(sofaEph.Status != 0,
            "SofaEphBody TopocentricGeometric status should not be 0.");

        sofaEph.Status = 0;
        pSofa = sofaBody.TopocentricAstrometric(goodTdb, observerPV.Position);
        assert(sofaEph.Status == 0,
            "SofaEphBody TopocentricAstrometric status should be 0.");
        pSofa = sofaBody.TopocentricAstrometric(badTdb, observerPV.Position);
        assert(sofaEph.Status != 0,
            "SofaEphBody TopocentricAstrometric status should not be 0.");

        sofaEph.Status = 0;
        pSofa = sofaBody.TopocentricApparent(goodTdb, observerPV);
        assert(sofaEph.Status == 0,
            "SofaEphBody TopocentricApparent status should be 0.");
        pSofa = sofaBody.TopocentricApparent(badTdb, observerPV);
        assert(sofaEph.Status != 0,
            "SofaEphBody TopocentricApparent status should not be 0.");

    }
#endif      // end self test code

}       // end class SofaBody
}       // end namespace HirosePS.SofaJpl
