﻿// SofaEphemeris.cs
// An ephemeris which calls the SOFA EPV00 and PLAN94 routines.
// Copyright 2018 by Paul S. Hirose

// revision history
#if false

2015-06-24  Created
2015-06-26  Finished. Self test OK.

#endif


using System;


namespace HirosePS.SofaJpl {
/// <summary>
/// Solar system ephemeris based on the SOFA EPV00 and PLAN94 routines.
/// </summary>
public class SofaEphemeris : Ephemeris {

    // Contains the cumulative OR of the status values returned by the SOFA routines. It is the user
    // responsibility to check and clear this value. It is not automatically cleared. Has no effect
    // on program operation.
    private int _status;


    /// <summary>
    /// The available bodies in the ephemeris.
    /// </summary>
    public enum Body {
        /// <summary>
        /// Mercury
        /// </summary>
        Mercury = 1,
        /// <summary>
        /// Venus
        /// </summary>
        Venus,
        /// <summary>
        /// Earth-Moon barycenter
        /// </summary>
        EarthMoonBary,
        /// <summary>
        /// Mars
        /// </summary>
        Mars,
        /// <summary>
        /// Jupiter
        /// </summary>
        Jupiter,
        /// <summary>
        /// Saturn
        /// </summary>
        Saturn,
        /// <summary>
        /// Uranus
        /// </summary>
        Uranus,
        /// <summary>
        /// Neptune
        /// </summary>
        Neptune,
        /// <summary>
        /// Sun
        /// </summary>
        Sun,
        /// <summary>
        /// Earth
        /// </summary>
        Earth
    };



    /// <summary>
    /// Construct a SOFA ephemeris.
    /// </summary>
    public SofaEphemeris() { }


    /// <summary>
    /// Cumulative OR of the status values returned by the SOFA routines.
    /// </summary>
    /// <remarks>It is the user responsibility to check and clear this value. It is not
    /// automatically cleared. Has no effect on program operation.</remarks>
    public int Status {
        get { return _status; }
        set { _status = value; }
    }


    /// <summary>
    /// Name of the ephemeris.
    /// </summary>
    /// <returns>ephemeris name</returns>
    public override string ToString() {
        return "SOFA EPV00/PLAN94 ephemeris";
    }


    /// <summary>
    /// Get the barycentric geometric BCRS position and velocity of a specified body.
    /// </summary>
    /// <param name="tdb">epoch (TDB)</param>
    /// <param name="targetBody">the body</param>
    /// <returns>geometric BCRS position and velocity (km, km/day)</returns>
    public override PVVector Barycentric(JulianDate tdb, int targetBody) {
        double[,] earthBaryArray = new double[2, 3];
        double[,] earthHelioArray = new double[2, 3];
        double[,] targetArray = new double[2, 3];
        double[,] sunArray = new double[2, 3];
        PVVector result;

        Body target = (Body)targetBody;
        if (target < Body.Mercury || target > Body.Earth)
            throw new ArgumentOutOfRangeException("targetBody", "is not a legal value");

        // Calculate Sun barycentric position and velocity from Earth barycentric PV minus
        // heliocentric PV.
        _status |= Sofa.Epv00(JulianDate.J2000Base.ToDouble(), tdb.ToJ2000(), earthHelioArray,
            earthBaryArray);
        Sofa.Pvmpv(earthBaryArray, earthHelioArray, sunArray);

        // Get target heliocentric position and velocity. Convert to barycentric.
        if (target < Body.Sun) {
            // Add Sun barycentric PV to target heliocentric PV.
            _status |= Sofa.Plan94(JulianDate.J2000Base.ToDouble(), tdb.ToJ2000(), targetBody,
                targetArray);
            Sofa.Pvppv(sunArray, targetArray, targetArray);
            result = new PVVector(targetArray);
        } else if (target == Body.Sun)
            result = new PVVector(sunArray);
        else
            result = new PVVector(earthBaryArray);

        // Au and au/day are the units of the SOFA ephemeris routines. Convert to km and km/day.
        return KmPerAu * result;
    }


    /// <summary>
    /// Get the heliocentric geometric ICRS position and velocity of a specified body.
    /// </summary>
    /// <param name="tdb">epoch (TDB)</param>
    /// <param name="targetBody">the body</param>
    /// <returns>ICRS geometric position and velocity (km, km/day)</returns>
    public override PVVector Heliocentric(JulianDate tdb, int targetBody) {
        double[,] targetArray = new double[2, 3];
        PVVector result;

        Body target = (Body)targetBody;
        if (target < Body.Mercury || target > Body.Earth)
            throw new ArgumentOutOfRangeException("targetBody", "is not a legal value");

        if (target < Body.Sun) {
            _status |= Sofa.Plan94(JulianDate.J2000Base.ToDouble(), tdb.ToJ2000(), targetBody,
                targetArray);
            result = new PVVector(targetArray);
        } else if (target == Body.Sun)
            result = new PVVector();    // zero PVVector
        else {      // Earth
            double[,] earthHelioArray = new double[2, 3];
            double[,] earthBaryArray = new double[2, 3];
            _status |= Sofa.Epv00(JulianDate.J2000Base.ToDouble(), tdb.ToJ2000(), earthHelioArray,
            earthBaryArray);
            result = new PVVector(earthHelioArray);
        }

        // Au and au/day are the units of the SOFA ephemeris routines. Convert to km and km/day.
        return KmPerAu * result;
    }


    /// <summary>
    /// Get the geocentric geometric ICRS position and velocity of a specified body.
    /// </summary>
    /// <param name="tdb">epoch (TDB)</param>
    /// <param name="targetBody">the body</param>
    /// <returns>geocentric geometric ICRS position and velocity (km, km/day)</returns>
    public override PVVector Geocentric(JulianDate tdb, int targetBody) {
        double[,] earthBaryArray = new double[2, 3];
        double[,] earthHelioArray = new double[2, 3];
        double[,] targetArray = new double[2, 3];
        PVVector result;

        Body target = (Body)targetBody;
        if (target < Body.Mercury || target > Body.Earth)
            throw new ArgumentOutOfRangeException("targetBody", "is not a legal value");

        // Earth barycentric PV.
        _status |= Sofa.Epv00(JulianDate.J2000Base.ToDouble(), tdb.ToJ2000(), earthHelioArray,
            earthBaryArray);

        if (target < Body.Sun) {
            // Target heliocentric PV.
            _status |= Sofa.Plan94(JulianDate.J2000Base.ToDouble(), tdb.ToJ2000(), targetBody,
                targetArray);
            // Subtract Earth heliocentric PV.
            Sofa.Pvmpv(targetArray, earthHelioArray, targetArray);
            result = new PVVector(targetArray);
        } else if (target == Body.Sun)
            result = new PVVector(earthHelioArray).Negate();
        else     // Earth is the target body. 
            result = new PVVector();    // zero PVVector

        // Au and au/day are the units of the SOFA ephemeris routines. Convert to km and km/day.
        return KmPerAu * result;
    }


    /// <summary>
    /// Earth barycentric ICRS position and velocity.</summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>Earth barycentric ICRS position and velocity (km, km/day)</returns>
    public override PVVector EarthBaryPV(JulianDate tdb) {
        return Barycentric(tdb, (int)SofaEphemeris.Body.Earth);
    }

    /// <summary>
    /// Sun barycentric ICRS position and velocity.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>Sun barycentric ICRS position and velocity (km, km/day)</returns>
    public override PVVector SunBaryPV(JulianDate tdb) {
        return Barycentric(tdb, (int)SofaEphemeris.Body.Sun);
    }



    // test code
#if DEBUG

    /// <summary>
    /// Run the class self test.
    /// </summary>
    public static void Test() {

        JulianDate tdb = new JulianDate(2000, 12, 7, 17, 30, 0);
        PVVector pv1, pv2;
        int sofaTarget, jplTarget;

        // The RMS errors for Jupiter, from the SOFA documentation, times 3.
        const double positionTolerance = 3.0 * 71700;       // km
        const double velocityTolerance = 3.0 * 7.70 * 86400.0 / 1000.0;     // km/day

        // Create the ephemerides. The JPL is independently tested, so here it's the gold standard.
        SofaEphemeris sofaEph = new SofaEphemeris();
        JplEphemeris jplEph = new JplEphemeris("binp2000_2099.422");


        // Barycentric Sun.
        sofaTarget = (int)SofaEphemeris.Body.Sun;
        jplTarget = (int)JplEphemeris.Body.Sun;
        pv1 = sofaEph.Barycentric(tdb, sofaTarget);
        pv1.EqualsApprox(sofaEph.SunBaryPV(tdb), 0, 0, "SunBaryPV() failed.");
        pv2 = jplEph.Barycentric(tdb, jplTarget);
        pv1.Position.EqualsApprox(pv2.Position, positionTolerance,
            "SofaEphemeris.Barycentric() Sun position failed.");
        pv1.Velocity.EqualsApprox(pv2.Velocity, velocityTolerance,
            "SofaEphemeris.Barycentric() Sun velocity failed.");

        // barycentric Earth
        sofaTarget = (int)SofaEphemeris.Body.Earth;
        jplTarget = (int)JplEphemeris.Body.Earth;
        pv1 = sofaEph.Barycentric(tdb, sofaTarget);
        pv1.Position.EqualsApprox(sofaEph.EarthBaryPV(tdb).Position, 0,
            "EarthBaryPV() position failed.");
        pv1.Velocity.EqualsApprox(sofaEph.EarthBaryPV(tdb).Velocity, 0,
            "EarthBaryPV() velocity failed.");
        pv2 = jplEph.Barycentric(tdb, jplTarget);
        pv1.Position.EqualsApprox(pv2.Position, positionTolerance,
            "SofaEphemeris.Barycentric() Earth position failed.");
        pv1.Velocity.EqualsApprox(pv2.Velocity, velocityTolerance,
            "SofaEphemeris.Barycentric() Earth velocity failed.");

        // barycentric Jupiter
        sofaTarget = (int)SofaEphemeris.Body.Jupiter;
        jplTarget = (int)JplEphemeris.Body.Jupiter;
        pv1 = sofaEph.Barycentric(tdb, sofaTarget);
        pv2 = jplEph.Barycentric(tdb, jplTarget);
        pv1.Position.EqualsApprox(pv2.Position, positionTolerance,
            "SofaEphemeris.Barycentric() Jupiter position failed.");
        pv1.Velocity.EqualsApprox(pv2.Velocity, velocityTolerance,
            "SofaEphemeris.Barycentric() Jupiter velocity failed.");

        // Heliocentric Sun
        sofaTarget = (int)SofaEphemeris.Body.Sun;
        jplTarget = (int)JplEphemeris.Body.Sun;
        pv1 = sofaEph.Heliocentric(tdb, sofaTarget);
        pv2 = jplEph.Heliocentric(tdb, jplTarget);
        pv1.Position.EqualsApprox(pv2.Position, positionTolerance,
            "SofaEphemeris.Heliocentric() Sun position failed.");
        pv1.Velocity.EqualsApprox(pv2.Velocity, velocityTolerance,
            "SofaEphemeris.Heliocentric() Sun velocity failed.");

        // Heliocentric Earth
        sofaTarget = (int)SofaEphemeris.Body.Earth;
        jplTarget = (int)JplEphemeris.Body.Earth;
        pv1 = sofaEph.Heliocentric(tdb, sofaTarget);
        pv2 = jplEph.Heliocentric(tdb, jplTarget);
        pv1.Position.EqualsApprox(pv2.Position, positionTolerance,
            "SofaEphemeris.Heliocentric() Earth position failed.");
        pv1.Velocity.EqualsApprox(pv2.Velocity, velocityTolerance,
            "SofaEphemeris.Heliocentric() Earth velocity failed.");

        // Heliocentric Jupiter
        sofaTarget = (int)SofaEphemeris.Body.Jupiter;
        jplTarget = (int)JplEphemeris.Body.Jupiter;
        pv1 = sofaEph.Heliocentric(tdb, sofaTarget);
        pv2 = jplEph.Heliocentric(tdb, jplTarget);
        pv1.Position.EqualsApprox(pv2.Position, positionTolerance,
            "SofaEphemeris.Heliocentric() Jupiter position failed.");
        pv1.Velocity.EqualsApprox(pv2.Velocity, velocityTolerance,
            "SofaEphemeris.Heliocentric() Jupiter velocity failed.");

        // Geocentric Sun
        sofaTarget = (int)SofaEphemeris.Body.Sun;
        jplTarget = (int)JplEphemeris.Body.Sun;
        pv1 = sofaEph.Geocentric(tdb, sofaTarget);
        pv2 = jplEph.Geocentric(tdb, jplTarget);
        pv1.Position.EqualsApprox(pv2.Position, positionTolerance,
            "SofaEphemeris.Geocentric() Sun position failed.");
        pv1.Velocity.EqualsApprox(pv2.Velocity, velocityTolerance,
            "SofaEphemeris.Geocentric() Sun velocity failed.");

        // Geocentric Earth
        sofaTarget = (int)SofaEphemeris.Body.Earth;
        jplTarget = (int)JplEphemeris.Body.Earth;
        pv1 = sofaEph.Geocentric(tdb, sofaTarget);
        pv2 = jplEph.Geocentric(tdb, jplTarget);
        pv1.Position.EqualsApprox(pv2.Position, positionTolerance,
            "SofaEphemeris.Geocentric() Earth position failed.");
        pv1.Velocity.EqualsApprox(pv2.Velocity, velocityTolerance,
            "SofaEphemeris.Geocentric() Earth velocity failed.");

        // Geocentric Jupiter
        sofaTarget = (int)SofaEphemeris.Body.Jupiter;
        jplTarget = (int)JplEphemeris.Body.Jupiter;
        pv1 = sofaEph.Geocentric(tdb, sofaTarget);
        pv2 = jplEph.Geocentric(tdb, jplTarget);
        pv1.Position.EqualsApprox(pv2.Position, positionTolerance,
            "SofaEphemeris.Geocentric() Jupiter position failed.");
        pv1.Velocity.EqualsApprox(pv2.Velocity, velocityTolerance,
            "SofaEphemeris.Geocentric() Jupiter velocity failed.");

        // Status. At this point it should be zero. Then request a remote epoch and verify it
        // becomes nonzero.
        if (sofaEph.Status != 0)
            throw new Exception("SofaEphemeris.Status should be 0.");
        sofaEph.Barycentric(new JulianDate(1800, 1, 1, 0), jplTarget);
        if (sofaEph.Status == 0)
            throw new Exception("SofaEphemeris.Status should be nonzero.");

        // km per au
        if (Math.Abs(sofaEph.KmPerAu - 1.496e8) > .001e8)
            throw new Exception("SofaEphemeris.KmPerAu property is wrong.");
    }

#endif      // end test code
}       // end class SofaEphemeris
}       // end namespace HirosePS.SofaJpl
