﻿// SolarSystemBody.cs
// Class to represent a solar system body in the SofaJpl project.
// Copyright 2018 by Paul S. Hirose

// modification history
#if false

2015-06-28  Rarranged astrometricPlace() computation sequence and apparentPlace() a little too.
2015-09-05  Changed TopocentricGeometric() to return a PVVector so it's consistent with the other
methods that return geometric place.

#endif


using System;
using System.Collections.Generic;


namespace HirosePS.SofaJpl {

/// <summary>
/// Base class of all solar system bodies.
/// </summary>
public abstract class SolarSystemBody : Body {

    // max no. of iterations allowed in the light time loop
    private const int MAX_ITERATIONS = 10;

    // Desired accuracy (radians) of the light time correction.
    private double _accuracy;

    // radii of the named solar system bodies (km)
    private static Dictionary<string, float> _radius;


    static SolarSystemBody() {
        _radius = new Dictionary<string, float>(11);
        // values from 2016 Almanac
        _radius.Add("Mercury", 2439.7f);
        _radius.Add("Venus", 6051.8f);
        _radius.Add("Earth", 6378.14f);
        _radius.Add("Mars", 3396.19f);
        _radius.Add("Jupiter", 71492.0f);
        _radius.Add("Saturn", 60268.0f);
        _radius.Add("Uranus", 25559.0f);
        _radius.Add("Neptune", 24764.0f);
        _radius.Add("Pluto", 1195.0f);
        _radius.Add("Moon", 1737.4f);
        _radius.Add("Sun", 696e3f);
    }

    

    // properties


    /// <summary>
    /// Get or set the desired accuracy of the light time correction. Does
    /// not affect geometric place computation.</summary>
    /// <value>desired accuracy (radians).</value>
    public double Accuracy {
        get { return _accuracy; }
        set { _accuracy = value; }
    }




    // instance methods


    /// <summary>
    /// Geocentric astrometric place.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>geocentric astrometric place (km), in the same coordinate reference
    /// system as the ephemeris</returns>
    /// <remarks>The algorithm is from The Astronomical Almanac 2016. Light time includes
    /// gravitational retardation due to the Sun. The returned vector's magnitude is the distance
    /// between the geometric place of the geocenter at <paramref name="tdb"/> and the body when it
    /// emitted the light.</remarks>
    public override Vector GeocentricAstrometric(JulianDate tdb) {
        if (this.IsEarth)
            return new Vector();
        Vector E, Q;
        return astrometricPlace(tdb, new Vector(), out E, out Q, true);
    }


    /// <summary>
    /// Geocentric apparent place, optionally including light deflection due to solar gravity.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="deflect">true = include light deflection due to solar gravity
    /// </param>
    /// <returns>geocentric apparent place (km), in the same coordinate reference
    /// system as the ephemeris</returns>
    /// <remarks>The algorithm is from The Astronomical Almanac 2016. It includes the gravitational
    /// retardation and (if requested) deflection of light due to the Sun. The returned vector's
    /// magnitude is the distance between the geometric place of the geocenter at
    /// <paramref name="tdb"/> and the body when it emitted the light.</remarks>
    public override Vector GeocentricApparent(JulianDate tdb, bool deflect) {
        if (this.IsEarth)
            return new Vector();
        return apparentPlace(tdb, new PVVector(), deflect, true);
    }


    /// <summary>
    /// Geocentric apparent place.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>geocentric apparent place (km), in the same coordinate reference
    /// system as the ephemeris</returns>
    /// <remarks>The algorithm is from The Astronomical Almanac 2016. It includes the gravitational
    /// retardation and deflection of light due to the Sun. The returned vector's
    /// magnitude is the distance between the geometric place of the geocenter at
    /// <paramref name="tdb"/> and the body when it emitted the light.</remarks>
    public override Vector GeocentricApparent(JulianDate tdb) {
        return GeocentricApparent(tdb, true);
    }


    // Replaced by a method which returns a PVVector.
#if false
    /// <summary>
    /// Topocentric geometric place.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observerP">observer position (km) with respect to the geocenter,
    /// in the same coordinate reference system as the ephemeris</param>
    /// <returns>topocentric geometric place (km), in the same coordinate reference
    /// system as the ephemeris</returns>
    public override Vector TopocentricGeometric(JulianDate tdb, Vector observerP) {
        // Result is correct even if the body is Earth, since GeocentricGeometric()
        // will return a zero vector.
        return GeocentricGeometric(tdb).Position - observerP;
    }
#else
    /// <summary>
    /// Topocentric geometric position and velocity.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observerPV">observer position (km) and velocity (km/day) with respect to the
    /// geocenter, in the same coordinate reference system as the ephemeris</param>
    /// <returns>topocentric geometric place (km) and velocity (km/day), in the same coordinate
    /// reference system as the ephemeris</returns>
    public override PVVector TopocentricGeometric(JulianDate tdb, PVVector observerPV) {
        // Result is correct even if the body is Earth, since GeocentricGeometric()
        // will return a zero vector.
        return GeocentricGeometric(tdb) - observerPV;
    }
#endif


    /// <summary>
    /// Topocentric astrometric place.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observerP">observer position (km) with respect to the geocenter,
    /// in the same coordinate reference system as the ephemeris</param>
    /// <returns>topocentric astrometric place (km), in the same coordinate reference
    /// system as the ephemeris</returns>
    /// <remarks>The algorithm is from The Astronomical Almanac 2016, except that the topocenter
    /// replaces the geocenter. Gravitational light retardation due to the Sun is included. The
    /// returned vector's magnitude is the distance between the geometric place of the topocenter at
    /// <paramref name="tdb"/> and the body when it emitted the light.</remarks>
    public override Vector TopocentricAstrometric(JulianDate tdb, Vector observerP) {
        if (this.IsEarth)
            return -observerP;
        Vector E, Q;
        return astrometricPlace(tdb, observerP, out E, out Q, true);
    }


    /// <summary>
    /// Topocentric apparent place, with optional deflection of light due to solar gravitation.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observer">observer's position (km) and velocity (km/day) with respect to the
    /// geocenter, in the same coordinate reference system as the ephemeris</param>
    /// <param name="deflect">true = include light deflection due to solar gravitation</param>
    /// <returns>topocentric apparent place (km), in the same coordinate
    /// reference system as the ephemeris</returns>
    /// <remarks>The algorithm is from The Astronomical Almanac 2016, except that the barycentric
    /// position and velocity of the topocenter replaces that of the geocenter. Retardation and (if
    /// requested) deflection of light due to solar gravitation are included. The returned vector's
    /// magnitude is the distance between the geometric positions of the observer at
    /// <paramref name="tdb"/> and the body when it emitted the light.</remarks>
    public override Vector TopocentricApparent(JulianDate tdb, PVVector observer,
        bool deflect) {

        if (this.IsEarth)
            return -observer.Position;
        return apparentPlace(tdb, observer, deflect, true);
    }


    /// <summary>
    /// Topocentric apparent place, including deflection of light due to solar gravitation.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observer">observer's position (km) and velocity (km/day) with respect to the
    /// geocenter, in the same coordinate reference system as the ephemeris</param>
    /// <returns>topocentric apparent place (km), in the same coordinate
    /// reference system as the ephemeris</returns>
    /// <remarks>The algorithm is from The Astronomical Almanac 2016, except that the barycentric
    /// position and velocity of the topocenter replaces that of the geocenter. Retardation and
    /// deflection of light due to solar gravitation are included. The returned vector's
    /// magnitude is the distance between the geometric positions of the observer at
    /// <paramref name="tdb"/> and the body when it emitted the light.</remarks>
    public override Vector TopocentricApparent(JulianDate tdb, PVVector observer) {
        return TopocentricApparent(tdb, observer, true);
    }


#if DEBUG
    /// <summary>
    /// Astrometric position of this instance.
    /// </summary>
    /// <remarks>The returned vector's modulus is the distance between the observer
    /// at <paramref name="tdb"/> and the body at (<paramref name="tdb"/> minus
    /// light time).</remarks>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observer">observer position (km) with respect to the geocenter,
    /// in the same coordinate reference system as the ephemeris</param>
    /// <param name="E">(output) observer heliocentric position (km) at
    /// <paramref name="tdb"/></param>
    /// <param name="Q">(output) target heliocentric position (km) at
    /// <paramref name="tdb"/> minus light time </param>
    /// <param name="retard">true = include the effect of gravitational retardation
    /// due to the Sun</param>
    /// <returns>astrometric position (km), in the same coordinate reference system
    /// as the ephemeris</returns>
#endif
    private Vector astrometricPlace(JulianDate tdb, Vector observer,
        out Vector E, out Vector Q, bool retard) {

        // observer barycentric position at tdb
        // Vector Eb = earthBaryPV(tdb).Position + observer;
        Vector Eb = Ephemeris.EarthBaryPV(tdb).Position + observer;

        // Sun barycentric position at tdb
        // Vector Sb = sunBaryP(tdb);
        Vector Sb = Ephemeris.SunBaryPV(tdb).Position;

        // observer heliocentric position at tdb
        E = Eb - Sb;
        double Emod = E.Modulus();

        // light time (first approximation = 0)
        Duration tau = new Duration(0.0);

        // observer to target vector from previous loop iteration
        Vector P1 = new Vector();

        // observer to target vector from current loop iteration
        Vector P2 = new Vector();
        double Pmod;    // will receive modulus of P2

        // Target heliocentric and barycentric positions at tdb minus tau. Initialize with tau = 0.
        Q = Heliocentric(tdb).Position;
        Vector Qb = Barycentric(tdb).Position;

        double Qmod;    // will receive modulus of Q

        int i;
        for (i = 1; i <= MAX_ITERATIONS; ++i) {
            // vector from observer (at tdb) to target (at tdb minus tau), and vector modulus
            P2 = Qb - Eb;
            Pmod = P2.Modulus();

            // If P2 is close enough (in radian measure) to the previous iteration, exit unless this
            // is the first iteration.
            if (i != 1 && new Vector(P2 - P1).Modulus() / Pmod < Accuracy)
                break;

            // Improve the estimate of tau.
            if (IsSun || retard == false)
                tau = new Duration(Pmod / cLight);
            else {
                Qmod = Q.Modulus();
                tau = new Duration((Pmod + 2.0 * muOverC2
                    * Math.Log((Emod + Pmod + Qmod) / (Emod - Pmod + Qmod))) /
                    cLight);
            }
            // Update target heliocentric and barycentric positions at tdb minus tau. Even if
            // retardation is not requested, update Q because it's one of the method outputs.
            Q = Heliocentric(tdb - tau).Position;
            Qb = Barycentric(tdb - tau).Position;
            P1 = P2;    // remember the previous vector
        }
        if (i > MAX_ITERATIONS)
            throw new Exception("light time loop failed to converge");

        return P2;
    }


#if DEBUG
    /// <summary>
    /// Apparent position.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observerPV">observer position (km) and velocity (km/day) with
    /// respect to the geocenter, in the same coordinate reference system as the
    /// ephemeris</param>
    /// <param name="deflect">true = apply light deflection due to solar gravitation</param>
    /// <param name="retard">true = apply gravitational retardation due to the Sun</param>
    /// <returns>apparent position (km), in the same coordinate frame as the
    /// ephemeris</returns>
    /// <remarks>The returned vector's modulus is the geometric distance between the observer
    /// position at <paramref name="tdb"/> and the body when it emitted the light.</remarks>
#endif
    private Vector apparentPlace(JulianDate tdb, PVVector observerPV, bool deflect,
        bool retard) {
        Vector E;       // observer heliocentric position at tdb
        Vector Q;       // target heliocentric position at when it emitted the light 
        Vector P = astrometricPlace(tdb, observerPV.Position, out E, out Q, retard);

        // form unit vectors
        Vector e = E.Unit();
        Vector p = P.Unit();
        Vector q = Q.Unit();

        // Calculate the direction of the target, corrected for light deflection due
        // to the Sun's gravity. Constant k includes scale factor for units of km.
        Vector p1;
        double k = 2.0 * muOverC2 / (Ephemeris.KmPerAu * Ephemeris.KmPerAu);
        if (IsSun || deflect == false)
            p1 = p;
        else
            p1 = p + k * E.Modulus() * (p.Dot(q) * e - e.Dot(p) * q) / (1.0 + q.Dot(e));

        // observer velocity (including the diurnal component) in terms of c
        Vector V = new PVVector(Ephemeris.EarthBaryPV(tdb) + observerPV).Velocity / cLight;
        double betaInv = Math.Sqrt(1.0 - V.Modulus() * V.Modulus());

        // Calculate the proper direction of the target by applying the observer's
        // velocity.
        Vector p2 = (betaInv * p1 + (1.0 + p1.Dot(V) / (1.0 + betaInv)) * V)
            / (1.0 + p1.Dot(V));

        // Make the modulus of the returned vector equal to the modulus of the
        // astrometric vector.
        return P.Modulus() * p2.Unit();
    }



    // static methods


    /// <summary>
    /// Get the equatorial radius of the named solar system body.
    /// </summary>
    /// <param name="name">name (case sensitive)</param>
    /// <returns>radius (km)</returns>
    /// <exception cref="KeyNotFoundException"> if name is not found</exception>
    /// <remarks>Names are Mercury, Venus, Earth, Mars, Jupiter, Saturn, Uranus, Neptune, Pluto,
    /// Moon, Sun (case sensitive).</remarks>
    public static float BodyRadius(string name) {
        return _radius[name];
    }

}       // end class SolarSystemBody
}       // end namespace HirosePS.SofaJpl
