// Spherical.cs
// Spherical coordinate class.
// Copyright 2018 by Paul S. Hirose.

// Revision history.
#if false

2007-10-24 Eliminated unsafe code and references to the Win32 Sofajpl DLL.
Many of the XML comments are from the IAU SOFA documentation, and the class uses the SOFA
library, so the SOFA software license applies to this class.

2007-12-05 Completed self-test code.

2008-03-13 Minor improvements to comments.

2008-06-20 New constructor for a unit vector from spherical coords. Tests OK.

2008-11-02 Deleted angle normalization in the constructors. Added El property.

2008-11-04 Added RA, HA, Dec properties. Test OK.

2009-01-12 Added El, Distance properties.

2012-03-30 Added PolarAngle, Colatitude, NPD properties. Deleted SOFA license. It's
available online. Self-test OK.

2015-05-10 I changed to C source for the SOFA library. Its different capitalization style compared
to the Fortran code required corresponding changes here. Storage order for matrices is reversed to
match the C convention.

2015-05-19 SeparationAngle and PositionAngle methods have overloads to accept a Vector.
Methods and properties are virtual now. Self test OK.

2015-06-26 De-virtualized the class.

2018-08-12 Improved the XML comments.
#endif


using System;


namespace HirosePS.SofaJpl
{
    /// <summary>
    /// A class to represent position as spherical coordinates.
    /// </summary>
    /// <remarks>
    /// <para>The internal representation stores radius, theta, and phi in double precision.</para>
    /// <para>The most common use of this class is to convert a Vector object to
    /// spherical coordinates.</para>
    /// <para>Instances of this class are immutable.</para>
    /// </remarks>
    public class Spherical
    {
        // the internal representation
        // longitude (increasing east), latitude (increasing north), radius
        private double theta, phi, r;


        // constructors

        /// <summary>
        /// Initialize from another Spherical instance.
        /// </summary>
        /// <param name="s">Spherical</param>
        public Spherical(Spherical s)
        {
            this.theta = s.theta;
            this.phi = s.phi;
            this.r = s.r;
        }

        /// <summary>
        /// Initialize from spherical angles and radius.
        /// </summary>
        /// <remarks>
        /// Angle parameters are stored in the object without change. For example,
        /// phi is not be normalized to the interval [-pi/2, +pi/2].
        /// </remarks>
        /// <param name="theta">Longitude angle (radians), increasing east.</param>
        /// <param name="phi">Latitude angle (radians), increasing north.</param>
        /// <param name="r">Radius.</param>
        public Spherical(
            double theta,
            double phi,
            double r)
        {
            this.theta = theta;
            this.phi = phi;
            this.r = r;
        }


        /// <summary>
        /// Initialize from spherical angles only. Radius is set to 1.
        /// </summary>
        /// <remarks>
        /// Angle parameters are stored in the object without change. For example,
        /// phi is not be normalized to the interval [-pi/2, +pi/2].
        /// </remarks>
        /// <param name="theta">Longitude angle (radians), increasing east.</param>
        /// <param name="phi">Latitude angle (radians), increasing north.</param>
        public Spherical(double theta, double phi)
            : this(theta, phi, 1.0)
        { }


        /// <summary>
        /// Construct a Spherical from a Vector instance.
        /// </summary>
        /// <remarks>If the Vector coincides with the +x axis, theta = 0 and phi = 0.
        /// If the Vector coincides with the +y axis, theta = pi/2 and phi = 0.
        /// If the Vector coincides with the +z axis, theta = 0 and phi = pi/2.
        /// If the Vector is zero, all components of the result are zero.
        /// Theta will be in the interval [-pi, +pi].
        /// </remarks>
        /// <param name="v">Vector instance</param>
        public Spherical(Vector v)
        {
            Sofa.P2s(v.ToArray(), out this.theta, out this.phi, out this.r);
        }


        /// <summary>
        /// Convert this instance into a string with longitude, latitude, radius.
        /// </summary>
        /// <returns>String representation of Spherical.</returns>
        public override string ToString()
        {
            return string.Format("{0:f6} {1:f6} {2:g6}", theta, phi, r);
        }


        /// <summary>
        /// Get the theta, the longitude angle. This is the internal
        /// representation, and not necessarily normalized.
        /// </summary>
        /// <value>theta (radians)</value>
        public double Theta { get { return this.theta; } }

        /// <summary>
        /// Get the longitude angle: positive east, increasing to 2pi radians.
        /// </summary>
        /// <value>east longitude angle in the interval [0, 2pi) radians</value>
        public double LonEast { get { return Angle.NormPlus(Theta); } }

        /// <summary>
        /// Get Right Ascension: positive east, increasing to 2pi radians.
        /// </summary>
        /// <value>Right Ascension in the interval [0, 2pi) radians</value>
        public double RA { get { return LonEast; } }

        /// <summary>
        /// Get the longitude angle: positive west, increasing to 2pi radians.
        /// </summary>
        /// <value>west longitude angle in the interval [0, 2pi) radians</value>
        public double LonWest { get { return Angle.NormPlus(Angle.Explement(Theta)); } }

        /// <summary>
        /// Get the hour angle: positive west, increasing to 2pi radians.
        /// </summary>
        /// <value>hour angle in the interval [0, 2pi) radians</value>
        public double HA { get { return LonWest; } }

        /// <summary>
        /// Get the longitude angle: positive east, negative west.
        /// </summary>
        /// <value>east longitude angle in the interval [-pi, +pi) radians</value>
        public double LonEastWest { get { return Angle.NormPlusMinus(Theta); } }

        /// <summary>
        /// Angle theta, adjusted to the normal azimuth convention: at theta = pi/2,
        /// azimuth = 0, increasing as theta decreases.
        /// </summary>
        /// <value>azimuth in the interval [0, 2pi) radians</value>
        public double Az
        {    get
            { return Angle.NormPlus(Angle.Explement(Theta - Angle.HalfPi)); }
        }

        /// <summary>
        /// Get angle phi (latitude angle).
        /// </summary>
        /// <value>angle phi (radians)</value>
        public double Phi { get { return this.phi; } }

        /// <summary>
        /// Get the elevation angle. Identical to property Phi.
        /// </summary>
        /// <value>elevation angle (radians)</value>
        public double El { get { return Phi; } }


        /// <summary>
        /// Get the altitude (elevation) angle. Identical to property Phi.
        /// </summary>
        /// <value>altitude (radians)</value>
        public double Alt { get { return Phi; } }


        /// <summary>
        /// Get the north latitude angle. Identical to property Phi.
        /// </summary>
        /// <value>north latitude angle (radians)</value>
        public double Lat { get { return Phi; } }

        /// <summary>
        /// Get the declination angle. Identical to property Phi.
        /// </summary>
        /// <value>declination angle (radians)</value>
        public double Dec { get { return Phi; } }

        /// <summary>
        /// Get the zenith distance (complement of elevation angle).
        /// </summary>
        /// <value>zenith distance (radians)</value>
        public double ZenithDistance { get { return Angle.Complement(Phi); } }

        /// <summary>
        /// Get the polar angle (= zenith distance) in radians.
        /// </summary>
        public double PolarAngle { get { return ZenithDistance; } }

        /// <summary>
        /// Get Colatitude (= zenith distance), radians.
        /// </summary>
        public double Colatitude { get { return ZenithDistance; } }

        /// <summary>
        /// Get north polar distance (= zenith distance), radians.
        /// </summary>
        public double NPD { get { return ZenithDistance; } }

        /// <summary>
        /// Get the radius.
        /// </summary>
        /// <value>radius</value>
        public double Radius { get { return this.r; } }


        /// <summary>
        /// Get the distance (radius) from the origin.
        /// </summary>
        /// <value>distance (radius) from origin.</value>
        public double Distance { get { return this.r; } }


        // methods

        /// <summary>
        /// Separation angle between this instance and another Spherical.
        /// </summary>
        /// <remarks>Uses a vector method for high accuracy regardless of the angle.</remarks>
        /// <param name="rhs">the other Spherical</param>
        /// <returns>separation angle (radians)</returns>
        public double SeparationAngle(Spherical rhs)
        {
            return Sofa.Seps(this.theta, this.phi, rhs.theta, rhs.phi);
        }


        /// <summary>
        /// Separation angle between this instance and a Vector.
        /// </summary>
        /// <remarks>Uses a vector method for high accuracy regardless of the angle.</remarks>
        /// <param name="vec">Vector to the other object</param>
        /// <returns>separation angle (radians)</returns>
        public double SeparationAngle(Vector vec) {
            return vec.SeparationAngle(new Vector(this));
        }


        /// <summary>
        /// Position angle of another Spherical with respect to this Spherical.
        /// Zero at north, increases east.
        /// </summary>
        /// <remarks>Uses a vector method for high accuracy at any angle. Returns zero if the
        /// points coincide.</remarks>
        /// <param name="rhs">the other point</param>
        /// <returns>position angle of the other point (radians, pi to -pi)</returns>
        public double PositionAngle(Spherical rhs)
        {
            return Sofa.Pas(this.theta, this.phi, rhs.theta, rhs.phi);
        }


        /// <summary>
        /// Position angle of a Vector with respect to this Spherical. Zero at north, increases
        /// east.
        /// </summary>
        /// <remarks>Uses a vector method for high accuracy at any angle. Returns zero if the
        /// points coincide.</remarks>
        /// <param name="vec">Vector to the other point</param>
        /// <returns>position angle of the other point (radians, pi to -pi)</returns>
        public double PositionAngle(Vector vec) {
            return this.PositionAngle(new Spherical(vec));
        }




        #region self-test code
#if DEBUG


        /// <summary>
        /// Self-test method. Runs silently if all tests pass.
        /// </summary>
        public static void Test()
        {
            Spherical sph1, sph2;
            Vector vec1, vec2;

            // Initialize from 2 spherical coordinates
            sph1 = new Spherical(2.5, 1.5);
            sph1.EqualsApprox(2.5, 1.5, 1.0, 0.0, 0.0);

            // Initialize from 3 spherical coordinates
            sph1 = new Spherical(6.350872, -.148138, 95.827);
            sph1.EqualsApprox(6.350872, -0.148138, 95.827, .000001, .001);

            // Initialize from another Spherical
            sph2 = new Spherical(sph1);
            sph1.EqualsApprox(6.350872, -0.148138, 95.827, .000001, .001);

            // Initialize from spherical coordinates at pole
            sph1 = new Spherical(Angle.DegreesToRadians(616.1), Angle.DegreesToRadians(-90.0), .9);
            sph1.EqualsApprox(10.752974, -1.570796, 0.9, .000001, .001);

            // Initialize from a Vector
            vec1 = new Vector(7.181, -3.048, 4.183);
            sph1 = new Spherical(vec1);
            sph1.EqualsApprox(-0.401408, 0.492192, 8.85181, .000001, .00001);

            // Initialize from a Vector at pole:");
            vec1 = new Vector(0.0, 0.0, 4.183);
            sph1 = new Spherical(vec1);
            assert(Math.Abs(sph1.phi - 1.570796) < .000001 && Math.Abs(sph1.r - 4.183) < .001);

            // ToString()
            assert(sph2.ToString().Equals("6.350872 -0.148138 95.827"));

            // Properties
            vec1 = new Vector(7.181, -3.048, 4.183);
            sph1 = new Spherical(vec1);
            assert(Math.Abs(sph1.Theta - (-0.401408)) < .000001);
            assert(Math.Abs(sph1.LonEast - 5.881778) < .000001);
            assert(Math.Abs(sph1.RA - 5.881778) < .000001);
            assert(Math.Abs(sph1.LonWest - 0.401408) < .000001);
            assert(Math.Abs(sph1.HA - 0.401408) < .000001);
            assert(Math.Abs(sph1.LonEastWest - (-0.401408)) < .000001);
            assert(Math.Abs(sph1.Az - 1.972204) < .000001);
            assert(Math.Abs(sph1.Phi - 0.492192) < .000001);
            assert(Math.Abs(sph1.Alt - 0.492192) < .000001);
            assert(Math.Abs(sph1.El - 0.492192) < .000001);
            assert(Math.Abs(sph1.Lat - 0.492192) < .000001);
            assert(Math.Abs(sph1.Dec - 0.492192) < .000001);
            assert(Math.Abs(sph1.ZenithDistance - 1.078604) < .000001);
            assert(sph1.ZenithDistance == sph1.PolarAngle);
            assert(sph1.ZenithDistance == sph1.NPD);
            assert(sph1.ZenithDistance == sph1.Colatitude);
            assert(Math.Abs(sph1.Radius - 8.851811) < .000001);
            assert(Math.Abs(sph1.Distance - 8.851811) < .000001);

            // separation angle

            sph1 = new Spherical(1.33, .630, 2.6);
            sph2 = new Spherical(1.33, -1.342, 2.6);
            assert(Math.Abs(sph1.SeparationAngle(sph2) - 1.972) < .001);
            vec2 = new Vector(sph2);
            assert(Math.Abs(sph1.SeparationAngle(vec2) - 1.972) < .001);

            // position angle

            sph2 = new Spherical(1.34, .630, 2.6);
            assert(Math.Abs(sph1.PositionAngle(sph2) - 1.571) < .01);
            vec2 = new Vector(sph2);
            assert(Math.Abs(sph1.PositionAngle(vec2) - 1.571) < .01);

            sph2 = new Spherical(1.32, .630, 2.6);
            assert(Math.Abs(sph1.PositionAngle(sph2) - (-1.571)) < .01);
            vec2 = new Vector(sph2);
            assert(Math.Abs(sph1.PositionAngle(vec2) - (-1.571)) < .01);
        }


        /// <summary>
        /// Test that the values of the instance are within tolerance.
        /// </summary>
        /// <param name="theta">expected longitude angle</param>
        /// <param name="phi">expected latitude angle</param>
        /// <param name="r">expected radius</param>
        /// <param name="toleranceAngle">max allowable error in angles</param>
        /// <param name="toleranceRadius">max allowable error in radius</param>
        public void EqualsApprox(double theta, double phi, double r,
            double toleranceAngle, double toleranceRadius)
        {
            if (Math.Abs(this.theta - theta) > toleranceAngle ||
                Math.Abs(this.phi - phi) > toleranceAngle ||
                Math.Abs(this.r - r) > toleranceRadius)
                throw new Exception("Spherical out of tolerance.");
        }

        private static void assert(bool assertion)
        {
            if (!assertion)
                throw new Exception("Assertion failed.");
        }

#endif
        #endregion
    }
}
