﻿// Star.cs
// Class to represent a star. Part of the SofaJpl fundamental astronomy toolbox.
// Copyright 2018 by Paul S. Hirose

// revision history
#if false

2015-07-02  Created.
2015-07-03  Self test OK.
2015-09-05  Changed TopocentricGeometric() to return a PVVector so it's consistent with the other
methods that return geometric place.

2018-04-20  The km/au value is a constant, not taken from the ephemeris. This lets you use null
for the ephemeris parameter of the constructor, if only the barycentric position of the star is
needed.
#endif


using System;


namespace HirosePS.SofaJpl {

/// <summary>
/// Class to represent a star.
/// </summary>
public class Star : Body {
    private const double secondsPerDay = 86400.0;
    private const double daysPerJy = 365.25;
    private const double kmPerAu = 0.149597870700000000e+09;

    // Smallest usable parallax. Any less parallax (including zero or negative)
    // is equivalent to this value. Initialized in static constructor.
    private static readonly double _minimumParallax;     // radians

    // epoch of the catalog data
    JulianDate _epoch;

    // catalog right ascension and declination
    private double _ra, _dec;   // radians

    // catalog proper motion in right ascension, in great circle units, i.e., the
    // cosine of declination is included.
    private double _raPM;      // radians per day

    // catalog proper motion in declination
    private double _decPM;      // radians per day

    // catalog parallax, radians
    private double _parallax;

    // catalog radial velocity, km per day, receding
    private double _rv;




    static Star() {
        // Max distance is 10 gigaparsecs, or about 3.1e23 km.
        _minimumParallax = Angle.DmsToRad(0, 0, 1e-10);
    }


    /// <summary>
    /// Construct a Star object.
    /// </summary>
    /// <remarks>
    /// <para>The star data and the ephemeris use the same coordinate reference system,
    /// e.g., the ICRS. Output will be in the same system.</para>
    /// <para>RA proper motion is in great circle units, i.e., includes the cos(declination) factor.
    /// </para>
    /// <para>If only the barycentric position of the star will be requested, the
    /// <paramref name="eph"/> parameter is not used and therefore may be null.</para>
    /// </remarks>
    /// <param name="catalogEpoch">the time (TDB) at which the parameters are valid
    /// </param>
    /// <param name="ra">Right Ascension (hours)</param>
    /// <param name="dec">declination (degrees)</param>
    /// <param name="parallax">parallax in milli arc seconds (mas)</param>
    /// <param name="pmRa">RA proper motion (great circle mas per Julian year)</param>
    /// <param name="pmDec">declination proper motion (mas/year)</param>
    /// <param name="radialVel">radial velocity, receding (km/s)</param>
    /// <param name="eph">solar system ephemeris</param>
    /// <param name="name">star name</param>
    public Star(JulianDate catalogEpoch, double ra, double dec, double parallax,
        double pmRa, double pmDec, double radialVel, Ephemeris eph, string name) {
        _epoch = catalogEpoch;
        _ra = Angle.HoursToRadians(ra);
        _dec = Angle.DegreesToRadians(dec);
        _parallax = Angle.DmsToRad(0, 0, parallax / 1000.0);
        _raPM = Angle.DmsToRad(0, 0, pmRa / 1000.0) / daysPerJy;
        _decPM = Angle.DmsToRad(0, 0, pmDec / 1000.0) / daysPerJy;
        _rv = radialVel * secondsPerDay;

        Name = name;
        Ephemeris = eph;
    }



    /// <summary>
    /// Minimum effective parallax (radians).
    /// </summary>
    /// <remarks>If the parallax supplied to the constructor is less than this value
    /// (including zero or negative), this value is used instead of the actual
    /// parallax. This value is a constant.
    /// </remarks>
    /// <value>get the minimum usable parallax (radians)</value>
    public double MinimumParallax { get { return _minimumParallax; } }


    /// <summary>
    /// The epoch of the star catalog data.
    /// </summary>
    /// <value>epoch of the star catalog data</value>
    public JulianDate CatalogEpoch {
        get { return this._epoch; }
    }


    /// <summary>
    /// right ascension (radians)
    /// </summary>
    /// <value>catalog right ascension (radians)</value>
    public double RightAsc {
        get { return this._ra; }
    }


    /// <summary>
    /// declination (radians)
    /// </summary>
    /// <value>sets/gets declination (radians)</value>
    public double Dec {
        get { return this._dec; }
    }


    /// <summary>
    /// Proper motion in right ascension (radians per day) in great circle units
    /// </summary>
    /// <value>proper motion in right ascension (great circle radians per day)
    /// </value>
    public double RightAscPropMot {
        get { return this._raPM; }
    }


    /// <summary>
    /// Proper motion in declination (radians per day)
    /// </summary>
    /// <value>proper motion in declination (radians per day)</value>
    public double DecPropMot {
        get { return this._decPM; }
    }


    /// <summary>
    /// Parallax (radians)
    /// </summary>
    /// <value>parallax (radians)</value>
    public double Para {
        get { return this._parallax; }
    }


    /// <summary>
    /// Radial receding velocity (km per day)
    /// </summary>
    /// <value>radial velocity (km per day, receding)</value>
    public double RadVel {
        get { return this._rv; }
    }


    /// <summary>
    /// Barycentric geometric position and velocity.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>barycentric geometric position (km) and velocity (km/day), in the
    /// same coordinate reference system as the catalog data
    /// </returns>
    public override PVVector Barycentric(JulianDate tdb) {
        return starBarycentricPV(tdb);
    }


    /// <summary>
    /// Heliocentric geometric position and velocity.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>heliocentric geometric position (km) and velocity (km/day), in the
    /// same coordinate reference system as the catalog data</returns>
    public override PVVector Heliocentric(JulianDate tdb) {
        PVVector starBaryPV = Barycentric(tdb);
        PVVector sunBaryPV = Ephemeris.SunBaryPV(tdb);
        return starBaryPV - sunBaryPV;
    }


    /// <summary>
    /// Geocentric geometric position.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>geocentric geometric position (km), in the same coordinate reference
    /// system as the catalog data</returns>
    public override PVVector GeocentricGeometric(JulianDate tdb) {
        return starBarycentricPV(tdb) - Ephemeris.EarthBaryPV(tdb);
    }


    /// <summary>
    /// Geocentric astrometric (= geometric) position.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>geocentric astrometric position (km) in the same coordinate reference
    /// system as the catalog data</returns>
    public override Vector GeocentricAstrometric(JulianDate tdb) {
        return GeocentricGeometric(tdb).Position;
    }


    /// <summary>
    /// Geocentric apparent place, with optional light deflection due to solar gravity.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="deflect">true = include light deflection due to solar gravity</param>
    /// <returns>geocentric apparent position (km) in the same coordinate reference
    /// system as the catalog data</returns>
    /// <remarks>The algorithm is from The Astronomical Almanac 2016. The modulus of the returned
    /// vector is the distance (km) to the star.</remarks>
    public override Vector GeocentricApparent(JulianDate tdb, bool deflect) {
        return apparentP(tdb, Ephemeris.EarthBaryPV(tdb), deflect);
    }


    /// <summary>
    /// Geocentric apparent place, including light deflection due to solar gravity.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <returns>geocentric apparent position (km) in the same coordinate reference
    /// system as the catalog data</returns>
    /// <remarks>The algorithm is from The Astronomical Almanac 2016. The modulus of the returned
    /// vector is the distance (km) to the star.</remarks>
    public override Vector GeocentricApparent(JulianDate tdb) {
        return GeocentricApparent(tdb, true);
    }


    // Replaced by a method that returns a PVVector.
#if false
    /// <summary>
    /// Topocentric geometric position.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observerP">observer position (km) with respect to the geocenter
    /// </param>
    /// <returns>topocentric geometric position (km)</returns>
    /// <remarks><paramref name="observerP"/> must use the same reference system as the catalog
    /// data. The output vector is in the same system. Its modulus is the distance (km) to the star.
    /// </remarks>
    public override Vector TopocentricGeometric(JulianDate tdb, Vector observerP) {
        return GeocentricGeometric(tdb).Position - observerP;
    }
#else
    /// <summary>
    /// Topocentric geometric position and velocity.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observerPV">observer position (km) and velocity (km/day) with respect to the
    /// geocenter</param>
    /// <returns>topocentric geometric position (km) and velocity (km/day)</returns>
    /// <remarks><paramref name="observerPV"/> must use the same reference system as the catalog
    /// data. The output is in the same system.
    /// </remarks>
    public override PVVector TopocentricGeometric(JulianDate tdb, PVVector observerPV) {
        return GeocentricGeometric(tdb) - observerPV;
    }
#endif


    /// <summary>
    /// Topocentric astrometric (= geometric) position.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observerP">observer position (km) with respect to the geocenter
    /// </param>
    /// <returns>topocentric astrometric (identical to geometric) position (km)
    /// </returns>
    /// <remarks><paramref name="observerP"/> must use the same reference system as the catalog
    /// data. The output vector is in the same system. Its modulus is the distance (km) to the
    /// star.</remarks>
    public override Vector TopocentricAstrometric(JulianDate tdb, Vector observerP) {
        return TopocentricGeometric(tdb, new PVVector(observerP, new Vector())).Position;
    }


    /// <summary>
    /// Topocentric apparent position, with optional light deflection due to solar gravity.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observer">observer position (km) and velocity (km/day) with
    /// respect to the geocenter</param>
    /// <param name="deflect">true = include light deflection due to solar gravity</param>
    /// <returns>topocentric apparent position (km)</returns>
    /// <remarks><paramref name="observer"/> must use the same coordinate reference system as the
    /// catalog data. Result is in the same system, and is computed with the method in the 2016
    /// Astronomical Almanac, Section B, including dinural parallax and aberration, and (if
    /// requested) light deflection due to the Sun's gravity. Modulus of the returned vector is the
    /// distance to the star.</remarks>
    public override Vector TopocentricApparent(JulianDate tdb, PVVector observer, bool deflect) {
        return apparentP(tdb, Ephemeris.EarthBaryPV(tdb) + observer, deflect);
    }


    /// <summary>
    /// Topocentric apparent position, including light deflection due to solar gravity.
    /// </summary>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observer">observer position (km) and velocity (km/day) with
    /// respect to the geocenter</param>
    /// <returns>topocentric apparent position (km)</returns>
    /// <remarks><paramref name="observer"/> must use the same coordinate reference system as the
    /// catalog data. Result is in the same system, and is computed with the method in the 2016
    /// Astronomical Almanac, Section B, including dinural parallax and aberration, and (if
    /// requested) light deflection due to the Sun's gravity. Modulus of the returned vector is the
    /// distance to the star.</remarks>
    public override Vector TopocentricApparent(JulianDate tdb, PVVector observer) {
        return TopocentricApparent(tdb, observer, true);
    }


#if DEBUG
    /// <summary>
    /// Helper method to compute star barycentric position and velocity.
    /// </summary>
    /// <param name="tdb">TDB</param>
    /// <returns>barycentric star position (km) and velocity (km/day), in the same
    /// reference system as the constructor parameters</returns>
#endif
    private PVVector starBarycentricPV(JulianDate tdb) {
        // Limit distance to a comfortable maximum and prevent division by zero.
        double pi;
        if (Para >= MinimumParallax)
            pi = Para;      // the catalog parallax
        else
            pi = MinimumParallax;   // the minimum allowable parallax

        // This changes the distance unit from au to km.
        pi /= kmPerAu;

        // unit vector to star's catalog position
        Vector r0 = new Vector(new Spherical(RightAsc, Dec, 1.0));

        // sines and cosines of RA and dec
        double sina = Math.Sin(RightAsc);
        double cosa = Math.Cos(RightAsc);
        double sind = Math.Sin(Dec);
        double cosd = Math.Cos(Dec);

        // normal triad unit vectors in the direction of increasing RA and dec
        Vector p0 = new Vector(-sina, cosa, 0.0);   // RA
        Vector q0 = new Vector(-sind * cosa, -sind * sina, cosd);   // dec

        // space motion vector (km per day)
        Vector mu0 = (RightAscPropMot * p0 + DecPropMot * q0) / pi + RadVel * r0;

        // position and velocity relative to the barycenter
        double t = new Duration(tdb - CatalogEpoch).ToDays();
        return new PVVector(r0 / pi + mu0 * t, mu0);
    }


#if DEBUG
    /// <summary>
    /// Helper method to compute star's apparent place.
    /// </summary>
    /// <remarks>The <paramref name="observer"/> parameter
    /// must use the same coordinate reference system as the constructor parameters.
    /// The output vector is in the same system. Its modulus is the distance (km) of
    /// date to the star.</remarks>
    /// <param name="tdb">time (TDB)</param>
    /// <param name="observer">barycentric observer position (km) and velocity
    /// (km/day)</param>
    /// <param name="deflect">true = apply light deflection due to solar gravitation
    /// </param>
    /// <returns>vector to apparent star</returns>
#endif
    private Vector apparentP(JulianDate tdb, PVVector observer, bool deflect) {
        // geocentric (or topocentric) geometric vector to star
        Vector P = starBarycentricPV(tdb).Position - observer.Position;

        // heliocentric vector to observer
        Vector E = observer.Position - this.Ephemeris.SunBaryPV(tdb).Position;

        // convert P and E to unit vectors
        Vector p = P.Unit();
        Vector e = E.Unit();

        // vector to star, optionally corrected for light deflection
        Vector p1;
        if (deflect)
            p1 = p + (2.0 * muOverC2 * E.Modulus() / (kmPerAu * kmPerAu)) *
                (e - p.Dot(e) * p) / (1.0 + p.Dot(e));
        else
            p1 = p;

        // velocity of observer with respect to the speed of light
        Vector V = observer.Velocity / cLight;
        double betaInv = Math.Sqrt(1.0 - V.Modulus() * V.Modulus());

        // proper direction to star
        Vector p2 = (betaInv * p1 + (1.0 + p1.Dot(V) / (1.0 + betaInv)) * V)
            / (1.0 + p1.Dot(V));

        // Make the vector modulus equal to the geometric distance to star.
        return P.Modulus() * p2.Unit();
    }



    
    // self-test code
#if DEBUG

    /// <summary>
    /// Run the class self test.
    /// </summary>
    public static void Test() {
        double masPerRadian = Angle.RadiansToDegrees(1.0) * 3600.0 * 1000.0;
        const int daysPerJcy = 36525;

        Vector p1;
        PVVector pv1, pv2;

        // duplicate the example from Astronomical Almanac 2012 p. B73

        JulianDate tt = new JulianDate(2012, 1, 1, 0.0);

        double ra = 14.0 + (39.0 + 36.4958 / 60.0) / 60.0;
        double dec = -(60.0 + (50.0 + 2.309 / 60.0) / 60.0);
        double pi = 742.0;
        double pmRa = -3678.06;
        double pmDec = 482.87;
        double radVel = -21.6;
        string name = "Almanac";

        SofaEphemeris sofaEph = new SofaEphemeris();
        SofaBody sun = new SofaBody(SofaEphemeris.Body.Sun, sofaEph, .01);

        Star star = new Star(JulianDate.J2000Base, ra, dec, pi, pmRa, pmDec, radVel, sofaEph, name);

        // Verify name.
        if (star.Name != name)
            throw new Exception("Star name wrong.");

        // barycentric position & velocity of date
        pv1 = star.Barycentric(tt);

        // J2000.0 q and m from 2012 Almanac, in km and km/day
        Vector almP =
            new Vector(-.373860494, -.312618798, -.873211210) *
            masPerRadian / pi * sofaEph.KmPerAu;
        Vector almV =
            new Vector(-.000687882, .001749237, .001545387) / daysPerJcy *
            masPerRadian / pi * sofaEph.KmPerAu;

        // Update almanac position to date.
        almP += tt.ToJ2000() * almV;

        // check direction
        testSeparationAngle(pv1.Position, almP, Angle.DmsToRad(0, 0, .001),
            "testStar() barycentric direction");

        // Distance should match Almanac within 1 ppm.
        double modulus = almP.Modulus();
        equalsApprox(pv1.Position.Modulus(), modulus, modulus * 1e-6,
            "testStar() barycentric distance");

        // Check space velocity.
        testSeparationAngle(pv1.Velocity, almV, Angle.DmsToRad(0, 0, 1),
            "testStar() barycentric velocity direction");
        modulus = almV.Modulus();
        equalsApprox(pv1.Velocity.Modulus(), modulus, 1e-6 * modulus,
            "testStar() barycentric velocity modulus");

        // Compare heliocentric PV to separately computed value.
        pv1 = star.Heliocentric(tt);
        pv2 = star.Barycentric(tt) - sun.Barycentric(tt);
        testSeparationAngle(pv1.Position, pv2.Position, Angle.DmsToRad(0, 0, .001),
            "testStar() heliocentric position direction");
        modulus = pv2.Position.Modulus();
        equalsApprox(pv1.Position.Modulus(), modulus, 1e-6 * modulus,
            "testStar() heliocentric position modulus");
        testSeparationAngle(pv1.Velocity, pv2.Velocity, Angle.DmsToRad(0, 1, 0),
            "testStar() heliocentric velocity direction");
        modulus = pv2.Velocity.Modulus();
        equalsApprox(pv1.Velocity.Modulus(), modulus, 1e-6 * modulus,
            "testStar() heliocentric velocity magnitude");

        // geocentric geometric direction, and corresponding vector from Almanac
        p1 = star.GeocentricGeometric(tt).Position;
        testSeparationAngle(p1, new Vector(-.374015228, -.312472937, -.873197156),
            Angle.DmsToRad(0, 0, .001), "testStar() geocentric geometric");

        // Verify geocentric astrometric is the same as geocentric geometric.
        equalsApprox(star.GeocentricAstrometric(tt), p1, 0.0,
            "testStar() geocentric astrometric");

        // geocentric apparent direction
        testSeparationAngle(star.GeocentricApparent(tt).Unit(),
            new Vector(-.374096719, -.312473742, -.873161958),
            Angle.DmsToRad(0, 0, .001), "testStar geocentric apparent");

        // Topocenter position and velocity in the GCRS. Assume delta T = 0.
        double lat = Angle.DegreesToRadians(-40.0);
        double lon = Angle.DegreesToRadians(50.0);
        Topocenter observer = new Topocenter(lat, lon, 0.0);
        double deltaPsi, deltaEpsilon;
        RMatrix.NutationAngles80(tt, out deltaPsi, out deltaEpsilon);
        RMatrix crsToTrs = RMatrix.GcrsToItrs(
            RMatrix.GcrsToTrue(RMatrix.Precess76(tt), RMatrix.MeanObliq80(tt),
            deltaPsi, deltaEpsilon), Angle.Gast94(tt), new RMatrix());
        PVVector observerPV = observer.ToPVVector(crsToTrs.Transpose());

        // Topocentric geometric and astrometric place should be identical, and
        // almost identical to geocentric geometric.
        Vector topocentricGeometric = star.TopocentricGeometric(tt, observerPV).Position;
        Vector topocentricAstrometric = star.TopocentricAstrometric(tt, observerPV.Position);
        equalsApprox(topocentricAstrometric, topocentricGeometric, 0.0,
            "testStar() topocentric geometric differs from astrometric");
        testSeparationAngle(topocentricGeometric, p1, Angle.DmsToRad(0, 0, .001),
            "testStar() topocentric geometric not close enough to geocentric geometric");

        // Topocentric apparent direction vs. MICA 2.0. Star is alpha Centauri.
        // SIMBAD catalog data.
        star = new Star(JulianDate.J2000Base, Angle.HmsToHours(14, 39, 36.49400),
            Angle.DmsToDeg(-60, 50, 02.3737), 754.81, -3679.25, 473.67, -24.6,
            sofaEph, "alpha Centauri");
        RMatrix trsToHor = RMatrix.ItrsToHor(lat, lon, 0.0, 0.0);
        Vector topocentricApparent = trsToHor * crsToTrs * star.TopocentricApparent(
            tt, observerPV);
        topocentricApparent.Unit().EqualsApprox(new Vector(new Spherical(
            Angle.Complement(Angle.DmsToRad(140, 52, 50.0)),
            Angle.Complement(Angle.DmsToRad(46, 28, 02.8)))),
            Angle.DmsToRad(0, 0, .2));
    }
#endif  // end self-test code

}       // end class Star
}       // end namespace HirosePS.SofaJpl
