// TimeFields.cs
// A class to break a Julian date into separate fields: year, month, etc.
// Copyright 2018 by Paul S. Hirose.

// revision history
#if false
2007-10-08 Rewrote to call my .NET Framework SofaJplNet DLL instead of Win32 sofajpl.dll.

2007-12-05 Completed self-test code.
2007-12-06 Self-test minor rewrite to take advantage of new Duration relational operators.
2008-05-15 JD to calendar date conversion now done entirely in-house.
2008-07-12 Added constructors which take date and time components.
2008-11-02 Corrected comments on date range.
2014-11-06 Rewrote dayNumToDate() with a new algorithm. Methods and properties made virtual.
Test OK.
2015-06-26 De-virtualized the class.

2018-04-19 New self test for rollover at year end. Test OK.
#endif


using System;
using System.Collections.Generic;
using System.Text;


namespace HirosePS.SofaJpl
{
    /// <summary>
    /// A class to convert a JulianDate or Utc object into Julian or Gregorian calendar date
    /// and time at a specified precision.
    /// </summary>
    /// <remarks>
    /// <para>Resolution has to be specified when constructing an object because it
    /// can affect all the components. For example, if the time is 23:59:59.9 on December
    /// 31, the date will round up to January 1 if one-second resolution is requested.
    /// </para>
    /// <para>This class correctly correctly handles UTC step adjustments (negative,
    /// positive, and fractional second), including times within a step adjustment.</para>
    /// <para>Year, month, and day are stored in Int32 variables. Time of day is stored in
    /// an instance of the Sexagesimal class.</para>
    /// <para>Any time in the years -5,880,000 to 5,870,000 (approx.), Gregorian or Julian
    /// calendar, is processed to full accuracy.</para>
    /// <para>Instances of this class are immutable.</para>
    /// </remarks>
    public class TimeFields
    {
        // number of days in a leap year cycle: 4 Julian years or 400 Gregorian years
        const int julianCycleLength = 4 * 365 + 1;
        const int gregorianCycleLength = 400 * 365 + 100 - 3;

        private bool julian;    // true if Julian calendar
        private int year, month, day;
        private Sexagesimal hms;      // always positive

        // .5 = 2 hour resolution, 1.0 = 1 hour resolution, 60.0 = 1 minute,
        // 360.0 = 10 second, 3600.0 = 1 second, 36000.0 = .1 sec resolution, etc.
        private double resolution;



        
        // constructors


        /// <summary>
        /// Initialize from date (Julian or Gregorian calendar) and time (in a Sexagesimal).
        /// </summary>
        /// <param name="year">year</param>
        /// <param name="month">month (1 = Jan)</param>
        /// <param name="day">day of month</param>
        /// <param name="hms">time of day</param>
        /// <param name="julian">true if date is in Julian calendar</param>
        /// <exception cref="ArgumentOutOfRangeException"> if illegal month.</exception>
        public TimeFields(int year, int month, int day, Sexagesimal hms, bool julian)
        {
            if (month < 1 || month > 12)
                throw new ArgumentOutOfRangeException("Illegal month.");
            this.year = year;
            this.month = month;
            this.day = day;
            this.hms = hms;
            this.julian = julian;
            this.resolution = hms.Resolution;
        }


        /// <summary>
        /// Initialize from date (Gregorian calendar) and time (in a Sexagesimal).
        /// </summary>
        /// <param name="year">year</param>
        /// <param name="month">month (1 = Jan)</param>
        /// <param name="day">day of month</param>
        /// <param name="hms">time of day</param>
        /// <exception cref="ArgumentOutOfRangeException"> if illegal month.</exception>
        public TimeFields(int year, int month, int day, Sexagesimal hms)
            : this(year, month, day, hms, false)
        { }


        /// <summary>
        /// Initialize from date (Julian or Gregorian calendar) and time (individual components).
        /// </summary>
        /// <param name="year">year</param>
        /// <param name="month">month (1 = Jan)</param>
        /// <param name="day">day of month</param>
        /// <param name="hour">hours</param>
        /// <param name="minute">minutes</param>
        /// <param name="second">seconds</param>
        /// <param name="resolution">desired time roundoff resolution: 60.0 = 1 minute,
        /// 3600.0 = 1 second, 36000.0 = .1 second, etc.</param>
        /// <param name="julian">true if date is in Julian calendar</param>
        /// <exception cref="ArgumentOutOfRangeException"> if illegal month.</exception>
        public TimeFields(int year, int month, int day, int hour, int minute, double second,
            double resolution, bool julian)
            : this(year, month, day,
            new Sexagesimal(Angle.HmsToHours(hour, minute, second), resolution), julian)
        { }


        /// <summary>
        /// Initialize from date (Gregorian calendar) and time (individual components).
        /// </summary>
        /// <param name="year">year</param>
        /// <param name="month">month (1 = Jan)</param>
        /// <param name="day">day of month</param>
        /// <param name="hour">hours</param>
        /// <param name="minute">minutes</param>
        /// <param name="second">seconds</param>
        /// <param name="resolution">Resolution units per hour, e.g., 60 = 1 minute, 36000 = .1
        /// second. The object is rounded to the nearest resolution unit. Negative or zero is
        /// illegal.</param>
        /// <exception cref="ArgumentOutOfRangeException"> if illegal month.</exception>
        public TimeFields(int year, int month, int day, int hour, int minute, double second,
            double resolution)
            : this(year, month, day, hour, minute, second, resolution, false)
        { }


        /// <summary>
        /// Initialize from a JulianDate. Date in Gregorian or Julian calendar, with provision for
        /// seconds to exceed 60 within a UTC step adjustment.
        /// </summary>
        /// <param name="jd">JulianDate instance</param>
        /// <param name="resolution">Resolution units per hour, e.g., 60 = 1 minute, 36000 = .1
        /// second. The object is rounded to the nearest resolution unit. Negative or zero is
        /// illegal.</param>
        /// <param name="julianFlag">If true, output is in the Julian calendar.</param>
        /// <param name="stepSeconds">quantity to add to seconds (ignored if negative)</param>
        /// <remarks>
        /// <para>Any time in the years -5,880,000 to +5,870,000 (approx.), Gregorian or
        /// Julian calendar, is processed to full accuracy.</para>
        /// <para>To allow seconds to exceed 60 within a UTC step adjustment,
        /// <paramref name="stepSeconds"/> adds directly to seconds without any "roll over" when
        /// they exceed 60. A non integer value is allowed for fractional second step adjustments.
        /// A negative value is ignored.</para>
        /// </remarks>
        public TimeFields(JulianDate jd, double resolution, bool julianFlag,
            double stepSeconds)
        {
            this.resolution = resolution;
            this.julian = julianFlag;

            // time of day
            this.hms = new Sexagesimal(jd.TimePart * 24.0, resolution, 24, stepSeconds);
            
            // increment date if hms rounded up to 0
            jd += new Duration(hms.Revolutions);

            // Compute calendar date.
            jdToDate(jd.DatePart, julianFlag, out this.year, out this.month, out this.day);
        }


        /// <summary>
        /// Initialize from a JulianDate. Date in Gregorian or Julian calendar.
        /// </summary>
        /// <param name="jd">JulianDate instance</param>
        /// <param name="resolution">Resolution units per hour, e.g., 60 = 1 minute, 36000 = .1
        /// second. The object is rounded to the nearest resolution unit. Negative or zero is
        /// illegal.</param>
        /// <param name="julianFlag">If true, output is in the Julian calendar.</param>
        /// <remarks>Any time in the years -5,880,000 to +5,870,000 (approx.), Gregorian or
        /// Julian calendar, is processed to full accuracy.
        /// </remarks>
        public TimeFields(JulianDate jd, double resolution, bool julianFlag)
            : this(jd, resolution, julianFlag, 0.0)
        { }


        /// <summary>
        /// Initialize from a JulianDate. Date in Gregorian calendar.
        /// </summary>
        /// <param name="jd">JulianDate instance</param>
        /// <param name="resolution">Resolution units per hour, e.g., 60 = 1 minute, 36000 = .1
        /// second. The object is rounded to the nearest resolution unit. Negative or zero is
        /// illegal.</param>
        /// <remarks>Any time in the years -5,880,000 to +5,870,000 (approx.), Gregorian or
        /// Julian calendar, is processed to full accuracy.
        /// </remarks>
        public TimeFields(JulianDate jd, double resolution)
            : this(jd, resolution, false, 0.0)
        { }




        // Properties

        /// <summary>
        /// Indicates whether or not the date is in the Julian calendar.
        /// </summary>
        /// <value>True if Julian calendar.</value>
        public bool Julian { get { return this.julian; } }

        /// <summary>
        /// Get the year.
        /// </summary>
        /// <value>year: -1 = 2 BC, 0 = 1 BC, 1 = 1 AD, etc.
        /// </value>
        public int Year { get { return year; } }

        /// <summary>
        /// Get the month.
        /// </summary>
        /// <value>month: 1 = January, etc.</value>
        public int Month { get { return month; } }

        /// <summary>
        /// Get the day of the month.
        /// </summary>
        /// <value>day of month</value>
        public int Day { get { return day; } }

        /// <summary>
        /// The hours, minutes, seconds, and fractional seconds, rounded to the precision
        /// specified in the constructor.
        /// </summary>
        /// <value>Time of day, in a new Sexagesimal instance. Always positive.</value>
        public Sexagesimal Hms { get { return hms; } }

        /// <summary>
        /// Get the time resolution to which this instance was rounded.
        /// </summary>
        /// <value>The resolution of this instance is one hour divided by the returned
        /// value. E.g., .5 = 2 hours, 60.0 = 1 minute, 360.0 = 10 seconds, etc.
        /// </value>
        public double Resolution { get { return this.resolution; } }


        // Methods

        /// <summary>
        /// Convert the value of this instance to a string with appropriate precision.
        /// </summary>
        /// <returns>String in ISO 8601 format</returns>
        public override string ToString()
        {
            return string.Format("{0:d4}-{1:d2}-{2:d2}T{3:02g:}",
                this.year, this.month, this.day, hms);
        }


        /// <summary>
        /// Convert this instance to a JulianDate.
        /// </summary>
        /// <returns>new JulianDate instance</returns>
        public JulianDate ToJulianDate()
        {
            return new JulianDate(this.year, this.month, this.day,
                this.hms.ToDouble() / 24.0, this.julian);
        }


#if DEBUG
        /// <summary>
        /// Convert Julian date to a date in the Julian or Gregorian calendar.
        /// </summary>
        /// <param name="jd">Julian date</param>
        /// <param name="julianCal">true to obtain date in the Julian calendar</param>
        /// <param name="y">year (0 = 1 BC, -1 = 2 BC, etc.)</param>
        /// <param name="m">month (1 - 12)</param>
        /// <param name="d">day of month</param>
#endif
        private static void jdToDate(double jd, bool julianCal, out int y, out int m, out int d)
        {
	        // convert jd to n, a day number which equals -1 on 1 BC Dec 30,
	        // 0 on 1 BC Dec 31, +1 on 1 AD Jan 1, etc., of the applicable calendar
            int n = (int) (jd - (julianCal ? 1721422.5 : 1721424.5));

            dayNumToDate(n, julianCal, out y, out m, out d);
        }


#if DEBUG
        /// <summary>
        /// Get number of days in the given year, Gregorian or Julian calendar.
        /// </summary>
        /// <param name="y">year (0 = 1 BC, -1 = 2 BC, etc.)</param>
        /// <param name="julianFlag">true if Julian calendar</param>
        /// <returns>365 or 366</returns>
#endif
        private static int daysInThisYear(int y, bool julianFlag) {
            return (JulianDate.IsLeapYear(y, julianFlag) == 0) ? 365 : 366;
        }


#if DEBUG
        /// <summary>
        /// Convert a day number in the Julian or Gregorian calendar to a date.
        /// </summary>
        /// <param name="n">day number (= 1 on January 1, 1 AD)</param>
        /// <param name="julianFlag">true if day number is in the Julian calendar</param>
        /// <param name="y">year (0 = 1 BC, -1 = 2 BC, etc.)</param>
        /// <param name="m">month, 1 - 12</param>
        /// <param name="d">day of month</param>
        /// <remarks>The algorithm itself has unlimited range, BC and AD. In this implementation the
        /// time span is limited by integer overflow, and exceeds the years -5e6 to +5e6.</remarks>
#endif
        private static void dayNumToDate(int n, bool julianFlag, out int y, out int m, out int d) {

            int[,] month_len = {
	            {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},    // normal year
	            {0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}};   // leap year
            int cycles, remainder, leap, daysInMonth;

            // Estimate the year by dividing n by the mean length of a year: 1461/4 or 146097/400.
            // The algorithm avoids an initial multiplication by 4 or 400, which would reduce the
            // usable range of n due to overflow.

            if (julianFlag) {
                cycles = n / julianCycleLength;
                remainder = n % julianCycleLength;
                y = cycles * 4 + remainder * 4 / julianCycleLength;
            }
            else {
                cycles = n / gregorianCycleLength;
                remainder = n % gregorianCycleLength;
                y = cycles * 400 + remainder * 400 / gregorianCycleLength;
            }

            // For AD dates, y is one less than the correct year. For BC dates it depends on the
            // behavior of integer division of a negative number. If division truncates down (-3/2
            // = -2), y is one less than the correct year. If it truncates up (-3/2 = -1), y is the
            // correct year. (Walk through the computation of y when n = +180 and n = -180 to
            // satisfy yourself on this.) The next part corrects for these issues.

            if (n > 0)
                ++y;
            else
                y = y - (-1 / 2);

            // One more correction to y. If n is the first or last day of a year, y can be wrong by
            // one year. E.g., at +4 Dec 31 it's too high, and at +204 Jan 1 it's too low.
            // (Both dates Gregorian calendar.) Either condition is detected by computing day of
            // year (d) and verifying it's within the legal range. If necessary, adjust y and
            // recompute day of year.

            while (true) {
                d = n - JulianDate.DateToDayNumber(y - 1, 12, 31, julianFlag);
                if (d < 1) {
                    --y;
                    continue;
                }
                if (d > daysInThisYear(y, julianFlag)) {
                    ++y;
                    continue;
                }
                else
                    break;
            }

            // 1 if leap year, 0 otherwise
            leap = JulianDate.IsLeapYear(y, julianFlag);

            // Determine month. At loop termination, whatever is left in d is the day of month.
            for (m = 1; m < 13; ++m) {
                daysInMonth = month_len[leap, m];
                if (d <= daysInMonth)
                    break;
                d -= daysInMonth;
            }
            if (m == 13)
                throw new Exception("Algorithm failed at month loop.");
        }



        #region self-test code
#if DEBUG

        /// <summary>
        /// Self-test method. Runs silently if all tests pass.
        /// </summary>
        public static void Test()
        {
            JulianDate jd1;
            TimeFields tf1, tf2;
            int y, m, d, yc, mc, dc, n;
            int[,] monthLength = {
                { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
                { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 } };

            // master constructor that takes JulianDate; step seconds = 1.0; Gregorian
            // This also checks the properties.
            jd1 = new JulianDate(2005, 12, 31, .999999);
            tf1 = new TimeFields(jd1, 36000.0, false, 1.0);
            assert(tf1.Year == 2005 && tf1.Month == 12 && tf1.Day == 31
                && (int)tf1.Hms.Hours == 23 && (int)tf1.Hms.Minutes == 59
                && Math.Abs(tf1.Hms.Seconds - 60.9) < .01 && !tf1.Julian
                && tf1.Resolution == 36000.0);

            // constructor with default Julian flag and step seconds
            tf1 = new TimeFields(jd1, 36000.0);
            assert(tf1.Year == 2005 && tf1.Month == 12 && tf1.Day == 31
                && (int)tf1.Hms.Hours == 23 && (int)tf1.Hms.Minutes == 59
                && Math.Abs(tf1.Hms.Seconds - 59.9) < .01 && !tf1.Julian);

            // constructor with Julian flag true and default step seconds
            tf2 = new TimeFields(jd1, 36000.0, true);
            assert(tf2.Year == 2005 && tf2.Month == 12 && tf2.Day == 18
                && (int)tf2.Hms.Hours == 23 && (int)tf2.Hms.Minutes == 59
                && Math.Abs(tf2.Hms.Seconds - 59.9) < .01 && tf2.Julian);

            // constructor that takes date, time (Sexagesimal), calendar
            tf1 = new TimeFields(2008, 7, 12, new Sexagesimal(10 + (36 + 41.67 / 60.0) / 60.0,
                36000.0), true);
            assert(tf1.Year == 2008 && tf1.Month == 7 && tf1.Day == 12 &&
                (int)tf1.Hms.Hours == 10 && (int)tf1.Hms.Minutes == 36 &&
                Math.Abs(tf1.Hms.Seconds - 41.7) < .01 && tf1.Julian &&
                tf1.Resolution == 36000.0);

            // constructor that's same as above, but assumes Gregorian calendar
            tf1 = new TimeFields(2008, 7, 12, new Sexagesimal(10 + (36 + 41.67 / 60.0) / 60.0,
                36000.0));
            assert(tf1.Year == 2008 && tf1.Month == 7 && tf1.Day == 12 &&
                (int)tf1.Hms.Hours == 10 && (int)tf1.Hms.Minutes == 36 &&
                Math.Abs(tf1.Hms.Seconds - 41.7) < .01 && !tf1.Julian &&
                tf1.Resolution == 36000.0);

            // constructor that takes date, time (components), calendar
            tf1 = new TimeFields(2008, 7, 12, 10, 36, 41.48, 3600.0, true);
            assert(tf1.Year == 2008 && tf1.Month == 7 && tf1.Day == 12 &&
                (int)tf1.Hms.Hours == 10 && (int)tf1.Hms.Minutes == 36 &&
                Math.Abs(tf1.Hms.Seconds - 41.0) < .1 && tf1.Julian &&
                tf1.Resolution == 3600.0);

            // constructor that's same as above, but assumes Gregorian calendar
            tf1 = new TimeFields(2008, 7, 12, 10, 36, 41.48, 3600.0);
            assert(tf1.Year == 2008 && tf1.Month == 7 && tf1.Day == 12 &&
                (int)tf1.Hms.Hours == 10 && (int)tf1.Hms.Minutes == 36 &&
                Math.Abs(tf1.Hms.Seconds - 41.0) < .1 && !tf1.Julian &&
                tf1.Resolution == 3600.0);

            // ToString()
            assert(tf2.ToString().Equals("2005-12-18T23:59:59.9"));

            // ToJulianDate()
            tf1 = new TimeFields(jd1, 36000.0);
            assert(tf1.ToJulianDate() - jd1 < (new Duration(.0000001)));

            // Round trip, date to day number to date, Gregorian calendar.
            for (y = -400; y <= 401; ++y) {
                for (m = 1; m <= 12; ++m) {
                    // first day of month
                    d = 1;
                    n = JulianDate.DateToDayNumber(y, m, d, false);
                    dayNumToDate(n, false, out yc, out mc, out dc);
                    if (yc != y || mc != m || dc != d)
                        throw new Exception(String.Format("Gregorian at {0}-{1}-{2}", y, m, d));
                    // last day of month
                    d = monthLength[JulianDate.IsLeapYear(y, false), m];
                    n = JulianDate.DateToDayNumber(y, m, d, false);
                    dayNumToDate(n, false, out yc, out mc, out dc);
                    if (yc != y || mc != m || dc != d)
                        throw new Exception(String.Format("Gregorian at {0}-{1}-{2}", y, m, d));
                }
            }

            // Round trip, date to day number to date, Julian calendar.
            for (y = -4; y <= 5; ++y) {
                for (m = 1; m <= 12; ++m) {
                    // first day of month
                    d = 1;
                    n = JulianDate.DateToDayNumber(y, m, d, true);
                    dayNumToDate(n, true, out yc, out mc, out dc);
                    if (yc != y || mc != m || dc != d)
                        throw new Exception(String.Format("Julian at {0}-{1}-{2}", y, m, d));
                    // last day of month
                    d = monthLength[JulianDate.IsLeapYear(y, true), m];
                    n = JulianDate.DateToDayNumber(y, m, d, true);
                    dayNumToDate(n, true, out yc, out mc, out dc);
                    if (yc != y || mc != m || dc != d)
                        throw new Exception(String.Format("Julian at {0}-{1}-{2}", y, m, d));
                }
            }


            // Roll over to next year if late in the last second.
            jd1 = new JulianDate(2017, 12, 31, 23, 59, 59.9);
            tf1 = new TimeFields(jd1, 3600);
            if (tf1.Year != 2018 || tf1.Month != 1 || tf1.Day != 1
                || tf1.Hms.Hours != 0 || tf1.Hms.Minutes != 0 || tf1.Hms.Seconds != 0.0)
                throw new Exception("Failed to roll over to 2018");

            // Roll over to next year if late in the leap second.
            jd1 = new JulianDate(2016, 12, 31, 23, 59, 59.9);
            tf1 = new TimeFields(jd1, 3600, false, 1.0);
            if (tf1.Year != 2017 || tf1.Month != 1 || tf1.Day != 1
                || tf1.Hms.Hours != 0 || tf1.Hms.Minutes != 0 || tf1.Hms.Seconds != 0.0)
                throw new Exception("Failed to roll over in leap second");
        }


        private static void assert(bool assertion)
        {
            if (!assertion)
                throw new Exception("Assertion failed.");
        }


#endif
        #endregion  // end self-test code
    }
}
