// Topocenter.cs
// Topocenter class.
// Copyright 2018 by Paul S. Hirose.

// revision history
#if false
2007-11-14 Added properties to get x, y, z.
2007-11-26 Test code completed. It runs OK.
2008-05-04 Method to obtain the instance's GCRS position and velocity.
2008-05-13 The lat/lon helper method no longer relies on a C++ routine.
2012-04-01 Eliminated latitude overshoot code from constructor, since a latitude slightly
in excess of exactly pi/2 is harmless. Changed the lat/lon helper method so it returns
a value instead of working via side effect. Added GcrsPV() as an alias to ToPVVector().
Test OK.
2012-06-18 Changing distance unit to km. Geodetic coordinates also stored in the instance.
Added properties to get them.
2012-06-19 Added ellipsoid parameters to the constructor that takes a Vector.
2012-07-11 Changed my mind about ellipsoid radius constructor parameter. Unit of measure
will be meters, not km.
2012-07-14 Changed my mind back about distance unit. It will be km everwhere.
2012-08-18 Finished the self-test. Needs a final check.
2012-08-29 Updated self test to work with my RMatrix changes.
2012-10-12 The constructors now take height above ellipsoid in meters, and angles in
degrees, since these quantities usually come from human input. However, the properties
will output the internal units of km and radians.
2012-10-13 Updated code to match revised Iso6709 class. Can't think of anything else to
do here. Renamed class to Topocenter. Self test OK.
2012-10-25 Constructor units changed to radians and km. Self test OK.
2012-12-22 Added constructor XML comment remark that height is km, not m.

2015-05-10 I now compile the SOFA library from their C source, which has a different capitalization
style compared to the Fortran code, requiring corresponding changes here.
Properties and methods are virtual now.
2015-05-19 Eliminated some Vector copy operations. This minor efficiency improvement is safe since
Vectors are immutable. Self test OK.
2015-06-26 De-virtualized the class.
2015-08-30 Height above ellipsoid changed to meters in constructors and property.

2018-08-12 Improvements to the XML comments.
#endif


using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;


namespace HirosePS.SofaJpl
{
    /// <summary>
    /// Position of an observer on or near the Earth's surface.
    /// </summary>
    /// <remarks>
    /// <para>After an instance is initialized from geodetic coordinates, it supplies the GCRS
    /// position and velocity of the observer for the computation of topocentric coordinates.</para>
    /// <para>Instances of this class are immutable.</para>
    /// </remarks>
    public class Topocenter
    {
        // GRS80 ellipsoid radius and flattening
        private const double grs80Radius = 6378.1370;
        private const double grs80FInverse = 298.257222101;

        // Earth nominal mean angular velocity, radians / day (2012 Almanac p. K6)
        private const double omegaEarth = 7.292115e-5 * 86400.0;

        // Internally, the point is represented by a Vector in the ITRS, and by
        // its geodetic latitude, longitude, and height.
        private Vector _vector;         // kilometers
        private double _lat, _lon;      // radians, positive north and east
        private double _height;         // kilometers above ellipsoid

        // size and shape of the ellipsoid
        private double _a;              // equatorial radius, km
        private double _FInverse;       // ca. 298




        // Constructors

        /// <summary>
        /// Copy from another instance.
        /// </summary>
        /// <param name="g">the instance to copy</param>
        public Topocenter(Topocenter g)
        {
            this._vector = g._vector;
            this._lat = g._lat;
            this._lon = g._lon;
            this._height = g._height;
            this._a = g._a;
            this._FInverse = g._FInverse;
        }


        /// <summary>
        /// Initialize from a Vector in the ITRS, and ellipsoid dimensions.
        /// </summary>
        /// <param name="v">ITRS Vector from Earth's center (km)</param>
        /// <param name="ellipsoidRadius">ellipsoid equatorial radius (km)</param>
        /// <param name="fInverse">ellipsoid 1/f (ca. 298)</param>
        public Topocenter(Vector v, double ellipsoidRadius, double fInverse)
        {
            this._vector = v;

            // geodetic coordinates.
            int status;
            this._a = ellipsoidRadius;
            this._FInverse = fInverse;
            status = Sofa.Gc2gde(ellipsoidRadius, 1.0 / fInverse, v.ToArray(),
                out this._lon, out this._lat, out this._height);
            if (status != 0)
                throw new Exception("Illegal ellipsoid radius or flattening.");
        }

        /// <summary>
        /// Initialize from a Vector in the ITRS. The GRS80 ellipsoid is assumed.
        /// </summary>
        /// <param name="v">ITRS Vector from Earth's center (km)</param>
        public Topocenter(Vector v)
            : this(v, grs80Radius, grs80FInverse)
        { }


        /// <summary>
        /// Initialize from latitude, longitude, height, equatorial radius, and
        /// flattening.
        /// </summary>
        /// <param name="lat">latitude, positive north (radians)</param>
        /// <param name="lon">longitude, positive east (radians)</param>
        /// <param name="height">height above ellipsoid (meters)</param>
        /// <param name="ellipsoidRadius">ellipsoid equatorial radius (km)</param>
        /// <param name="fInverse">ellipsoid 1/f (ca. 298)</param>
        public Topocenter(double lat, double lon, double height,
            double ellipsoidRadius, double fInverse)
        {
            this._lat = lat;
            this._lon = lon;
            this._height = height / 1000.0;
            this._a = ellipsoidRadius;
            this._FInverse = fInverse;

            this._vector = latLonToVector(this._lat, this._lon, this._height,
                this._a, this._FInverse);
        }

        /// <summary>
        /// Initialize from geodetic latitude, longitude, and height. The GRS80 ellipsoid
        /// is assumed.
        /// </summary>
        /// <param name="lat">latitude, positive north (radians)</param>
        /// <param name="lon">longitude, positive east (radians)</param>
        /// <param name="height">height above ellipsoid (meters)</param>
        public Topocenter(double lat, double lon, double height)
            : this(lat, lon, height, grs80Radius, grs80FInverse)
        { }

#if false
        /// <summary>
        /// Initialize from an Iso6709 object and the ellipsoid dimensions.
        /// </summary>
        /// <param name="iso6709Obj">Iso6709 object</param>
        /// <param name="ellipsoidRadius">ellipsoid equatorial radius (km)</param>
        /// <param name="fInverse">ellipsoid 1/f (ca. 298)</param>
        public Topocenter(Iso6709 iso6709Obj, double ellipsoidRadius,
            double fInverse)
            : this(Angle.DegreesToRadians(iso6709Obj.Latitude),
            Angle.DegreesToRadians(iso6709Obj.Longitude), iso6709Obj.Altitude,
            ellipsoidRadius, fInverse)
        { }

        /// <summary>
        /// Initialize from an Iso6709 object. The GRS80 ellipsoid is assumed.
        /// </summary>
        /// <param name="iso6709Obj">Iso6709 instance</param>
        public Topocenter(Iso6709 iso6709Obj)
            : this(iso6709Obj, grs80Radius, grs80FInverse)
        { }
#endif


#if DEBUG
        /// <summary>
        /// Convert geodetic coordinates to Vector.
        /// </summary>
        /// <param name="lat">latitude, positive north, radians</param>
        /// <param name="lon">longitude, positive east, radians</param>
        /// <param name="h">height above ellipsoid, km</param>
        /// <param name="ellipsoidRadius">ellipsoid equatorial radius, km</param>
        /// <param name="fInverse">ellipsoid 1/f (ca. 298)</param>
        /// <returns>Vector from geocenter (km)</returns>
        /// <exception cref="ArgumentException">if <paramref name="fInverse"/> &lt;= 1
        /// </exception>
#endif
        private static Vector latLonToVector(double lat, double lon, double h,
            double ellipsoidRadius, double fInverse)
        {
            if (fInverse <= 1.0)
                throw new ArgumentException("Flattening must be greater than 1.");
            double cosphi = Math.Cos(lat);
            double sinphi = Math.Sin(lat);
            double omf2 = (1.0 - 1.0 / fInverse);
            omf2 *= omf2;
            double C = 1.0 / Math.Sqrt(cosphi*cosphi + omf2*sinphi*sinphi);
            double S = C * omf2;

            double x = (ellipsoidRadius * C + h) * cosphi * Math.Cos(lon);
            double y = (ellipsoidRadius * C + h) * cosphi * Math.Sin(lon);
            double z = (ellipsoidRadius * S + h) * sinphi;
            return new Vector(x, y, z);
        }




        // properties

        /// <summary>
        /// Get the x coordinate (km).
        /// </summary>
        public double X { get { return this._vector.X; } }

        /// <summary>
        /// Get the y coordinate (km).
        /// </summary>
        public double Y { get { return this._vector.Y; } }

        /// <summary>
        /// Get the z coordinate (km).
        /// </summary>
        public double Z { get { return this._vector.Z; } }

        /// <summary>
        /// Get north latitude (radians).
        /// </summary>
        public double Lat { get { return this._lat; } }

        /// <summary>
        /// Get east longitude (radians).
        /// </summary>
        public double Lon { get { return this._lon; } }

        /// <summary>
        /// Get height above ellipsoid (meters).
        /// </summary>
        public double Height { get { return _height * 1000.0; } }

        /// <summary>
        /// Get ellipsoid equatorial radius (km).
        /// </summary>
        public double EllipsoidRadius { get { return this._a; } }

        /// <summary>
        /// Get ellipsoid 1/f (ca. 298)
        /// </summary>
        public double FInverse { get { return this._FInverse; } }



        // methods

        /// <summary>
        /// Put the x, y, and z components into a String.
        /// </summary>
        /// <returns>xyz in f3 format (km)</returns>
        public override string ToString()
        {
            return String.Format("{0:f3} {1:f3} {2:f3}",
                _vector.X, _vector.Y, _vector.Z);
        }

        /// <summary>
        /// Convert this instance to a Vector (in km) from the center of Earth.
        /// </summary>
        /// <returns>new geocentric Vector to the geographic position (km)</returns>
        public Vector ToVector()
        {
            return this._vector;
        }


        /// <summary>
        /// Get the GCRS position and diurnal velocity of the geographic position, given
        /// the ITRS to GCRS rotation matrix.
        /// </summary>
        /// <remarks>The velocity (but not the position) is slightly non-rigorous
        /// because it's calculated as if Earth rotates about the ITRS pole.</remarks>
        /// <param name="itrsToGcrs">rotation matrix, ITRS to GCRS</param>
        /// <returns>position (km) and velocity (km/day) in the GCRS</returns>
        public PVVector ToPVVector(RMatrix itrsToGcrs)
        {
            // Topocenter position and velocity in an inertial frame coincident with the
            // instantaneous orientation of the ITRS.
            PVVector pv = new PVVector(new double[,] {
                {this.X, this.Y, this.Z},
                {-this.Y * omegaEarth, this.X * omegaEarth, 0.0}});

            // rotate from ITRS to GCRS
            return itrsToGcrs * pv;
        }

        /// <summary>
        /// Get the GCRS position and diurnal velocity of the geographic position, given
        /// the ITRS to GCRS rotation matrix. Alternate name for ToPVVector().
        /// </summary>
        /// <param name="itrsToGcrs">rotation matrix, ITRS to GCRS</param>
        /// <returns>position (km) and velocity (km/day) in the GCRS</returns>
        public PVVector ToGcrsPV(RMatrix itrsToGcrs)
        {
            return ToPVVector(itrsToGcrs);
        }




        #region self-test code
#if DEBUG

        /// <summary>
        /// Self-test method.
        /// </summary>
        public static void Test()
        {
            // geodetic coords
            double lat = Angle.DegreesToRadians(34.0);
            double lon = Angle.DegreesToRadians(-118.0);
            double h = 200.0;
            double a = 6.4e3;               // equatorial radius, km
            double flattening = 300.0;      // 1/f

            // GRS80 a and f from SOFA routine
            int status;
            double sofaRadius, sofaFlattening;
            status = Sofa.Eform(Sofa.Ellipsoid.GRS80, out sofaRadius, out sofaFlattening);
            sofaRadius /= 1000.0;   // convert m to km
            sofaFlattening = 1.0 / sofaFlattening;

#if false
            // ISO 6709 object corresponding to the lat and lon above
            Iso6709 parsed6709 = new Iso6709("+34-118+200/");
#endif

            // Corresponding xyz. Will be computed independently by the SOFA routine.
            double[] xyz = new double[3];


            // construct from lat/lon & ellipsoid dimensions
            Topocenter gp1 = new Topocenter(lat, lon, h, a, flattening);
            status = Sofa.Gd2gce(a, 1.0 / flattening, lon, lat, h / 1000.0, xyz);
            gp1.xyzEqualsApprox(xyz, 1e-6, "construct from lat/lon & ellipsoid");

            // Test the properties.
            equalsApprox(gp1.X, xyz[0], 1e-6, "X property, specified ellipsoid");
            equalsApprox(gp1.Y, xyz[1], 1e-6, "Y property, specified ellipsoid");
            equalsApprox(gp1.Z, xyz[2], 1e-6, "Z property, specified ellipsoid");
            equalsApprox(gp1.Lat, lat, 0.0, "Lat property, specified ellipsoid");
            equalsApprox(gp1.Lon, lon, 0.0, "Lon property, specified ellipsoid");
            equalsApprox(gp1.Height, h, .001, "Height property, specified ellipsoid");
            equalsApprox(gp1.EllipsoidRadius, a, 0.0,
                "EllipsoidRadius property, specified ellipsoid");
            equalsApprox(gp1.FInverse, flattening, 0.0,
                "FInverse property, specified ellipsoid");

            // construct from lat/lon & default (GRS80) ellipsoid
            gp1 = new Topocenter(lat, lon, h);
            status = Sofa.Gd2gc(Sofa.Ellipsoid.GRS80, lon, lat, h, xyz);
            Sofa.Sxp(.001, xyz, xyz);   // convert m to km
            gp1.xyzEqualsApprox(xyz, 1e-6, "construct from lat/lon & default ellipsoid");
            equalsApprox(gp1.X, xyz[0], 1e-6, "X property, default ellipsoid");
            equalsApprox(gp1.Y, xyz[1], 1e-6, "Y property, default ellipsoid");
            equalsApprox(gp1.Z, xyz[2], 1e-6, "Z property, default ellipsoid");
            equalsApprox(gp1.Lat, lat, 0.0, "Lat property, default ellipsoid");
            equalsApprox(gp1.Lon, lon, 0.0, "Lon property, default ellipsoid");
            equalsApprox(gp1.Height, h, .001, "Height property, default ellipsoid");
            equalsApprox(gp1.EllipsoidRadius, sofaRadius, 1e-6,
                "EllipsoidRadius property, default ellipsoid");
            equalsApprox(gp1.FInverse, sofaFlattening, 1e-6,
                "FInverse property, default ellipsoid");

            // construct at the poles
            double polarRadius = a * (1.0 - 1.0 / flattening);
            xyz[0] = 0.0;
            xyz[1] = 0.0;
            xyz[2] = polarRadius + h / 1000.0;
            gp1 = new Topocenter(Angle.HalfPi, 0.0, h, a, flattening);
            gp1.xyzEqualsApprox(xyz, 1e-6, "construct at north pole");
            xyz[2] = -xyz[2];
            gp1 = new Topocenter(-Angle.HalfPi, 0.0, h, a, flattening);
            gp1.xyzEqualsApprox(xyz, 1e-6, "construct at south pole");

            // copy constructor
            gp1 = new Topocenter(lat, lon, h, a, flattening);
            Topocenter gp2 = new Topocenter(gp1);
            gp2.xyzEqualsApprox(gp1._vector.ToArray(), 0.0, "copy constructor xyz");
            equalsApprox(gp2.Lat, lat, 0.0, "copy constructor lat");
            equalsApprox(gp2.Lon, lon, 0.0, "copy constructor lon");
            equalsApprox(gp2.Height, h, .001, "copy constructor height");
            equalsApprox(gp2.EllipsoidRadius, a, 0.0,
                "copy constructor ellipsoid radius");
            equalsApprox(gp2.FInverse, flattening, 0.0,
                "copy constructor FInverse");

            // Construct from a Vector, verify geodetic coordinate round trip.
            status = Sofa.Gd2gce(a, 1.0 / flattening, lon, lat, h / 1000.0, xyz);
            gp1 = new Topocenter(new Vector(xyz), a, flattening);
            gp1.xyzEqualsApprox(xyz, 0.0, "construct from Vector (specified ellipsoid)");
            equalsApprox(gp1.EllipsoidRadius, a, 0.0,
                "EllipsoidRadius from Vector (specified ellipsoid)");
            equalsApprox(gp1.FInverse, flattening, 0.0,
                "FInverse from Vector (specified ellipsoid)");
            gp1.geodeticEqualsApprox(lat, lon, h, .1,
                "geodetic coords from Vector constuctor (specified ellipsoid)");

            // Construct from Vector and default (GRS80) ellipsoid.
            status = Sofa.Gd2gc(Sofa.Ellipsoid.GRS80, lon, lat, h, xyz);
            Sofa.Sxp(.001, xyz, xyz);   // convert m to km
            gp1 = new Topocenter(new Vector(xyz));
            gp1.xyzEqualsApprox(xyz, 0.0, "construct from Vector (default ellipsoid)");
            equalsApprox(gp1.EllipsoidRadius, sofaRadius, 1e-6,
                "EllipsoidRadius from Vector (default ellipsoid)");
            equalsApprox(gp1.FInverse, sofaFlattening, 1e-6,
                "FInverse from Vector (default ellipsoid)");
            gp1.geodeticEqualsApprox(lat, lon, h, .1,
                "geodetic coords from Vector constuctor (default ellipsoid)");

#if false
            // ISO 6709 constructor, specified ellipsoid.
            status = Sofa.Gd2gce(a, 1.0 / flattening, lon, lat, h, xyz);
            gp1 = new Topocenter(parsed6709, a, flattening);
            gp1.xyzEqualsApprox(xyz, 1e-6,
                "construct from Iso6709 and ellipsoid dimensions");

            // ISO 6709 constructor, default ellipsoid.
            Sofa.GD2GC(2, lon, lat, h * 1000.0, xyz, out status);
            Sofa.SXP(.001, xyz, xyz);   // convert m to km
            gp1 = new Topocenter(parsed6709);
            gp1.xyzEqualsApprox(xyz, 1e-6,
                "construct from Iso6709, default ellipsoid");
#endif

            
            // methods

            // ToString()
            assertTrue(gp1.ToString().Equals(String.Format("{0:f3} {1:f3} {2:f3}",
                xyz[0], xyz[1], xyz[2])));

            // ToVector()
            Vector v = gp1.ToVector();
            equalsApprox(v.X, xyz[0], 1e-6, "ToVector() X");
            equalsApprox(v.Y, xyz[1], 1e-6, "ToVector() Y");
            equalsApprox(v.Z, xyz[2], 1e-6, "ToVector() Z");


            // ToPVVector

            // MJD at J2000.0
            double mjdBase, mjd, mjdAtJ2000;
            status = Sofa.Cal2jd(2000, 1, 1, out mjdBase, out mjd);
            mjdAtJ2000 = mjd + .5;

            // MJD at 0 h on the test date.
            status = Sofa.Cal2jd(2012, 7, 19, out mjdBase, out mjd);

            // fractional days
            double fracDays = 17.0 / 24.0;

            // ITRS to GCRS rotation matrix. Assume TT = UT1 and
            // zero polar motion. Because the output of C2T06A() has the reverse sense of
            // what's needed, it must be transposed.
            double[,] itrsToGcrs = new double[3, 3];
            Sofa.C2t06a(mjdBase + mjdAtJ2000, mjd - mjdAtJ2000 + fracDays,
                mjdBase + mjd, fracDays, 0.0, 0.0, itrsToGcrs);
            Sofa.Tr(itrsToGcrs, itrsToGcrs);

            // Observer's rectangular coordinates (km) in the ITRS. This is already
            // proven correct.
            gp1 = new Topocenter(lat, lon, h, a, flattening);
            double[] observerPosition = gp1.ToVector().ToArray();

            // Vector of length equal to Earth rotation rate (radians/day), directed to
            // the ITRS Z axis.
            double[] earthRotationVector = { 0.0, 0.0, omegaEarth };

            // Observer velocity in an inertial reference system oriented to the ITRS.
            double[] observerVelocity = new double[3];
            Sofa.Pxp(earthRotationVector, observerPosition, observerVelocity);

            // Rotate position and velocity from the ITRS to the GCRS, convert to a
            // PVVector.
            Sofa.Rxp(itrsToGcrs, observerPosition, observerPosition);
            Sofa.Rxp(itrsToGcrs, observerVelocity, observerVelocity);
            PVVector pv1 = new PVVector(
                new Vector(observerPosition), new Vector(observerVelocity));

            // observer's position and velocity computed via the class method
            PVVector pv2 = gp1.ToGcrsPV(new RMatrix(itrsToGcrs));

            // Compare results. For the velocity, tolerance is
            // 1e-6 of the expected velocity.
            double r = Math.Sqrt(observerPosition[0] * observerPosition[0] +
                observerPosition[1] * observerPosition[1]);
            pv1.EqualsApprox(pv2, 1e-6, 1e-6 * omegaEarth * r, "ToGcrsPV()");
        }


        private static void equalsApprox(double a, double b, double tolerance, string s)
        {
            if (Math.Abs(a - b) > tolerance)
                throw new Exception(s);
        }


        /// <summary>
        /// Verify the vector part of an instance is within tolerance.
        /// </summary>
        /// <param name="xyz">correct xyz coordinates (km)</param>
        /// <param name="tolerance">max allowable error (km)</param>
        /// <param name="message">error message</param>
        private void xyzEqualsApprox(double[] xyz, double tolerance, string message)
        {
            if (xyz.Length != 3)
                throw new Exception(message + ": xyz.Length wrong");
            if (
                Math.Abs(this._vector.X - xyz[0]) > tolerance ||
                Math.Abs(this._vector.Y - xyz[1]) > tolerance ||
                Math.Abs(this._vector.Z - xyz[2]) > tolerance)
                throw new Exception(message);
        }


        /// <summary>
        /// Verify latitude, longitude, and height above ellipsoid within tolerance.
        /// </summary>
        /// <param name="lat">correct north latitude (radians)</param>
        /// <param name="lon">correct east correct longitude (radians)</param>
        /// <param name="height">correct height above ellipsoid (meters)</param>
        /// <param name="toleranceMeters">max allowable error (meters)</param>
        /// <param name="message">error message</param>
        private void geodeticEqualsApprox(double lat, double lon, double height,
            double toleranceMeters, string message)
        {
            double a = this._a * 1000.0;
            // Longitude error is adjusted for latitude.
            if (Math.Abs(this._lat - lat) * a > toleranceMeters
            || Math.Abs(this._lon - lon) * a * Math.Abs(Math.Cos(lat)) > toleranceMeters
            || Math.Abs(this._height * 1000.0 - height) > toleranceMeters)
                throw new Exception(message);
        }


        /// <summary>
        /// Throw exception if argument is false.
        /// </summary>
        /// <param name="assertion">true = OK</param>
        private static void assertTrue(bool assertion)
        {
            if (!assertion)
                throw new Exception("Assertion was false.");
        }


#endif      // end class self-test code
        #endregion
    }
}
