// Utc.cs
// Class to represent Universal Coordinated Time.
// Copyright 2018 by Paul S. Hirose

// revision history
#if false
// 2007-12-05 Got a good start on the Test() method. Constructor doesn't detect step adjustment
// greater than 1 second, but it appears that wasn't part of its design.

// 2008-01-15 Converted to use my UtcTable class. Self-test OK, but need to add more methods
// and tests.

// 2008-01-17 Class seems to have everything it needs. Self-test OK.

// 2008-12-09 Added constructor that takes a TimeFields.

// 2008-12-27 In the constructor that takes Terrestrial Time, added exception if it's
// outside the table range. Also added TableIsLoaded property.

// 2009-01-06 Added XML comments to LoadTableFromFile() to say exception is thrown
// when load fails.

2014-10-07  Deleted supportt for ISO 8601.

2015-04-14 Revisions necessary because the UtcTable class was rewritten to convert between UTC and
TAI, not TT. Also, I'm adding provision for each instance to have its own UtcTable instead of
the static table (which is still available).

2015-04-19 Methods and properties made virtual. Self test OK. Ready for release.

2015-06-26 De-virtualized the class.

2015-09-03 In ToTimeFields() the "fudge factor" of 1/4 of the resolution, added to TAI, causes a
bug at say Dec 31 23:59:59.9 when the year ends in a leap second. It puts TAI inside the leap second
and thus a step second is applied incorrectly to the TimeFields object. Removed the fudge factor.

2018-06-04 New constructor which takes date & time & julian flag. Uses default table.
#endif


using System;
using System.Collections.Generic;
using System.Text;
using System.IO;


namespace HirosePS.SofaJpl
{
    /// <summary>
    /// A class to represent Universal Coordinated Time.
    /// </summary>
    /// <remarks>
    /// <para>This is a complete UTC implementation, including the "rubber second" era before 1972.
    /// It obtains the history of UTC from an external table which is an ASCII file and easily
    /// updated by the user when a leap second occurs.</para>
    /// <para>Instances of this class are immutable.</para>
    /// </remarks>
    public class Utc
    {
        // internally, the UTC instance is stored as Terrestrial Time
        private JulianDate _tt;

        // The table used by this instance. Not necessarily the shared table.
        private UtcTable _table;

        // shared UTC to TT conversion table
        private static UtcTable _sharedTable;



        // Constructors


        /// <summary>
        /// Construct from a Gregorian or Julian calendar date, UTC fractional days, and the
        /// supplied UtcTable.
        /// </summary>
        /// <param name="year">0 = 1 BC, -1 = 2 BC, etc.</param>
        /// <param name="month">1 = January</param>
        /// <param name="day">day of month (not restricted to 1 - 31)</param>
        /// <param name="fractionalDay">0.0 to approximately 1.0</param>
        /// <param name="julian">true if the date is in the Julian calendar</param>
        /// <param name="table">UTC table</param>
        /// <exception cref="ArgumentOutOfRangeException">if the date is outside the UTC
        /// table limits</exception>
        /// <exception cref="ArgumentNullException">if <paramref name="table"/> is null</exception>
        /// <remarks>
        /// <para>
        /// Day of month is not range checked. E.g., January 180 may be used for the 180th day of
        /// the year. The date is normalized before UTC adjustments are applied, so 2006
        /// January -1 is equivalent to 2005 December 31.
        /// </para>
        /// <para>Formally, the maximum <paramref name="fractionalDay"/> is 1.0, or slightly
        /// different if the day ends with a step adjustment. However, the constructor does not
        /// apply a range check. Excess time overflows into the next day.</para>
        /// </remarks>
        public Utc(int year, int month, int day, double fractionalDay, bool julian, UtcTable table)
        {
            if (table == null)
                throw new ArgumentNullException("table");
            _table = table;
            JulianDate date = new JulianDate(year, month, day, 0.0, julian);
            Duration time = new Duration(fractionalDay);
            _tt = date + time + _table.TaiMinusUtc(date, time) + Duration.TTMinusTai;
        }


        /// <summary>
        /// Construct from a Gregorian or Julian calendar date, time of day, and the
        /// default UtcTable.
        /// </summary>
        /// <param name="year">0 = 1 BC, -1 = 2 BC, etc.</param>
        /// <param name="month">1 = January</param>
        /// <param name="day">day of month (not restricted to 1 - 31)</param>
        /// <param name="hour">hours</param>
        /// <param name="minute">minutes</param>
        /// <param name="second">seconds</param>
        /// <param name="julian">true if the date is in the Julian calendar</param>
        /// <exception cref="ArgumentOutOfRangeException">if the date is outside the UTC
        /// table limits</exception>
        /// <remarks>
        /// <para>
        /// Day of month is not range checked. E.g., January 180 may be used for the 180th day of
        /// the year. The date is normalized before UTC adjustments are applied, so 2006
        /// January -1 is equivalent to 2005 December 31.
        /// </para>
        /// <para>Formally, time of day should not exceed 24 hours, or slightly
        /// different if the day ends with a step adjustment. However, the constructor does not
        /// apply a range check. Excess time overflows into the next day.</para>
        /// </remarks>
        public Utc(int year, int month, int day, int hour, int minute, double second, bool julian)
            : this(year, month, day, Angle.HmsToDays(hour, minute, second), julian, _sharedTable)
        { }


        /// <summary>
        /// Construct from a Gregorian calendar date and UTC time in fractional days. The
        /// default UTC table is used.
        /// </summary>
        /// <param name="year">0 = 1 BC, -1 = 2 BC, etc.</param>
        /// <param name="month">1 = January</param>
        /// <param name="day">day of month (not restricted to 1 - 31)</param>
        /// <param name="fractionalDay">0.0 to 1.0 (not exactly 1.0 if the day ends with a step
        /// adjustment).</param>
        /// <exception cref="ArgumentOutOfRangeException">if the date is outside the UTC
        /// table limits, or <paramref name="fractionalDay"/> outside the legal range.</exception>
        /// <exception cref="ArgumentNullException">if a UTC table is not loaded</exception>
        /// <remarks>
        /// <para>A default UTC table must be loaded before calling this constructor.</para>
        /// <para>Day of month is not range checked. E.g., January 180 may be used for the 180th day
        /// of the year. The date is normalized before UTC adjustments are applied, so 2006
        /// January -1 is equivalent to 2005 December 31.</para>
        /// <para>Maximum permissible <paramref name="fractionalDay"/> is slightly more or less than
        /// 1.0 if the day ends with a step adjustment.</para>
        /// </remarks>
        public Utc(int year, int month, int day, double fractionalDay)
            : this(year, month, day, fractionalDay, false, _sharedTable)
        { }


        /// <summary>
        /// Construct from date (Gregorian calendar) and UTC time. The default UTC table is used.
        /// </summary>
        /// <param name="year">0 = 1 BC, -1 = 2 BC, etc.</param>
        /// <param name="month">1 = January</param>
        /// <param name="day">day of month (not restricted to 1 - 31)</param>
        /// <param name="hour">hour</param>
        /// <param name="minute">minute</param>
        /// <param name="second">second</param>
        /// <exception cref="ArgumentOutOfRangeException">if the date is outside the UTC
        /// table limits, or time is outside the legal range.</exception>
        /// <exception cref="ArgumentNullException">if a UTC table is not loaded</exception>
        /// <remarks>
        /// <para>A default UTC table must be loaded before calling this constructor.</para>
        /// <para>Day of month is not range checked. E.g., January 180 may be used for the 180th day
        /// of the year. The date is normalized before UTC adjustments are applied, so 2006
        /// January -1 is equivalent to 2005 December 31.</para>
        /// <para>Maximum permissible time of day is slightly more or less than 24 hours if the day
        /// ends with a step adjustment.</para>
        /// </remarks>
        public Utc(int year, int month, int day, int hour, int minute, double second)
            : this(year, month, day, Angle.HmsToDays(hour, minute, second), false, _sharedTable)
        { }


        /// <summary>
        /// Construct from a JulianDate in the Terrestrial Time scale and a UtcTable.
        /// </summary>
        /// <param name="terrestrialTime">Terrestrial Time</param>
        /// <param name="table">UTC table</param>
        /// <exception cref="ArgumentNullException"> if <paramref name="table"/> is null</exception>
        /// <exception cref="ArgumentOutOfRangeException">if time is outside the UTC table limits
        /// </exception>
        public Utc(JulianDate terrestrialTime, UtcTable table)
        {
            if (table == null)
                throw new ArgumentNullException("table");
            JulianDate tai = terrestrialTime - Duration.TTMinusTai;
            if (tai < table.FirstTai || tai > table.LastTai)
                throw new ArgumentOutOfRangeException(
                    "terrestrialTime", "outside the UTC table limits");
            this._table = table;
            this._tt = terrestrialTime;
        }


        /// <summary>
        /// Construct from a JulianDate in the Terrestrial Time scale and the default UTC table.
        /// </summary>
        /// <param name="terrestrialTime">Terrestrial Time</param>
        /// <exception cref="ArgumentNullException"> if no default UTC table is loaded</exception>
        /// <exception cref="ArgumentOutOfRangeException">if time is outside the UTC table limits
        /// </exception>
        /// <remarks>A default UTC table must be loaded before calling this constructor.</remarks>
        public Utc(JulianDate terrestrialTime)
            : this(terrestrialTime, _sharedTable) { }


        /// <summary>
        /// Construct from UTC in a TimeFields instance and the default UTC table.
        /// </summary>
        /// <param name="tf">UTC</param>
        /// <exception cref="ArgumentNullException"> if no default UTC table is loaded</exception>
        /// <exception cref="ArgumentOutOfRangeException">if time is outside the UTC table limits
        /// </exception>
        /// <remarks>A default UTC table must be loaded before calling this constructor.</remarks>
        public Utc(TimeFields tf)
            : this(tf.Year, tf.Month, tf.Day, tf.Hms.ToDouble() / 24.0, tf.Julian, _sharedTable)
        { }




        // Properties


        /// <summary>
        /// Get the Terrestrial Time of this instance.
        /// </summary>
        /// <value>Terrestrial Time</value>
        public JulianDate TerrestrialTime { get { return this._tt; } }


        /// <summary>
        /// Get the UTC table used by this instance.
        /// </summary>
        public UtcTable Table { get { return _table; } }


        /// <summary>
        /// Get or set the default table of the class (not necessarily used by any instance).
        /// Get returns null if not loaded.
        /// </summary>
        public static UtcTable DefaultTable {
            get { return _sharedTable; }
            set { _sharedTable = value; }
        }



        // Methods


        /// <summary>
        /// Load the default UTC table from a file in USNO leap second table format.
        /// </summary>
        /// <param name="filename">name of file</param>
        /// <exception cref="ArgumentException">if unable to open file, or table format
        /// is incorrect</exception>
        /// <remarks>
        /// <para>Extensions to the USNO format are recognized:</para>
        /// <para>1. The Julian date field is ignored and may be blank. The line is assumed to
        /// take effect at 0 hours UTC of the year, month, and day in the line.</para>
        /// <para>2. Any line beginning with a semicolon is ignored.</para>
        /// <para>3. Blank lines and leading and trailing white space are ignored.</para>
        /// </remarks>
        public static void LoadTableFromFile(string filename)
        {
            using (StreamReader sr = new StreamReader(filename)) {
                if (sr == null)
                    throw new ArgumentException("can't open", "filename");
                _sharedTable = new UtcTable(sr);
            }
        }


        /// <summary>
        /// Load the default UTC table from a string in USNO leap second table format.
        /// </summary>
        /// <param name="s">UTC table, as a string</param>
        /// <exception cref="ArgumentException">if table format is incorrect</exception>
        /// <remarks>
        /// <para>Extensions to the USNO format are recognized:</para>
        /// <para>1. The Julian date field is ignored and may be blank. The line is assumed to
        /// take effect at 0 hours UTC of the year, month, and day in the line.</para>
        /// <para>2. Any line beginning with a semicolon is ignored.</para>
        /// <para>3. Blank lines and leading and trailing white space are ignored.</para>
        /// </remarks>
        public static void LoadTableFromString(string s)
        {
            _sharedTable = new UtcTable(new StringReader(s));
        }


        /// <summary>
        /// Convert this instance into UTC in a TimeFields object, Gregorian or Julian calendar,
        /// rounded to the specified resolution.
        /// </summary>
        /// <param name="resolution">60 = 1 minute resolution, 360 = 10 second, 3600 = 1 second,
        /// etc. The result is rounded to the nearest multiple of the resolution.</param>
        /// <param name="julian">true to generate result in the Julian calendar</param>
        /// <returns>UTC in a TimeFields instance.</returns>
        /// <remarks>
        /// <para>The result is correct within a step adjustment, including the fractional second
        /// adjustments before 1972. A negative leap second will also be handled properly, should
        /// one ever be used.</para>
        /// <para>The roundoff may not be strictly correct if the day ends with a step adjustment
        /// (positive or negative) and resolution is more coarse than 1 second. For example,
        /// if time is 23:59:30.1, and 1 minute resolution is selected, and the day will end
        /// with a positive leap second, correct rounding to the nearest minute should yield
        /// 23:59, with a roundoff error of 30.1 s. The apparently correct round up to 00:00 (which
        /// this routine does) increases roundoff error to 30.9 s due to the leap second.</para>
        /// <para>That anomaly won't occur if resolution is equal to or finer than the size
        /// of the step adjustment.</para>
        /// </remarks>
        public TimeFields ToTimeFields(double resolution, bool julian)
        {
            // Get UTC-TAI and the duration of the time step.
            double stepSeconds;     // step adjustment, if any
            Duration utcMinusTai = _table.UtcMinusTai(this._tt - Duration.TTMinusTai,
                out stepSeconds);

            // Convert TT to UTC. Pass that to the TimeFields constructor, including
            // the UTC step adjustment if applicable.
            return new TimeFields(this._tt - Duration.TTMinusTai + utcMinusTai, resolution, julian,
                stepSeconds);
        }


        /// <summary>
        /// Convert this instance into UTC in a TimeFields object in the Gregorian calendar,
        /// rounded to the specified resolution.
        /// </summary>
        /// <param name="resolution">60 = 1 minute resolution, 360 = 10 second, 3600 = 1 second,
        /// etc. The result is rounded to the nearest multiple of the resolution.</param>
        /// <returns>UTC in a TimeFields instance.</returns>
        /// <remarks>
        /// <para>The result is correct within a step adjustment, including the fractional second
        /// adjustments before 1972. A negative leap second will also be handled properly, should
        /// one ever be used.</para>
        /// <para>The roundoff may not be strictly correct if the day ends with a step adjustment
        /// (positive or negative) and resolution is more coarse than 1 second. For example,
        /// if time is 23:59:30.1, and 1 minute resolution is selected, and the day will end
        /// with a positive leap second, correct rounding to the nearest minute should yield
        /// 23:59, with a roundoff error of 30.1 s. The apparently correct round up to 00:00 (which
        /// this routine does) increases roundoff error to 30.9 s due to the leap second.</para>
        /// <para>That anomaly won't occur if resolution is equal to or finer than the size
        /// of the step adjustment.</para>
        /// </remarks>
        public TimeFields ToTimeFields(double resolution)
        {
            return ToTimeFields(resolution, false);
        }


        /// <summary>
        /// Convert to a UTC string with .1-second resolution.
        /// </summary>
        /// <returns>UTC string</returns>
        public override string ToString()
        {
            return this.ToTimeFields(36000).ToString();
        }




        #region self-test code
#if DEBUG

        /// <summary>
        /// Test the class. Runs silently if all tests pass.
        /// </summary>
        public static void Test()
        {
            string tableString =
 @"1961 JAN  1 =JD 2437300.5  TAI-UTC=   1.4228180 S + (MJD - 37300.) X 0.001296 S
 1961 AUG  1 =JD 2437512.5  TAI-UTC=   1.3728180 S + (MJD - 37300.) X 0.001296 S
 1962 JAN  1 =JD 2437665.5  TAI-UTC=   1.8458580 S + (MJD - 37665.) X 0.0011232S
 1963 NOV  1 =JD 2438334.5  TAI-UTC=   1.9458580 S + (MJD - 37665.) X 0.0011232S
 1964 JAN  1 =JD 2438395.5  TAI-UTC=   3.2401300 S + (MJD - 38761.) X 0.001296 S
 1964 APR  1 =JD 2438486.5  TAI-UTC=   3.3401300 S + (MJD - 38761.) X 0.001296 S
 1964 SEP  1 =JD 2438639.5  TAI-UTC=   3.4401300 S + (MJD - 38761.) X 0.001296 S
 1965 JAN  1 =JD 2438761.5  TAI-UTC=   3.5401300 S + (MJD - 38761.) X 0.001296 S
 1965 MAR  1 =JD 2438820.5  TAI-UTC=   3.6401300 S + (MJD - 38761.) X 0.001296 S
 1965 JUL  1 =JD 2438942.5  TAI-UTC=   3.7401300 S + (MJD - 38761.) X 0.001296 S
 1965 SEP  1 =JD 2439004.5  TAI-UTC=   3.8401300 S + (MJD - 38761.) X 0.001296 S
 1966 JAN  1 =JD 2439126.5  TAI-UTC=   4.3131700 S + (MJD - 39126.) X 0.002592 S
 1968 FEB  1 =JD 2439887.5  TAI-UTC=   4.2131700 S + (MJD - 39126.) X 0.002592 S
 1972 JAN  1 =JD 2441317.5  TAI-UTC=  10.0       S + (MJD - 41317.) X 0.0      S
 1972 JUL  1 =JD 2441499.5  TAI-UTC=  11.0       S + (MJD - 41317.) X 0.0      S
 1973 JAN  1 =JD 2441683.5  TAI-UTC=  12.0       S + (MJD - 41317.) X 0.0      S
 1974 JAN  1 =JD 2442048.5  TAI-UTC=  13.0       S + (MJD - 41317.) X 0.0      S
 1975 JAN  1 =JD 2442413.5  TAI-UTC=  14.0       S + (MJD - 41317.) X 0.0      S
 1976 JAN  1 =JD 2442778.5  TAI-UTC=  15.0       S + (MJD - 41317.) X 0.0      S
 1977 JAN  1 =JD 2443144.5  TAI-UTC=  16.0       S + (MJD - 41317.) X 0.0      S
 1978 JAN  1 =JD 2443509.5  TAI-UTC=  17.0       S + (MJD - 41317.) X 0.0      S
 1979 JAN  1 =JD 2443874.5  TAI-UTC=  18.0       S + (MJD - 41317.) X 0.0      S
 1980 JAN  1 =JD 2444239.5  TAI-UTC=  19.0       S + (MJD - 41317.) X 0.0      S
 1981 JUL  1 =JD 2444786.5  TAI-UTC=  20.0       S + (MJD - 41317.) X 0.0      S
 1982 JUL  1 =JD 2445151.5  TAI-UTC=  21.0       S + (MJD - 41317.) X 0.0      S
 1983 JUL  1 =JD 2445516.5  TAI-UTC=  22.0       S + (MJD - 41317.) X 0.0      S
 1985 JUL  1 =JD 2446247.5  TAI-UTC=  23.0       S + (MJD - 41317.) X 0.0      S
 1988 JAN  1 =JD 2447161.5  TAI-UTC=  24.0       S + (MJD - 41317.) X 0.0      S
 1990 JAN  1 =JD 2447892.5  TAI-UTC=  25.0       S + (MJD - 41317.) X 0.0      S
 1991 JAN  1 =JD 2448257.5  TAI-UTC=  26.0       S + (MJD - 41317.) X 0.0      S
 1992 JUL  1 =JD 2448804.5  TAI-UTC=  27.0       S + (MJD - 41317.) X 0.0      S
 1993 JUL  1 =JD 2449169.5  TAI-UTC=  28.0       S + (MJD - 41317.) X 0.0      S
 1994 JUL  1 =JD 2449534.5  TAI-UTC=  29.0       S + (MJD - 41317.) X 0.0      S
 1996 JAN  1 =JD 2450083.5  TAI-UTC=  30.0       S + (MJD - 41317.) X 0.0      S
 1997 JUL  1 =JD 2450630.5  TAI-UTC=  31.0       S + (MJD - 41317.) X 0.0      S
 1999 JAN  1 =JD 2451179.5  TAI-UTC=  32.0       S + (MJD - 41317.) X 0.0      S
 2006 JAN  1 =JD 2453736.5  TAI-UTC=  33.0       S + (MJD - 41317.) X 0.0      S
 2009 JAN  1                TAI-UTC=  34.0       S + (MJD - 41317.) X 0.0      S
 2010 JAN  1                TAI-UTC=  34.0       S + (MJD - 41317.) X 0.0      S";

            string utcTableFilename = @"C:\Users\Stan\Documents\astro\IERS\leapSecTable.txt";
            const double sPerDay = 86400.0;
            Duration tolerance = new Duration(1e-9 / sPerDay);
            UtcTable table, defaultTable;
            Utc utc1;
            Duration utcTime;
            JulianDate tt1, tt2, utcDate;
            TimeFields tf1;
            bool wrongException;
            int year, month, day, hour, minute;
            double second, decimalDays;

            year = 2005;
            month = 12;
            day = 31;
            hour = 23;
            minute = 59;
            second = 60.9;
            decimalDays = Angle.HmsToDays(hour, minute, second);

            
            // Detect table not loaded.

            wrongException = false;

            if (DefaultTable != null)
                throw new Exception("SharedTable failed to return null when no table was loaded.");

            try {
                utc1 = new Utc(year, month, day, decimalDays, false, null);
                wrongException = true;
                throw new Exception();
            } catch (Exception e) {
                if (wrongException)
                    throw e;
            }

            try {
                utc1 = new Utc(year, month, day, decimalDays);
                wrongException = true;
                throw new Exception();
            }
            catch (Exception e) {
                if (wrongException)
                    throw e;
            }

            try {
                tt1 = new JulianDate(year, month, day, decimalDays);
                utc1 = new Utc(tt1, null);
                wrongException = true;
                throw new Exception();
            }
            catch (Exception e) {
                if (wrongException)
                    throw e;
            }

            try {
                tt1 = new JulianDate(year, month, day, decimalDays);
                utc1 = new Utc(tt1);
                wrongException = true;
                throw new Exception();
            }
            catch (Exception e) {
                if (wrongException)
                    throw e;
            }

            try {
                tf1 = new TimeFields(year, month, day, hour, minute, second, 3600, false);
                utc1 = new Utc(tf1);
                wrongException = true;
                throw new Exception();
            }
            catch (Exception e) {
                if (wrongException)
                    throw e;
            }


            // Construct a UTC table from a string.
            table = new UtcTable(new StringReader(tableString));


            // Verify the constructors that take a UTC table parameter accept the table, and the
            // ones that use the default table still fail.
            utc1 = new Utc(year, month, day, decimalDays, false, table);
            tt1 = new JulianDate(year, month, day, decimalDays);
            utc1 = new Utc(tt1, table);

            try {
                utc1 = new Utc(year, month, day, decimalDays);
                wrongException = true;
                throw new Exception();
            }
            catch (Exception e) {
                if (wrongException)
                    throw e;
            }

            try {
                tt1 = new JulianDate(year, month, day, decimalDays);
                utc1 = new Utc(tt1);
                wrongException = true;
                throw new Exception();
            }
            catch (Exception e) {
                if (wrongException)
                    throw e;
            }

            try {
                tf1 = new TimeFields(year, month, day, hour, minute, second, 3600, false);
                utc1 = new Utc(tf1);
                wrongException = true;
                throw new Exception();
            }
            catch (Exception e) {
                if (wrongException)
                    throw e;
            }


            // load the default UTC table from a file
            Utc.LoadTableFromFile(utcTableFilename);

            // verify the property
            if (DefaultTable == null)
                throw new Exception("SharedTable returned null when a table was loaded.");

            // Verify the constructors that rely on the default table no longer throw exceptions.

            utc1 = new Utc(year, month, day, decimalDays);

            tt1 = new JulianDate(year, month, day, decimalDays);
            utc1 = new Utc(tt1);

            tf1 = new TimeFields(year, month, day, hour, minute, second, 3600, false);
            utc1 = new Utc(tf1);

            // Verify accurate conversion to TT.

            // date & decimal days, Gregorian calendar
            utc1 = new Utc(year, month, day, decimalDays, false, table);
            tt1 = utc1.TerrestrialTime;
            utcDate = new JulianDate(year, month, day, 0);
            utcTime = new Duration(decimalDays);
            tt2 = utcDate + utcTime + table.TaiMinusUtc(utcDate, utcTime) + Duration.TTMinusTai;
            equalsApprox(tt1, tt2, tolerance);
            utc1 = new Utc(year, month, day, decimalDays);
            tt1 = utc1.TerrestrialTime;
            equalsApprox(tt1, tt2, tolerance);

            // date & decimal days, Julian calendar
            utc1 = new Utc(year, month, day - 13, decimalDays, true, table);
            tt1 = utc1.TerrestrialTime;
            equalsApprox(tt1, tt2, tolerance);

            // date and time, Julian calendar, default table
            utc1 = new Utc(year, month, day - 13, hour, minute, second, true);
            tt1 = utc1.TerrestrialTime;
            equalsApprox(tt1, tt2, tolerance);

            // date & hms, Gregorian
            utc1 = new Utc(year, month, day, hour, minute, second);
            tt1 = utc1.TerrestrialTime;
            equalsApprox(tt1, tt2, tolerance);

            // TT & UTC table
            utc1 = new Utc(tt2, table);
            tt1 = utc1.TerrestrialTime;
            equalsApprox(tt1, tt2, tolerance);
            
            // TT & default table
            utc1 = new Utc(tt2);
            tt1 = utc1.TerrestrialTime;
            equalsApprox(tt1, tt2, tolerance);

            // UTC (Gregorian calendar) in a TimeFields
            tf1 = new TimeFields(year, month, day, hour, minute, second, 36000);
            utc1 = new Utc(tf1);
            equalsApprox(utc1.TerrestrialTime, tt2, tolerance);

            // UTC (Julian calendar) in a TimeFields
            tf1 = new TimeFields(year, month, day - 13, hour, minute, second, 36000, true);
            utc1 = new Utc(tf1);
            equalsApprox(utc1.TerrestrialTime, tt2, tolerance);


            // properties

            // The TerrestrialTime property was just tested.

            // Default table property. This also tests the methods for loading the table from
            // string or file.

            // Load default table from string.
            Utc.LoadTableFromString(tableString);
            defaultTable = Utc.DefaultTable;
            utc1 = new Utc(year, month, day, decimalDays);
            tt1 = utc1.TerrestrialTime;
            tt2 = utcDate + utcTime + defaultTable.TaiMinusUtc(utcDate, utcTime) +
                Duration.TTMinusTai;
            equalsApprox(tt1, tt2, tolerance);

            // Load default table from file.
            Utc.LoadTableFromFile(utcTableFilename);
            defaultTable = Utc.DefaultTable;
            utc1 = new Utc(year, month, day, decimalDays);
            tt1 = utc1.TerrestrialTime;
            tt2 = utcDate + utcTime + defaultTable.TaiMinusUtc(utcDate, utcTime) +
                Duration.TTMinusTai;
            equalsApprox(tt1, tt2, tolerance);

            // Load default table via property.
            Utc.DefaultTable = table;
            defaultTable = Utc.DefaultTable;
            utc1 = new Utc(year, month, day, decimalDays);
            tt1 = utc1.TerrestrialTime;
            tt2 = utcDate + utcTime + defaultTable.TaiMinusUtc(utcDate, utcTime) +
                Duration.TTMinusTai;
            equalsApprox(tt1, tt2, tolerance);

            // The instance property that returns the UTC table. This relies on the previous test.
            assert(defaultTable.Equals(table));


            // ToTimeFields, Gregorian calendar
            tf1 = utc1.ToTimeFields(36000, false);
            assert(tf1.Year == year);
            assert(tf1.Month == month);
            assert(tf1.Day == day);
            assert(Math.Floor(tf1.Hms.Hours) == hour);
            assert(Math.Floor(tf1.Hms.Minutes) == minute);
            assert(Math.Abs(tf1.Hms.Seconds - second) < .1);
            assert(tf1.Julian == false);

            // ToTimeFields, Juian calendar
            tf1 = utc1.ToTimeFields(36000, true);
            assert(tf1.Year == year);
            assert(tf1.Month == month);
            assert(tf1.Day == day - 13);
            assert(Math.Floor(tf1.Hms.Hours) == hour);
            assert(Math.Floor(tf1.Hms.Minutes) == minute);
            assert(Math.Abs(tf1.Hms.Seconds - second) < .1);
            assert(tf1.Julian == true);

            // ToTimeFields, default Gregorian calendar
            tf1 = utc1.ToTimeFields(36000);
            assert(tf1.Year == year);
            assert(tf1.Month == month);
            assert(tf1.Day == day);
            assert(Math.Floor(tf1.Hms.Hours) == hour);
            assert(Math.Floor(tf1.Hms.Minutes) == minute);
            assert(Math.Abs(tf1.Hms.Seconds - second) < .1);
            assert(tf1.Julian == false);

            // At 1 s resolution, the TimeFields should roll over to the next year.
            tf1 = utc1.ToTimeFields(3600);
            assert(tf1.Year == year + 1);
            assert(tf1.Month == 1);
            assert(tf1.Day == 1);
            assert(Math.Floor(tf1.Hms.Hours) == 0);
            assert(Math.Floor(tf1.Hms.Minutes) == 0);
            assert(Math.Abs(tf1.Hms.Seconds) < .1);

            // Verify the "fudge factor" bug is gone.

            utc1 = new SofaJpl.Utc(2005, 12, 31, 23, 59, 59.9);
            tf1 = utc1.ToTimeFields(3.6e4);
            assert(tf1.Year == 2005);
            assert(tf1.Month == 12);
            assert(tf1.Day == 31);
            assert(Math.Floor(tf1.Hms.Hours) == 23);
            assert(Math.Floor(tf1.Hms.Minutes) == 59);
            assert(Math.Abs(tf1.Hms.Seconds - 59.9) < .01);

            tf1 = utc1.ToTimeFields(3.6e3);
            assert(tf1.Year == 2006);
            assert(tf1.Month == 1);
            assert(tf1.Day == 1);
            assert(Math.Floor(tf1.Hms.Hours) == 0);
            assert(Math.Floor(tf1.Hms.Minutes) == 0);
            assert(Math.Abs(tf1.Hms.Seconds - 0) < .01);
        }


        // Helper method to verify the difference between two JulianDates doesn't exceed a given
        // tolerance.
        private static void equalsApprox(JulianDate jd1, JulianDate jd2, Duration tolerance) {
            if (tolerance.ToDays() < 0.0)
                throw new ArgumentException("must not be negative", "tolerance");
            if ((jd1 - jd2).Abs() > tolerance)
                throw new Exception("difference between JulianDates out of tolerance");
        }


        private static void assert(bool assertion)
        {
            if (!assertion)
                throw new Exception("Assertion failed.");
        }

#endif
        #endregion  // end self-test code
    }
}
