// UtcTable.cs
// Class to represent a table for converting between the UTC and TAI time scales.
// Copyright 2018 by Paul S. Hirose

// revision history
#if false
// 2007-12-30  Created.
// 2008-01-14  Self-test method completed.
// 2008-05-15  UTC table may contain comment lines beginning with semicolon.
// New tests for dates before and after table span.
2014-11-11  Redesigning class to convert between UTC and TAI (not TT).

2014-11-15  Change from TT to TAI was successful, but an unrelated problem appeared in testing.
In the pre-1972 "rubber second" era of UTC, the rate component of TAI-UTC involves subtraction of
one UTC from another. There's a basic flaw: due to step adjustments, the difference obtained
thereby is not the number of elapsed UTC seconds. Thus a UTC step adjustment causes a step
discontinuity in UTC-TAI. It's a few nanosec, but worse than the 1-ns accuracy that I think
is attainable.

2015-04-13  The problem mentioned on 11-15 is solved, at a cost of considerable complication.
The table constructor now provides an argument to select whether or not this solution is applied.
Tests pass within 1 ns if yes, .1 us if no, as expected.

2015-04-19 Made the properties and methods virtual. Test OK. Ready for release.

2015-06-26 De-virtualized the class.

2018-06-03  Bug fix to UtcMinusTai(). At 23:59:60 on a day ending with a leap second, the round trip
to TAI and back yielded UTC at 00:00:00 of the next day. I had used > when I should have used >=.
The check of an illegal (too great) UTC time of day in TaiMinusUtc() has been removed, on the
principle of "liberal on input, conservative on output."
#endif


using System;
using System.Collections.Generic;
using System.Text;
using System.IO;


namespace HirosePS.SofaJpl
{
    /// <summary>
    /// A table for converting between the UTC and TAI scales.
    /// </summary>
    /// <remarks>
    /// <para>Rate offsets, fractional second step adjustments, and even multi-second step
    /// adjustments are allowed.</para>
    /// <para>Normally it's more convenient to use the Utc class instead of this class.</para>
    /// <para>Instances of this class are immutable.</para>
    /// </remarks>
    public class UtcTable
    {
        // each entry in the UTC/TAI conversion table is an instance of the tableEntry class
        private class tableEntry
        {
            // For brevity, I often say "this line" instead of "the time period
            // applicable to this line".

            // The epoch (UTC and TAI) when this line becomes effective.
            public JulianDate beginUtc, beginTai;

            // The rate correction base epoch. At this epoch the rate difference between UTC
            // and TAI has no effect on the time scale conversion. Irrelevant after 1971.

            // Before 1972, baseUtc (and the corresponding baseTai) are not necessarily the exact
            // value in the USNO table. They're adjusted to yield the true amount of elapsed "rubber
            // time" (including step adjustments) from the base epoch.
            
            // For instance, the base epoch for the first two lines of the USNO table is MJD 37300.
            // But there's a -.05 s step at the end of the first line (the last minute has 59.95 s).
            // Thus for a UTC within the span of the second line, if we subtract 37300 the
            // difference will be .05 s more than the true amount of elapsed UTC. To compensate, the
            // base epoch in the second line is adjusted to MJD 37300 - (-.05 s).
            public JulianDate baseUtc, baseTai;

            // TAI-UTC at the rate correction base epoch.
            public Duration offset;

            // Number of days UTC gains on TAI per TAI day. Zero after 1971; less than 0 before.
            public double rateUtc;
            // Number of days TAI gains on UTC per 86400 UTC seconds. Zero after 1971; greater
            // than 0 before.
            public double rateTai;

            // Additional number of days UTC gains on TAI per TAI day. Zero after 1971.
            public double rateUtcCorrection;
            // Additional number of days TAI gains on UTC per 86400 UTC sec. Zero after 1971
            public double rateTaiCorrection;

            // Amount of UTC and TAI inserted at the end of this line. Negative if the last minute
            // has less than 60 seconds.
            public Duration stepDurationUtc, stepDurationTai;

            // The UTC and TAI when the step adjustment begins. If the step adjustment deletes time,
            // beginStepUtc is at 23:59:59 point something. If the step inserts time, or there is
            // no step, beginStepUtc is 00:00:00 on the next day. The ambiguity is harmless because
            // the way the variable is used (23:59:60 has the same effect as 00:00:00 on the
            // next day).
            public JulianDate beginStepUtc, beginStepTai;

            // Algebraic sum of all step adjustments since the last rate change. Does not include
            // the step at the end of this line.
            public Duration accumulatedSteps;

            // Last possible UTC epoch of this line. lastUtcTime is one day plus stepDurationUtc.
            public JulianDate lastUtcDate;
            public Duration lastUtcTime;

        }   // end class tableEntry


        // the UTC/TAI conversion table is a List of instances of the above class.
        List<tableEntry> table;


        // After this epoch (which is at midnight) the table is invalid. Actually it could be used
        // almost 24 more hours. But since the step adjustment at the end of this day is unknown,
        // to be safe an exception occurs if a UTC or TAI parameter is after this epoch.
        JulianDate tableFinalEpoch;


        // miscellaneous statics

        private static string correctFormat =
            "2006 JAN  1                TAI-UTC=  33         S + (MJD - 41317.) X 0.0      S";

        private static string[] month =
        { "JAN", "FEB", "MAR", "APR", "MAY", "JUN", "JUL", "AUG", "SEP", "OCT", "NOV", "DEC" };

        private static readonly Duration zeroDays = new Duration(0.0);
        private static readonly Duration oneDay = new Duration(1.0);
        private const double secPerDay = 86400.0;

        // A step adjustment with absolute value less than this is forced to zero. Prevents
        // annoyance of trivial "step adjustments" which are artifacts of computational error.
        private static readonly Duration minimumStep = new Duration(.01 / secPerDay);



        // constructor

        /// <summary>
        /// Construct from a TextReader which provides the leap second table.
        /// </summary>
        /// <param name="tr">a set of Lines in the USNO leap second table format</param>
        /// <param name="modifyRates">If true, modify the tabular rates (when nonzero) to eliminate
        /// small (few nanosec) discontinuities in the UTC / TAI conversion.</param>
        /// <exception cref="ArgumentException">if incorrect format.</exception>
        /// <remarks>
        /// <para>Expects the USNO leap second table format, with some extensions.</para>
        /// <para>1. The Julian date field is ignored and may be blank. The line is assumed to
        /// take effect at 0 hours UTC of the year, month, and day in the line.</para>
        /// <para>2. Any line beginning with a semicolon is a comment and is ignored.</para>
        /// <para>3. Blank lines and leading and trailing white space are ignored.</para>
        /// <para>Rate offsets and fractional second step adjustments before 1972 are recognized.
        /// Re-introduction of these features, and even multi-second step adjustments, is allowed.
        /// </para>
        /// </remarks>
        public UtcTable(TextReader tr, bool modifyRates)
        {
            this.table = new List<tableEntry>(60);
            String line;
            while ((line = tr.ReadLine()) != null)
                processLine(line);
            computeSteps();
            if (modifyRates)
                adjustRates();
        }


        /// <summary>
        /// Construct from a TextReader which provides the leap second table.
        /// </summary>
        /// <param name="tr">a set of Lines in the USNO leap second table format</param>
        /// <exception cref="ArgumentException">if incorrect format.</exception>
        /// <remarks>
        /// <para>Expects the USNO leap second table format, with some extensions.</para>
        /// <para>1. The Julian date field is ignored and may be blank. The line is assumed to
        /// take effect at 0 hours UTC of the year, month, and day in the line.</para>
        /// <para>2. Any line beginning with a semicolon is a comment and is ignored.</para>
        /// <para>3. Blank lines and leading and trailing white space are ignored.</para>
        /// <para>Rate offsets and fractional second step adjustments before 1972 are recognized.
        /// Re-introduction of these features, and even multi-second step adjustments, is allowed.
        /// </para>
        /// <para>Small adjustments are automatically applied to rate (if nonzero) to eliminate
        /// discontinuities of a few nanoseconds in the conversion to TAI.</para>
        /// </remarks>
        public UtcTable(TextReader tr)
            : this(tr, true) { }


        private void processLine(string s)
        {
            s = s.Trim().ToUpper();

            if (s[0] == ';')
                return;     // the line is a comment; ignore

            if (
                s.Length != 79
                || string.Compare(s, 27, correctFormat, 27, 8) != 0
                || string.Compare(s, 48, correctFormat, 48, 10) != 0
                || string.Compare(s, 67, correctFormat, 67, 1) != 0
                || string.Compare(s, 78, correctFormat, 78, 1) != 0)
                throw new ArgumentException("UTC table format is wrong.");

            int y = int.Parse(s.Substring(0, 4));
            int m = 1 + Array.IndexOf(month, s.Substring(5, 3));
            if (m < 1)
                throw new ArgumentException(string.Format("{0} is not a month name.",
                    s.Substring(5, 3)));
            int d = int.Parse(s.Substring(9, 2));

            tableEntry entry = new tableEntry();

            entry.beginUtc = new JulianDate(y, m, d, 0.0);
            entry.offset = new Duration(double.Parse(s.Substring(35, 13)) / secPerDay);
            entry.baseUtc = new JulianDate(2400000.5, double.Parse(s.Substring(58, 6)));
            entry.rateTai = double.Parse(s.Substring(68, 10)) / secPerDay;

            // entry.rateUtc = -entry.rateTai / (1.0 + entry.rateTai);
            // The above statement is the obvious version. By taking the reciprocal of both sides,
            // simplifying, and again taking the reciprocal, we get the statement below.
            // It avoids the loss of significant digits from entry.rateTai when 1 is added.
            if (entry.rateTai != 0.0)
                entry.rateUtc = 1.0 / (1.0 / -entry.rateTai - 1.0);

            // Provisional value, subject to change if rate correction is applied.
            entry.beginTai = convertTimeScale(entry.beginUtc, entry.offset, entry.baseUtc,
                entry.rateTai, 0.0);

            // the remaining fields are initialized after all table entries are added
            table.Add(entry);
        }


        // Put step adjustment data and other stuff into the UTC table.
        private void computeSteps() {
            JulianDate oldBeginUtc, newBeginUtc;
            Duration offset1, offset2, step, accumulatedSteps;
            double previousTaiRate;
            tableEntry baseLine;

            if (table.Count < 2)
                throw new Exception("UTC table must have at least 2 entries.");

            // The last date in the table is unusable. So is the date before that, since any step
            // adjustment at the end of that date is unknown. Therefore, a table inquiry in that day
            // is prohibited.
            this.tableFinalEpoch = table[table.Count - 1].beginUtc - oneDay;

            // Set up initial conditions of the table processing loop.
            // newBeginUtc and oldBeginUtc will verify table has increasing chronological order.
            newBeginUtc = new JulianDate(0.0, 0.0);     // must predate first table entry
            previousTaiRate = table[0].rateTai;         // rate of first line
            accumulatedSteps = zeroDays;                // initialize to zero

            // Add step adjustment data to the table. The last line is excluded because it's
            // impossible to determine step adjustment duration without data from a next line.
            int i = 0;
            for (; i < table.Count - 1; ++i) {
                oldBeginUtc = newBeginUtc;
                newBeginUtc = table[i].beginUtc;
                if (newBeginUtc <= oldBeginUtc)
                    throw new Exception("UTC table not in increasing chronological order");

                // TAI when this line becomes effective.
                table[i].beginTai = convertTimeScale(table[i].beginUtc, table[i].offset,
                    table[i].baseUtc, table[i].rateTai, 0.0);

                // Rate adjustment base epoch, in TAI.
                baseLine = findUtc(table[i].baseUtc);
                table[i].baseTai = convertTimeScale(table[i].baseUtc, baseLine.offset,
                    baseLine.baseUtc, baseLine.rateTai, 0.0);

                // Use the current line to compute the TAI-UTC offset when the next line becomes
                // effective.
                offset1 = computeOffset(table[i + 1].beginUtc, table[i].offset, table[i].baseUtc,
                    table[i].rateTai, 0.0);
                // Use the next line to repeat the computation.
                offset2 = computeOffset(table[i + 1].beginUtc, table[i + 1].offset,
                    table[i + 1].baseUtc, table[i + 1].rateTai, 0.0);
                // The difference in offsets is the length of the step adjustment (UTC units)
                // at the end of this line. A very small value is assumed to be due to trivial
                // computational error, so it's replaced with zero.
                step = offset2 - offset1;
                if (step.Abs() < minimumStep)
                    step = zeroDays;
                table[i].stepDurationUtc = step;
                table[i].stepDurationTai = step * (1.0 + table[i].rateTai);

                // UTC epoch when step adjustment begins. If last minute has at least 60 s, it's
                // 23:59:60, but I use 0 h on the next day, which is synonymous due to the way this
                // variable is used.
                if (step >= zeroDays)
                    table[i].beginStepUtc = table[i + 1].beginUtc;
                else      // last minute has less than 60 s
                    table[i].beginStepUtc = table[i + 1].beginUtc + table[i].stepDurationUtc;

                // TAI epoch when step adjustment begins.
                table[i].beginStepTai = convertTimeScale(table[i].beginStepUtc,
                    table[i].offset, table[i].baseUtc, table[i].rateTai, 0.0);

                // last epoch to which this line applies
                table[i].lastUtcDate = table[i + 1].beginUtc - oneDay;
                table[i].lastUtcTime = oneDay + table[i].stepDurationUtc;

                if (table[i].rateTai == previousTaiRate) {
                    // Same rate as the previous line (or this is first line of table). The
                    // accumulatedSteps for this line includes all step adjustments since the
                    // last rate change, but not the step at the end of the line.
                    table[i].accumulatedSteps = accumulatedSteps;
                    accumulatedSteps += step;       // add the step at end of current line
                } else {
                    // This line introduces a new rate.
                    previousTaiRate = table[i].rateTai;     // remember the new rate
                    table[i].accumulatedSteps = zeroDays;
                    accumulatedSteps = step;    // the step at the end of this line
                }
            }
            // At this point, i indexes into the last line. Verify it's in chronological order.
            if (table[i].beginUtc <= table[i - 1].beginUtc)
                throw new Exception("UTC table not in increasing chronological order");
        }



        private void adjustRates() {
            tableEntry baseLine;
            JulianDate baseUtc, baseTai;
            Duration additionalOffset, accumulatedSteps;
            double previousTaiRate, rateTaiCorrection, rateUtcCorrection;
            // An "anchor" is a table entry where a change of rate occurs.
            int anchorIndex;

            // Initialize to the first line.
            anchorIndex = 0;
            previousTaiRate = table[0].rateTai;

            for (int i = 0; i < table.Count; ++i) {
                if (table[i].rateTai != previousTaiRate || i == table.Count - 1) {
                    // This line introduces a new rate or is the last line. First multiply the
                    // accumulated step adjustments (during the previous rate) by the rate, to get
                    // the effect on the TAI-UTC offset.
                    accumulatedSteps = table[i - 1].stepDurationUtc + table[i - 1].accumulatedSteps;
                    additionalOffset = accumulatedSteps * previousTaiRate;
                    // Negate and divide by the number of UTC days the old rate was in effect,
                    // including step adjustments. Result is a rate correction which will cancel
                    // additionalOffset. To derive the corresponding UTC rate correction I use the
                    // formula described in processLine().
                    rateTaiCorrection = -additionalOffset /
                        (table[i].beginUtc - table[anchorIndex].beginUtc + accumulatedSteps);
                    if (rateTaiCorrection != 0.0)
                        rateUtcCorrection = 1.0 / (1.0 / -rateTaiCorrection - 1.0);
                    else
                        rateUtcCorrection = 0.0;
                    // Finally, go back to update the lines affected by the corrected rates.
                    baseUtc = table[anchorIndex].baseUtc;
                    // This guards against a mistake in the file.
                    if (baseUtc < table[anchorIndex].beginUtc || baseUtc >= table[i].beginUtc)
                        throw new Exception(string.Format(
                            "base UTC for lines {0} to {1} is outside their time span.",
                            anchorIndex + 1, i));
                    baseLine = findUtc(baseUtc);    // not necessarily at the anchor index
                    baseTai = convertTimeScale(baseUtc, baseLine.offset, baseUtc, baseLine.rateTai,
                        rateTaiCorrection);
                    for (int j = anchorIndex; j < i; ++j) {
                        if (table[j].baseUtc != baseUtc)
                            throw new Exception(string.Format(
                                "base UTC in lines {0} and {1} are not equal",
                                anchorIndex + 1, j + 1));
                        // Adjust baseUtc so subtracting it yields actual elapsed time (including
                        // step adjustments) from the base epoch.
                        table[j].baseUtc += baseLine.accumulatedSteps - table[j].accumulatedSteps;
                        table[j].baseTai = baseTai;
                        table[j].rateUtcCorrection = rateUtcCorrection;
                        table[j].rateTaiCorrection = rateTaiCorrection;
                        table[j].beginTai = convertTimeScale(table[j].beginUtc, table[j].offset,
                            table[j].baseUtc, table[j].rateTai, rateTaiCorrection);
                        table[j].stepDurationTai = table[j].stepDurationUtc *
                            (1.0 + table[j].rateTai + rateTaiCorrection);
                        table[j].beginStepTai = convertTimeScale(table[j].beginStepUtc,
                            table[j].offset, table[j].baseUtc, table[j].rateTai, rateTaiCorrection);
                    }
                    anchorIndex = i;     // this line is the new anchor
                    // Remember the new rate.
                    previousTaiRate = table[i].rateTai;
                }
            }
        }


        // Compute the offset between UTC and TAI.
        private static Duration computeOffset(JulianDate epoch, Duration tabularOffset,
            JulianDate baseEpoch, double rate, double rateAdjustment) {
                return tabularOffset + (epoch - baseEpoch) * (rate + rateAdjustment);
        }


        // Convert from UTC to TAI or vice versa (depends on the parameters).
        private static JulianDate convertTimeScale(JulianDate epoch, Duration tabularOffset,
            JulianDate baseEpoch, double rate, double rateAdjustment) {
                return epoch + computeOffset(epoch, tabularOffset, baseEpoch, rate, rateAdjustment);
        }


#if DEBUG
        /// <summary>
        /// Predicate object to determine if the tableEntry passed to Match() becomes valid
        /// before (or exactly at) a given UTC, which must be at midnight so it doesn't alias
        /// to a different date when inside a step adjustment.
        /// </summary>
#endif
        private class utcLessEqual
        {
            private JulianDate utc;

            public utcLessEqual(JulianDate utc)
            {
                if (utc.TimePart != 0.0)
                    throw new ArgumentException(
                        "utcTimeLessEqual constructor received a UTC not at midnight.");
                this.utc = utc;
            }

            public bool Match(tableEntry entry)
            {
                return entry.beginUtc.CompareTo(this.utc) <= 0;
            }
        }


#if DEBUG
        /// <summary>
        /// Predicate object to determine if the tableEntry passed to Match() becomes valid
        /// before (or exactly at) a given TAI.
        /// </summary>
#endif
        private class taiLessEqual
        {
            private JulianDate tai;

            public taiLessEqual(JulianDate tai)
            {
                this.tai = tai;
            }

            public bool Match(tableEntry entry)
            {
                return entry.beginTai.CompareTo(this.tai) <= 0;
            }
        }


#if DEBUG
        /// <summary>
        /// Find the table entry applicable to a UTC date.
        /// </summary>
        /// <param name="utc">UTC at midnight</param>
        /// <returns>the matching table entry</returns>
        /// <exception cref="ArgumentOutOfRangeException">if the time is outside the table's
        /// range or UTC is not at midnight.</exception>
#endif
        private tableEntry findUtc(JulianDate utc)
        {
            // A JulianDate can safely represent UTC if time is restricted to midnight.
            if (utc.TimePart != 0.0)
                throw new ArgumentOutOfRangeException("utc", "not at midnight");
            if (utc < this.table[0].beginUtc)
                throw new ArgumentOutOfRangeException("utc", "before first date in table");
            if (utc > this.tableFinalEpoch)
                throw new ArgumentOutOfRangeException("utc", "after last valid date in table");

            // Find applicable entry in table.
            utcLessEqual utcPredicate = new utcLessEqual(utc);
            return table.FindLast(utcPredicate.Match);
        }


#if DEBUG
        /// <summary>
        /// Find the table entry applicable to a TAI Julian date.
        /// </summary>
        /// <param name="tai">TAI</param>
        /// <returns>the matching table entry</returns>
        /// <exception cref="ArgumentOutOfRangeException">if the time is outside the table's
        /// range.</exception>
#endif
        private tableEntry findTai(JulianDate tai)
        {
            if (tai < this.table[0].beginTai)
                throw new ArgumentOutOfRangeException("tai", "before first date in table");
            if (tai > this.tableFinalEpoch)
                throw new ArgumentOutOfRangeException("tai", "after last valid date in table");

            taiLessEqual taiPredicate = new taiLessEqual(tai);
            return table.FindLast(taiPredicate.Match);
        }


        /// <summary>
        /// Get the earliest UTC in the table.
        /// </summary>
        public JulianDate FirstUtc { get { return table[0].beginUtc; } }


        /// <summary>
        /// Get the earliest TAI in the table.
        /// </summary>
        public JulianDate FirstTai { get { return table[0].beginTai; } }


        /// <summary>
        /// Get the last UTC in the table.
        /// </summary>
        public JulianDate LastUtc { get { return this.tableFinalEpoch; } }


        /// <summary>
        /// Get the last TAI in the table.
        /// </summary>
        public JulianDate LastTai { get { return this.tableFinalEpoch; } }


        /// <summary>
        /// Get number of entries in the table.
        /// </summary>
        public int Count { get { return table.Count; } }


        /// <summary>
        /// Returns the duration (possibly zero) of the step adjustment at the end of a UTC date.
        /// </summary>
        /// <param name="utc">UTC at midnight</param>
        /// <returns>duration of step adjustment (negative if last minute is less than 60 s,
        /// zero if no step)
        /// </returns>
        /// <exception cref="ArgumentOutOfRangeException"> if <paramref name="utc"/> is not at
        /// midnight or the step would occur after the last valid epoch of the table</exception>
        public Duration StepDuration(JulianDate utc)
        {
            if (utc >= this.tableFinalEpoch)
                throw new ArgumentOutOfRangeException("utc", "after last permissible epoch");
            tableEntry entry = findUtc(utc);
            if (utc == entry.lastUtcDate)
                return entry.stepDurationUtc;
            // not the last date of a table entry, so no step adjustment occurs
            return zeroDays;
        }


        /// <summary>
        /// Shows the correct format for a constructor input line.
        /// </summary>
        public static string Example { get { return correctFormat; } }


        /// <summary>
        /// Computes TAI-UTC given UTC.
        /// </summary>
        /// <param name="utcDate">UTC at 0 hours</param>
        /// <param name="utcTime">Amount of time past 0 hours. Can exceed 1 day if the day ends with
        /// a step adjustment.</param>
        /// <returns>TAI - UTC</returns>
        /// <exception cref="ArgumentOutOfRangeException">if <paramref name="utcTime"/> is
        /// negative or <paramref name="utcDate"/> is outside the table range.</exception>
        public Duration TaiMinusUtc(JulianDate utcDate, Duration utcTime)
        {
            tableEntry entry = findUtc(utcDate);
            if (utcTime < zeroDays)
                throw new ArgumentOutOfRangeException("utcTime", "was negative.");
            if (utcDate != entry.lastUtcDate && utcTime > oneDay)
                throw new ArgumentOutOfRangeException("utcTime",
                        "exceeds 1 day but is not in a step adjustment");
            return computeOffset(utcDate + utcTime, entry.offset, entry.baseUtc, entry.rateTai,
                entry.rateTaiCorrection);
        }


        /// <summary>
        /// Computes UTC-TAI, and the duration of the step adjustment if applicable, given TAI.
        /// </summary>
        /// <param name="tai">TAI</param>
        /// <param name="stepSeconds">Zero except inside a UTC step adjustment. Then
        /// it's the duration of the adjustment, e.g., 1.0 in a positive leap second.</param>
        /// <returns>Usually UTC-TAI. Within a UTC step adjustment, UTC - TAI - (duration of step
        /// adjustment).</returns>
        /// <remarks>
        /// <para>Adding the returned value to TAI produces quasi UTC, which may be converted to
        /// sexagesimal in the normal manner. Then, adding <paramref name="stepSeconds"/> to the
        /// seconds part of the sexagesimal produces a true sexagesimal UTC in which seconds can
        /// exceed 60.</para>
        /// <para>It's impossible to be inside a step adjustment when a minute has less
        /// than 60 seconds, so <paramref name="stepSeconds"/> is never negative</para>
        /// </remarks>
        public Duration UtcMinusTai(JulianDate tai, out double stepSeconds)
        {
            tableEntry entry = findTai(tai);
            Duration step;
            if (tai >= entry.beginStepTai)
                step = entry.stepDurationUtc;
            else
                step = zeroDays;
            stepSeconds = step.ToSeconds();
            return computeOffset(tai, -entry.offset, entry.baseTai, entry.rateUtc,
                entry.rateUtcCorrection) - step;
        }




        #region self-test code
#if DEBUG


        // Class to hold data applicable to one line of the UTC table. For testing. Not in
        // production code.
        private class lineData {
            public JulianDate BeginUtc;
            public Duration Offset;
            public JulianDate BaseEpoch;
            public double Rate;
            public double RateCorrection;
            public Duration Step;
            public double StepSeconds;

            /// <summary>
            /// Construct an object to hold data applicable to one line of the UTC table.
            /// </summary>
            /// <param name="beginUtc">UTC when the line becomes effective</param>
            /// <param name="offsetSec">constant part of TAI-UTC computation (sec)</param>
            /// <param name="baseMjd">base epoch (MJD) of rate part of TAI-UTC computation
            /// (has no effect if <paramref name="rate"/> = 0)</param>
            /// <param name="rate">number of seconds TAI gains on UTC, per 86400 UTC seconds</param>
            /// <param name="stepSec">step adjustment at end of period covered by this line (sec)
            /// (negative if last minute has less than 60 sec)</param>
            /// <param name="beginAccumulatedStepsSec">algebraic sum of step adjustments since this
            /// rate became effective until this line became effective (sec) (ignored if
            /// <paramref name="rate"/> = 0)</param>
            /// <param name="baseAccumulatedStepsSec">algebraic sum of step adjustments since this
            /// rate became effective until rate base epoch (sec) (ignored if
            /// <paramref name="rate"/> = 0)</param>
            /// <param name="rateAccumulatedStepsSec">algebraic sum of step adjustments since this
            /// rate became effective until next rate becomes effective (sec) (ignored if
            /// <paramref name="rate"/> = 0)</param>
            /// <param name="thisAnchor">UTC when this rate became effective (ignored if
            /// <paramref name="rate"/> = 0)</param>
            /// <param name="nextAnchor">UTC when next rate becomes effective (ignored if
            /// <paramref name="rate"/> = 0)</param>
            public lineData(JulianDate beginUtc, double offsetSec, double baseMjd, double rate,
            double stepSec, double beginAccumulatedStepsSec, double baseAccumulatedStepsSec,
            double rateAccumulatedStepsSec, JulianDate thisAnchor, JulianDate nextAnchor) {

                Duration beginAccumulatedSteps, baseEpochAccumulatedSteps, rateAccumulatedSteps,
                    extraDivergence, anchorDifference;

                // Convert parameters to standard units.
                this.BeginUtc = beginUtc;
                this.Offset = new Duration(offsetSec / secPerDay);
                this.BaseEpoch = new JulianDate(JulianDate.MjdBase.ToDouble(), baseMjd);
                this.Rate = rate / secPerDay;
                this.Step = new Duration(stepSec / secPerDay);
                this.StepSeconds = stepSec;

                // If rate is zero, the rest of the method is unnecessary.
                if (rate == 0.0)
                    return;

                beginAccumulatedSteps = new Duration(beginAccumulatedStepsSec / secPerDay);
                baseEpochAccumulatedSteps = new Duration(baseAccumulatedStepsSec / secPerDay);
                rateAccumulatedSteps = new Duration(rateAccumulatedStepsSec / secPerDay);


                // Adjust the base epoch of the rate correction for step adjustments.
                // (The convertTimeScale() method assumes this adjustment has been accomplished.)
                this.BaseEpoch += baseEpochAccumulatedSteps - beginAccumulatedSteps;

                // The additional amount TAI gains on UTC due to step adjustments while this rate
                // is in effect.
                extraDivergence = rateAccumulatedSteps * this.Rate;

                // Negate the extra divergence (because the goal is to eliminate it) then divide by
                // the UTC duration (including step adjustments) the rate is in effect, to obtain
                // TAI rate correction. Zero after 1971.
                anchorDifference = nextAnchor - thisAnchor + rateAccumulatedSteps;
                if (anchorDifference != zeroDays)
                    this.RateCorrection = -extraDivergence / anchorDifference;
                else
                    this.RateCorrection = 0.0;
            }
        }   // end class lineData

        
        // Helper method to verify the difference between two JulianDates doesn't exceed a given
        // tolerance.
        private static void equalsApprox(JulianDate jd1, JulianDate jd2, Duration tolerance) {
            if (tolerance < zeroDays)
                throw new ArgumentException("must not be negative", "tolerance");
            if ((jd1 - jd2).Abs() > tolerance)
                throw new Exception("difference between JulianDates out of tolerance");
        }


        /// <summary>
        /// Test one line of the table.
        /// </summary>
        /// <param name="thisLine">the line to test</param>
        /// <param name="nextLine">the line after the line to test</param>
        /// <param name="table">the UtcTable to test</param>
        /// <param name="dt">When a tiny interval of time is needed, e.g., when epoch must be a
        /// little bit before the end of a leap second, dt is the size of the "little bit".</param>
        /// <param name="toleranceSeconds">Max permissible error for UTC -> TAI conversions etc.
        /// </param>
        /// <exception cref="Exception"> if any test fails</exception>
        private static void testLine(lineData thisLine, lineData nextLine, UtcTable table,
            Duration dt, double toleranceSeconds) {
            JulianDate tai1, tai2, beginTai, beginStep, lastUtcDate;
            // In general, a JulianDate can't fully represent UTC. E.g., 23:59:60.5 looks the same
            // as 00:00:00.5 the next day. However, these tests are written with that in mind.
            JulianDate utc1, utc2;
            Duration lastUtcTime, timeDiff;
            double stepSeconds, utcTime;

            // Compute several quantities which will be needed for the tests.

            // Max permissible error.
            double toleranceDays = toleranceSeconds / secPerDay;
            Duration tolerance = new Duration(toleranceDays);

            // Epoch when step adjusment begins. If no step, or positive step (last minute longer
            // than 60 seconds) this is the beginning of next day. If negative step, it's a
            // little before end of day.
            if (thisLine.Step < zeroDays)
                beginStep = nextLine.BeginUtc + thisLine.Step;
            else
                beginStep = nextLine.BeginUtc;

            // TAI when this line becomes effective.
            beginTai = convertTimeScale(thisLine.BeginUtc, thisLine.Offset, thisLine.BaseEpoch,
                thisLine.Rate, thisLine.RateCorrection);

            // Last UTC covered by this line. The lastUtcTime is not strictly correct at the end of
            // the table since the last legal date in the whole table is 0 hours UTC or TAI, on the
            // date before the date of the final line.
            lastUtcDate = nextLine.BeginUtc - oneDay;
            lastUtcTime = oneDay + thisLine.Step;


            // Now begin the test of one line of the table. First, test at beginning epoch of line.

            // Convert UTC beginning epoch to TAI, verify it matches TAI already computed.
            utc1 = thisLine.BeginUtc;
            tai1 = utc1 + table.TaiMinusUtc(utc1, zeroDays);
            equalsApprox(tai1, beginTai, tolerance);

            // Verify reverse conversion yields the original UTC and correct step.
            utc2 = tai1 + table.UtcMinusTai(tai1, out stepSeconds);
            equalsApprox(utc2, utc1, tolerance);
            assert(Math.Abs(stepSeconds) < toleranceSeconds);

            // Should be no step adjustment on this date.
            assert(table.StepDuration(utc1) == zeroDays);

            // Verify UTC to TAI just after line begins
            utc1 = thisLine.BeginUtc + dt;
            tai1 = utc1 + table.TaiMinusUtc(new JulianDate(utc1.DatePart, 0.0), dt);
            tai2 = convertTimeScale(utc1, thisLine.Offset, thisLine.BaseEpoch, thisLine.Rate,
                thisLine.RateCorrection);
            equalsApprox(tai1, tai2, tolerance);

            // Verify round trip and zero time step.
            utc2 = tai1 + table.UtcMinusTai(tai1, out stepSeconds);
            equalsApprox(utc2, utc1, tolerance);
            assert(Math.Abs(stepSeconds) < toleranceSeconds);

            // Test inside the step adjustment ("inside" includes 23:59:60). If step is negative or
            // there's no step, skip these tests. Also avoid testing at the end the table, where the
            // result would be invalid.
            if (thisLine.Step > dt && lastUtcDate < table.tableFinalEpoch) {

                // Verify conversion to TAI exactly when the step adjustment begins, i.e., at
                // 23:59:60. Note that utc1 represents this epoch exactly: the date portion of
                // the JulianDate instance ends in .5, and the time portion is zero.
                utc1 = lastUtcDate + oneDay;
                tai1 = utc1 + table.TaiMinusUtc(lastUtcDate, oneDay);
                tai2 = convertTimeScale(utc1, thisLine.Offset, thisLine.BaseEpoch,
                    thisLine.Rate, thisLine.RateCorrection);
                equalsApprox(tai1, tai2, tolerance);

                // Verify round trip. Due to the inexactness of floating point math,
                // it is possible the round trip will yield a UTC a little before 23:59:60
                utc2 = tai1 + table.UtcMinusTai(tai1, out stepSeconds);
                if (stepSeconds != 0.0)     // inside the step adjustment
                    equalsApprox(utc1 - thisLine.Step, utc2, tolerance);
                else
                    // In this case utc2 shall be a little before 23:59:60. If it were equal,
                    // the round trip would comes out wrong (00:00:00 next day).
                    assert(utc2 < utc1 && utc1 - utc2 < tolerance);

                // Verify conversion to TAI just inside the step adjustment.
                // utc1 is a quasi UTC. Can't use that in the TaiMinusUtc() call because it's
                // in the wrong (next) day. But elsewhere the ambiguity is harmless.
                utc1 = lastUtcDate + oneDay + dt;
                tai1 = utc1 + table.TaiMinusUtc(lastUtcDate, oneDay + dt);
                tai2 = convertTimeScale(utc1, thisLine.Offset, thisLine.BaseEpoch,
                    thisLine.Rate, thisLine.RateCorrection);
                equalsApprox(tai1, tai2, tolerance);

                // Verify round trip and correct time step. Note that a correct round trip should
                // yield the true UTC minus the duration of the step adjustment.
                utc2 = tai1 + table.UtcMinusTai(tai1, out stepSeconds);
                equalsApprox(utc1 - thisLine.Step, utc2, tolerance);
                assert(Math.Abs(stepSeconds - thisLine.Step.ToSeconds()) < toleranceSeconds);

                // Verify conversion to TAI just before end of step adjustment.
                utc1 = lastUtcDate + lastUtcTime - dt;
                tai1 = utc1 + table.TaiMinusUtc(lastUtcDate, lastUtcTime - dt);
                tai2 = convertTimeScale(utc1, thisLine.Offset, thisLine.BaseEpoch,
                    thisLine.Rate, thisLine.RateCorrection);
                equalsApprox(tai1, tai2, tolerance);

                // Verify round trip and correct time step.
                utc2 = tai1 + table.UtcMinusTai(tai1, out stepSeconds);
                equalsApprox(utc1 - thisLine.Step, utc2, tolerance);
                assert(Math.Abs(stepSeconds - thisLine.Step.ToSeconds() ) < toleranceSeconds);

                // Verify conversion to TAI at end of step adjustment. Should be identical to
                // beginning of next minute.
                utcTime = Angle.HmsToDays(23, 59, 60 + thisLine.StepSeconds);
                utc1 = new JulianDate(lastUtcDate.DatePart, utcTime);
                tai1 = utc1 + table.TaiMinusUtc(lastUtcDate, lastUtcTime);
                tai2 = convertTimeScale(utc1, thisLine.Offset, thisLine.BaseEpoch,
                    thisLine.Rate, thisLine.RateCorrection);
                equalsApprox(tai1, tai2, tolerance);

                // Verify round trip and correct time step.
                utc2 = tai1 + table.UtcMinusTai(tai1, out stepSeconds);
                equalsApprox(utc1 - thisLine.Step, utc2, tolerance);
                // Due to the inexactness of floating point math, utc2 may fall inside or outside of
                // the step adjustment. In the latter case, stepSeconds is zero and the test of its
                // accuracy is not applicable.
                if (stepSeconds != 0.0)
                    assert(Math.Abs(stepSeconds - thisLine.Step.ToSeconds()) < toleranceSeconds);
            }


            // At the end of the table these tests are inhibited since they would exceed the last
            // valid date.
            if (lastUtcDate < table.tableFinalEpoch) {

                // UTC to TAI just before step adjustment (if last second has less than 60 seconds)
                // or just before end of 59th second (if last second has at least 60 seconds). Also
                // tests convertTimeScale() since it's not used here but is used in the code under
                // test.
                utc1 = beginStep - dt;
                tai1 = utc1 + table.TaiMinusUtc(new JulianDate(utc1.DatePart, 0.0),
                    new Duration(utc1.TimePart));
                tai2 = utc1 + thisLine.Offset + (utc1 - thisLine.BaseEpoch) *
                    (thisLine.Rate + thisLine.RateCorrection);
                equalsApprox(tai1, tai2, tolerance);

                // Verify round trip and zero time step.
                utc2 = tai1 + table.UtcMinusTai(tai1, out stepSeconds);
                equalsApprox(utc1, utc2, tolerance);
                assert(Math.Abs(stepSeconds) < toleranceSeconds);

                // Test at the exact end of the UTC day (as close as can be approximated). This test
                // is not always duplicated by the the code that tests inside a step adjustment,
                // since a line in the UTC table doesn't necessarily end in a step adjustment.
                // In the rubber second era, the purpose of the next line may be to introduce a new
                // rate with no step adjustment.
                utcTime = Angle.HmsToDays(23, 59, 60 + thisLine.StepSeconds);
                utc1 = new JulianDate(lastUtcDate.DatePart, utcTime);
                tai1 = utc1 + table.TaiMinusUtc(lastUtcDate, lastUtcTime);
                tai2 = convertTimeScale(utc1, thisLine.Offset, thisLine.BaseEpoch,
                    thisLine.Rate, thisLine.RateCorrection);
                equalsApprox(tai1, tai2, tolerance);

                // Verify round trip and correct time step.
                utc2 = tai1 + table.UtcMinusTai(tai1, out stepSeconds);
                equalsApprox(utc1 - thisLine.Step, utc2, tolerance);
                // Due to the inexactness of floating point math, utc2 may fall inside or outside of
                // the step adjustment. In the latter case, stepSeconds is zero and the test of its
                // accuracy is not applicable.
                if (stepSeconds != 0.0)
                    assert(Math.Abs(stepSeconds - thisLine.Step.ToSeconds()) < toleranceSeconds);

                // Test for discontinuity across the dividing line between this table entry and the
                // next. The UTCs are dt either side of the line. The corresponding TAIs should
                // differ by 2*dt. (I assume dt is such a short duration, the rate difference
                // between UTC and TAI is neglibible.)
                utc1 = lastUtcDate + lastUtcTime - dt;
                tai1 = utc1 + table.TaiMinusUtc(lastUtcDate, lastUtcTime - dt);

                utc2 = nextLine.BeginUtc + dt;
                tai2 = utc2 + table.TaiMinusUtc(nextLine.BeginUtc, dt);

                timeDiff = tai2 - tai1;
                timeDiff.EqualApprox(0.0, (2.0 * dt).ToDays(), toleranceDays);

                // Verify StepDuration() on last date applicable to this line.
                table.StepDuration(lastUtcDate).EqualApprox(thisLine.Step.IntegerPart,
                    thisLine.Step.FractionPart, toleranceDays);
            }
        }


        /// <summary>
        /// Test the class. Runs silently if all tests pass.
        /// </summary>
        public static void Test()
        {
            // the last 3 lines have some minor variations in format but should be accepted
            string goodData1 =
 @"1961 JAN  1 =JD 2437300.5  TAI-UTC=   1.4228180 S + (MJD - 37300.) X 0.001296 S
 1961 AUG  1 =JD 2437512.5  TAI-UTC=   1.3728180 S + (MJD - 37300.) X 0.001296 S
 1962 JAN  1 =JD 2437665.5  TAI-UTC=   1.8458580 S + (MJD - 37665.) X 0.0011232S
 1963 NOV  1 =JD 2438334.5  TAI-UTC=   1.9458580 S + (MJD - 37665.) X 0.0011232S
 1964 JAN  1 =JD 2438395.5  TAI-UTC=   3.2401300 S + (MJD - 38761.) X 0.001296 S
 1964 APR  1 =JD 2438486.5  TAI-UTC=   3.3401300 S + (MJD - 38761.) X 0.001296 S
 1964 SEP  1 =JD 2438639.5  TAI-UTC=   3.4401300 S + (MJD - 38761.) X 0.001296 S
 1965 JAN  1 =JD 2438761.5  TAI-UTC=   3.5401300 S + (MJD - 38761.) X 0.001296 S
 1965 MAR  1 =JD 2438820.5  TAI-UTC=   3.6401300 S + (MJD - 38761.) X 0.001296 S
 1965 JUL  1 =JD 2438942.5  TAI-UTC=   3.7401300 S + (MJD - 38761.) X 0.001296 S
 1965 SEP  1 =JD 2439004.5  TAI-UTC=   3.8401300 S + (MJD - 38761.) X 0.001296 S
 1966 JAN  1 =JD 2439126.5  TAI-UTC=   4.3131700 S + (MJD - 39126.) X 0.002592 S
 1968 FEB  1 =JD 2439887.5  TAI-UTC=   4.2131700 S + (MJD - 39126.) X 0.002592 S
 1972 JAN  1 =JD 2441317.5  TAI-UTC=  10.0       S + (MJD - 41317.) X 0.0      S
 1972 JUL  1 =JD 2441499.5  TAI-UTC=  11.0       S + (MJD - 41317.) X 0.0      S
 1973 JAN  1 =JD 2441683.5  TAI-UTC=  12.0       S + (MJD - 41317.) X 0.0      S
 1974 JAN  1 =JD 2442048.5  TAI-UTC=  13.0       S + (MJD - 41317.) X 0.0      S
 1975 JAN  1 =JD 2442413.5  TAI-UTC=  14.0       S + (MJD - 41317.) X 0.0      S
 1976 JAN  1 =JD 2442778.5  TAI-UTC=  15.0       S + (MJD - 41317.) X 0.0      S
 1977 JAN  1 =JD 2443144.5  TAI-UTC=  16.0       S + (MJD - 41317.) X 0.0      S
 1978 JAN  1 =JD 2443509.5  TAI-UTC=  17.0       S + (MJD - 41317.) X 0.0      S
 1979 JAN  1 =JD 2443874.5  TAI-UTC=  18.0       S + (MJD - 41317.) X 0.0      S
 1980 JAN  1 =JD 2444239.5  TAI-UTC=  19.0       S + (MJD - 41317.) X 0.0      S
 1981 JUL  1 =JD 2444786.5  TAI-UTC=  20.0       S + (MJD - 41317.) X 0.0      S
 1982 JUL  1 =JD 2445151.5  TAI-UTC=  21.0       S + (MJD - 41317.) X 0.0      S
 1983 JUL  1 =JD 2445516.5  TAI-UTC=  22.0       S + (MJD - 41317.) X 0.0      S
 1985 JUL  1 =JD 2446247.5  TAI-UTC=  23.0       S + (MJD - 41317.) X 0.0      S
 1988 JAN  1 =JD 2447161.5  TAI-UTC=  24.0       S + (MJD - 41317.) X 0.0      S
 1990 JAN  1 =JD 2447892.5  TAI-UTC=  25.0       S + (MJD - 41317.) X 0.0      S
 1991 JAN  1 =JD 2448257.5  TAI-UTC=  26.0       S + (MJD - 41317.) X 0.0      S
 1992 JUL  1 =JD 2448804.5  TAI-UTC=  27.0       S + (MJD - 41317.) X 0.0      S
 1993 JUL  1 =JD 2449169.5  TAI-UTC=  28.0       S + (MJD - 41317.) X 0.0      S
 1994 JUL  1 =JD 2449534.5  TAI-UTC=  29.0       S + (MJD - 41317.) X 0.0      S
 1996 JAN  1 =JD 2450083.5  TAI-UTC=  30.0       S + (MJD - 41317.) X 0.0      S
 1997 JUL  1 =JD 2450630.5  TAI-UTC=  31.0       S + (MJD - 41317.) X 0.0      S
; a line begining with a semicolon is ignored
 1999 Jan  1 =JD 2451179.5  TAI-UTC=  32.0       S + (MJD - 41317.) X 0.0      S
  2006 JAN  1 =JD 2453736.5  TAI-UTC=  33         S + (MJD - 41317.) X 0.0      S
  2008 JUL  1                TAI-UTC=  33.0       S + (MJD - 41317.) X 0.0      S ";


            string badData1 =
@" JD 2451179.5  TAI-UTC=  32.0       S + (MJD - 41317.) X 0.0      S
 2006 JAN  1 =JD 2453736.5  TAI-UTC=  33.0       S + (MJD - 41317.) X 0.0      S
 2008 JUL  1                TAI-UTC=  33.0       S + (MJD - 41317.) X 0.0      S";


            UtcTable table1, table2;
            Duration dt = new Duration(1e-6 / secPerDay);
            double toleranceSeconds = 1e-9;
            Duration oneMicrosec = new Duration(1e-6 / secPerDay);
            Duration oneNanosec = new Duration(1e-9 / secPerDay);

            lineData line61jan = new lineData(new JulianDate(1961, 1, 1, 0.0), 1.4228180, 37300.0,
                0.001296, -.05, 0.0, 0.0, -.05, new JulianDate(1961, 1, 1, 0.0),
                new JulianDate(1962, 1, 1, 0.0));
            lineData line61aug = new lineData(new JulianDate(1961, 8, 1, 0.0), 1.3728180, 37300.0,
                0.001296, 0.0, -.05, 0.0, -.05, new JulianDate(1961, 1, 1, 0.0),
                new JulianDate(1962, 1, 1, 0.0));
            lineData line62jan = new lineData(new JulianDate(1962, 1, 1, 0), 1.8458580, 37665,
                0.0011232, .1, 0, 0, .1, new JulianDate(1962, 1, 1, 0),
                new JulianDate(1964, 1, 1, 0));
            lineData line63nov = new lineData(new JulianDate(1963, 11, 1, 0), 1.9458580, 37665,
                0.0011232, 0, .1, 0, .1, new JulianDate(1962, 1, 1, 0),
                new JulianDate(1964, 1, 1, 0));
            lineData line64jan = new lineData(new JulianDate(1964, 1, 1, 0), 3.2401300, 38761,
                .001296, .1, 0, .3, .6, new JulianDate(1964, 1, 1, 0),
                new JulianDate(1966, 1, 1, 0));
            lineData line64apr = new lineData(new JulianDate(1964, 4, 1, 0), 3.3401300, 38761,
                .001296, .1, .1, .3, .6, new JulianDate(1964, 1, 1, 0),
                new JulianDate(1966, 1, 1, 0));
            lineData line64sep = new lineData(new JulianDate(1964, 9, 1, 0), 3.4401300, 38761,
                .001296, .1, .2, .3, .6, new JulianDate(1964, 1, 1, 0),
                new JulianDate(1966, 1, 1, 0));
            lineData line65jan = new lineData(new JulianDate(1965, 1, 1, 0), 3.5401300, 38761,
                .001296, .1, .3, .3, .6, new JulianDate(1964, 1, 1, 0),
                new JulianDate(1966, 1, 1, 0));
            lineData line65mar = new lineData(new JulianDate(1965, 3, 1, 0), 3.6401300, 38761,
                .001296, .1, .4, .3, .6, new JulianDate(1964, 1, 1, 0),
                new JulianDate(1966, 1, 1, 0));
            lineData line65jul = new lineData(new JulianDate(1965, 7, 1, 0), 3.7401300, 38761,
                .001296, .1, .5, .3, .6, new JulianDate(1964, 1, 1, 0),
                new JulianDate(1966, 1, 1, 0));
            lineData line65sep = new lineData(new JulianDate(1965, 9, 1, 0), 3.8401300, 38761,
                .001296, 0, .6, .3, .6, new JulianDate(1964, 1, 1, 0),
                new JulianDate(1966, 1, 1, 0));
            lineData line66jan = new lineData(new JulianDate(1966, 1, 1, 0), 4.3131700, 39126,
                .002592, -.1, 0, 0, -.1 + .107758, new JulianDate(1966, 1, 1, 0),
                new JulianDate(1972, 1, 1, 0));
            lineData line68feb = new lineData(new JulianDate(1968, 2, 1, 0), 4.2131700, 39126,
                .002592, .107758, -.1, 0, -.1 + .107758, new JulianDate(1966, 1, 1, 0),
                new JulianDate(1972, 1, 1, 0));
            lineData line72jan = new lineData(new JulianDate(1972, 1, 1, 0), 10, 41317, 0,
                1, 0, 0, 0, new JulianDate(1972, 1, 1, 0), new JulianDate(1972, 1, 1, 0));
            lineData line72jul = new lineData(new JulianDate(1972, 7, 1, 0), 11, 41317, 0,
                1, 0, 0, 0, new JulianDate(1972, 1, 1, 0), new JulianDate(1972, 1, 1, 0));
            lineData line73jan = new lineData(new JulianDate(1973, 1, 1, 0), 12, 41317, 0,
                1, 0, 0, 0, new JulianDate(1972, 1, 1, 0), new JulianDate(1972, 1, 1, 0));
            lineData line74jan = new lineData(new JulianDate(1974, 1, 1, 0), 13, 41317, 0,
                1, 0, 0, 0, new JulianDate(1972, 1, 1, 0), new JulianDate(1972, 1, 1, 0));
            lineData line75jan = new lineData(new JulianDate(1975, 1, 1, 0), 14, 41317, 0,
                1, 0, 0, 0, new JulianDate(1972, 1, 1, 0), new JulianDate(1972, 1, 1, 0));
            lineData line97jul = new lineData(new JulianDate(1997, 7, 1, 0), 31, 41317, 0,
                1, 0, 0, 0, new JulianDate(1972, 1, 1, 0), new JulianDate(1972, 1, 1, 0));
            lineData line99jan = new lineData(new JulianDate(1999, 1, 1, 0), 32, 41317, 0,
                1, 0, 0, 0, new JulianDate(1972, 1, 1, 0), new JulianDate(1972, 1, 1, 0));
            lineData line06jan = new lineData(new JulianDate(2006, 1, 1, 0), 33, 41317, 0,
                1, 0, 0, 0, new JulianDate(1972, 1, 1, 0), new JulianDate(1972, 1, 1, 0));
            lineData line08jul = new lineData(new JulianDate(2008, 7, 1, 0), 33, 41317, 0,
                1, 0, 0, 0, new JulianDate(1972, 1, 1, 0), new JulianDate(1972, 1, 1, 0));
            

            // bad input data
            try {
                table1 = new UtcTable(new StringReader(badData1));
                Console.WriteLine("failed to detect bad UTC table format");
            } catch {
                ;
            }

            // good input data
            table1 = new UtcTable(new StringReader(goodData1));

            // test the UTC limits of the table
            assert(table1.FirstUtc.CompareTo(new JulianDate(1961, 1, 1, 0.0)) == 0);
            assert(table1.LastUtc.CompareTo(new JulianDate(2008, 6, 30, 0.0)) == 0);

            // test the TAI limits of the table
            table1.FirstTai.EqualsApprox(2437300.5, 1.4228180 / secPerDay, 1e-12);
            assert(table1.LastTai.CompareTo(new JulianDate(2008, 6, 30, 0.0)) == 0);

            // number of table entries
            assert(table1.Count == 38);

            // step adjustment at end of day
            assert(Math.Abs(table1.StepDuration(new JulianDate(1965, 8, 31, 0.0)).ToSeconds()
                - .1) < 1e-7);
            assert(Math.Abs(table1.StepDuration(new JulianDate(1968, 1, 31, 0.0)).ToSeconds()
                + .1) < 1e-7);
            assert(Math.Abs(table1.StepDuration(new JulianDate(1971, 12, 31, 0.0)).ToSeconds()
                - .1077580) < 1e-7);
            assert(Math.Abs(table1.StepDuration(new JulianDate(2005, 12, 31, 0.0)).ToSeconds()
                - 1.0) < 1e-7);
            assert(Math.Abs(table1.StepDuration(new JulianDate(1968, 2, 1, 0.0)).ToSeconds())
                < 1e-7);
            assert(Math.Abs(table1.StepDuration(new JulianDate(2006, 1, 1, 0.0)).ToSeconds())
                < 1e-7);

            // example constructor input line
            assert(Example.Equals(
                "2006 JAN  1                TAI-UTC=  33         S + (MJD - 41317.) X 0.0      S"));

            testLine(line61jan, line61aug, table1, dt, toleranceSeconds);
            testLine(line61aug, line62jan, table1, dt, toleranceSeconds);
            testLine(line62jan, line63nov, table1, dt, toleranceSeconds);
            testLine(line63nov, line64jan, table1, dt, toleranceSeconds);
            testLine(line64jan, line64apr, table1, dt, toleranceSeconds);
            testLine(line64apr, line64sep, table1, dt, toleranceSeconds);
            testLine(line64sep, line65jan, table1, dt, toleranceSeconds);
            testLine(line65jan, line65mar, table1, dt, toleranceSeconds);
            testLine(line65mar, line65jul, table1, dt, toleranceSeconds);
            testLine(line65jul, line65sep, table1, dt, toleranceSeconds);
            testLine(line65sep, line66jan, table1, dt, toleranceSeconds);
            testLine(line66jan, line68feb, table1, dt, toleranceSeconds);
            testLine(line68feb, line72jan, table1, dt, toleranceSeconds);

            // After testing the first few lines of the leap second era I think it's safe to skip to
            // the last few lines.
            testLine(line72jan, line72jul, table1, dt, toleranceSeconds);
            testLine(line72jul, line73jan, table1, dt, toleranceSeconds);
            testLine(line73jan, line74jan, table1, dt, toleranceSeconds);
            testLine(line74jan, line75jan, table1, dt, toleranceSeconds);

            testLine(line97jul, line99jan, table1, dt, toleranceSeconds);
            testLine(line99jan, line06jan, table1, dt, toleranceSeconds);
            testLine(line06jan, line08jul, table1, dt, toleranceSeconds);


            // Construct a table with no rate adjustment, i.e., rates exactly from the file.
            // Then repeat the tests. Error tolerance must be relaxed to prevent failure.
            table2 = new UtcTable(new StringReader(goodData1), false);
            toleranceSeconds = 1e-7;

            testLine(line61jan, line61aug, table2, dt, toleranceSeconds);
            testLine(line61aug, line62jan, table2, dt, toleranceSeconds);
            testLine(line62jan, line63nov, table2, dt, toleranceSeconds);
            testLine(line63nov, line64jan, table2, dt, toleranceSeconds);
            testLine(line64jan, line64apr, table2, dt, toleranceSeconds);
            testLine(line64apr, line64sep, table2, dt, toleranceSeconds);
            testLine(line64sep, line65jan, table2, dt, toleranceSeconds);
            testLine(line65jan, line65mar, table2, dt, toleranceSeconds);
            testLine(line65mar, line65jul, table2, dt, toleranceSeconds);
            testLine(line65jul, line65sep, table2, dt, toleranceSeconds);
            testLine(line65sep, line66jan, table2, dt, toleranceSeconds);
            testLine(line66jan, line68feb, table2, dt, toleranceSeconds);
            testLine(line68feb, line72jan, table2, dt, toleranceSeconds);

            testLine(line72jan, line72jul, table2, dt, toleranceSeconds);
            testLine(line72jul, line73jan, table2, dt, toleranceSeconds);
            testLine(line73jan, line74jan, table2, dt, toleranceSeconds);
            testLine(line74jan, line75jan, table2, dt, toleranceSeconds);

            testLine(line97jul, line99jan, table2, dt, toleranceSeconds);
            testLine(line99jan, line06jan, table2, dt, toleranceSeconds);
            testLine(line06jan, line08jul, table2, dt, toleranceSeconds);
        }


        private static void assert(bool assertion)
        {
            if (!assertion)
                throw new Exception("Assertion failed.");
        }

#endif
        #endregion  // end self-test code
    }       // end class UtcTable
}       // end namespace HirosePS
