// Vector.cs
// A 3-dimensional vector class based on the IAU SOFA library.
// Copyright 2018 by Paul S. Hirose

// revision history
#if false
2007-10-19 Finished rewriting so the internal representation is a double array instead of
separate x, y, and z fields. Eliminated unsafe code and references to the Win32 Sofajpl
DLL. Many of the XML comments are from the IAU SOFA documentation, and the class uses the
SOFA library, so the SOFA software license applies to this class. There is a copy of that
license at the end of this file.

2008-01-19 Self-test code complete. All tests OK.
2008-02-11 Clarified a comment.
2008-05-03 Added an EqualsApprox() overload to compare to a Vector instance.
2009-01-07 Added more detail to the XML comments for the constructor that takes a
Spherical.
2012-07-19 Added an EqualsApprox() test method overload which takes a error message
argument. Deleted SOFA license.
2012-08-30 Added an EqualsApprox() which accepts three doubles and an error message.
2015-05-10 I now compile the SOFA library from their C source, which has a different capitalization
style compared to the Fortran code, requiring corresponding change here.
2015-05-18 Improved the EqualsApprox() methods to report the amount of error and the element
where it occurred.
2015-05-19 Properties and methods changed to virtual. Added overloads for SeparationAngle and
PositionAngle which take a parameter of type Spherical. Self test OK.
2015-06-26 De-virtualized the class.
2015-07-02 EqualsApprox() methods throw exception if a NaN argument is detected.
#endif


using System;


namespace HirosePS.SofaJpl
{
    /// <summary>
    /// A 3-dimensional vector in rectangular coordinates.
    /// </summary>
    /// <remarks>The internal representation employs three double precision values.</remarks>
    public class Vector
    {
        // The vector is stored in rectangular coordinates here.
        private double[] xyz;




        /// <summary>
        /// Construct a null vector.
        /// </summary>
        public Vector() { this.xyz = new double[3]; }


        /// <summary>
        /// Construct vector from individual x, y, z components.
        /// </summary>
        /// <param name="x">x coordinate</param>
        /// <param name="y">y coordinate</param>
        /// <param name="z">z coordinate</param>
        public Vector(double x, double y, double z)
        {
            this.xyz = new double[3] { x, y, z };
        }


        /// <summary>
        /// Initialize from another vector.
        /// </summary>
        /// <param name="v">another Vector</param>
        public Vector(Vector v)
        {
            this.xyz = v.xyz;
        }


        /// <summary>
        /// Construct from array of {x, y, z}.
        /// </summary>
        /// <param name="arr">array of {x, y, z}</param>
        public Vector(double[] arr)
        {
            this.xyz = new double[3] { arr[0], arr[1], arr[2] };
        }


        /// <summary>
        /// Construct vector from a Spherical object.
        /// </summary>
        /// <param name="sph">Spherical instance</param>
        /// <remarks>If spherical angles theta and phi are zero, the result is a +x
        /// vector. If theta is pi/2 and phi is zero, the result is a +y vector.
        /// </remarks>
        public Vector(Spherical sph)
        {
            this.xyz = new double[3];
            Sofa.S2p(sph.Theta, sph.Phi, sph.Radius, this.xyz);
        }

        /// <summary>
        /// Construct a new Vector (of modulus 1) at a given position angle and separation angle
        /// from an existing Vector. If the latter is null, a null Vector is constructed.
        /// </summary>
        /// <param name="vec">vector to the center point</param>
        /// <param name="positionAngle">position angle from <paramref name="vec"/>, 0 at north,
        /// increasing east (radians)</param>
        /// <param name="separationAngle">separation angle from <paramref name="vec"/> (radians)
        /// </param>
        public Vector(Vector vec, double positionAngle, double separationAngle)
        {
            // convert vec to a unit vector
            double[] unitVec = { vec.xyz[0], vec.xyz[1], vec.xyz[2] };
            double r;   // magnitude
            Sofa.Pn(unitVec, out r, unitVec);
            if (r == 0.0) {     // handle the case of a null vector
                this.xyz = new double[3];
                return;
            }

            // Create a unit vector to the desired point in the PA / separation angle coordinate
            // frame by treating the position angle and separation as azimuth and zenith distance,
            // respectively.

            double[] resultVec = new double[3];
            // spherical to rectangular conversion
            Sofa.S2c(
                Angle.TwoPi + Angle.HalfPi - positionAngle,
                Angle.Complement(separationAngle),
                resultVec);

            // Rotate resultVec into the same frame as unitVec (the reference direction). This
            // is most easily visualized by imagining the opposite rotation (from the latter to the
            // former), creating the corresponding rotation matrix, then finally using the transpose
            // of that matrix for the actual rotation.

            // To get the rotation angles, convert unitVec to spherical coords.
            double theta, phi;      // longitude & latitude angles
            Sofa.C2s(unitVec, out theta, out phi);

            // Build the rotation matrix.
            double[,] rotationMat = new double[3, 3];
            Sofa.Ir(rotationMat);   // initialize to identity matrix
            Sofa.Rz(theta + Angle.HalfPi, rotationMat);     // z rotate by theta + 90 deg
            Sofa.Rx(Angle.Complement(phi), rotationMat);    // x rotate by complement of phi

            // Rotate resultVec into the final frame.
            Sofa.Trxp(rotationMat, resultVec, resultVec);
            this.xyz = resultVec;
        }


        /// <summary>
        /// Construct a rotation vector from a rotation matrix, as described in the IAU SOFA
        /// documentation.
        /// </summary>
        /// <remarks>
        /// <para>1) A rotation matrix describes a rotation about some arbitrary axis.
        /// The axis is called the Euler axis, and the angle through which
        /// the reference frame rotates is called the Euler angle.  The
        /// "rotator vector" returned by this constructor has the same direction
        /// as the Euler axis, and its magnitude is the Euler angle in radians.
        /// </para>
        /// <para>2) If rmat is a zero matrix, the result is a zero vector.
        /// If rmat is not a rotation matrix the result is undefined.  rmat must be proper
        /// (i.e. have a positive determinant) and real orthogonal (inverse = transpose).
        /// </para>
        /// <para>3) The reference frame rotates clockwise as seen looking along
        /// the rotation vector from the origin.
        /// </para>
        /// <para>Because the amount of rotation can be scaled up and down simply by
        /// multiplying the vector by a scalar, r-vectors are useful for
        /// representing spins about an axis which is fixed.
        /// </para>
        /// </remarks>
        /// <param name="rmat">rotation matrix</param>
        public Vector(RMatrix rmat)
        {
            this.xyz = new double[3];
            Sofa.Rm2v(rmat.ToArray(), this.xyz);
        }


        /// <summary>
        /// Write this instance's x,y,z components to string.
        /// </summary>
        /// <returns>String formatted as "{0:g6} {1:g6} {2:g6}".</returns>
        public override string ToString()
        {
            return string.Format("{0:g6} {1:g6} {2:g6}", this.xyz[0], this.xyz[1], this.xyz[2]);
        }


        /// <summary>
        /// Put the x, y, z coordinates of this instance into an new array in that order.
        /// </summary>
        /// <returns>new array of x, y, z</returns>
        public double[] ToArray()
        {
            return new double[] { this.xyz[0], this.xyz[1], this.xyz[2] };
        }


        /// <summary>
        /// The x coordinate of this instance.
        /// </summary>
        public double X
        {
            get { return this.xyz[0]; }
        }

        /// <summary>
        /// The y coordinate of this instance.
        /// </summary>
        public double Y
        {
            get { return this.xyz[1]; }
        }

        /// <summary>
        /// The z coordinate of this instance.
        /// </summary>
        public double Z
        {
            get { return this.xyz[2]; }
        }


        /// <summary>
        /// Additive inverse of this instance.
        /// </summary>
        /// <returns>new Vector: negative of this instance.</returns>
        public Vector Negate()
        {
            return new Vector(-this.xyz[0], -this.xyz[1], -this.xyz[2]);
        }

        /// <summary>
        /// Additive inverse operator.
        /// </summary>
        /// <param name="v">Vector</param>
        /// <returns>new Vector: negative <paramref name="v"/></returns>
        public static Vector operator -(Vector v)
        {
            return v.Negate();
        }

        /// <summary>
        /// Add this Vector to another.
        /// </summary>
        /// <param name="rhs">another Vector</param>
        /// <returns>sum of the Vectors</returns>
        public Vector Add(Vector rhs)
        {
            return new Vector(this.xyz[0] + rhs.xyz[0], this.xyz[1] + rhs.xyz[1],
                this.xyz[2] + rhs.xyz[2]);
        }

        /// <summary>
        /// Vector addition operator.
        /// </summary>
        /// <param name="lhs">Vector</param>
        /// <param name="rhs">Vector</param>
        /// <returns>sum of the Vectors</returns>
        public static Vector operator +(Vector lhs, Vector rhs)
        {
            return lhs.Add(rhs);
        }

        /// <summary>
        /// Subtract a Vector from this instance.
        /// </summary>
        /// <param name="rhs">Vector to subtract</param>
        /// <returns>difference between Vectors</returns>
        public Vector Subtract(Vector rhs)
        {
            return new Vector(this.xyz[0] - rhs.xyz[0], this.xyz[1] - rhs.xyz[1],
                this.xyz[2] - rhs.xyz[2]);
        }

        /// <summary>
        /// Vector subtraction operator.
        /// </summary>
        /// <param name="lhs">Vector</param>
        /// <param name="rhs">Vector</param>
        /// <returns>difference between Vectors</returns>
        public static Vector operator -(Vector lhs, Vector rhs)
        {
            return lhs.Subtract(rhs);
        }

        /// <summary>
        /// Multiply this Vector by a scalar.
        /// </summary>
        /// <param name="rhs">multiplier</param>
        /// <returns>this Vector times the scalar</returns>
        public Vector Multiply(double rhs)
        {
            return new Vector(rhs * this.xyz[0], rhs * this.xyz[1], rhs * this.xyz[2]);
        }

        /// <summary>
        /// Operator: Vector times a scalar.
        /// </summary>
        /// <param name="lhs">Vector</param>
        /// <param name="rhs">scalar</param>
        /// <returns>this Vector times the scalar</returns>
        public static Vector operator *(Vector lhs, double rhs)
        {
            return lhs.Multiply(rhs);
        }

        /// <summary>
        /// Operator: scalar times a Vector.
        /// </summary>
        /// <param name="lhs">scalar</param>
        /// <param name="rhs">Vector</param>
        /// <returns>this Vector times the scalar</returns>
        public static Vector operator *(double lhs, Vector rhs)
        {
            return rhs.Multiply(lhs);
        }

        /// <summary>
        /// Divide this Vector by a scalar.
        /// </summary>
        /// <param name="d">scalar</param>
        /// <returns>quotient</returns>
        /// <exception cref="ArgumentOutOfRangeException">thrown if d is 0.</exception>
        public Vector Divide(double d)
        {
            if (d == 0.0)
                throw new ArgumentOutOfRangeException("Attempt to divide Vector by 0.");
            return new Vector(this.xyz[0] / d, this.xyz[1] / d, this.xyz[2] / d);
        }

        /// <summary>
        /// Operator: divide a Vector by a scalar.
        /// </summary>
        /// <param name="lhs">Vector</param>
        /// <param name="rhs">scalar</param>
        /// <returns>quotient</returns>
        /// <exception cref="ArgumentOutOfRangeException">if rhs is zero.</exception>
        public static Vector operator /(Vector lhs, double rhs)
        {
            return lhs.Divide(rhs);
        }

        /// <summary>
        /// Modulus (length) of this Vector.
        /// </summary>
        /// <returns>modulus</returns>
        public double Modulus()
        {
            return Sofa.Pm(this.xyz);
        }

        /// <summary>
        /// Create a unit vector (modulus = 1) with the same orientation as this instance.
        /// Result has modulus 0 if input has modulus 0.
        /// </summary>
        /// <returns>new Vector of modulus 1</returns>
        public Vector Unit()
        {
            Vector result = new Vector();
            double r;   // modulus
            Sofa.Pn(this.xyz, out r, result.xyz);
            return result;
        }
        
        /// <summary>
        /// Dot product of this Vector and another Vector.
        /// </summary>
        /// <param name="vec">the other Vector</param>
        /// <returns>dot product</returns>
        public double Dot(Vector vec)
        {

            return Sofa.Pdp(this.xyz, vec.xyz);
        }

        /// <summary>
        /// Cross product of this Vector and another Vector.
        /// </summary>
        /// <param name="vec">the other Vector</param>
        /// <returns>vector cross product</returns>
        public Vector Cross(Vector vec)
        {
            Vector result = new Vector();
            Sofa.Pxp(this.xyz, vec.xyz, result.xyz);
            return result;
        }

        /// <summary>
        /// Separation angle between two Vectors. If either vector is null, zero is returned.
        /// </summary>
        /// <param name="vec">the other Vector</param>
        /// <returns>separation angle (radians)</returns>
        /// <remarks>The algorithm has high accuracy regardless of the angle.</remarks>
        public double SeparationAngle(Vector vec)
        {
            return Sofa.Sepp(this.xyz, vec.xyz);
        }

        /// <summary>
        /// Separation angle between a Vector and a Spherical. If the vector is null, or the
        /// Spherical has zero radius, zero is returned.
        /// </summary>
        /// <param name="sph">a point in spherical coordinates</param>
        /// <returns>separation angle (radians)</returns>
        /// <remarks>The algorithm has high accuracy regardless of the angle.</remarks>
        public double SeparationAngle(Spherical sph) {
            return sph.SeparationAngle(this);
        }


        /// <summary>
        /// Position angle of another vector with respect to this vector.
        /// </summary>
        /// <param name="vec">the other point</param>
        /// <returns>position angle (radians)</returns>
        /// <remarks>
        /// Angle is zero at north, increasing east, and in the interval [-pi, +pi].
        /// Result is zero if the points have identical direction or if either has zero magnitude.
        /// </remarks>
        public double PositionAngle(Vector vec)
        {
            return Sofa.Pap(this.xyz, vec.xyz);
        }


        /// <summary>
        /// Position angle of a point in spherical coordinates with respect to this vector.
        /// </summary>
        /// <param name="sph">the other point</param>
        /// <returns>position angle (radians)</returns>
        /// <remarks>
        /// Angle is zero at north, increasing east, and in the interval [-pi, +pi].
        /// Result is zero if the points have identical direction or if either has zero magnitude.
        /// </remarks>
        public double PositionAngle(Spherical sph) {
            return this.PositionAngle(new Vector(sph));
        }



        #region self-test code
#if DEBUG


        /// <summary>
        /// Self-test method. Runs silently if all tests pass.
        /// </summary>
        public static void Test()
        {
            Vector vec1, vec2;
            Spherical sph2;

            // Test EqualsApprox().

            vec1 = new Vector(1.0, 2.0, 3.0);
            vec2 = new Vector(1.0, 2.0, 3.0);
            vec1.EqualsApprox(vec2, 0.0, "EqualsApprox() failed");

            // Should throw exception.
#if false
            vec1 = new Vector(1.0, 2.0, 4.0);
            vec1.EqualsApprox(vec2, 0.0, "EqualsApprox() test");
#endif

            // Construct null vector.
            vec1 = new Vector();
            vec1.EqualsApprox(0.0, 0.0, 0.0, 1e-6);

            // Construct from individual x, y, z components.
            vec1 = new Vector(1.0, -2.0, 3);
            vec1.EqualsApprox(1.0, -2.0, 3.0, 1e-6);

            // Construct from another Vector.
            vec2 = new Vector(vec1);
            vec1.EqualsApprox(1.0, -2.0, 3.0, 1e-6);

            // Construct from array.
            double[] arr1 = { 1.0, 2.0, 3.0 };
            vec1 = new Vector(arr1);
            vec1.EqualsApprox(1.0, 2.0, 3.0, 1e-6);

            // Construct from Spherical
            new Vector(new Spherical(-2.0, 1.0, 3.0)).EqualsApprox(-.675, -1.474, 2.524, .001);

            // Construct from a Vector, position angle, and separation angle.
            vec1 = new Vector(new Spherical(-Math.PI / 4.0, 0.0, 1.0));
            vec1 = new Vector(vec1, 0.0, Math.PI / 4.0);
            vec1.EqualsApprox(.5, -.5, Math.Sqrt(2.0) / 2.0, 1e-6);

            // Construct from a rotation matrix.
            vec1 = new Vector(new RMatrix().ZRotate(1.0));
            vec1.EqualsApprox(0.0, 0.0, 1.0, 1e-6);

            // ToString()
            assert(vec1.ToString().Equals("0 0 1"));

            // ToArray()
            arr1 = new Vector(0.0, 1.0, 2.0).ToArray();
            assert(arr1[0] == 0.0 && arr1[1] == 1.0 && arr1[2] == 2.0);

            // x, y, and z properties
            vec1 = new Vector(0.0, 1.0, 2.0);
            assert(vec1.X == 0.0 && vec1.Y == 1.0 && vec1.Z == 2.0);

            // Negate() and unary -
            vec1 = new Vector(-1.0, -2.0, -3.0);
            vec1.Negate().EqualsApprox(1.0, 2.0, 3.0, 1e-6);
            vec2 = -vec1;
            vec2.EqualsApprox(1.0, 2.0, 3.0, 1e-6);

            // Add() and + operator
            vec1 = new Vector(2.0, 4.0, 6.0);
            vec2 = new Vector(1.0, 2.0, 3.0);
            vec1.Add(vec2).EqualsApprox(3.0, 6.0, 9.0, 1e-6);
            new Vector(vec1 + vec2).EqualsApprox(3.0, 6.0, 9.0, 1e-6);

            // Subtract() and - operator
            vec1.Subtract(vec2).EqualsApprox(1.0, 2.0, 3.0, 1e-6);
            new Vector(vec1 - vec2).EqualsApprox(1.0, 2.0, 3.0, 1e-6);

            // Multiply() and * operators
            vec1.Multiply(-.5).EqualsApprox(-1.0, -2.0, -3.0, 1e-6);
            new Vector(vec1 * .5).EqualsApprox(1.0, 2.0, 3.0, 1e-6);
            new Vector(.5 * vec1).EqualsApprox(1.0, 2.0, 3.0, 1e-6);

            // Divide() and / operator
            vec1.Divide(2.0).EqualsApprox(1.0, 2.0, 3.0, 1e-6);
            new Vector(vec1 / 2.0).EqualsApprox(1.0, 2.0, 3.0, 1e-6);

            // Modulus() and Unit()
            assert(vec1.Modulus() - 7.483 < .001);
            vec2 = vec1.Unit();
            new Vector(vec2 * vec1.Modulus()).EqualsApprox(2.0, 4.0, 6.0, 1e-6);

            // dot and cross products
            vec1 = new Vector(-.046, .831, -.086);
            vec2 = new Vector(-.946, -.448, -.391);
            assert(Math.Abs(vec1.Dot(vec2) - (-.295)) < .001);
            vec1.Cross(vec2).EqualsApprox(-.363, .063, .807, .001);

            // separation angle
            assert(Math.Abs(vec1.SeparationAngle(vec2) - 1.892) < .001);
            sph2 = new Spherical(vec2);
            assert(Math.Abs(vec1.SeparationAngle(sph2) - 1.892) < .001);

            // position angle
            vec1 = new Vector(1.0, 0.0, 0.0);
            vec2 = new Vector(1.0, .1, 0.0);
            sph2 = new Spherical(vec2);
            assert(Math.Abs(vec1.PositionAngle(vec2) - Math.PI / 2.0) < 1e-6);
            assert(Math.Abs(vec1.PositionAngle(sph2) - Math.PI / 2.0) < 1e-6);
            vec2 = new Vector(1.0, 0.0, .1);
            sph2 = new Spherical(vec2);
            assert(Math.Abs(vec1.PositionAngle(vec2)) < 1e-6);
            assert(Math.Abs(vec1.PositionAngle(sph2)) < 1e-6);
        }


        /// <summary>
        /// Throw an exception if any of the Vector's components is out of tolerance.
        /// </summary>
        /// <param name="x">correct x</param>
        /// <param name="y">correct y</param>
        /// <param name="z">correct z</param>
        /// <param name="tolerance">max allowable error</param>
        /// <param name="message">error message</param>
        /// <exception cref="Exception">if out of tolerance.</exception>
        public void EqualsApprox(double x, double y, double z, double tolerance,
            string message) {
            equalsApprox(this.xyz[0], x, tolerance, 0, message);
            equalsApprox(this.xyz[1], y, tolerance, 1, message);
            equalsApprox(this.xyz[2], z, tolerance, 2, message);
        }


        /// <summary>
        /// Throw an exception if any of the Vector's components is out of tolerance.
        /// </summary>
        /// <param name="x">correct x</param>
        /// <param name="y">correct y</param>
        /// <param name="z">correct z</param>
        /// <param name="tolerance">max allowable error</param>
        /// <exception cref="Exception">if out of tolerance.</exception>
        public void EqualsApprox(double x, double y, double z, double tolerance)
        {
            EqualsApprox(x, y, z, tolerance, "Vector");
        }


        /// <summary>
        /// Throw an exception if any of the Vector's components is out of tolerance.
        /// </summary>
        /// <param name="v">correct values</param>
        /// <param name="tolerance">max allowable error</param>
        /// <param name="message">error message</param>
        /// <exception cref="Exception">if out of tolerance.</exception>
        public void EqualsApprox(Vector v, double tolerance, string message) {
            EqualsApprox(v.xyz[0], v.xyz[1], v.xyz[2], tolerance, message);
        }


        /// <summary>
        /// Throw an exception if any of the Vector's components is out of tolerance.
        /// </summary>
        /// <param name="v">correct values</param>
        /// <param name="tolerance">max allowable error</param>
        /// <exception cref="Exception">if out of tolerance.</exception>
        public void EqualsApprox(Vector v, double tolerance)
        {
            EqualsApprox(v, tolerance, "Vector");
        }


        /// <summary>
        /// Throw an exception if the separation angle from another vector is excessive.
        /// </summary>
        /// <param name="v">the other vector</param>
        /// <param name="tolerance">max permissible separation angle (radians)</param>
        /// <param name="message">message to display in exception ("Separation angle excessive: "
        /// is prepended)</param>
        /// <exception cref="Exception"> if separation angle exceeds tolerance</exception>
        /// <exception cref="ArgumentException"> if either vector has zero modulus</exception>
        public void EqualsApproxDirection(Vector v, double tolerance, string message) {
            if (this.Modulus() == 0.0 || v.Modulus() == 0.0)
                throw new ArgumentException("Null vector passed to EqualsApproxDirection().");
            if (this.SeparationAngle(v) > tolerance)
                throw new Exception("Separation angle excessive: " + message);
        }


        private static void assert(bool assertion)
        {
            if (!assertion)
                throw new Exception("Assertion false.");
        }


        private static void equalsApprox(double a, double b, double tolerance, int i,
            string message) {

            if (double.IsNaN(a) || double.IsNaN(b))
                throw new ArgumentException("NaN passed to Vector.equalsApprox()");

            double error = a - b;
            if (Math.Abs(error) > tolerance)
                throw new Exception(message + string.Format(
                    ": error = {0:e2} at element [{1:d}]", error, i));
        }

#endif
        #endregion  // end self-test code
    }
}
